Package: maestro
Type: Package
Title: Orchestration of Data Pipelines
Version: 1.2.0
Authors@R: 
    c(
      person("Will", "Hipson", , "will.e.hipson@gmail.com", role = c("cre", "aut", "cph"), 
      comment = c(ORCID = "0000-0002-3931-2189")),
      person("Ryan", "Garnett", , "ryangarnett78@gmail.com", role = c("aut", "ctb", "cph"))
    )
Maintainer: Will Hipson <will.e.hipson@gmail.com>
Description: Framework for creating and orchestrating data pipelines. Organize, orchestrate, and monitor multiple pipelines in a single project. Use tags to decorate functions with scheduling parameters and configuration.
License: MIT + file LICENSE
URL: https://github.com/whipson/maestro,
        https://whipson.github.io/maestro/
BugReports: https://github.com/whipson/maestro/issues
Encoding: UTF-8
Imports: cli (>= 3.3.0), dplyr (>= 1.1.0), glue, lifecycle, logger,
        lubridate (>= 1.9.1), purrr (>= 1.0.0), R.utils, R6, rlang (>=
        1.0.0), roxygen2, tictoc, timechange, utils
Depends: R (>= 4.1.0)
Suggests: asciicast, furrr, future, knitr, quarto, rmarkdown,
        rstudioapi, testthat (>= 3.0.0), withr
Config/testthat/edition: 3
VignetteBuilder: knitr, quarto
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-02 12:11:28 UTC; willhipson
Author: Will Hipson [cre, aut, cph] (ORCID:
    <https://orcid.org/0000-0002-3931-2189>),
  Ryan Garnett [aut, ctb, cph]
Repository: CRAN
Date/Publication: 2026-07-02 12:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(roclet_process,roclet_maestro)
S3method(roclet_process,roclet_maestroDays)
S3method(roclet_process,roclet_maestroFlags)
S3method(roclet_process,roclet_maestroFrequency)
S3method(roclet_process,roclet_maestroHours)
S3method(roclet_process,roclet_maestroInputs)
S3method(roclet_process,roclet_maestroLabel)
S3method(roclet_process,roclet_maestroLogLevel)
S3method(roclet_process,roclet_maestroMap)
S3method(roclet_process,roclet_maestroMonths)
S3method(roclet_process,roclet_maestroOutputs)
S3method(roclet_process,roclet_maestroPriority)
S3method(roclet_process,roclet_maestroRunIf)
S3method(roclet_process,roclet_maestroSkip)
S3method(roclet_process,roclet_maestroStartTime)
S3method(roclet_process,roclet_maestroTz)
S3method(roxy_tag_parse,roxy_tag_maestro)
S3method(roxy_tag_parse,roxy_tag_maestroDays)
S3method(roxy_tag_parse,roxy_tag_maestroFlags)
S3method(roxy_tag_parse,roxy_tag_maestroFrequency)
S3method(roxy_tag_parse,roxy_tag_maestroHours)
S3method(roxy_tag_parse,roxy_tag_maestroInputs)
S3method(roxy_tag_parse,roxy_tag_maestroLabel)
S3method(roxy_tag_parse,roxy_tag_maestroLogLevel)
S3method(roxy_tag_parse,roxy_tag_maestroMap)
S3method(roxy_tag_parse,roxy_tag_maestroMonths)
S3method(roxy_tag_parse,roxy_tag_maestroOutputs)
S3method(roxy_tag_parse,roxy_tag_maestroPriority)
S3method(roxy_tag_parse,roxy_tag_maestroRunIf)
S3method(roxy_tag_parse,roxy_tag_maestroSkip)
S3method(roxy_tag_parse,roxy_tag_maestroStartTime)
S3method(roxy_tag_parse,roxy_tag_maestroTz)
export(build_schedule)
export(create_maestro)
export(create_pipeline)
export(get_artifacts)
export(get_flags)
export(get_labels)
export(get_network)
export(get_run_sequence)
export(get_schedule)
export(get_slot_usage)
export(get_status)
export(invoke)
export(last_build_errors)
export(last_run_errors)
export(last_run_messages)
export(last_run_warnings)
export(run_schedule)
export(suggest_orch_frequency)
import(R6)
import(lifecycle)
import(logger)
import(tictoc)
import(utils)
importFrom(R.utils,countLines)
importFrom(roxygen2,roclet_output)
importFrom(roxygen2,roclet_process)
importFrom(roxygen2,roxy_tag_parse)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# maestro <a href="https://whipson.github.io/maestro/"><img src="man/figures/logo.png" alt="maestro website" align="right" height="138"/></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/whipson/maestro/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/whipson/maestro/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/whipson/maestro/branch/main/graph/badge.svg)](https://app.codecov.io/gh/whipson/maestro?branch=main)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/maestro)](https://cran.r-project.org/package=maestro)
[![Total
Downloads](https://cranlogs.r-pkg.org/badges/grand-total/maestro)](https://CRAN.R-project.org/package=maestro)
[![Monthly
Downloads](https://cranlogs.r-pkg.org/badges/maestro)](https://cran.r-project.org/package=maestro)
<!-- badges: end -->

`maestro` is a lightweight framework for creating and orchestrating data
pipelines in R. At its core, maestro is an R script scheduler that is
unique in two ways:

1.  Stateless: It does not need to be continuously running - it can be
    run in a serverless architecture
2.  Use of *rounded scheduling*: The timeliness of pipeline executions
    depends on how often you run your orchestrator

In `maestro` you create **pipelines** (functions) and schedule them
using `roxygen2` tags - these are special comments (decorators) above
each function. Then you create an **orchestrator** containing `maestro`
functions for scheduling and invoking the pipelines.

## Installation

`maestro` is available on CRAN and can be installed via:

``` r
install.packages("maestro")
```

Or, try out the development version via:

``` r
devtools::install_github("https://github.com/whipson/maestro")
```

## Big Picture

`maestro` is a framework for orchestrating multiple pipelines. The
picture below imagines multiple weather-based ingestion pipelines,
managed by a single orchestrator.

<img src="man/figures/orchestrator.svg" alt="Diagram showing the Maestro orchestrator firing three pipelines (temp_readings every 15 min, weather_bulletins every 30 min, satellite_imagery every hour) with a downstream send_alert pipeline." width="100%"/>

## Project Setup

A `maestro` project needs at least two components:

1.  A collection of R pipelines (functions) that you want to schedule
2.  A single orchestrator script that kicks off the scripts when they’re
    scheduled to run

The project file structure will look something like this:

    sample_project
    ├── orchestrator.R
    └── pipelines
        ├── satellite_imagery.R
        ├── temp_readings.R
        └── weather_bulletins.R

Use `maestro::create_maestro()` to easily create this project structure
in a blank R project.

### Orchestrator

The orchestrator is a script that checks the schedules of all the
pipelines in a `maestro` project and executes them. The orchestrator
also handles global execution tasks such as collecting logs and managing
shared resources like global objects and custom functions.

You have the option of using Quarto, RMarkdown, or a straight-up R
script for the orchestrator, but the former two have some advantages
with respect to deployment on Posit Connect.

A simple orchestrator looks like this:

``` r
library(maestro)

schedule <- build_schedule()

output <- run_schedule(
  schedule, 
  orch_frequency = "15 minutes"
)
```

<img src="man/figures/README-/unnamed-chunk-2.svg"
style="width:100.0%" />

The function `build_schedule()` scours through all the pipelines in the
project and builds a schedule. Then `run_schedule()` checks each
pipeline’s scheduled time against the system time within some margin of
rounding and calls those pipelines to run.

### Pipelines

A pipeline is task we want to run. This task may involve retrieving data
from a source, performing cleaning and computation on the data, then
sending it to a destination. `maestro` is not concerned with what your
pipeline does, but rather *when* you want to run it. Here’s a simple
pipeline in `maestro`:

``` r
#' Example ETL pipeline
#' @maestroFrequency daily
#' @maestroStartTime 12:30:00
my_etl <- function() {
  
  # Pretend we're getting data from a source
  message("Get data")
  extracted <- mtcars
  
  # Transform
  message("Transforming")
  transformed <- extracted |> 
    dplyr::mutate(hp_deviation = hp - mean(hp))
  
  # Load - write to a location
  message("Writing")
  # write.csv(transformed, file = paste0("transformed_mtcars_", Sys.Date(), ".csv"))
}
```

What makes this a `maestro` pipeline is the use of special
*roxygen*-style comments above the function definition:

- `#' @maestroFrequency daily` indicates that this function should
  execute at a daily frequency.

- `#' @maestroStartTime 12:30:00` indicates that it runs at 12:30.

In other words, we’d expect it to run every day at 12:30. There are more
`maestro` tags than these ones and all follow the camelCase convention
established by `roxygen2`.

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroOutputs high_road low_road
  start <- function() {
    c('a', 'A')
  }
  
  #' @maestroInputs start
  high_road <- function(.input) {
    toupper(.input)
  }
  
  #' @maestroInputs start
  low_road <- function(.input) {
    tolower(.input)
  }",
  con = "pipelines/dags.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' ./pipelines/dags.R
# #' @maestroOutputs high_road low_road
# start <- function() {
#   c('a', 'A')
# }
# 
# #' @maestroInputs start
# high_road <- function(.input) {
#   toupper(.input)
# }
# 
# #' @maestroInputs start
# low_road <- function(.input) {
#   tolower(.input)
# }

## ----eval=FALSE---------------------------------------------------------------
# #' Example ETL pipeline
# #' @maestroFrequency 1 day
# #' @maestroStartTime 2024-03-25 12:30:00
# my_etl <- function() {
# 
#   # Pretend we're getting data from a source
#   message("Get data")
#   extracted <- mtcars
# 
#   # Transform
#   message("Transforming")
#   transformed <- extracted |>
#     dplyr::mutate(hp_deviation = hp - mean(hp))
# 
#   # Load - write to a location
#   message("Writing")
#   write.csv(transformed, file = paste0("transformed_mtcars_", Sys.Date(), ".csv"))
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(file.remove("pipelines/dags.R"))
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroFrequency 1 day
  #' @maestroStartTime 2024-03-25 12:30:00
  #' @maestroOutputs transform
  extract <- function() {
    # Imagine this is something way more complicated, like a database call
    mtcars
  }
  
  #' @maestroOutputs load
  transform <- function(.input) {
    .input |> 
      dplyr::mutate(hp_deviation = hp - mean(hp))
  }
  
  #' @maestro
  load <- function(.input) {
    write.csv(.input, file = paste0('transformed_mtcars.csv'))
  }",
  con = "pipelines/etl.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency 1 day
# #' @maestroStartTime 2024-03-25 12:30:00
# #' @maestroOutputs transform
# extract <- function() {
#   # Imagine this is something way more complicated, like a database call
#   mtcars
# }
# 
# #' @maestroOutputs load
# transform <- function(.input) {
#   .input |>
#     dplyr::mutate(hp_deviation = hp - mean(hp))
# }
# 
# #' @maestro
# load <- function(.input) {
#   write.csv(.input, file = paste0("transformed_mtcars.csv"))
# }

## ----cleanup, echo=FALSE, message=FALSE, warning=FALSE------------------------
invisible(unlink("pipelines", recursive = TRUE))
invisible(unlink("transformed_mtcars.csv"))

test_that(".prev_on_cycle returns NA when current <= start (POSIXct)", {
  start <- as.POSIXct("2024-01-01 09:00:00", tz = "UTC")

  expect_true(is.na(.prev_on_cycle(start, current = start, unit = "hour")))
  expect_true(is.na(.prev_on_cycle(start, current = start - 1, unit = "hour")))
})

test_that(".prev_on_cycle returns NA when current <= start (Date)", {
  start <- as.Date("2024-01-01")

  expect_true(is.na(.prev_on_cycle(start, current = start, unit = "day")))
  expect_true(is.na(.prev_on_cycle(start, current = start - 1, unit = "day")))
})

test_that(".prev_on_cycle returns NA for invalid start", {
  expect_true(is.na(.prev_on_cycle(
    as.POSIXct(NA), current = as.POSIXct("2024-01-02"), unit = "day"
  )))
})

test_that(".prev_on_cycle works for seconds", {
  start <- as.POSIXct("2024-01-01 00:00:00", tz = "UTC")
  current <- as.POSIXct("2024-01-01 00:00:45", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 10L, unit = "second"),
    as.POSIXct("2024-01-01 00:00:40", tz = "UTC")
  )
})

test_that(".prev_on_cycle works for minutes", {
  start <- as.POSIXct("2024-01-01 09:00:00", tz = "UTC")
  current <- as.POSIXct("2024-01-01 09:47:00", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 15L, unit = "minute"),
    as.POSIXct("2024-01-01 09:45:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle works for hours", {
  start <- as.POSIXct("2024-01-01 06:00:00", tz = "UTC")
  current <- as.POSIXct("2024-01-01 20:00:01", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 2L, unit = "hour"),
    as.POSIXct("2024-01-01 20:00:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle works for days", {
  start <- as.POSIXct("2024-01-01 09:35:00", tz = "UTC")
  current <- as.POSIXct("2024-01-08 09:35:01", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 3L, unit = "day"),
    as.POSIXct("2024-01-07 09:35:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle works for weeks", {
  start <- as.POSIXct("2024-01-01 00:00:00", tz = "UTC")
  current <- as.POSIXct("2024-01-22 00:00:01", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 2L, unit = "week"),
    as.POSIXct("2024-01-15 00:00:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle does not return current when it is exactly on cycle", {
  start <- as.POSIXct("2024-01-01 09:00:00", tz = "UTC")
  current <- as.POSIXct("2024-01-01 11:00:00", tz = "UTC")

  result <- .prev_on_cycle(start, current, amount = 1L, unit = "hour")

  # Must return the previous slot, not current
  expect_equal(result, as.POSIXct("2024-01-01 10:00:00", tz = "UTC"))
})

test_that(".prev_on_cycle returns current slot when nudged by 1 second", {
  start <- as.POSIXct("2024-01-01 09:00:00", tz = "UTC")
  current <- as.POSIXct("2024-01-01 11:00:00", tz = "UTC")

  result <- .prev_on_cycle(start, current + 1, amount = 1L, unit = "hour")

  expect_equal(result, as.POSIXct("2024-01-01 11:00:00", tz = "UTC"))
})

test_that(".prev_on_cycle works for months", {
  start <- as.POSIXct("2024-01-15 09:00:00", tz = "UTC")
  current <- as.POSIXct("2024-04-20 09:00:00", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 1L, unit = "month"),
    as.POSIXct("2024-04-15 09:00:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle clamps month-end dates for months (POSIXct)", {
  start <- as.POSIXct("2024-01-31 00:00:00", tz = "UTC")
  current <- as.POSIXct("2024-03-01 00:00:00", tz = "UTC")

  result <- .prev_on_cycle(start, current, amount = 1L, unit = "month")

  expect_equal(result, as.POSIXct("2024-02-29 00:00:00", tz = "UTC"))
})

test_that(".prev_on_cycle works for multi-month step", {
  start <- as.POSIXct("2024-01-01 00:00:00", tz = "UTC")
  current <- as.POSIXct("2024-07-15 00:00:00", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 3L, unit = "month"),
    as.POSIXct("2024-07-01 00:00:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle works for years", {
  start <- as.POSIXct("2020-06-15 12:00:00", tz = "UTC")
  current <- as.POSIXct("2024-06-16 00:00:00", tz = "UTC")

  expect_equal(
    .prev_on_cycle(start, current, amount = 1L, unit = "year"),
    as.POSIXct("2024-06-15 12:00:00", tz = "UTC")
  )
})

test_that(".prev_on_cycle clamps leap-day start for years", {
  start <- as.POSIXct("2020-02-29 00:00:00", tz = "UTC")
  current <- as.POSIXct("2023-03-01 00:00:00", tz = "UTC")

  result <- .prev_on_cycle(start, current, amount = 1L, unit = "year")

  expect_equal(result, as.POSIXct("2023-02-28 00:00:00", tz = "UTC"))
})

test_that(".prev_on_cycle preserves the start timezone", {
  start <- as.POSIXct("2024-01-01 09:00:00", tz = "America/Halifax")
  current <- as.POSIXct("2024-01-01 12:00:00", tz = "America/Halifax")

  result <- .prev_on_cycle(start, current, amount = 1L, unit = "hour")

  expect_equal(attr(result, "tzone"), "America/Halifax")
  expect_equal(result, as.POSIXct("2024-01-01 11:00:00", tz = "America/Halifax"))
})

test_that(".prev_on_cycle works for days (Date)", {
  start <- as.Date("2024-01-01")
  current <- as.Date("2024-01-10")

  expect_equal(
    .prev_on_cycle(start, current, amount = 3L, unit = "day"),
    as.Date("2024-01-07")
  )
})

test_that(".prev_on_cycle works for weeks (Date)", {
  start <- as.Date("2024-01-01")
  current <- as.Date("2024-01-20")

  expect_equal(
    .prev_on_cycle(start, current, amount = 1L, unit = "week"),
    as.Date("2024-01-15")
  )
})

test_that(".prev_on_cycle works for months (Date)", {
  start <- as.Date("2024-01-15")
  current <- as.Date("2024-04-20")

  expect_equal(
    .prev_on_cycle(start, current, amount = 1L, unit = "month"),
    as.Date("2024-04-15")
  )
})

test_that(".prev_on_cycle clamps month-end dates (Date)", {
  start <- as.Date("2024-01-31")
  current <- as.Date("2024-03-01")

  expect_equal(
    .prev_on_cycle(start, current, amount = 1L, unit = "month"),
    as.Date("2024-02-29")
  )
})

test_that(".prev_on_cycle works for years (Date)", {
  start <- as.Date("2020-03-10")
  current <- as.Date("2024-03-11")

  expect_equal(
    .prev_on_cycle(start, current, amount = 1L, unit = "year"),
    as.Date("2024-03-10")
  )
})

test_that(".prev_on_cycle errors on sub-day unit with Date input", {
  start <- as.Date("2024-01-01")

  expect_error(.prev_on_cycle(start, as.Date("2024-01-02"), unit = "hour"))
  expect_error(.prev_on_cycle(start, as.Date("2024-01-02"), unit = "minute"))
  expect_error(.prev_on_cycle(start, as.Date("2024-01-02"), unit = "second"))
})

test_that(".prev_on_cycle errors on non-Date/POSIXct start", {
  expect_error(.prev_on_cycle("2024-01-01", Sys.time(), unit = "day"))
})

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    options(maestro.check_datetime_override = as.POSIXct("2026-04-21 03:00:00", tz = "UTC"))
    set.seed(1) 
  })
)

options(asciicast_theme = "pkgdown")

## -----------------------------------------------------------------------------
library(maestro)

## ----echo=FALSE, error=FALSE, warning=FALSE, message=FALSE--------------------
dir.create("pipelines")
writeLines(
  "
  #' my_pipe maestro pipeline
  #'
  #' @maestroFrequency 1 day
  #' @maestroStartTime 03:00:00
  #' @maestroTz UTC
  #' @maestroLogLevel INFO
  
  my_pipe <- function() {
  
    # Pipeline code
  }",
  con = "pipelines/my_pipe.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' my_pipe maestro pipeline
# #'
# #' @maestroFrequency 1 day
# #' @maestroStartTime 03:00:00
# #' @maestroTz UTC
# #' @maestroLogLevel INFO
# 
# my_pipe <- function() {
# 
#   # Pipeline code
# }

## -----------------------------------------------------------------------------
#' my_pipe maestro pipeline
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 03:00:00
#' @maestroTz UTC
#' @maestroLogLevel INFO

my_pipe <- function() {

  random_data <- data.frame(
    letters = sample(letters, 10),
    numbers = sample.int(10)
  )
  
  write.csv(random_data, file = tempfile())
}

## ----example-pipe-------------------------------------------------------------
# create_pipeline(
#   pipe_name = "another_pipeline",
#   pipeline_dir = "pipelines",
#   frequency = "1 hour",
#   tz = "America/Halifax",
#   log_level = "ERROR"
# )

## ----cleanup, echo=FALSE, message=FALSE, warning=FALSE------------------------
unlink("pipelines", recursive = TRUE)

## -----------------------------------------------------------------------------
#| echo: false
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")


## -----------------------------------------------------------------------------
#| echo: false
#| output: false
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency hourly
  hello <- function() {
    message('hello')
  }
  
  #' @maestroFrequency hourly
  uhh <- function() {
    warning('this could be a problem')
  }
  
  #' @maestroFrequency hourly
  oh_no <- function() {
    stop('this is bad')
  }",
  con = "pipelines/logs.R"
)


## -----------------------------------------------------------------------------
#| eval: false
# #' pipelines/logs.R
# #' @maestroFrequency hourly
# hello <- function() {
#   message('hello')
# }
# 
# #' @maestroFrequency hourly
# uhh <- function() {
#   warning('this could be a problem')
# }
# 
# #' @maestroFrequency hourly
# oh_no <- function() {
#   stop('this is bad')
# }


## # orchestrator.R
## library(maestro)
## 
## schedule <- build_schedule(quiet = TRUE)
## 
## status <- run_schedule(
##   schedule,
##   run_all = TRUE,
##   log_to_file = TRUE
## )

## readLines("maestro.log")

## # orchestrator.R
## schedule <- build_schedule(quiet = TRUE)
## 
## status <- run_schedule(
##   schedule,
##   run_all = TRUE,
##   log_to_console = TRUE
## )

## -----------------------------------------------------------------------------
#| echo: false
#| output: false
file.remove("maestro.log")


## -----------------------------------------------------------------------------
#| echo: false
#| output: false
unlink("pipelines", recursive = TRUE)
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency hourly
  #' @maestroLogLevel ERROR
  hello <- function() {
    message('hello')
  }
  
  #' @maestroFrequency hourly
  #' @maestroLogLevel ERROR
  uhh <- function() {
    warning('this could be a problem')
  }
  
  #' @maestroFrequency hourly
  #' @maestroLogLevel ERROR
  oh_no <- function() {
    stop('this is bad')
  }",
  con = "pipelines/logs.R"
)


## -----------------------------------------------------------------------------
#| eval: false
# #' pipelines/logs.R
# #' @maestroFrequency hourly
# #' @maestroLogLevel ERROR
# hello <- function() {
#   message('hello')
# }
# 
# #' @maestroFrequency hourly
# #' @maestroLogLevel ERROR
# uhh <- function() {
#   warning('this could be a problem')
# }
# 
# #' @maestroFrequency hourly
# #' @maestroLogLevel ERROR
# oh_no <- function() {
#   stop('this is bad')
# }


## # orchestrator.R
## schedule <- build_schedule(quiet = TRUE)
## 
## status <- run_schedule(
##   schedule,
##   run_all = TRUE,
##   log_to_console = TRUE
## )

## -----------------------------------------------------------------------------
#| echo: false
#| output: false
unlink("pipelines", recursive = TRUE)
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency hourly
  hello <- function() {
    logger::log_info('hi', namespace = 'hello')
  }
  
  #' @maestroFrequency hourly
  uhh <- function() {
    logger::log_warn('this could be a problem', namespace = 'uhh')
  }
  
  #' @maestroFrequency hourly
  oh_no <- function() {
    logger::log_error('this is bad', namespace = 'oh_no')
  }",
  con = "pipelines/logs.R"
)


## -----------------------------------------------------------------------------
#| eval: false
# #' pipelines/logs.R
# #' @maestroFrequency hourly
# hello <- function() {
#   logger::log_info("hi", namespace = "hello")
# }
# 
# #' @maestroFrequency hourly
# uhh <- function() {
#   logger::log_warn("this could be a problem", namespace = "uhh")
# }
# 
# #' @maestroFrequency hourly
# oh_no <- function() {
#   logger::log_error("this is bad", namespace = "oh_no")
# }


## # orchestrator.R
## schedule <- build_schedule(quiet = TRUE)
## 
## status <- run_schedule(
##   schedule,
##   run_all = TRUE,
##   log_to_console = TRUE
## )

## ----cleanup------------------------------------------------------------------
#| echo: false
#| output: false
unlink("pipelines", recursive = TRUE)
file.remove("maestro.log")

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")

## ----eval=FALSE---------------------------------------------------------------
# #' ./pipelines/conditional1.R
# #' @maestroFrequency 1 day
# #' @maestroRunIf sample(c(TRUE, FALSE), size = 1)
# random_execution <- function() {
#   message("Maybe, maybe not")
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(unlink("pipelines", recursive = TRUE))
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency 1 day
  #' @maestroRunIf TRUE
  random_execution <- function() {
    message('Maybe, maybe not')
  }",
  con = "pipelines/conditional1.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' ./pipelines/conditional2.R
# #' @maestroFrequency 1 hour
# extract_flights <- function() {
# 
#   # Pretends to fetch data from an API
#   data.frame(
#     flight_id = 1:5,
#     airline_code = c("UA", "AC", "AC", "AA", "DE"),
#     departure_time = as.POSIXct("2025-10-14 12:00:00") + c(100, 450, 750, 1450, 1750)
#   )
# }
# 
# #' @maestroInputs extract_flights
# #' @maestroRunIf
# #' is.data.frame(.input) && nrow(.input) > 0
# transform_flights <- function(.input) {
# 
#   proc_data <- .input
# 
#   proc_data <- proc_data |>
#     dplyr::filter(
#       flight_id > 5
#     ) |>
#     dplyr::mutate(
#       departing_from = "YHZ"
#     )
# 
#   proc_data
# }
# 
# #' @maestroInputs transform_flights
# #' @maestroRunIf
# #' is.data.frame(.input) && nrow(.input) > 0
# load_flights <- function(.input) {
# 
#   write.csv("flights.csv")
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(unlink("pipelines", recursive = TRUE))
dir.create("pipelines")
writeLines(
  "
  #' ./pipelines/conditional1.R
  #' @maestroFrequency 1 hour
  extract_flights <- function() {
    
    # Pretends to fetch data from an API
    data.frame(
      flight_id = 1:5,
      airline_code = c('UA', 'AC', 'AC', 'AA', 'DE'),
      departure_time = as.POSIXct('2025-10-14 12:00:00') + c(100, 450, 750, 1450, 1750)
    )
  }

  #' @maestroInputs extract_flights
  #' @maestroRunIf 
  #' is.data.frame(.input) && nrow(.input) > 0
  transform_flights <- function(.input) {
    
    proc_data <- .input

    proc_data <- proc_data |> 
      dplyr::filter(
        flight_id > 5
      ) |> 
      dplyr::mutate(
        departing_from = 'YHZ'
      )

    proc_data
  }

  #' @maestroInputs transform_flights
  #' @maestroRunIf 
  #' is.data.frame(.input) && nrow(.input) > 0
  load_flights <- function(.input) {

    write.csv('flights.csv')
  }
  ",
  con = "pipelines/conditional2.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' ./pipelines/conditional3.R
# #' @maestroFrequency 1 day
# #' @maestroRunIf prod
# process_payments <- function() {
# 
#   # Some code that does important payment processing stuff
#   # but only in production!
# 
#   message("Payments processed")
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(unlink("pipelines", recursive = TRUE))
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency 1 day
  #' @maestroRunIf prod
  process_payments <- function() {
    
    # Some code that does important payment processing stuff 
    # but only in production!

    message('Payments processed')
  }",
  con = "pipelines/conditional3.R"
)

## ----cleanup, echo=FALSE, message=FALSE, warning=FALSE------------------------
invisible(unlink("pipelines", recursive = TRUE))

## -----------------------------------------------------------------------------
#| eval: false
# library(cronR)
# library(here) # makes it easier to deal with path issues
# 
# cmd <- cron_rscript(
#   here("orchestrator.R"),
#   workdir = here()
# )
# 
# cron_add(
#   cmd,
#   frequency = "daily",
#   id = "maestro",
#   ask = FALSE
# )


## -----------------------------------------------------------------------------
#| eval: false
# library(taskscheduleR)
# library(here) # makes it easier to deal with path issues
# 
# taskscheduler_create(
#   taskname = "maestro",
#   rscript = here("orchestrator.R"),
#   schedule = "DAILY",
#   exec_path = here(),
#   startdate = format(Sys.Date() - 1, "%m/%d/%Y")
# )

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-01 09:00:00
#' @maestroTz UTC
#'
#'
#' @export
get_mtcars2 <- function() {
  mtcars
}

## -----------------------------------------------------------------------------
#| echo: false
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")


## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
dir.create("pipelines")
writeLines(
  "
  # ./pipelines/daily_example.R
  #' daily_example maestro pipeline
  #'
  #' @maestroFrequency 1 day
  #' @maestroStartTime 09:20:00
  daily_example <- function() {
  
    # Pipeline code
  }
  ",
  con = "pipelines/daily_example.R"
)


## ----eval=FALSE---------------------------------------------------------------
# # ./pipelines/daily_example.R
# #' daily_example maestro pipeline
# #'
# #' @maestroFrequency 1 day
# #' @maestroStartTime 09:20:00
# daily_example <- function() {
# 
#   # Pipeline code
# }


## # ./orchestrator.R
## 
## library(maestro)
## 
## schedule <- build_schedule()
## 
## status <- run_schedule(
##   schedule,
##   orch_frequency = "1 day",
##   check_datetime = as.POSIXct("2024-06-20 08:00:00", tz = "UTC")
## )

## # ./orchestrator.R
## status <- run_schedule(
##   schedule,
##   orch_frequency = "15 minutes",
##   check_datetime = as.POSIXct("2024-06-20 08:00:00", tz = "UTC")
## )

## ----echo=FALSE---------------------------------------------------------------
data.frame(
  name = paste0("pipe", 1:3),
  frequency = c("1 hour", "2 days", "4 months"),
  start_time = as.POSIXct(c("2024-06-18 12:30:00", "2024-06-18 06:00:00", "2024-06-20 00:00:00"))
)


## ----echo=FALSE---------------------------------------------------------------
data.frame(
  name = paste0("pipe", 4:6),
  frequency = c("1 hour", "1 hour", "1 hour"),
  start_time = as.POSIXct(c("2024-06-18 00:00:00", "2024-06-18 00:10:00", "2024-06-18 00:20:00"))
)


## ----eval=FALSE---------------------------------------------------------------
# #' specific_hours maestro pipeline
# #'
# #' @maestroFrequency hourly
# #' @maestroHours 6 9 12 15 18
# some_hours_example <- function() {
# 
#   # Pipeline code
# }


## ----eval=FALSE---------------------------------------------------------------
# #' specific_days_of_week maestro pipeline
# #'
# #' @maestroFrequency daily
# #' @maestroDays Mon Tue Wed Thu Fri
# some_dow_example <- function() {
# 
#   # Pipeline code
# }
# 
# #' specific_days_of_week maestro pipeline
# #'
# #' @maestroFrequency daily
# #' @maestroDays 1 14 28
# some_dom_example <- function() {
# 
#   # Pipeline code
# }


## ----eval=FALSE---------------------------------------------------------------
# #' specific_months maestro pipeline
# #'
# #' @maestroFrequency months
# #' @maestroMonths 1 7 10
# some_months <- function() {
# 
#   # Pipeline code
# }


## ----eval=FALSE---------------------------------------------------------------
# #' business_hours maestro pipeline
# #'
# #' @maestroFrequency hourly
# #' @maestroHours 9 10 11 12 13 14 15 16 17
# #' @maestroDays Mon Tue Wed Thu Fri
# during_work_hours_example <- function() {
# 
#   # Pipeline code
# }


## ----cleanup, echo=FALSE, message=FALSE, warning=FALSE------------------------
unlink("pipelines", recursive = TRUE)

#' @maestroFrequency daily
#' @maestroDays 1 8 24 28
specific_days <- function() {

}

#' @maestroFrequency hourly
#' @maestroHours 0 4 8 12 16 20
specific_hours <- function() {

}

#' @maestroFrequency biweekly
#' @maestroMonths 1 5 10
specific_months1 <- function() {

}

#' @maestroFrequency hourly
#' @maestroDays Mon Wed Fri
specific_days2 <- function() {

}

#' @maestroFrequency hourly
#' @maestroDays Sun Sat
specific_days3 <- function() {

}

#' @maestroFrequency monthly
#' @maestroMonths 1 5 10
specific_months3 <- function() {

}

#' @maestroFrequency hourly
#' @maestroHours 1 12
#' @maestroDays 1
#' @maestroMonths 1 5 10
specific_multi <- function() {

}

#' Create a new orchestrator
#'
#' @param open whether or not to open the script upon creation
#' @param path file path for the orchestrator script
#' @param type file type for the orchestrator (supports R, Quarto, and RMarkdown)
#' @param quiet whether to silence messages in the console (default = `FALSE`)
#' @param overwrite whether to overwrite an existing orchestrator or maestro project
#'
#' @return invisible
create_orchestrator <- function(
    path,
    type = c("R", "Quarto", "RMarkdown"),
    open = interactive(),
    quiet = FALSE,
    overwrite = FALSE
) {

  type <- match.arg(type, choices = c("R", "Quarto", "RMarkdown"))

  template <- ifelse(type == "R", "orchestrator_template", "orchestrator_template_qmd")

  script <- readLines(system.file(template, package = "maestro")) |>
    paste(collapse = "\n") |>
    glue::glue(
      .open = "{{",
      .close = "}}",
      .null = NULL
    )

  extension <- switch (type,
    R = "R",
    Quarto = "qmd",
    RMarkdown = "Rmd"
  )

  path <- paste0(path, ".", extension)

  if (file.exists(path)) {
    if (!overwrite) {
      cli::cli_abort(
        c("File {.file {path}} already exists.",
          "Set {.code maestro::create_orchestrator(overwrite = TRUE)} to overwrite anyway."),
        call = NULL
      )
    } else {
      if (!quiet) cli::cli_alert_warning("Overwriting existing orchestrator at {.file {path}}.")
    }
  }

  writeLines(
    script,
    path
  )

  if (open) {
    rlang::check_installed("rstudioapi")
    rstudioapi::documentOpen(path)
  }

  if(!quiet) {
    cli::cli_alert_success("Created orchestrator at {.file {path}}")
  }
}

test_that("create_orchestrator creates a new orchestrator", {
  withr::with_tempdir({
    create_orchestrator("orchestrator" ,open = FALSE)
    expect_true(file.exists("orchestrator.R"))
  }) |>
    expect_message()
})

test_that("create_orchestrator aborts if file already exists", {

  withr::with_tempdir({
    expect_error({
      create_orchestrator(".", open = FALSE)
      create_orchestrator(".", open = FALSE)
    }, regexp = "already exists.")

    # Works if overwrite = TRUE
    expect_message(
      create_orchestrator(".", open = FALSE, overwrite = TRUE),
      regexp = "Overwriting existing"
    )
  })
}) |>
  suppressMessages()

# maestroFrequency ----------------------------------------------------------
#' @importFrom roxygen2 roclet_output roclet_process roxy_tag_parse
#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroFrequency <- function(x) {

  # Cast lower and trim whitespace
  x$raw <- x$raw |>
    tolower() |>
    trimws()

  if (x$raw == "") {
    x$val <- "1 day"

  } else if (grepl("^[0-9]", x$raw)) {

    tryCatch({
      # Try to parse it using our function
      parse_rounding_unit(x$raw)

      # Try to parse using timechange
      timechange::time_round(Sys.time(), unit = x$raw)

      x$val <- x$raw
    }, error = \(e) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroFrequency `{x$raw}`.
          Must have a format of [number] [units] (e.g., 1 day, 1 week, 4 months, etc)
          or one of hourly, daily, weekly, etc."
        )
      )
      return()
    })

  } else {

    if (!x$raw %in% c("hourly", "daily", "weekly", "biweekly", "monthly", "quarterly", "yearly")) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroFrequency `{x$raw}`.
          Must have a format of [number] [units] (e.g., 1 day, 1 week, 4 months, etc)
          or one of hourly, daily, weekly, etc."
        )
      )
      return()
    }

    x$val <- x$raw
  }

  x
}

maestroFrequency_roclet <- function() {
  roxygen2::roclet("maestroFrequency")
}

#' @exportS3Method
roclet_process.roclet_maestroFrequency <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroFrequency")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroStartTime ----------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroStartTime <- function(x) {

  x$raw <- x$raw |>
    trimws()

  .weekday_abbrs <- c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
  weekday_pattern <- sprintf(
    "^(%s)\\s+[0-9]{2}:[0-9]{2}:[0-9]{2}$",
    paste(.weekday_abbrs, collapse = "|")
  )
  monthday_pattern <- "^[0-9]{1,2}(\\s+[0-9]{2}:[0-9]{2}:[0-9]{2})?$"

  tryCatch({

    if (x$raw == "") {
      x$val <- "1970-01-01 00:00:00"
    } else if (grepl("^[0-9]{2}:[0-9]{2}:[0-9]{2}$", x$raw)) {
      as.POSIXct(x$raw, format = "%H:%M:%S")  # validate format
      x$val <- x$raw
    } else if (grepl(weekday_pattern, x$raw)) {
      x$val <- x$raw
    } else if (grepl(monthday_pattern, x$raw)) {
      mday <- as.integer(strsplit(x$raw, "\\s+")[[1]][1])
      if (is.na(mday) || mday < 1L || mday > 31L) {
        roxygen2::roxy_tag_warning(
          x,
          glue::glue("Invalid maestroStartTime `{x$raw}`. Month-day must be an integer in [1-31].")
        )
        return()
      }
      x$val <- x$raw
    } else {
      x_ts <- as.POSIXct(x$raw, format = "%Y-%m-%d %H:%M:%S")
      if (is.na(x_ts)) x_ts <- as.POSIXct(x$raw, format = "%Y-%m-%d")
      if (is.na(x_ts)) stop("could not parse")
      x$val <- strftime(x_ts, format = "%Y-%m-%d %H:%M:%S")
    }

  }, error = \(e) {
    roxygen2::roxy_tag_warning(
      x,
      glue::glue(
        "Invalid maestroStartTime `{x$raw}`.
        Must be a datetime (yyyy-mm-dd HH:MM:SS), HH:MM:SS, Mon HH:MM:SS, or DD HH:MM:SS."
      )
    )
    return()
  })


  x
}

maestroStartTime_roclet <- function() {
  roxygen2::roclet("maestroStartTime")
}

#' @exportS3Method
roclet_process.roclet_maestroStartTime <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroStartTime")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroTz -----------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroTz <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw == "") {
    x$val <- "UTC"
  } else {
    if(!x$raw %in% OlsonNames()) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroTz `{x$raw}`.
          Must be a valid timezone string. See valid tzs with `OlsonNames()`"
        )
      )
      return()
    }
    x$val <- x$raw
  }
  x
}

maestroTz_roclet <- function() {
  roxygen2::roclet("maestroTz")
}

#' @exportS3Method
roclet_process.roclet_maestroTz <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroTz")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroSkip ---------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroSkip <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw != "") {
    roxygen2::roxy_tag_warning(
      x,
      "Invalid maestroSkip. Use tag name with no value to indicate pipeline should be skipped (e.g., `#' @maestroSkip`)."
    )
  }

  x$val <- TRUE
  x
}

maestroSkip_roclet <- function() {
  roxygen2::roclet("maestroSkip")
}

#' @exportS3Method
roclet_process.roclet_maestroSkip <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroSkip")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroLogLevel ----------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroLogLevel <- function(x) {

  log_levels <- c(
    "INFO", "WARN", "ERROR", "OFF", "FATAL", "SUCCESS", "DEBUG", "TRACE"
  )

  x$raw <- x$raw |>
    toupper() |>
    trimws()

  if (x$raw == "") {
    x$val <- "INFO"
  } else if (x$raw %in% log_levels) {
    x$val <- x$raw
  } else {
    roxygen2::roxy_tag_warning(
      x,
      glue::glue(
        "Invalid maestroLogLevel `{x$raw}`.
        Must be one of {paste(log_levels, collapse = ', ')}"
      )
    )
    return()
  }

  x
}

maestroLogLevel_roclet <- function() {
  roxygen2::roclet("maestroLogLevel")
}

#' @exportS3Method
roclet_process.roclet_maestroLogLevel <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroLogLevel")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroHours --------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroHours <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw != "") {
    x_sep <- tryCatch({
      as.numeric(strsplit(x$raw, "\\s+")[[1]])
    }, warning = function(w) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroHours `{x$raw}`.
        Must be one or more integers [0-23] separated by spaces (e.g., 1 4 7)"
        )
      )
      return()
    })

    # Check for values [0-23]
    if (!all(x_sep %in% 0:23)) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroHours `{x$raw}`.
        Must be one or more integers [0-23] separated by spaces (e.g., 1 4 7)"
        )
      )
      return()
    }

    x$val <- x_sep
  }

  x
}

maestroHours_roclet <- function() {
  roxygen2::roclet("maestroHours")
}

#' @exportS3Method
roclet_process.roclet_maestroHours <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroHours")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroDays --------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroDays <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw != "") {

    x_sep <- strsplit(x$raw, "\\s+")[[1]]

    if (all(grepl("[0-9]", x_sep))) {

      x_sep <- as.numeric(x_sep)

      # Check for values [1-31]
      if (!all(x_sep %in% 1:31)) {
        roxygen2::roxy_tag_warning(
          x,
          glue::glue(
            "Invalid maestroDays `{x$raw}`.
            All must be either integers [1-31] or values Mon, Tue, etc. separated by spaces"
          )
        )
        return()
      }

    } else if (all(x_sep %in% c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"))) {

      x_sep <- factor(x_sep, levels = c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"))

    } else {

      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroDays `{x$raw}`.
            All must be either integers [1-31] or values Mon, Tue, etc. separated by spaces"
        )
      )
      return()
    }
  }

  x$val <- x_sep

  x
}

maestroDays_roclet <- function() {
  roxygen2::roclet("maestroDays")
}

#' @exportS3Method
roclet_process.roclet_maestroDays <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroDays")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroMonths -----------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroMonths <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw != "") {
    x_sep <- tryCatch({
      as.numeric(strsplit(x$raw, "\\s+")[[1]])
    }, warning = function(w) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroMonths `{x$raw}`.
        Must be one or more integers [1-12] separated by spaces (e.g., 1 6 12)"
        )
      )
      return()
    })

    # Check for values [1-12]
    if (!all(x_sep %in% 1:12)) {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroMonths `{x$raw}`.
        Must be one or more integers [1-12] separated by spaces (e.g., 1 6 12)"
        )
      )
      return()
    }

    x$val <- x_sep
  }

  x
}

maestroMonths_roclet <- function() {
  roxygen2::roclet("maestroMonths")
}

#' @exportS3Method
roclet_process.roclet_maestroMonths <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroMonths")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroInputs -----------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroInputs <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw == "") {
    roxygen2::roxy_tag_warning(
      x,
      glue::glue(
        "Empty maestroInputs. If pipeline doesn't take inputs, remove maestroInputs tag."
      )
    )
    return(x)
  }

  collect_match <- regmatches(x$raw, regexpr("^collect\\(([^)]+)\\)$", x$raw))

  if (length(collect_match) == 1L) {
    inner <- trimws(regmatches(collect_match, regexpr("(?<=\\()([^)]+)(?=\\))", collect_match, perl = TRUE)))
    pipe_names <- strsplit(inner, "[[:space:],]+")[[1]]
    pipe_names <- pipe_names[nzchar(pipe_names)]
    x$val <- list(inputs = pipe_names, is_collect = TRUE)
  } else {
    # Plain list of pipeline names (existing behaviour)
    x$val <- list(
      inputs = strsplit(x$raw, "\\s+")[[1]],
      is_collect = FALSE
    )
  }

  x
}

maestroInputs_roclet <- function() {
  roxygen2::roclet("maestroInputs")
}

#' @exportS3Method
roclet_process.roclet_maestroInputs <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroInputs")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroMap ------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroMap <- function(x) {
  # Split on whitespace to support pmap-style multiple iterators.
  # Each element is an R expression string referencing `.input`.
  # A single expression is stored as a length-1 character vector,
  # keeping this fully backward-compatible with the old single-iterator form.
  parts <- strsplit(trimws(x$raw), "\\s+")[[1]]
  parts <- parts[nzchar(parts)]
  if (length(parts) == 0) {
    parts <- ".input"
  }
  x$val <- parts
  x
}

#' @exportS3Method
roclet_process.roclet_maestroMap <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroMap")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}

# maestroOutputs ----------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroOutputs <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw == "") {
    roxygen2::roxy_tag_warning(
      x,
      glue::glue(
        "Empty maestroOutputs. If pipeline doesn't take inputs, remove maestroOutputs tag."
      )
    )
    return(x)
  }

  x$val <- strsplit(x$raw, "\\s+")[[1]]

  x
}

maestroOutputs_roclet <- function() {
  roxygen2::roclet("maestroOutputs")
}

#' @exportS3Method
roclet_process.roclet_maestroOutputs <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroOutputs")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}


# maestro -----------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestro <- function(x) {
  x$val <- TRUE
  x
}

maestro_roclet <- function() {
  roxygen2::roclet("maestro")
}

#' @exportS3Method
roclet_process.roclet_maestro <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestro")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}


# maestroPriority ---------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroPriority <- function(x) {

  x$raw <- x$raw |>
    trimws()

  if (x$raw == "") {
    roxygen2::roxy_tag_warning(
      x,
      "Empty maestroPriority. If pipeline doesn't have a specified priority remove maestroPriority tag."
    )
    return(x)
  } else {
    if (!grepl("^\\d+$", x$raw) || x$raw == "0") {
      roxygen2::roxy_tag_warning(
        x,
        glue::glue(
          "Invalid maestroPriority `{x$raw}`.
          Must be a single positive integer from 1 on."
        )
      )
      return(x)
    }
    x$val <- x$raw
  }

  x
}

maestroPriority_roclet <- function() {
  roxygen2::roclet("maestroPriority")
}

#' @exportS3Method
roclet_process.roclet_maestroPriority <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroPriority")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}


# maestroFlags --------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroFlags <- function(x) {

  x$raw <- x$raw |>
    trimws()

  x_sep <- strsplit(x$raw, "\\s+")[[1]]
  x$val <- x_sep
  x
}

maestroFlags_roclet <- function() {
  roxygen2::roclet("maestroFlags")
}

#' @exportS3Method
roclet_process.roclet_maestroFlags <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tag(blocks[[1]], "maestroFlags")
  list(
    val = tags$val,
    node = blocks[[1]]$object$topic
  )
}


# maestroLabel ------------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroLabel <- function(x) {

  x$raw <- x$raw |>
    trimws()

  x_sep <- strsplit(x$raw, "\\s+")[[1]]
  if (length(x_sep) != 2) {
    roxygen2::roxy_tag_warning(
      x,
      "Invalid maestroLabel. Should follow the format `#' @maestroLabel key value`."
    )
    return(x)
  }
  x$val <- x_sep
  x
}

maestroLabel_roclet <- function() {
  roxygen2::roclet("maestroLabel")
}

#' @exportS3Method
roclet_process.roclet_maestroLabel <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tags(blocks[[1]], "maestroLabel")
  list(
    val = purrr::map(tags, ~{
      .x$val
    }),
    node = blocks[[1]]$object$topic
  )
}


# maestroRunIf -----------------------------------------------------------

#' @exportS3Method
roxy_tag_parse.roxy_tag_maestroRunIf <- function(x) {
  x$val <- x$raw
  x
}

maestroRunIf_roclet <- function() {
  roxygen2::roclet("maestroRunIf")
}

#' @exportS3Method
roclet_process.roclet_maestroRunIf <- function(x, blocks, env, base_path) {
  tags <- roxygen2::block_get_tags(blocks[[1]], "maestroRunIf")
  list(
    val = purrr::map(tags, ~{
      .x$val
    }),
    node = blocks[[1]]$object$topic
  )
}

test_that("parse maestroFrequency tag works with '1 day'", {
  res <- roxygen2::roc_proc_text(
    maestroFrequency_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_good.R"))
  ) |>
    expect_no_message()
  expect_type(res$val, "character")
})

test_that("parse maestroFrequency tag works with 'daily'", {
  res <- roxygen2::roc_proc_text(
    maestroFrequency_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_single_good.R"))
  ) |>
    expect_no_message()
  expect_type(res$val, "character")
})

test_that("maestroFrequency default value is expected", {
  res <- roxygen2::roc_proc_text(
    maestroFrequency_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_default.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
  expect_equal(res$val, "1 day")
})

test_that("bad usage of maestroFrequency warns and gives no val", {
  res <- roxygen2::roc_proc_text(
    maestroFrequency_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_bad.R"))
  ) |>
    expect_warning(regexp = "Must have a format")

  expect_null(res$val)
})

test_that("bad usage of maestroFrequency warns and gives no val", {
  res <- roxygen2::roc_proc_text(
    maestroFrequency_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_single_bad.R"))
  ) |>
    expect_warning(regexp = "Must have a format")

  expect_null(res$val)
})

test_that("maestroStartTime default value is returned", {
  res <- roxygen2::roc_proc_text(
    maestroStartTime_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_default.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
  expect_s3_class(as.POSIXct(res$val), "POSIXct")
})

test_that("integer maestroStartTime fails", {
  res <- roxygen2::roc_proc_text(
    maestroStartTime_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_start_time_int.R"))
  ) |>
    expect_warning(regexp = "Invalid maestroStartTime")

  expect_null(res$val)
})

test_that("partial maestroStartTime works", {
  res <- roxygen2::roc_proc_text(
    maestroStartTime_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_start_time_date.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
  expect_s3_class(as.POSIXct(res$val), "POSIXct")
})

test_that("maestroStartTime accepts weekday format", {
  res <- roxygen2::roc_proc_text(
    maestroStartTime_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_weekly_weekday.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
  expect_equal(res$val, "Mon 04:00:00")
})

test_that("maestroStartTime accepts month-day format", {
  res <- roxygen2::roc_proc_text(
    maestroStartTime_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_monthly_monthday.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
  expect_equal(res$val, "15 04:00:00")
})

test_that("build_schedule_entry errors on weekday format with non-weekly frequency", {
  expect_error(
    build_schedule_entry(
      test_path("test_pipelines/test_pipeline_weekday_wrong_freq.R")
    ),
    regexp = "weekday format"
  )
})

test_that("build_schedule_entry errors on month-day format with non-monthly frequency", {
  expect_error(
    build_schedule_entry(
      test_path("test_pipelines/test_pipeline_monthday_wrong_freq.R")
    ),
    regexp = "month-day format"
  )
})

test_that("build_schedule_entry stores weekday start_time as raw string", {
  res <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_weekly_weekday.R")
  )
  pipeline <- res$MaestroPipelines[[1]]
  expect_match(pipeline$.__enclos_env__$private$start_time_raw, "^Mon ")
  # resolve_start_time should produce a Monday
  resolved <- pipeline$.__enclos_env__$private$resolve_start_time(lubridate::now())
  expect_s3_class(resolved, "POSIXct")
  expect_equal(lubridate::wday(resolved, week_start = 1), 1L)
})

test_that("build_schedule_entry stores month-day start_time as raw string", {
  res <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_monthly_monthday.R")
  )
  pipeline <- res$MaestroPipelines[[1]]
  expect_match(pipeline$.__enclos_env__$private$start_time_raw, "^15")
  # resolve_start_time should produce the 15th of the month
  resolved <- pipeline$.__enclos_env__$private$resolve_start_time(lubridate::now())
  expect_s3_class(resolved, "POSIXct")
  expect_equal(lubridate::mday(resolved), 15L)
})

test_that("parse maestroTz tag works", {
  res <- roxygen2::roc_proc_text(
    maestroTz_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_good.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
})

test_that("maestroTz default returns value", {
  res <- roxygen2::roc_proc_text(
    maestroStartTime_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_default.R"))
  ) |>
    expect_no_message()

  expect_type(res$val, "character")
})

test_that("bad usage of maestroTz warns and returns null val", {
  res <- roxygen2::roc_proc_text(
    maestroTz_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_tz_bad.R"))
  ) |>
    expect_warning(regexp = "Must be a valid timezone")

  expect_null(res$val)
})

test_that("parse maestroSkip works", {
  res <- roxygen2::roc_proc_text(
    maestroSkip_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_skip.R"))
  )
  expect_true(res$val)
})

test_that("Nonexistent maestroSkip is NULL", {
  res <- roxygen2::roc_proc_text(
    maestroSkip_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_daily_good.R"))
  )
  expect_null(res$val)
})

test_that("Invalid usage of maestroSkip warns but still returns a value of TRUE", {

  expect_warning({
    res <- roxygen2::roc_proc_text(
      maestroSkip_roclet(),
      readLines(test_path("test_pipelines/test_pipeline_skip_bad.R"))
    )
  })

  expect_true(res$val)
})

test_that("parse maestroLogLevel works", {
  res <- roxygen2::roc_proc_text(
    maestroLogLevel_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_loglevel_good.R"))
  )

  expect_type(res$val, "character")
})

test_that("invalid maestroLogLevel warns", {
  res <- roxygen2::roc_proc_text(
    maestroLogLevel_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_loglevel_bad.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("parse maestroHours works", {
  res <- roxygen2::roc_proc_text(
    maestroHours_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_hours_good.R"))
  )

  expect_type(res$val, "double")
})

test_that("invalid maestroHours warns", {
  res <- roxygen2::roc_proc_text(
    maestroHours_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_hours_bad.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("invalid maestroHours warns", {
  res <- roxygen2::roc_proc_text(
    maestroHours_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_hours_bad2.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("parse maestroDays works for days of month", {
  res <- roxygen2::roc_proc_text(
    maestroDays_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_days_good.R"))
  )

  expect_type(res$val, "double")
})

test_that("parse maestroDays works for days of week", {
  res <- roxygen2::roc_proc_text(
    maestroDays_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_days_good2.R"))
  )

  expect_type(res$val, "integer")
})

test_that("invalid maestroDays warns for invalid days of month", {
  res <- roxygen2::roc_proc_text(
    maestroDays_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_days_bad2.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("invalid maestroDays warns for invalid days of month", {
  res <- roxygen2::roc_proc_text(
    maestroDays_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_days_bad3.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("invalid maestroDays warns for invalid days of week", {
  res <- roxygen2::roc_proc_text(
    maestroDays_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_days_bad2.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("parse maestroMonths works", {
  res <- roxygen2::roc_proc_text(
    maestroMonths_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_months_good.R"))
  )

  expect_type(res$val, "double")
})

test_that("invalid maestroMonths warns", {
  res <- roxygen2::roc_proc_text(
    maestroMonths_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_months_bad.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("invalid maestroMonths warns", {
  res <- roxygen2::roc_proc_text(
    maestroMonths_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_months_bad2.R"))
  ) |>
    expect_warning(regexp = "Invalid")

  expect_null(res$val)
})

test_that("parse maestroInputs works", {
  res <- roxygen2::roc_proc_text(
    maestroInputs_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_inputs_good.R"))
  )

  expect_type(res$val$inputs, "character")
  expect_length(res$val$inputs, 3L)
})

test_that("parse maestroInputs warns if empty", {
  res <- roxygen2::roc_proc_text(
    maestroInputs_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_inputs_bad.R"))
  ) |>
    expect_warning()

  expect_null(res$val)
})

test_that("parse maestroOutputs works", {
  res <- roxygen2::roc_proc_text(
    maestroOutputs_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_outputs_good.R"))
  )

  expect_type(res$val, "character")
  expect_length(res$val, 3L)
})

test_that("parse maestroOutputs warns if empty", {
  res <- roxygen2::roc_proc_text(
    maestroOutputs_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_outputs_bad.R"))
  ) |>
    expect_warning()

  expect_null(res$val)
})

test_that("parse maestro works", {
  res <- roxygen2::roc_proc_text(
    maestro_roclet(),
    readLines(test_path("test_pipelines/test_pipeline_maestro.R"))
  )
  expect_true(res$val)
})

test_that("maestroStartTime formatted as HH:MM:SS is valid", {
  withr::with_tempdir({
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroStartTime 10:00:00
      hhmmss <- function() {

      }
      ",
      con = "hhmmss.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroStartTime_roclet(),
      readLines("hhmmss.R")
    )
  })

  expect_type(res$val, "character")
  time <- as.POSIXct(res$val, format = "%H:%M:%S")
  expect_s3_class(time, "POSIXct")
  expect_equal(lubridate::year(time), lubridate::year(lubridate::today()))
})

test_that("maestroPriority works", {
  withr::with_tempdir({
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroPriority 1
      priority <- function() {

      }
      ",
      con = "priority.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroPriority_roclet(),
      readLines("priority.R")
    )
  })
  expect_equal(res$val, "1")

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroPriority
      priority <- function() {

      }
      ",
      con = "priority.R"
    )

    expect_warning({
      res <- roxygen2::roc_proc_text(
        maestroPriority_roclet(),
        readLines("priority.R")
      )
    }, regexp = "Empty maestroPriority")

    expect_null(res$val)
  })

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroPriority 1.5
      priority <- function() {

      }
      ",
      con = "priority.R"
    )

    expect_warning({
      res <- roxygen2::roc_proc_text(
        maestroPriority_roclet(),
        readLines("priority.R")
      )
    }, regexp = "Invalid maestroPriority")
  })
})

test_that("parse maestroFlags works", {

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroFlags critical
      tagged <- function() {

      }
      ",
      con = "tag.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroFlags_roclet(),
      readLines("tag.R")
    )

    expect_equal(res$val, "critical")
  })

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroFlags critical awesome
      tagged <- function() {

      }
      ",
      con = "tag.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroFlags_roclet(),
      readLines("tag.R")
    )

    expect_equal(length(res$val), 2)
  })
})

test_that("parse maestroLabel works", {

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroLabel severity critical
      labelled <- function() {

      }
      ",
      con = "label.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroLabel_roclet(),
      readLines("label.R")
    )

    expect_equal(res$val[[1]], c("severity", "critical"))
  })

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroLabel a lot of labels
      labelled <- function() {

      }
      ",
      con = "label.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroLabel_roclet(),
      readLines("label.R")
    ) |>
      expect_warning()

    expect_null(res$val[[1]])
  })

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroLabel severity critical
      #' @maestroLabel domain transportation
      labelled <- function() {

      }
      ",
      con = "label.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroLabel_roclet(),
      readLines("label.R")
    )

    expect_equal(length(res$val), 2)
  })
})

test_that("parse maestroRunIf works by creating evaluable R code", {

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroRunIf 
      #' # some logic that will evaluate to TRUE/FALSE
      #' x <- TRUE
      #' x
      conditional <- function() {

      }
      ",
      con = "conditional.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroRunIf_roclet(),
      readLines("conditional.R")
    )

    expect_true(eval_code_str(res$val[[1]]))
  })

  withr::with_tempdir({
    writeLines(
      "
      #' @maestroRunIf 
      #' test_fun <- function(x = 1) {
      #'   if(x == 1) {
      #'     return(TRUE)
      #'   } else {
      #'     return(FALSE)
      #'   }
      #' }
      #' test_fun()
      conditional <- function() {

      }
      ",
      con = "conditional.R"
    )

    res <- roxygen2::roc_proc_text(
      maestroRunIf_roclet(),
      readLines("conditional.R")
    )

    expect_true(eval_code_str(res$val[[1]]))
  })
})

#' Manually run a pipeline regardless of schedule
#'
#' Instantly run a single pipeline from the schedule. This is useful for testing
#' purposes or if you want to just run something one-off.
#'
#' Scheduling parameters such as the frequency, start time, and specifiers are ignored.
#' The pipeline will be run even if `maestroSkip` is present. If the pipeline is a DAG
#' pipeline, `invoke` will attempt to execute the full DAG.
#'
#' @inheritParams run_schedule
#' @param pipe_name name of a single pipe name from the schedule
#' @param ... other arguments passed to `run_schedule()`
#' @param quiet silence metrics to the console (default = `FALSE`). Note this does not affect messages generated from pipelines when `log_to_console = TRUE`.
#' @param log_to_console whether or not to include pipeline messages, warnings, errors to the console (default = `FALSE`) (see Logging & Console Output section)
#'
#' @return invisible
#' @export
#'
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   invoke(schedule, "my_new_pipeline")
#' }
invoke <- function(schedule, pipe_name, resources = list(), ..., quiet = TRUE, log_to_console = FALSE) {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  pipe_names <- schedule$PipelineList$get_pipe_names()

  if (!rlang::is_scalar_character(pipe_name)) {
    cli::cli_abort(
      c("`pipe_name` must be a single character string referencing the name of a pipeline in the schedule.",
        "i" = "Available pipe_names are {.pkg {sort(pipe_names)}}"),
      call = rlang::caller_env()
    )
  }

  # Make sure pipe name is in the schedule
  if (!pipe_name %in% pipe_names) {
    cli::cli_abort(
      c("{.code {pipe_name}} is not the name of a pipeline in the schedule.",
        "i" = "Available pipe_names are {.pkg {sort(pipe_names)}}"),
      call = rlang::caller_env()
    )
  }

  # Ensure that elements in resources are named
  if (length(resources) > 0) {
    resources_length <- length(resources)
    n_named <- sum(names(resources) != "")
    if (resources_length > n_named) {
      cli::cli_abort(
        "All elements in `resources` must be named."
      )
    }

    n_uniq_names <- length(unique(names(resources)))
    if (resources_length > n_uniq_names) {
      cli::cli_abort(
        "All elements in `resources` must have unique names."
      )
    }
  }

  tryCatch({
    schedule$PipelineList$run(
      ..., 
      resources = resources, 
      pipes_to_run = pipe_name,
      quiet = quiet,
      log_to_console = log_to_console
    )
  }, error = function(e) {

    if (e$message == "unused argument (`NA` = NULL)") {
      cli::cli_abort(
        c(
          "Failed to invoke pipeline {.code {pipe_name}}",
          "i" = "Did you forget to pass pipeline arguments as `resources = list(name = val)`?"
        ),
        call = NULL
      )
    } else {
      cli::cli_abort(
        "Failed to invoke pipeline {.code {pipe_name}}",
        call = NULL
      )
    }
  })

  return(invisible())
}

schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)

test_that("invoke errors if schedule is not a MaestroSchedule", {
  expect_error({
    invoke(iris, pipe_name = "hello")
  }, regexp = "Schedule must be an object")
})

test_that("invoke errors if pipe_name is not a single character", {

  expect_error({
    invoke(schedule, pipe_name = letters)
  }, regexp = "must be a single character")
})

test_that("invoke errors if pipe_name is not in the schedule", {
  expect_error({
    invoke(schedule, pipe_name = "I don't exist")
  }, regexp = "is not the name of a pipeline in the schedule")
})

test_that("errors if resources are unnamed or non unique", {

  expect_error({
    invoke(schedule, "chatty", quiet = TRUE, resources = list(4))
  }, regexp = "All elements")

  expect_error({
    invoke(schedule, "chatty", quiet = TRUE, resources = list(a = 1, a = 2))
  }, regexp = "All elements")
})

test_that("invoke triggers the pipeline", {

  expect_no_error({
    invoke(schedule, "chatty", quiet = TRUE)
  })

  status <- schedule$get_status()
  expect_true(status$invoked[status$pipe_name == "chatty"])
})

test_that("invoke properly passes resources", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      times2 <- function(val) {
        val * 2
      }
      ",
      con = "pipelines/invoked.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    invoke(schedule, "times2", resources = list(val = 2), quiet = TRUE)

    expect_true(get_status(schedule)$success)
  })
})

test_that("invoke gives informative error message on failure", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      i_fail <- function() {
        stop()
      }
      ",
      con = "pipelines/invoked.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    invoke(schedule, "i_fail", quiet = TRUE)

    expect_snapshot(schedule$get_status()[, c("invoked", "success")])
  })
})

test_that("invoke triggers DAG pipelines", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      p1 <- function() {
        2
      }

      #' @maestroInputs p1
      p2 <- function(.input) {
        .input * 2
      }
      ",
      con = "pipelines/dag.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    invoke(schedule, "p1", quiet = TRUE)

    expect_snapshot(schedule$get_status()[, c("invoked", "success")])
  })
})

test_that("invoke only runs the pipeline selected", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      pipe1 <- function() {
        2
      }

      #' @maestroFrequency daily
      pipe2 <- function() {
        4
      }
      ",
      con = "pipelines/invoked.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    invoke(schedule, "pipe2", quiet = TRUE)

    expect_equal(sum(schedule$get_status()$invoked), 1)
  })
})

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroSkip
#'
#' @export
get_mtcars <- function() {
  mtcars
}

#' Maestro Tags
#'
#' maestro tags are roxygen2 comments for configuring the scheduling and execution
#' of pipelines.
#'
#' @md
#' @details
#'
#' maestro tags follow the format: `#' @@maestroTagName value`
#'
#' Some tags may not take a value.
#'
#' maestro tags must be written above the function that is to be included as a pipeline.
#' A typical pipeline with tags could look like this:
#'
#' ```
#' #' @@maestroFrequency 1 hour
#' #' @@maestroStartTime 12:30:00
#' #' @@maestroLogLevel WARN
#' my_pipeline <- function() {
#'
#'    # Pipeline code
#' }
#' ```
#' Below are descriptions of all the tags currently available in maestro along
#' with examples.
#'
#'
#' # maestroFrequency
#'
#' How often to run the pipeline. This tag takes a time unit indicating how long
#' to wait between subsequent runs of the pipeline. Acceptable values include
#' an integer value followed by one of minute(s), hour(s), day(s), week(s), month(s),
#' and year(s). Note that some combinations of integer + unit may be invalid.
#' Adverbs like 'hourly', 'daily', 'weekly', etc. are also valid.
#'
#' Default: `1 day`
#'
#' Examples:
#' - `#' @maestroFrequency 1 hour`
#' - `#' @maestroFrequency daily`
#' - `#' @maestroFrequency 2 weeks`
#' - `#' @maestroFrequency 3 months`
#'
#'
#' # maestroStartTime
#'
#' Timestamp, date, or time corresponding to the start time of the pipeline. This
#' also sets the cadence of the pipeline in some cases. For instance, if the start time
#' is `2025-03-18 03:00:00` and the frequency is daily, the pipeline will run at 03:00
#' every day. A value in the future prevents the pipeline from running until that time
#' has been reached.
#'
#' For pipelines with a frequency lower than daily, partial anchor formats are supported
#' to make it easier to express natural cycle points without choosing a specific date:
#'
#' - **`HH:MM:SS`** — time-of-day anchor for `minute`, `hour`, or `day` frequencies.
#'   e.g., `08:00:00` runs every day at 8am.
#' - **`Mon HH:MM:SS`** (weekday abbreviation + time) — anchor for `week` or multi-week
#'   frequencies. Resolved to that weekday of the current ISO week.
#'   e.g., `Mon 04:00:00` with `@maestroFrequency 1 week` runs every Monday at 4am.
#'   Valid abbreviations: `Mon`, `Tue`, `Wed`, `Thu`, `Fri`, `Sat`, `Sun`.
#' - **`DD HH:MM:SS`** or **`DD`** (month-day + optional time) — anchor for `month`
#'   frequencies. Resolved to that day of the current month.
#'   e.g., `15 04:00:00` with `@maestroFrequency 1 month` runs on the 15th of every month at 4am.
#'
#' Default: `2024-01-01 00:00:00`
#'
#' Examples:
#' - `#' @maestroStartTime 2025-02-05 12:00:00`
#' - `#' @maestroStartTime 2025-01-01`
#' - `#' @maestroStartTime 08:00:00`
#' - `#' @maestroStartTime Mon 04:00:00`
#' - `#' @maestroStartTime Wed 09:30:00`
#' - `#' @maestroStartTime 15 04:00:00`
#' - `#' @maestroStartTime 1`
#'
#'
#' # maestroTz
#'
#' Timezone in which the maestroStartTime is to be considered. Takes any valid timezone
#' string that can be found in `OlsonNames()`.
#'
#' Default: `UTC`
#'
#' Examples:
#' - `#' @maestroTz Europe/Paris`
#' - `#' @maestroTz America/Halifax`
#' - `#' @maestroTz US/Pacific`
#'
#'
#' # maestroLogLevel
#'
#' Minimum logging threshold for messages, warnings, and errors that come from the pipeline.
#' These levels correspond to those in `logger:::log_levels_supported` but most commonly
#' used are ERROR, WARN, INFO. For example, if you use `WARN` then any messages of lower
#' urgency (i.e., INFO) will be suppressed, but errors will be logged.
#'
#' Default: `INFO`
#'
#' Examples:
#' - `#' @maestroLogLevel ERROR`
#' - `#' @maestroLogLevel WARN`
#'
#'
#' # maestroHours
#'
#' Specific hours of the day in which to run the pipeline. This only applies for pipelines that run
#' at an hourly or minutely frequency. Acceptable values are integers from 0-23 separated
#' by spaces. If empty, pipeline runs at all hours. This tag uses the timezone specified by
#' `maestroTz` (will be UTC if empty).
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroHours 1 4 7 10`
#' - `#' @maestroHours 0 5 20`
#'
#'
#' # maestroDays
#'
#' Specific days of week or days of month on which to run the pipeline. This only applies
#' for pipelines that run at a minutely, hourly, or daily frequency. Acceptable values are
#' either integers from 1-31 or day of week strings like Mon, Tue, Wed, etc. These two
#' options cannot be used in combination.
#'
#' Default:
#'
#' Examples
#' - `#' @maestroDays 1 11 21 31`
#' - `#' @maestroDays Mon Tue Wed Thu Fri`
#'
#'
#' # maestroMonths
#'
#' Specific months of the year on which to run the pipeline. This only applies for
#' pipelines that do run at least monthly. Acceptable values are integers (1-12) corresponding
#' to the month of the year (e.g., 1 = January, 2 = February, etc.).
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroMonths 3 8 12`
#' - `#' @maestroMonths 10`
#'
#'
#' # maestroInputs
#'
#' For a DAG style pipeline, the names of pipelines that input into the pipeline.
#' These names must match the function name defining the inputting pipeline. Multiple
#' pipelines can be used as inputs and the input value is used in the target pipeline
#' via the required `.input` parameter. Note that this tag could be redundant if the
#' inputting pipeline uses `maestroOutputs`.
#'
#' To enable **fan-in (collect)**, wrap one or more input names with `collect()`.
#' The downstream pipeline receives a named list as `.input`, where each name
#' corresponds to an upstream pipeline and each value is that pipeline's return
#' value. All listed upstream pipelines must have succeeded before the collect
#' pipeline fires. Works with both static multi-source inputs and dynamic fan-out
#' followed by fan-in (`@maestroMap` → `collect()`): in the latter case `.input`
#' is a list of all successful iteration results.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroInputs extract verify`
#' - `#' @maestroInputs collect(letter_a, letter_b)`
#' - `#' @maestroInputs collect(multiply)` *(collect all iterations of a `@maestroMap` upstream)*
#'
#'
#' # maestroOutputs
#'
#' For a DAG style pipeline, the names of pipelines that receive the return value of
#' this pipeline as input. These names must match the function name defining the
#' outputting pipeline. Multiple pipelines can be outputted into. The return value
#' of the pipeline will be passed into the receiving pipeline. Note that this tag
#' could be redundant if pipeline to be inputted into uses `maestroInputs`.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroOutputs transform`
#'
#'
#' # maestroSkip
#'
#' Flags a pipeline to never be executed even if it is scheduled to run. This can
#' be useful when developing or testing a pipeline. This tag takes no value, instead
#' the presence of the tag indicates whether it is skipped. This tag is ignored when
#' `run_schedule(..., run_all = TRUE)` or when using `invoke()`.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroSkip`
#'
#'
#' # maestroRunIf
#'
#' An R expression that is evaluated at run time to determine whether the pipeline
#' should execute. The expression must return a single `TRUE` or `FALSE`. If the
#' expression returns `FALSE`, the pipeline is silently skipped for that run. Resources
#' passed via `run_schedule(..., resources = list(...))` are available inside the
#' expression, as is `.input` for DAG pipelines.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroRunIf Sys.getenv("RUN_PIPELINE") == "true"`
#' - `#' @maestroRunIf lubridate::wday(lubridate::now()) == 2`
#' - `#' @maestroRunIf !is.null(.input)`
#'
#'
#' # maestroMap
#'
#' Enables **dynamic fan-out (scatter)**: the downstream pipeline executes once
#' per element of the upstream return value. When the tag value is empty, each
#' element of the upstream return value is passed directly as `.input`. When one
#' or more R expressions referencing `.input` fields are provided (space-separated),
#' those fields are scattered over and zipped together `pmap`-style: each iteration
#' receives `.input` with all specified fields replaced by their i-th element.
#' The full list remains accessible as `.input` in every branch.
#'
#' Vectors must all have the same length, unless a vector has length 1, in which
#' case it is recycled across all iterations. Mismatched lengths produce a
#' pipeline error.
#'
#' Requires `@maestroInputs` to reference an upstream pipeline.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroMap` *(iterate over each element of the upstream return value)*
#' - `#' @maestroMap .input$items`
#' - `#' @maestroMap .input$ids .input$labels`
#' - `#' @maestroMap .input$model .input$label .input$threshold`
#'
#'
#' # maestroPriority
#'
#' Determines the order in which pipelines that run at the same scheduled instance
#' are executed. Values are positive integers from 1 to N. Order is determined in
#' descending order such that 1 indicates the highest priority. Pipelines with the
#' same priority run in the order in which `build_schedule()` parses the pipeline
#' (usually alphabetical according to file name and then line number within file).
#' By default, all pipelines are given equal priority.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroPriority 1`
#' - `#' @maestroPriority 3`
#'
#'
#' # maestroFlags
#'
#' Arbitrary strings which are then made accessible via `get_flags()`. A pipeline
#' can have multiple tags separated by spaces.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestroFlags critical etl cloud`
#' - `#' @maestroFlags aviation`
#'
#' 
#' # maestroLabel
#' 
#' Key-value pairs used for attributing metadata for a pipeline. Multiple labels
#' can be assigned to a pipeline using multiple instances of `@maestroLabel`. These
#' labels can be extracted as a data.frame. Each label must follow the pattern of 
#' `key` `value`.
#' 
#' Default:
#' 
#' Examples:
#' - `#' @maestroLabel domain transportation`
#' - `#' @maestroLabel author will.hipson`
#'
#' # maestro
#'
#' Generic tag for identifying a maestro pipeline with all the defaults. Useful
#' when you want a pipeline to be scheduled via maestro that accepts all default
#' tag values. Only use this tag if you have no other maestro tags. The tag takes
#' no value.
#'
#' Default:
#'
#' Examples:
#' - `#' @maestro`
#'
#' @name maestro_tags
NULL

#' Run a schedule
#'
#' Given a schedule in a `maestro` project, runs the pipelines that are scheduled to execute
#' based on the current time.
#'
#' @md
#' @details
#'
#' ## Pipeline schedule logic
#'
#' The function `run_schedule()` examines each pipeline in the schedule table and determines
#' whether it is scheduled to run at the current time using some simple time arithmetic. We assume
#' `run_schedule(schedule, check_datetime = Sys.time())`, but this need not be the case.
#'
#' ## Output
#'
#' `run_schedule()` returns the same MaestroSchedule object with modified attributes. Use `get_status()`
#' to examine the status of each pipeline and use `get_artifacts()` to get any return values from the
#' pipelines as a list.
#'
#' ## Pipelines with arguments (resources)
#'
#' If a pipeline takes an argument that doesn't include a default value, these can be supplied
#' in the orchestrator via `run_schedule(resources = list(arg1 = val))`. The name of the argument
#' used by the pipeline must match the name of the argument in the list. Currently, each named
#' resource must refer to a single object. In other words, you can't have two pipes using
#' the same argument but requiring different values.
#'
#' ## Running in parallel
#'
#' Pipelines can be run in parallel using the `cores` argument. First, you must run `future::plan(future::multisession)`
#' in the orchestrator. Then, supply the desired number of cores to the `cores` argument. Note that
#' console output appears different in multicore mode.
#'
#' ## Logging & Console Output
#'
#' By default, `maestro` suppresses pipeline messages, warnings, and errors from appearing in the console, but
#' messages coming from `print()` and other console logging packages like `cli` and `logger` are not suppressed
#' and will be interwoven into the output generated from `run_schedule()`. Messages from `cat()` and related functions are always suppressed
#' due to the nature of how those functions operate with standard output.
#'
#' Users are advised to make use of R's `message()`, `warning()`, and `stop()` functions in their pipelines
#' for managing conditions. Use `log_to_console = TRUE` to print these to the console.
#' 
#' Note that `log_to_console` does not work when running multicore. In the case of multicore, it is recommended to use
#' `log_to_file`.
#'
#' Maestro can generate a log file that is appended to each time the orchestrator is run. Use `log_to_file = TRUE` or `log_to_file = '[path-to-file]'` and
#' maestro will create/append to a file in the project directory.
#' This log file will be appended to until it exceeds the byte size defined in `log_file_max_bytes` argument after which
#' the log file is deleted.
#'
#' @param schedule object of type MaestroSchedule created using `build_schedule()`
#' @inheritParams get_pipeline_run_sequence
#' @param orch_frequency of the orchestrator, a single string formatted like "1 day", "2 weeks", "hourly", etc.
#' @param resources named list of shared resources made available to pipelines as needed
#' @param run_all run all pipelines regardless of the schedule (default is `FALSE`) - useful for testing.
#' Does not apply to pipes with a `maestroSkip` tag. Conditional pipelines using `maestroRunIf` still behave according to their condition.
#' @param n_show_next show the next n scheduled pipes
#' @param cores number of cpu cores to run if running in parallel. If > 1, `furrr` is used and a multisession plan must be executed in the orchestrator (see details)
#' @param log_file_max_bytes numeric specifying the maximum number of bytes allowed in the log file before purging the log (within a margin of error)
#' @param quiet silence metrics to the console (default = `FALSE`). Note this does not affect messages generated from pipelines when `log_to_console = TRUE`.
#' @param log_to_console whether or not to include pipeline messages, warnings, errors to the console (default = `FALSE`) (see Logging & Console Output section)
#' @param log_to_file either a boolean to indicate whether to create and append to a `maestro.log` or a character path to a specific log file. If `FALSE` or `NULL` it will not log to a file.
#'
#' @return MaestroSchedule object
#' @importFrom R.utils countLines
#' @export
#' @examples
#'
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   # Runs the schedule every 1 day
#'   run_schedule(
#'     schedule,
#'     orch_frequency = "1 day",
#'     quiet = TRUE
#'   )
#'
#'   # Runs the schedule every 15 minutes
#'   run_schedule(
#'     schedule,
#'     orch_frequency = "15 minutes",
#'     quiet = TRUE
#'   )
#' }
run_schedule <- function(
    schedule,
    orch_frequency = "1 day",
    check_datetime = lubridate::now(tzone = "UTC"),
    resources = list(),
    run_all = FALSE,
    n_show_next = 5,
    cores = 1,
    log_file_max_bytes = 1e6,
    quiet = FALSE,
    log_to_console = FALSE,
    log_to_file = FALSE
) {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  if (!(lubridate::is.Date(check_datetime) | lubridate::is.POSIXct(check_datetime))) {
    cli::cli_abort("`check_datetime` must be a {.cls Date} or {.cls POSIXct} type.")
  }

  # Allow a session-level override of check_datetime via options()
  override <- getOption("maestro.check_datetime_override")
  if (!is.null(override)) {
    if (!lubridate::is.POSIXct(override)) {
      cli::cli_abort(
        "{.code getOption('maestro.check_datetime_override')} must be a {.cls POSIXct} value."
      )
    }
    check_datetime <- override
  }

  # Get the orchestrator nunits
  orch_nunits <- validate_orch_frequency(orch_frequency)

  # Warn if running the orchestrator less frequent than the highest frequency pipeline
  pipelines <- schedule$PipelineList$MaestroPipelines
  if (length(pipelines) > 0 && !run_all) {
    pipe_names <- purrr::map_chr(pipelines, ~.x$get_pipe_name())
    freq_nunits <- purrr::map(pipelines, ~.x$get_frequency_nunits())
    frequency_n <- purrr::map_int(freq_nunits, "n")
    frequency_unit <- purrr::map_chr(freq_nunits, "unit")
    not_na <- !is.na(frequency_n)
    if (any(not_na)) {
      sch_units_lt_orch_unit <- units_lt_units(frequency_unit[not_na], standardize_units(orch_nunits$unit))
      if (any(sch_units_lt_orch_unit)) {
        offending_pipe_names <- pipe_names[not_na][sch_units_lt_orch_unit]
        offending_pipe_units <- frequency_unit[not_na][sch_units_lt_orch_unit]
        cli::cli_warn(
          c(
            "Pipeline{?s} {.pkg {offending_pipe_names}} {?has/have} {? a frequency/frequencies} higher (i.e., more often)
        than the frequency of the orchestrator. This means the pipeline{?s} will not run as frequently as specified.",
            "i" = "Consider increasing the frequency of the orchestrator or decreasing the pipeline frequency."
          )
        )
      }
    }
  }

  # Check if we're logging to a file
  if (!rlang::is_scalar_logical(log_to_file) && !rlang::is_scalar_character(log_to_file)) {
    cli::cli_abort(
      "When not NULL, {.code log_to_file} must be a single boolean or a single character string."
    )
  }

  if (is.logical(log_to_file) && log_to_file) {
    log_to_file <- "maestro.log"
  }

  if (is.character(log_to_file)) {
    if (!file.exists(log_to_file)) file.create(log_to_file)
  } else {
    log_to_file <- tempfile()
  }

  # Ensure that elements in resources are named
  if (length(resources) > 0) {
    resources_length <- length(resources)
    n_named <- sum(names(resources) != "")
    if (resources_length > n_named) {
      cli::cli_abort(
        "All elements in `resources` must be named."
      )
    }

    n_uniq_names <- length(unique(names(resources)))
    if (resources_length > n_uniq_names) {
      cli::cli_abort(
        "All elements in `resources` must have unique names."
      )
    }
  }

  schedule$run(
    orch_n = orch_nunits$n,
    orch_unit = orch_nunits$unit,
    check_datetime = check_datetime,
    resources = resources,
    run_all = run_all,
    n_show_next = n_show_next,
    cores = cores,
    log_file = log_to_file,
    log_file_max_bytes = log_file_max_bytes,
    quiet = quiet,
    log_to_console = log_to_console
  )

  return(schedule)
}

test_that("create_pipeline creates a new pipeline", {
  withr::with_tempdir({
    create_pipeline("new_pipe", open = FALSE)

    expect_no_error({
      schedule <- build_schedule()

      run_schedule(schedule)
    })

    expect_true(file.exists("pipelines/new_pipe.R"))
  }) |>
    expect_message()
}) |>
  suppressMessages()

test_that("create_pipeline fixes bad names", {
  withr::with_tempdir({
    create_pipeline("new-pipe", open = FALSE)

    expect_no_error({
      schedule <- build_schedule()

      run_schedule(schedule)
    })

    expect_true(file.exists("pipelines/new_pipe.R"))
  }) |>
    expect_message()
}) |>
  suppressMessages()

test_that("create_pipeline with POSIXct works and can be run", {

  withr::with_tempdir({
    create_pipeline("new-pipe", start_time = as.POSIXct("2024-05-01 12:00:00"), open = FALSE)

    expect_no_error({
      schedule <- build_schedule()

      run_schedule(schedule)
    })

    expect_true(file.exists("pipelines/new_pipe.R"))
  }) |>
    expect_message()

  withr::with_tempdir({
    create_pipeline("new-pipe", start_time = "10:00:00", open = FALSE)

    expect_no_error({
      schedule <- build_schedule()

      run_schedule(schedule)
    })

    expect_true(file.exists("pipelines/new_pipe.R"))
  }) |>
    expect_message()
}) |>
  suppressMessages()

test_that("create_pipeline aborts if pipeline already exists", {

  withr::with_tempdir({
    expect_error({
      create_pipeline("new-pipe", pipeline_dir = ".", open = FALSE)
      create_pipeline("new-pipe", pipeline_dir = ".", open = FALSE)
    }, regexp = "already exists")

    # Works if overwrite = TRUE
    expect_message(
      create_pipeline("new-pipe", pipeline_dir = ".", open = FALSE, overwrite = TRUE),
      regexp = "Overwriting existing"
    )
  })
}) |>
  suppressMessages()

test_that("create_pipeline with priority", {
  withr::with_tempdir({
    create_pipeline("new-pipe", start_time = "10:00:00", open = FALSE, priority = 1, quiet = TRUE)

    expect_no_error({
      schedule <- build_schedule(quiet = TRUE)

      run_schedule(schedule, quiet = TRUE)
    })

    expect_true(file.exists("pipelines/new_pipe.R"))
  })
})

test_that("Conditional pipes work with DAG pipelines via .input", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroOutputs p2
      p1 <- function() {
        TRUE
      }

      #' @maestroRunIf .input
      p2 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = FALSE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroOutputs p2
      p1 <- function() {
        FALSE
      }

      #' @maestroRunIf .input
      p2 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)

  # More complex multiline eval
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroOutputs p2
      p1 <- function() {
        7
      }

      #' @maestroRunIf
      #' is_mod <- (.input %% 2) == 0
      #' (is_mod * 4) == 4
      p2 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)
})

test_that("DAGs with a conditional pipe in the middle by default halt further execution of the DAG", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroOutputs p2
      p1 <- function() {
        FALSE
      }

      #' @maestroRunIf .input
      p2 <- function() {
        TRUE
      }

      #' @maestroInputs p2
      p3 <- function(.input) {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)
})

test_that("Conditional pipes work using resources", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroRunIf var == 4
      p1 <- function() {
        TRUE
      }

      #' @maestroFrequency 1 day
      #' @maestroRunIf var == 3
      p2 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE,
      resources = list(var = 4)
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)
})

test_that("Conditions with errors are handled", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroRunIf stop('oh no')
      p1 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$success)
  expect_snapshot(status$invoked)
  expect_snapshot(last_run_errors())
})

test_that("Conditions that don't return a single boolean are handled", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroRunIf c(1, 2)
      p1 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$success)
  expect_snapshot(status$invoked)
  expect_snapshot(last_run_errors())
})

test_that("Empty maestroRunIf is ignored", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroRunIf
      p1 <- function() {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = FALSE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$success)
  expect_snapshot(status$invoked)
})

test_that("Branching pipelines execute with conditionals", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      a <- function() {
        'b'
      }

      #' @maestroInputs a
      #' @maestroRunIf .input == 'b'
      b <- function(.input) {
        TRUE
      }

      #' @maestroInputs a
      #' @maestroRunIf .input == 'c'
      c <- function(.input) {
        TRUE
      }
      ",
      con = "pipelines/conditionals.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = FALSE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)
})

test_that("get_status returns a data.frame", {
  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)
  schedule <- run_schedule(schedule, orch_frequency = "hourly")

  expect_s3_class(
    get_status(schedule),
    "data.frame"
  )
}) |>
  suppressMessages()

test_that("errors if schedule is not a MaestroSchedule", {
  expect_error({
    get_status(iris)
  }, regexp = "Schedule must be an object")
})

test_that("get_artifacts returns artifacts", {
  schedule <- build_schedule(test_path("test_pipelines_run_artifacts"))

  output <- run_schedule(
    schedule,
    run_all = TRUE
  )

  artifacts <- get_artifacts(output)

  expect_type(artifacts, "list")
  expect_gt(length(artifacts), 0)
}) |>
  suppressMessages()

test_that("errors if schedule is not a MaestroSchedule", {
  expect_error({
    get_artifacts(iris)
  }, regexp = "Schedule must be an object")
})

#' Create a new pipeline in a pipelines directory
#'
#' Allows the creation of new pipelines (R scripts) and fills in the maestro tags as specified.
#'
#' @param pipe_name name of the pipeline and function
#' @param pipeline_dir directory containing the pipeline scripts
#' @param frequency how often the pipeline should run (e.g., 1 day, daily, 3 hours, 4 months). Fills in maestroFrequency tag
#' @param start_time start time of the pipeline schedule. Fills in maestroStartTime tag
#' @param tz timezone that pipeline will be scheduled in. Fills in maestroTz tag
#' @param log_level log level for the pipeline (e.g., INFO, WARN, ERROR). Fills in maestroLogLevel tag
#' @param open whether or not to open the script upon creation
#' @param quiet whether to silence messages in the console (default = `FALSE`)
#' @param overwrite whether or not to overwrite an existing pipeline of the same name and location.
#' @param skip whether to skip the pipeline when running in the orchestrator (default = `FALSE`)
#' @param inputs vector of names of pipelines that input into this pipeline (default = `NULL` for no inputs)
#' @param outputs vector of names of pipelines that receive output from this pipeline (default = `NULL` for no outputs)
#' @param priority a single positive integer corresponding to the order in which this pipeline will be invoked in the presence of other simultaneously invoked pipelines.
#'
#' @return invisible
#' @export
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline(
#'     "extract_data",
#'     pipeline_dir = pipeline_dir,
#'     frequency = "1 hour",
#'     open = FALSE,
#'     quiet = TRUE,
#'     overwrite = TRUE
#'   )
#'
#'   create_pipeline(
#'     "new_job",
#'     pipeline_dir = pipeline_dir,
#'     frequency = "20 minutes",
#'     start_time = as.POSIXct("2024-06-21 12:20:00"),
#'     log_level = "ERROR",
#'     open = FALSE,
#'     quiet = TRUE,
#'     overwrite = TRUE
#'   )
#' }
create_pipeline <- function(
    pipe_name,
    pipeline_dir = "pipelines",
    frequency = "1 day",
    start_time = Sys.Date(),
    tz = "UTC",
    log_level = "INFO",
    quiet = FALSE,
    open = interactive(),
    overwrite = FALSE,
    skip = FALSE,
    inputs = NULL,
    outputs = NULL,
    priority = NULL
  ) {

  skip <- if (skip) {
    "\n#' @maestroSkip"
  } else {
    ""
  }

  inputs <- if (!is.null(inputs)) {
    paste("\n#'", paste(inputs, collapse = " "))
  } else {
    ""
  }

  outputs <- if (!is.null(outputs)) {
    paste("\n#'", paste(outputs, collapse = " "))
  } else {
    ""
  }

  priority <- if (!is.null(priority)) {
    if (!(rlang::is_scalar_integerish(priority) && priority > 0)) {
      cli::cli_abort(
        c("`priority` must be a single positive whole number."),
        call = NULL
      )
    }
    paste("\n#' @maestroPriority", priority)
  } else {
    ""
  }

  # Makes a valid name for a pipe
  pipe_name <- gsub("\\.", "_", make.names(pipe_name))

  script <- readLines(system.file("pipeline_template", package = "maestro")) |>
    paste(collapse = "\n") |>
    glue::glue(
      .open = "{{",
      .close = "}}",
      .null = NULL
    )

  path <- file.path(pipeline_dir, paste0(pipe_name, ".R"))

  if (!dir.exists(pipeline_dir)) {
    dir.create(pipeline_dir)
  }

  if (file.exists(path)) {
    if (!overwrite) {
      cli::cli_abort(
        c("File {.file {path}} already exists.",
          "Set {.code create_pipeline(overwrite = TRUE)} to overwrite anyway."),
        call = NULL
      )
    } else {
      if (!quiet) cli::cli_alert_warning("Overwriting existing pipeline at {.file {path}}.")
    }
  }

  writeLines(
    script,
    path
  )

  if (open) {
    rstudioapi::documentOpen(path)
  }

  if(!quiet) {
    cli::cli_alert_success("Created pipeline at {.file {path}}")
  }
}

#' Build a schedule
#'
#' Builds a `MaestroSchedule` object for use in `run_schedule()`.
#'
#' This function parses the maestro tags of functions located in `pipeline_dir` which is
#' conventionally called 'pipelines'. An orchestrator requires a `MaestroSchedule`
#' to determine which pipelines are to run and when. Each pipeline in the schedule
#' is a parsed function and its scheduling parameters such as its frequency.
#'
#' The `MaestroSchedule` is mostly intended to be passed directly to `run_schedule()`.
#' In other words, it is not recommended to make changes to it.
#'
#' @param pipeline_dir path to directory containing the pipeline scripts
#' @param cores number of cpu cores to use when parsing pipeline scripts. If > 1,
#'   `furrr` is used and a multisession plan must be set in the orchestrator
#'   (e.g. `future::plan(future::multisession)`). See `run_schedule()` for details.
#' @param quiet silence metrics to the console (default = `FALSE`)
#'
#' @return MaestroSchedule
#' @export
#' @examples
#'
#' # Creating a temporary directory for demo purposes! In practice, just
#' # create a 'pipelines' directory at the project level.
#'
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   build_schedule(pipeline_dir = pipeline_dir)
#' }
build_schedule <- function(pipeline_dir = "./pipelines", cores = 1L, quiet = FALSE) {

  if (!dir.exists(pipeline_dir)) {
    cli::cli_abort("No directory called {.emph {pipeline_dir}}")
  }

  # Parse all the .R files in the `pipeline_dir` directory
  pipelines <- list.files(
    pipeline_dir,
    pattern = "*.R$",
    full.names = TRUE
  )

  # Error if the directory has no .R scripts
  if (length(pipelines) == 0) {
    cli::cli_inform(
      "No R scripts in {pipeline_dir}."
    )
    return(invisible())
  }

  if (cores < 1 || (cores %% 1) != 0) cli::cli_abort("`cores` must be a positive integer")
  is_multicore <- FALSE
  if (cores > 1) {
    tryCatch({
      rlang::check_installed("furrr")
      is_multicore <- TRUE
    }, error = \(e) {
      cli::cli_warn("{.pkg furrr} is required for running on multiple cores.")
    })
  }

  mapper_fun <- if (is_multicore) {
    function(...) {
      furrr::future_map(
        ...,
        .options = furrr::furrr_options(
          packages = c("maestro", "logger"),
          stdout = FALSE,
          seed = NULL
        )
      )
    }
  } else {
    purrr::map
  }

  pipeline_attempts <- mapper_fun(
    pipelines, purrr::safely(build_schedule_entry)
  ) |>
    stats::setNames(basename(pipelines))

  pipeline_results <- purrr::map(
    pipeline_attempts,
    ~.x$result
  ) |>
    purrr::discard(is.null)

  pipeline_errors <- purrr::map(
    pipeline_attempts,
    ~.x$error
  ) |>
    purrr::discard(is.null)

  # Assign the errors to the pkgenv
  maestro_pkgenv$last_build_errors <- pipeline_errors

  # Check uniqueness of the function names
  pipe_names <- purrr::map(pipeline_results, ~{
    .x$get_pipe_names()
  }) |>
    purrr::list_c()

  # Check for uniqueness of pipe names
  if (length(unique(pipe_names)) < length(pipe_names)) {
    non_unique_names <- pipe_names[duplicated(pipe_names)]
    cli::cli_abort(
      c("Function names must all be unique",
        "i" = "{.code {non_unique_names}} used more than once.")
    )
  }

  schedule <- MaestroSchedule$new(Pipelines = pipeline_results)

  schedule$PipelineList$validate_network()

  if (!quiet) {
    maestro_parse_cli(pipeline_results, pipeline_errors)
  }

  schedule
}

test_that("build_schedule works on a directory of all good pipelines", {
  res <- build_schedule(test_path("test_pipelines_parse_all_good"))
  expect_s3_class(res, "MaestroSchedule")
  schedule <- res$get_schedule()
  expect_s3_class(schedule, "data.frame")
}) |>
  suppressMessages()

test_that("build_schedule works on a directory of some good pipelines, warns", {

  expect_warning({
    res <- build_schedule(test_path("test_pipelines_parse_some_good"))
  }, regexp = "failed to parse")

  expect_s3_class(res, "MaestroSchedule")
  expect_gte(length(res$PipelineList), 1)
}) |>
  suppressMessages()

test_that("build_schedule errors on a directory of all bad pipelines", {
  expect_error({
    res <- build_schedule(test_path("test_pipelines_parse_all_bad"))
  }, regexp = "All scripts failed to parse")

  errors <- last_build_errors()
  expect_type(errors, "list")
  expect_gt(length(errors), 0)
}) |>
  suppressMessages()

test_that("build_schedule errors on a nonexistent directory with no .R scripts", {
  expect_error({
    build_schedule(test_path("directory_that_doesnt_exist"))
  }, regexp = "No directory called")
})

test_that("informs if referencing a folder with no R scripts", {
  withr::with_tempdir({
    expect_message({
      build_schedule(tempdir())
    }, regexp = "No R scripts in")
  })
})

test_that("errors if there are duplicate pipeline names", {
  expect_error({
    build_schedule(test_path("test_pipelines_dup_names"), quiet = TRUE)
  }, "Function names must all be unique")
})

test_that("show_network works on all independent pipes", {
  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)
  expect_warning(
    vis <- show_network(schedule),
    class = "lifecycle_warning_deprecated"
  )
  expect_s3_class(vis, "grViz")
}) |>
  suppressMessages()

test_that("show_network works on DAG pipelines", {
  schedule <- build_schedule(test_path("test_pipelines_dags_good"), quiet = TRUE)
  expect_warning(
    vis <- show_network(schedule),
    class = "lifecycle_warning_deprecated"
  )
  expect_s3_class(vis, "grViz")
}) |>
  suppressMessages()

test_that("errors if schedule is not a MaestroSchedule", {
  expect_error({
    suppressWarnings(show_network(iris))
  }, regexp = "Schedule must be an object")
})

test_that("get_flags works as expected", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      #' @maestroFlags experimental dev
      p1 <- function() {
      }

      #' @maestroFrequency 2 hours
      #' @maestroFlags critical aviation
      p2 <- function() {
      }

      #' @maestroFrequency 3 hours
      #' @maestroFlags critical
      p3 <- function() {
      }
      ",
      con = "pipelines/flags.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  flags_df <- get_flags(schedule)

  expect_snapshot(flags_df)
})

test_that("returns empty data.frame if there are no flags", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      p1 <- function() {
      }

      #' @maestroFrequency 2 hours
      p2 <- function() {
      }

      #' @maestroFrequency 3 hours
      p3 <- function() {
      }
      ",
      con = "pipelines/flags.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  flags_df <- get_flags(schedule)

  expect_snapshot(flags_df)
})

test_that("get_schedule returns a data.frame", {
  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)
  expect_s3_class(
    get_schedule(schedule),
    "data.frame"
  )

  expect_snapshot(schedule)
}) |>
  suppressMessages()

test_that("errors if schedule is not a MaestroSchedule", {
  expect_error({
    get_schedule(iris)
  }, regexp = "Schedule must be an object")
})

test_that("get_schedule works with DAG schedules", {
  schedule <- build_schedule(test_path("test_pipelines_dags_good"))
  expect_snapshot(get_schedule(schedule)[, c("script_path", "pipe_name", "frequency", "tz", "skip", "log_level")])
}) |>
  suppressMessages()

test_that("get_slot_usage works as expected", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      #' @maestroStartTime 10:00:00
      hourly1 <- function() {
      }

      #' @maestroFrequency 2 hours
      #' @maestroStartTime 10:00:00
      hourly2 <- function() {
      }

      #' @maestroFrequency 3 hours
      #' @maestroStartTime 10:00:00
      #' @maestroTz America/Halifax
      hourly3 <- function() {
      }
      ",
      con = "pipelines/hourlies.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  avail_hour <- get_slot_usage(
    schedule,
    orch_frequency = "1 hour"
  )
  expect_snapshot(avail_hour)

  avail_day <- get_slot_usage(
    schedule,
    orch_frequency = "1 hour",
    slot_interval = "day"
  )

  expect_snapshot(avail_day)
})

test_that("get_slot_usage works with variety of frequencies", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      #' @maestroStartTime 10:00:00
      hourly1 <- function() {
      }

      #' @maestroFrequency 3 days
      #' @maestroStartTime 2025-01-01 12:00:00
      daily1 <- function() {
      }

      #' @maestroFrequency weekly
      #' @maestroStartTime 2025-01-02 10:00:00
      #' @maestroTz UTC
      weekly1 <- function() {
      }
      ",
      con = "pipelines/multi.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  avail_hour <- get_slot_usage(
    schedule,
    orch_frequency = "1 hour"
  )
  expect_snapshot(avail_hour)

  avail_day <- get_slot_usage(
    schedule,
    orch_frequency = "1 hour",
    slot_interval = "day"
  )

  expect_snapshot(avail_day)

  avail_week <- get_slot_usage(
    schedule,
    orch_frequency = "1 hour",
    slot_interval = "week"
  )

  expect_snapshot(avail_week)
})

test_that("get_slot_usage informs on empty schedule", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "",
      con = "pipelines/multi.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  expect_message({
    get_slot_usage(
      schedule,
      "1 hour"
    )
  })
})

test_that("get_run_sequence works", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      p1 <- function() {
      }

      #' @maestroFrequency 2 hours
      p2 <- function() {
      }

      #' @maestroFrequency 3 hours
      p3 <- function() {
      }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  run_seq <- get_run_sequence(schedule, n = 10)
  expect_in(c("p1", "p2", "p3"), unique(run_seq$pipe_name))
  expect_equal(nrow(run_seq), 30L)
  expect_s3_class(run_seq$scheduled_time, "POSIXct")
  expect_true("is_primary" %in% names(run_seq))
  expect_true(all(run_seq$is_primary))
})

test_that("informative error messages on invalid params", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      p1 <- function() {
      }

      #' @maestroFrequency 2 hours
      p2 <- function() {
      }

      #' @maestroFrequency 3 hours
      p3 <- function() {
      }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  expect_error(
    get_run_sequence(schedule, n = -1),
    regexp = "must be a positive integer"
  )

  expect_error(
    get_run_sequence(schedule, min_datetime = 4),
    regexp = "must be a Date"
  )

  expect_error(
    get_run_sequence(schedule, min_datetime = as.Date("2025-01-01"), max_datetime = as.Date("2024-12-31")),
    regexp = "cannot be greater than"
  )

  expect_error(
    get_run_sequence(schedule, include_only_primary = "a"),
    regexp = "must be a boolean"
  )

  expect_error(
    get_run_sequence(schedule, include_skipped = "a"),
    regexp = "must be a boolean"
  )
})

test_that("dags are properly represented", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      p1 <- function() {
      }

      #' @maestroInputs p1
      p2 <- function(.input) {
      }

      #' @maestroInputs p2
      p3 <- function(.input) {
      }
      ",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  run_seq <- get_run_sequence(schedule, n = 10)
  expect_in(c("p1", "p2", "p3"), unique(run_seq$pipe_name))
  expect_equal(nrow(run_seq), 30L)
  expect_s3_class(run_seq$scheduled_time, "POSIXct")
  expect_true(!any(is.na(run_seq$scheduled_time)))

  # is_primary: p1 is root, p2 and p3 are downstream
  expect_true(all(run_seq$is_primary[run_seq$pipe_name == "p1"]))
  expect_true(all(!run_seq$is_primary[run_seq$pipe_name %in% c("p2", "p3")]))

  run_seq <- get_run_sequence(schedule, n = 10, include_only_primary = TRUE)
  expect_true(all(run_seq$pipe_name == "p1"))
  expect_equal(nrow(run_seq), 10L)
  expect_s3_class(run_seq$scheduled_time, "POSIXct")
  expect_true(!any(is.na(run_seq$scheduled_time)))
})

test_that("include_skipped = FALSE excludes skipped pipelines", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      p1 <- function() {
      }

      #' @maestroFrequency 2 hours
      #' @maestroSkip
      p2 <- function() {
      }

      #' @maestroFrequency 3 hours
      p3 <- function() {
      }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  run_seq_all <- get_run_sequence(schedule, n = 10, include_skipped = TRUE)
  expect_in(c("p1", "p2", "p3"), unique(run_seq_all$pipe_name))

  run_seq_no_skip <- get_run_sequence(schedule, n = 10, include_skipped = FALSE)
  expect_false("p2" %in% unique(run_seq_no_skip$pipe_name))
  expect_in(c("p1", "p3"), unique(run_seq_no_skip$pipe_name))
})

test_that("backward-looking min_datetime returns historic run sequences", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroStartTime 04:00:00
      daily_early <- function() {
      }

      #' @maestroFrequency 1 hour
      hourly <- function() {
      }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
  })

  now <- lubridate::now()
  min_dt <- now - lubridate::days(7)
  max_dt <- now

  run_seq <- get_run_sequence(schedule, min_datetime = min_dt, max_datetime = max_dt)

  expect_in(c("daily_early", "hourly"), unique(run_seq$pipe_name))
  expect_s3_class(run_seq$scheduled_time, "POSIXct")
  expect_true(all(run_seq$scheduled_time >= min_dt))
  expect_true(all(run_seq$scheduled_time <= max_dt))

  # daily pipeline at 04:00:00 should have ~7 occurrences over 7 days
  daily_seq <- run_seq[run_seq$pipe_name == "daily_early", ]
  expect_gte(nrow(daily_seq), 6L)
  expect_lte(nrow(daily_seq), 8L)
})

test_that("can get a single labeled pipeline", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      #' @maestroLabel domain transportation
      labelled <- function() {

      }
      ",
      con = "pipelines/label.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    expect_snapshot(schedule$get_labels())
  })
})

test_that("can get a double labeled pipeline", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      #' @maestroLabel domain transportation
      #' @maestroLabel author will
      labelled <- function() {

      }
      ",
      con = "pipelines/label.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    expect_snapshot(schedule$get_labels())
  })
})

test_that("works when there are no labeled pipelines", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      not_labeled <- function() {

      }
      ",
      con = "pipelines/label.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    expect_equal(nrow(schedule$get_labels()), 0)
  })
})

test_that("can get a double labeled pipeline with the same key", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      #' @maestroLabel author will
      #' @maestroLabel author kaz
      labelled <- function() {

      }
      ",
      con = "pipelines/label.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    expect_snapshot(schedule$get_labels())
  })
})

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-01 09:00:00
#' @maestroTz UTC
#'
#'
#' @export
get_mtcars2 <- function(vals) {
  if (is.null(vals)) stop("Null vals")
  mean(vals)
}

#' @maestroFrequency daily
numbers <- function() {
  1:3
}

#' @maestroInputs numbers
#' @maestroMap
multiply <- function(.input) {
  .input * 3
}

#' @maestroInputs collect(multiply)
add <- function(.input) {
  sum(unlist(.input))
}

#' Get the statuses of the pipelines in a MaestroSchedule object
#'
#' A status data.frame contains the names and locations of the pipelines as
#' well as information around whether they were invoked, the status (error, warning, etc.),
#' and the run time.
#'
#' @inheritParams run_schedule
#'
#' @return data.frame
#' @export
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   schedule <- run_schedule(
#'     schedule,
#'     orch_frequency = "1 day",
#'     quiet = TRUE
#'   )
#'
#'   get_status(schedule)
#'
#'   # Alternatively, use the underlying R6 method
#'   schedule$get_status()
#' }
get_status <- function(schedule) {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  schedule$get_status()
}

#' Class for an individual maestro pipeline
#' A pipeline is defined as a single R script with a schedule or input
#' @keywords internal
MaestroPipeline <- R6::R6Class(
  "MaestroPipeline",

  public = list(
    #' @description
    #' Create a new Pipeline object
    #' @param script_path path to the script
    #' @param pipe_name name of the pipeline
    #' @param frequency frequency of the pipeline (e.g., 1 day)
    #' @param start_time_raw start time as a raw string from the @maestroStartTime tag
    #' @param tz time zone of the pipeline
    #' @param hours specific hours of the day
    #' @param days specific days of week or month
    #' @param months specific months of year
    #' @param skip whether to skip the pipeline regardless of scheduling
    #' @param log_level log level of the pipeline
    #' @param inputs names of pipelines that this pipeline is dependent on for input
    #' @param outputs names of pipelines for which this pipeline is a dependency
    #' @param priority priority of the pipeline
    #' @param flags arbitrary pipelines flags
    #' @param run_if string representing an R expression that can be evaluated and returns TRUE or FALSE; or NULL
    #' @param is_collect logical; TRUE when @maestroInputs uses collect() fan-in marker
    #' @param map named list of key=expr_string pairs from @maestroMap, or NULL
    #' @param labels list of key-value pairs for pipeline labeling
    #'
    #' @return MaestroPipeline object
    initialize = function(
      script_path,
      pipe_name,
      frequency = NA_character_,
      start_time_raw = NA_character_,
      tz = NA_character_,
      hours = NULL,
      days = NULL,
      months = NULL,
      skip = FALSE,
      log_level = "INFO",
      inputs = NULL,
      outputs = NULL,
      priority = Inf,
      flags = c(),
      run_if = NULL,
      is_collect = FALSE,
      map = NULL,
      labels = list()
    ) {

      # Update the private attributes
      private$script_path <- script_path
      private$pipe_name <- pipe_name
      private$skip <- skip
      private$log_level <- log_level
      private$inputs <- inputs
      private$outputs <- outputs
      private$priority <- priority
      private$flags <- flags
      private$labels <- labels
      private$run_if <- if (!is.null(run_if) && trimws(run_if) == "") {
        NULL
      } else {
        run_if
      }
      private$is_collect <- is_collect
      private$map <- map

      if (is.null(inputs)) {
        private$tz <- tz
        private$frequency <- frequency
        private$start_time_raw <- start_time_raw

        private$hours <- hours
        private$months <- months

        # Create units and n
        withCallingHandlers(
          {
            nunits <- purrr::map(
              frequency,
              purrr::possibly(
                ~ {
                  parse_rounding_unit(.x)
                },
                otherwise = list(n = NA, unit = NA)
              )
            )
          },
          purrr_error_indexed = function(err) {
            rlang::cnd_signal(err$parent)
          }
        )

        # Create days_of_week and days_of_month from days
        days_of_week <- purrr::map_if(
          days,
          is.factor,
          as.numeric,
          .else = ~NA_real_
        ) |>
          purrr::discard(is.na) |>
          purrr::list_c()
        days_of_month <- purrr::map_if(
          days,
          is.numeric,
          ~.x,
          .else = ~NA_real_
        ) |>
          purrr::discard(is.na) |>
          purrr::list_c()

        # Update the transformed private attributes
        private$frequency_n <- purrr::map_int(nunits, "n")
        private$frequency_unit <- purrr::map_chr(nunits, "unit")
        private$days_of_week <- days_of_week %n% 1:7
        private$days_of_month <- days_of_month %n% 1:31
      }
    },

    #' @description
    #' Prints the pipeline
    print = function() {
      cli::cli_h3("Maestro Pipeline: {.emph {private$pipe_name}}")
      cli::cli_text("{.file {private$script_path}}")
      cli::cli_h3("Schedule")

      if (!is.null(private$inputs)) {
        cli::cli_li("Dependent on: {private$inputs}")
      } else {
        cli::cli_li("Frequency: {private$frequency}")
        cli::cli_li("Start Time: {private$start_time_raw}")
      }

      if (!is.null(private$hours)) {
        cli::cli_li("Hours: {private$hours}")
      }

      if (!is.null(private$days_of_week)) {
        cli::cli_li("Days of week: {private$days_of_week}")
      }

      if (!is.null(private$days_of_month)) {
        cli::cli_li("Days of month: {private$days_of_month}")
      }

      if (!is.null(private$months)) {
        cli::cli_li("Months: {private$months}")
      }

      cli::cli_h3("Status")
      switch(
        private$status,
        `Not Run` = cli::cli_li(cli::col_magenta(private$status)),
        `Success` = cli::cli_li(cli::col_green(private$status)),
        `Warning` = cli::cli_li(cli::col_yellow(private$status)),
        `Error` = cli::cli_li(cli::col_red(private$status)),
        cli::cli_li("Unknown")
      )

      if (!is.na(private$run_time_start)) {
        cli::cli_li("On: {private$run_time_start}")
      }
    },

    #' @description
    #' Runs the pipeline
    #' @param resources named list of arguments and values to pass to the pipeline
    #' @param log_file path to the log file for logging
    #' @param quiet whether to silence console output
    #' @param log_file_max_bytes maximum bytes of the log file before trimming
    #' @param .input input values from upstream pipelines
    #' @param depth number of inputting pipelines above the current
    #' @param log_to_console whether or not to output statements in the console (FALSE is to suppress and append to log)
    #' @param run_id unique id for the run
    #' @param input_run_id unique id of the run that inputted into the current run (NA if there is no input)
    #' @param lineage character vector of upstream pipeline names ordered from first to latest (or empty if no upstream pipes)
    #' @param iter iteration number for dynamic fanout
    #' @param pre_error trigger error from outside execution
    #' @param ... additional arguments (unused)
    #'
    #' @return invisible
    run = function(
      resources = list(),
      log_file = tempfile(),
      quiet = FALSE,
      log_file_max_bytes = 1e6,
      .input = NULL,
      depth = 0,
      log_to_console = FALSE,
      run_id = NA_character_,
      input_run_id = NA_character_,
      lineage = c(),
      iter = NULL,
      pre_error = NULL,
      ...
    ) {
      internal_run_id <- make_id()

      pipe_name <- private$pipe_name
      script_path <- private$script_path
      log_level <- private$log_level
      lineage <- paste0(c(lineage, pipe_name), collapse = "->")

      private$insert_run_time_attributes(
        internal_run_id,
        list(
          invoked = FALSE,
          success = NA,
          errors = 0L,
          warnings = 0L,
          messages = 0L
        )
      )

      if (!is.null(pre_error)) {
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            run_id = run_id,
            invoked = TRUE,
            input_run_id = input_run_id,
            lineage = lineage
          )
        )
        prepend <- if (depth == 0) {
          ""
        } else {
          marker <- if (private$is_collect) "|-+" else "|-"
          cli::format_inline(rep("  ", times = depth - 1), marker)
        }
        iter_label <- if (!is.null(iter)) cli::format_inline(cli::col_blue("[{iter}]")) else ""
        if (!quiet) {
          cli::cli_progress_step("{prepend}{cli::col_blue(pipe_name)}{iter_label}")
        }
        tryCatch(
          stop(pre_error),
          error = private$pre_error_handler(internal_run_id = internal_run_id)
        )
        if (!quiet) cli::cli_progress_done(result = "failed")
        return(invisible())
      }

      if (log_to_console) {
        logger_fun <- logger::appender_tee
      } else {
        logger_fun <- logger::appender_file
      }

      # Set the logger to null - we just want the text in a variable
      logger::log_threshold(level = log_level, namespace = pipe_name)
      logger::log_appender(
        appender = logger_fun(log_file, max_bytes = log_file_max_bytes),
        namespace = pipe_name
      )
      logger::log_layout(maestro_logger, namespace = pipe_name)

      # Create a context (environment) for running the pipeline
      if (is.null(private$sourced_context)) {
        maestro_context <- new.env()
        # Source the script
        withCallingHandlers(
          source(script_path, local = maestro_context),
          error = private$error_handler,
          warning = private$warning_handler,
          message = private$message_handler
        )
        private$sourced_context <- maestro_context
      } else {
        maestro_context <- private$sourced_context
      }

      resources <- append(resources, list(.input = .input))
      args <- formals(pipe_name, envir = maestro_context)

      prepend <- if (depth == 0) {
        ""
      } else {
        marker <- if (private$is_collect) "|-+" else "|-"
        cli::format_inline(rep("  ", times = depth - 1), marker)
      }

      iter_label <- if (!is.null(iter)) cli::format_inline(cli::col_blue("[{iter}]")) else ""

      do_run <- TRUE
      if (!is.null(private$run_if)) {
        if (!quiet) {
          cli::cli_progress_step("{prepend}{cli::col_blue(pipe_name)}{iter_label} (?)")
        }

        cond <- withCallingHandlers(
          eval_code_str(
            private$run_if,
            vars = resources,
            inherit = maestro_context
          ),
          error = private$cond_error_handler(internal_run_id = internal_run_id),
          warning = private$cond_warning_handler(
            internal_run_id = internal_run_id
          ),
          message = private$cond_message_handler(
            internal_run_id = internal_run_id
          )
        )

        if (!rlang::is_scalar_logical(cond)) {
          withCallingHandlers(
            stop(
              glue::glue("`{private$run_if}` did not return a single boolean."),
              call. = FALSE
            ),
            error = private$cond_error_handler(
              internal_run_id = internal_run_id
            )
          )
        }

        do_run <- cond
      }

      if (!do_run) {
        return(invisible())
      }

      private$status <- "Success"

      run_time_start <- lubridate::now()
      private$run_time_start <- run_time_start

      if (!quiet) {
        cli::cli_progress_step("{prepend}{cli::col_blue(pipe_name)}{iter_label}")
      }

      private$insert_run_time_attributes(
        internal_run_id,
        list(
          run_id = run_id,
          invoked = TRUE,
          pipeline_started = run_time_start,
          input_run_id = input_run_id,
          lineage = lineage
        )
      )

      results <- withCallingHandlers(
        {
          # Identify args with no default value (i.e. required args)
          required_args <- names(args)[sapply(args, function(x) identical(x, quote(expr = )))]
          missing_args <- setdiff(required_args, names(resources))
          if (length(missing_args) > 0) {
            stop(
              paste0(
                "missing argument",
                if (length(missing_args) > 1) "s" else "",
                " ",
                paste0("`", missing_args, "`", collapse = ", "),
                " - pass via `resources = list(",
                paste0(missing_args, " = ...", collapse = ", "),
                ")`"
              ),
              call. = FALSE
            )
          }
          do.call(
            pipe_name,
            args = resources[intersect(names(args), names(resources))],
            envir = maestro_context
          )
        },
        error = private$error_handler(internal_run_id = internal_run_id),
        warning = private$warning_handler(internal_run_id = internal_run_id),
        message = private$message_handler(internal_run_id = internal_run_id)
      )

      private$returns <- results

      run_time_end <- lubridate::now()

      private$run_time_artifacts[[run_id]] <- results
      private$insert_run_time_attributes(
        internal_run_id,
        list(
          success = TRUE,
          pipeline_ended = run_time_end
        )
      )

      invisible()
    },

    #' @description
    #' Get the pipeline name
    #' @return pipeline_name
    get_pipe_name = function() {
      private$pipe_name
    },

    #' @description
    #' Get the frequency n and unit as a list
    #' @return list with n and unit
    get_frequency_nunits = function() {
      list(n = private$frequency_n, unit = private$frequency_unit)
    },

    #' @description
    #' Get the schedule as a data.frame
    #' @return data.frame
    get_schedule = function() {
      dplyr::tibble(
        script_path = private$script_path %n% NA_character_,
        pipe_name = private$pipe_name %n% NA_character_,
        frequency = private$frequency %n% NA_character_,
        start_time = if (!is.null(private$inputs)) {
          lubridate::NA_POSIXct_
        } else {
          private$resolve_start_time(lubridate::now())
        },
        tz = private$tz %n% NA_character_,
        skip = private$skip %n% NA,
        log_level = private$log_level %n% NA_character_,
        frequency_n = private$frequency_n %n% NA_integer_,
        frequency_unit = private$frequency_unit %n% NA_character_,
        priority = private$priority %n% NA_integer_
      )
    },

    #' @description
    #' Check whether a pipeline is scheduled to run based on orchestrator frequency and current time
    #' @param orch_unit unit of the orchestrator (e.g., day)
    #' @param orch_n number of units of the frequency
    #' @param check_datetime datetime against which to check the timeliness of the pipeline (should almost always be now)
    #' @param ... unused
    #' @return MaestroPipeline
    check_timeliness = function(
      orch_unit,
      orch_n,
      check_datetime = lubridate::now(),
      ...
    ) {
      if (private$skip) {
        return(FALSE)
      }
      if (!is.null(private$inputs)) {
        return(TRUE)
      } # pipes with a dependency are always timely

      orch_string <- paste(orch_n, orch_unit)
      check_datetime_round <- timechange::time_round(
        lubridate::with_tz(check_datetime, private$tz),
        unit = orch_string
      )

      start_time <- private$resolve_start_time(check_datetime)

      # One step in pipeline frequency as a difftime/duration
      .one_freq_step <- function() {
        unit <- private$frequency_unit
        n <- private$frequency_n
        switch(
          unit,
          second = lubridate::seconds(n),
          minute = lubridate::minutes(n),
          hour = lubridate::hours(n),
          day = lubridate::days(n),
          week = lubridate::weeks(n),
          month = months(n),
          year = lubridate::years(n)
        )
      }

      # Check if a datetime passes all sub-day/calendar filters
      .passes_filters <- function(dt) {
        if (!all(0:23 %in% private$hours) && !lubridate::hour(dt) %in% private$hours) return(FALSE)
        if (!all(1:7 %in% private$days_of_week) && !lubridate::wday(dt, week_start = 1) %in% private$days_of_week) return(FALSE)
        if (!all(1:31 %in% private$days_of_month) && !lubridate::mday(dt) %in% private$days_of_month) return(FALSE)
        if (!all(1:12 %in% private$months) && !lubridate::month(dt) %in% private$months) return(FALSE)
        TRUE
      }

      prev <- .prev_on_cycle(
        start_time,
        current = check_datetime,
        amount = private$frequency_n,
        unit = private$frequency_unit
      )

      if (is.na(prev)) {
        # Could be: (a) pipeline hasn't started yet, or (b) check_datetime is
        # exactly on a cycle point (epsilon in .prev_on_cycle pushes it to NA).\
        # Detect case (b) by nudging forward 1 second.
        prev_eps <- .prev_on_cycle(
          start_time,
          current = check_datetime + lubridate::seconds(1),
          amount = private$frequency_n,
          unit = private$frequency_unit
        )
        if (!is.na(prev_eps) && as.integer(prev_eps) == as.integer(check_datetime)) {
          # check_datetime is exactly on a cycle point — treat it as prev
          prev <- prev_eps
        } else {
          # Pipeline truly hasn't started yet
          private$next_run <- start_time
          return(FALSE)
        }
      }

      # .prev_on_cycle returns strictly before check_datetime.
      # The current cycle slot could be either prev (when the orch frequency is
      # finer than the pipeline frequency and prev rounds into the current slot)
      # or prev + one_step (when the pipeline runs exactly at check_datetime).
      # Check both candidates and use whichever matches.
      current_cycle <- lubridate::as_datetime(prev + .one_freq_step())

      prev_round <- timechange::time_round(prev, unit = orch_string)
      current_cycle_round <- timechange::time_round(current_cycle, unit = orch_string)

      as_comparable <- function(dt) as.integer(dt)
      check_int <- as_comparable(check_datetime_round)

      matched_slot <- if (as_comparable(prev_round) == check_int) {
        prev
      } else if (as_comparable(current_cycle_round) == check_int) {
        current_cycle
      } else {
        NULL
      }

      is_scheduled_now <- !is.null(matched_slot) && .passes_filters(matched_slot)

      next_seq_start <- if (!is.null(matched_slot) && identical(as.integer(matched_slot), as.integer(current_cycle))) {
        current_cycle + .one_freq_step()
      } else {
        current_cycle
      }

      next_seq <- get_pipeline_run_sequence(
        pipeline_n = private$frequency_n,
        pipeline_unit = private$frequency_unit,
        pipeline_datetime = next_seq_start,
        check_datetime = next_seq_start + lubridate::days(.run_sequence_min_days_out(private$frequency_unit)),
        pipeline_hours = private$hours,
        pipeline_days_of_week = private$days_of_week,
        pipeline_days_of_month = private$days_of_month,
        pipeline_months = private$months
      )

      private$next_run <- if (length(next_seq) > 0) {
        timechange::time_round(next_seq[[1]], unit = orch_string)
      } else {
        NULL
      }

      return(is_scheduled_now)
    },

    #' @description
    #' Get status of the pipeline as a data.frame
    #' @return data.frame
    get_status = function() {
      dplyr::tibble(
        pipe_name = private$pipe_name,
        script_path = private$script_path,
        private$run_time_attributes,
        next_run = private$next_run
      )
    },

    #' @description
    #' Get status of the pipeline as a string
    #' @return character
    get_status_chr = function() {
      private$status
    },

    #' @description
    #' Names of pipelines that receive input from this pipeline
    #' @return character
    get_outputs = function() {
      private$outputs
    },

    #' @description
    #' Names of pipelines that input into this pipeline
    #' @return character
    get_inputs = function() {
      private$inputs
    },

    #' @description
    #' Get priority of the pipeline
    #' @return numeric
    get_priority = function() {
      private$priority
    },

    #' @description
    #' Get immediate return values from the pipeline for downstream pipelines
    #' @return list
    get_returns = function() {
      private$returns
    },

    #' @description
    #' Get artifacts (return values) from the pipeline
    #' @return list
    get_artifacts = function() {
      if (length(private$run_time_artifacts) == 1) {
        return(private$run_time_artifacts[[1]])
      }
      private$run_time_artifacts
    },

    #' @description
    #' Get the number of completed artifact runs (iterations) for this pipeline
    #' @return integer
    get_n_artifacts = function() {
      length(private$run_time_artifacts)
    },

    #' @description
    #' Get the number of times this pipeline was invoked (successes + errors).
    #' For @maestroMap pipelines this equals the number of iterations that have
    #' finished, regardless of outcome.
    #' @return integer
    get_n_invocations = function() {
      sum(private$run_time_attributes$invoked, na.rm = TRUE)
    },

    #' @description
    #' Get all artifact values as a plain unnamed list, regardless of how many
    #' iterations have run. Used by collect() to gather each-pipe outputs.
    #' @return list
    get_all_returns = function() {
      unname(as.list(private$run_time_artifacts))
    },

    #' @description
    #' Get list of errors from the pipeline
    #' @return list
    get_errors = function() {
      private$errors
    },

    #' @description
    #' Get list of warnings from the pipeline
    #' @return list
    get_warnings = function() {
      private$warnings
    },

    #' @description
    #' Get list of messages from the pipeline
    #' @return list
    get_messages = function() {
      private$messages
    },

    #' @description
    #' Get the flags of a pipeline as a vector
    #' @return character
    get_flags = function() {
      private$flags
    },

    #' @description
    #' Get the labels of a pipeline as a data.frame
    #' @return data.frame
    get_labels = function() {
      purrr::map(private$labels, ~{
        dplyr::tibble(
          label = .x[[1]],
          value = .x[[2]]
        )
      }) |> 
        purrr::list_rbind()
    },

    #' @description
    #' Get whether the pipeline uses `collect` for fan in
    #' @return logical
    get_is_collect = function() {
      private$is_collect
    },

    #' @description
    #' Get the .input value to iterate over
    #' @return character
    get_map = function() {
      private$map
    },

    #' @description
    #' Get whether the pipeline uses an iterator for dynamic fan out
    #' @return logical
    get_is_map = function() {
      !is.null(private$map)
    },

    #' @description
    #' Get the number of iterations expected for this @maestroMap pipeline.
    #' Set by run_pipe() just before the scatter loop so that resolve_collect_input()
    #' can compare against get_n_invocations() without depending on the upstream
    #' return value's length (which is wrong when a field selector is used).
    #' @return integer or NULL
    get_n_expected_iterations = function() {
      private$n_expected_iterations
    },

    #' @description
    #' Record the number of iterations that will be dispatched for this @maestroMap pipeline.
    #' @param n integer
    #' @return invisible
    set_n_expected_iterations = function(n) {
      private$n_expected_iterations <- as.integer(n)
      invisible()
    },

    #' @description
    #' Update the inputs of a pipeline
    #' @param inputs character vector of inputting pipeline names
    #' @return vector
    update_inputs = function(inputs) {
      private$inputs <- inputs
    },

    #' @description
    #' Update the outputs of a pipeline
    #' @param outputs character vector of outputting pipeline names
    #' @return vector
    update_outputs = function(outputs) {
      private$outputs <- outputs
    },

    #' @description
    #' Resets run time attributes
    #' @return invisible
    reset_run_time_attributes = function() {
      private$run_time_attributes <- dplyr::tibble(
        internal_run_id = NA_character_,
        invoked = FALSE,
        success = NA,
        pipeline_started = lubridate::NA_POSIXct_,
        pipeline_ended = lubridate::NA_POSIXct_,
        errors = 0L,
        warnings = 0L,
        messages = 0L,
        run_id = NA_character_,
        input_run_id = NA_character_,
        lineage = NA_character_
      )
      private$status <- "Not Run"
      private$errors <- NULL
      private$warnings <- NULL
      private$messages <- NULL
      private$returns <- NULL
      private$run_time_start <- lubridate::NA_POSIXct_
      private$n_expected_iterations <- NULL
    },

    #' @description
    #' Get the run sequence of a pipeline
    #' @param n optional sequence limit
    #' @param min_datetime optional minimum datetime
    #' @param max_datetime optional maximum datetime
    #' @return vector
    get_run_sequence = function(
      n = NULL,
      min_datetime = NULL,
      max_datetime = NULL
    ) {
      if (!is.null(private$inputs)) return(NULL)

      now <- lubridate::now()

      anchor <- if (!is.null(min_datetime)) {
        lubridate::as_datetime(min_datetime)
      } else {
        now
      }

      extend_to <- if (!is.null(max_datetime)) {
        lubridate::as_datetime(max_datetime)
      } else {
        now + lubridate::days(.run_sequence_days_out(private$frequency_unit))
      }

      resolved_start <- private$resolve_start_time(anchor)

      start_time_adj <- .prev_on_cycle(
        resolved_start,
        current = anchor,
        amount = private$frequency_n,
        unit = private$frequency_unit
      )
      if (is.na(start_time_adj)) start_time_adj <- resolved_start

      seq <- get_pipeline_run_sequence(
        pipeline_n = private$frequency_n,
        pipeline_unit = private$frequency_unit,
        pipeline_datetime = start_time_adj,
        check_datetime = extend_to,
        pipeline_hours = private$hours,
        pipeline_days_of_week = private$days_of_week,
        pipeline_days_of_month = private$days_of_month,
        pipeline_months = private$months
      )

      if (!is.null(n)) {
        seq <- seq[seq_len(min(length(seq), as.integer(n)))]
      }

      if (!is.null(min_datetime)) {
        min_datetime <- lubridate::as_datetime(min_datetime)
        seq <- seq[seq >= min_datetime]
      }

      if (!is.null(max_datetime)) {
        max_datetime <- lubridate::as_datetime(max_datetime)
        seq <- seq[seq <= max_datetime]
      }

      seq
    }
  ),

  private = list(
    script_path = NA_character_,
    pipe_name = NA_character_,
    frequency = NA_character_,
    start_time_raw = NA_character_,
    tz = NA_character_,
    hours = NULL,
    months = NULL,
    skip = NA,
    log_level = NA_character_,
    inputs = NULL,
    outputs = NULL,
    priority = Inf,
    flags = c(),
    labels = list(),
    run_if = NULL,
    is_collect = FALSE,
    map = NULL,
    n_expected_iterations = NULL,

    # Transformed attributes
    days_of_week = NULL,
    days_of_month = NULL,
    frequency_n = NA_integer_,
    frequency_unit = NA_character_,

    # Dynamic attributes
    status = "Not Run",
    run_time_start = lubridate::NA_POSIXct_,
    returns = NULL,
    next_run = NULL,
    errors = NULL,
    warnings = NULL,
    messages = NULL,
    sourced_context = NULL,

    # Resolve start_time for a given reference datetime.
    # All formats are parsed on demand via parse_maestro_start_time:
    # - Relative formats (HH:MM:SS, weekday, month-day): now anchors the result
    # - Absolute formats (YYYY-...): now is ignored by parse_maestro_start_time
    # - NA (no @maestroStartTime tag): floor now to the pipeline's frequency
    resolve_start_time = function(now) {
      if (!is.na(private$start_time_raw)) {
        return(parse_maestro_start_time(
          private$start_time_raw,
          tz = private$tz,
          now = lubridate::with_tz(now, private$tz)
        ))
      }

      # No start_time tag: floor now to the pipeline's frequency granularity
      lubridate::floor_date(
        lubridate::with_tz(now, private$tz),
        unit = switch(
          private$frequency_unit %n% "day",
          "year" = , "quarter" = "year",
          "month" = "month",
          "week"  = "week",
          "day"
        )
      )
    },

    escape_for_glue = function(msg) {
      msg <- trimws(msg)
      logger::skip_formatter(msg)
    },

    run_time_attributes = dplyr::tibble(
      internal_run_id = NA_character_,
      invoked = FALSE,
      success = NA,
      pipeline_started = lubridate::NA_POSIXct_,
      pipeline_ended = lubridate::NA_POSIXct_,
      errors = 0L,
      warnings = 0L,
      messages = 0L,
      run_id = NA_character_,
      input_run_id = NA_character_,
      lineage = NA_character_
    ),

    run_time_artifacts = list(),

    insert_run_time_attributes = function(internal_run_id, attributes) {
      row_idx <- which(
        private$run_time_attributes$internal_run_id == internal_run_id
      )

      if (length(row_idx) == 0) {
        placeholder_idx <- which(is.na(
          private$run_time_attributes$internal_run_id
        ))

        if (length(placeholder_idx) > 0) {
          row_idx <- placeholder_idx[1]
          private$run_time_attributes$internal_run_id[
            row_idx
          ] <- internal_run_id
        } else {
          new_row <- dplyr::tibble(internal_run_id = internal_run_id)
          for (attr_name in names(attributes)) {
            new_row[[attr_name]] <- NA
          }
          private$run_time_attributes <- dplyr::bind_rows(
            private$run_time_attributes,
            new_row
          )
          row_idx <- nrow(private$run_time_attributes)
        }
      }

      for (i in seq_along(attributes)) {
        private$run_time_attributes[[names(attributes)[i]]][
          row_idx
        ] <- attributes[[i]]
      }
    },

    error_handler = function(internal_run_id = NULL) {
      function(e) {
        private$errors <- c(private$errors, e$message)
        private$status <- "Error"
        logger::log_error(
          private$escape_for_glue(conditionMessage(e)),
          namespace = private$pipe_name
        )
        run_time_end <- lubridate::now()
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            pipeline_ended = run_time_end,
            success = FALSE,
            errors = 1L
          )
        )
      }
    },

    warning_handler = function(internal_run_id = NULL) {
      function(w) {
        warning_log <- logger::log_warn(
          private$escape_for_glue(conditionMessage(w)),
          namespace = private$pipe_name
        )
        private$warnings <- c(private$warnings, w$message)
        private$status <- "Warning"
        cur_n_warnings <- private$run_time_attributes$warnings[
          private$run_time_attributes$internal_run_id == internal_run_id
        ] %n%
          0L
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            warnings = cur_n_warnings + 1L
          )
        )
        invokeRestart("muffleWarning")
      }
    },

    message_handler = function(internal_run_id = NULL) {
      function(m) {
        message_log <- logger::log_info(
          private$escape_for_glue(conditionMessage(m)),
          namespace = private$pipe_name
        )
        private$messages <- c(private$messages, m$message)
        cur_n_messages <- private$run_time_attributes$messages[
          private$run_time_attributes$internal_run_id == internal_run_id
        ] %n%
          0L
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            messages = cur_n_messages + 1L
          )
        )
        invokeRestart("muffleMessage")
      }
    },

    pre_error_handler = function(internal_run_id) {
      function(e) {
        e$message <- paste("Error before pipeline execution:", e$message)
        private$errors <- c(private$errors, e$message)
        private$status <- "Error"
        logger::log_error(
          private$escape_for_glue(e$message),
          namespace = private$pipe_name
        )
        run_time_end <- lubridate::now()
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            pipeline_ended = run_time_end,
            success = FALSE,
            errors = 1L
          )
        )
      }
    },

    cond_error_handler = function(internal_run_id) {
      function(e) {
        e$message <- paste("Error evaluating condition:", e$message)
        private$errors <- e
        private$status <- "Error"
        logger::log_error(
          private$escape_for_glue(
            "Error evaluating condition: {conditionMessage(e)}"
          ),
          namespace = private$pipe_name
        )
        run_time_end <- lubridate::now()
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            invoked = TRUE,
            pipeline_ended = run_time_end,
            errors = 1L
          )
        )
      }
    },

    cond_warning_handler = function(internal_run_id) {
      function(w) {
        warning_log <- logger::log_warn(
          private$escape_for_glue(
            "Warning evaluating condition: {conditionMessage(w)}"
          ),
          namespace = private$pipe_name
        )
        w$message <- paste("Warned while evaluating condition:", w$message)
        private$warnings <- c(private$warnings, w$message)
        private$status <- "Warning"
        cur_n_warnings <- private$run_time_attributes$warnings[
          private$run_time_attributes$internal_run_id == internal_run_id
        ] %n%
          0L
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            warnings = cur_n_warnings + 1L
          )
        )
        invokeRestart("muffleWarning")
      }
    },

    cond_message_handler = function(internal_run_id) {
      function(m) {
        message_log <- logger::log_info(
          private$escape_for_glue(conditionMessage(m)),
          namespace = private$pipe_name
        )
        private$messages <- c(private$messages, m$message)
        cur_n_messages <- private$run_time_attributes$messages[
          private$run_time_attributes$internal_run_id == internal_run_id
        ] %n%
          0L
        private$insert_run_time_attributes(
          internal_run_id,
          list(
            messages = cur_n_messages + 1L
          )
        )
        invokeRestart("muffleMessage")
      }
    }
  )
)

test_that("collect() with fewer than two inputs errors at build_schedule()", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      src_a <- function() 'a'

      #' @maestroInputs collect(src_a)
      downstream <- function(.input) .input$src_a
      ",
      con = "pipelines/bad_collect.R"
    )
    expect_error(build_schedule(), regexp = "collect\\(\\).*at least two input pipelines or a single")
  })
})

test_that("Simple collect from distinct upstreams", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      letter_a <- function() {
        'a'
      }

      #' @maestroFrequency daily
      letter_b <- function() {
        'b'
      }
        
      #' @maestroInputs collect(letter_a, letter_b)
      ab <- function(.input) {
        paste0(.input$letter_a, .input$letter_b)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
  })

  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$ab)), "ab")
})

test_that("Simple collect from distinct upstreams - single error case", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      letter_a <- function() {
        'a'
      }

      #' @maestroFrequency daily
      letter_b <- function() {
        stop()
      }
        
      #' @maestroInputs collect(letter_a, letter_b)
      ab <- function(.input) {
        paste0(.input$letter_a, .input$letter_b)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
  })

  status <- get_status(schedule)
  expect_equal(status$invoked, c(TRUE, TRUE, FALSE))
})

test_that("Collect does not run when a non-each input errored", {

  # This tests gap #2: the non_each_pending guard uses != "Not Run", so an
  # errored input passes through and collect would run with NULL as that input.
  # A single primary fans out to two branches (so both are always visited before
  # collect is evaluated, removing any traversal-order ambiguity).
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      source <- function() {
        'a'
      }

      #' @maestroInputs source
      branch_ok <- function(.input) {
        paste0(.input, 'b')
      }

      #' @maestroInputs source
      branch_err <- function(.input) {
        stop('deliberate error')
      }

      #' @maestroInputs collect(branch_ok, branch_err)
      collector <- function(.input) {
        paste0(.input$branch_ok, .input$branch_err)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(schedule)
  })

  status <- get_status(schedule)
  expect_false(status$invoked[status$pipe_name == "collector"])
})

test_that("Simple collect from distinct upstreams - multi-output from upstream", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      letter_a <- function() {
        'a'
      }

      #' @maestroFrequency daily
      letter_b <- function() {
        list(
          letter = 'b',
          another_var = 1:4
        )
      }
        
      #' @maestroInputs collect(letter_a, letter_b)
      ab <- function(.input) {
        paste0(.input$letter_a, .input$letter_b$letter, .input$letter_b$another_var)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
  })

  status <- get_status(schedule)
  expect_all_equal(status$invoked, TRUE)
  expect_all_equal(status$success, TRUE)
  expect_equal(get_artifacts(schedule)$ab, c("ab1", "ab2", "ab3", "ab4"))
})

test_that("Simple collect from distinct upstreams with conditional on the collect", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      letter_a <- function() {
        'a'
      }

      #' @maestroFrequency daily
      letter_b <- function() {
        'b'
      }
        
      #' @maestroInputs collect(letter_a, letter_b)
      #' @maestroRunIf .input$letter_a == 'a' && .input$letter_b == 'b'
      ab <- function(.input) {
        paste0(.input$letter_a, .input$letter_b)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
  })

  status <- get_status(schedule)
  expect_all_equal(status$invoked, TRUE)
  expect_all_equal(status$success, TRUE)

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      letter_a <- function() {
        'a'
      }

      #' @maestroFrequency daily
      letter_b <- function() {
        'b'
      }
        
      #' @maestroInputs collect(letter_a, letter_b)
      #' @maestroRunIf .input$letter_a == 'a' && .input$letter_b == 'c'
      ab <- function(.input) {
        paste0(.input$letter_a, .input$letter_b)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
  })

  status <- get_status(schedule)
  expect_equal(status$invoked, c(TRUE, TRUE, FALSE))
})

test_that("Complex collect from distinct upstreams where an upstream is triggered twice", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      start <- function() {
        'a'
      }

      #' @maestroInputs start
      mid1 <- function(.input) {
        paste0(.input, 'b')
      }

      #' @maestroInputs start
      mid2 <- function(.input) {
        paste0(.input, 'c')
      }

      #' @maestroInputs mid1 mid2
      end <- function(.input) {
        paste0(.input, 'd')
      }

      #' @maestroInputs collect(mid2, end)
      collector <- function(.input) {
        paste0(.input$mid2, .input$end)
      }",
      con = "pipelines/fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
  })

  status <- get_status(schedule)
  expect_length(status$pipe_name[status$pipe_name == "collector"], 1)
})


test_that("Collect pipeline with its own downstream output runs the downstream", {

  schedule <- build_schedule(test_path("test_pipelines_fan_in_chained"))
  run_schedule(schedule)

  status <- get_status(schedule)
  expect_true(all(status$invoked))
  expect_true(all(status$success))
  expect_equal(get_artifacts(schedule)$ab_upper, "AB")
})

test_that("Dynamic fan out followed by fan in", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      numbers <- function() {
        1:3
      }

      #' @maestroInputs numbers
      #' @maestroMap
      multiply <- function(.input) {
        .input * 3
      }
      
      #' @maestroInputs collect(multiply)
      add <- function(.input) {
        sum(unlist(.input))
      }
      ",
      con = "pipelines/fanout-fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
  })
  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$add)), 18)
})

test_that("Dynamic fan out followed by fan in - partial error case", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      numbers <- function() {
        1:3
      }

      #' @maestroInputs numbers
      #' @maestroMap
      multiply <- function(.input) {
        if (.input == 2) stop()
        .input * 3
      }
      
      #' @maestroInputs collect(multiply)
      add <- function(.input) {
        sum(unlist(.input))
      }
      ",
      con = "pipelines/fanout-fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
  })
  status <- get_status(schedule)
  # multiply runs 3 times (1 error on input=2, 2 successes); add proceeds with the successful results
  expect_true(status$invoked[status$pipe_name == "add"])
  expect_true(status$success[status$pipe_name == "add"])
  # 1*3 + 3*3 = 3 + 9 = 12
  expect_equal(unlist(unname(get_artifacts(schedule)$add)), 12)
})

test_that("Dynamic fan out followed by fan in - all error case", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      numbers <- function() {
        1:3
      }

      #' @maestroInputs numbers
      #' @maestroMap
      multiply <- function(.input) {
        stop()
      }
      
      #' @maestroInputs collect(multiply)
      add <- function(.input) {
        sum(unlist(.input))
      }
      ",
      con = "pipelines/fanout-fanin.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
  })
  status <- get_status(schedule)
  expect_false(status$invoked[status$pipe_name == "add"])
})

test_that("Fan-out collect with iterateOver", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      get_letters <- function() {
        list(
          letter = letters[1:3],
          greeting = 'hello'
        )
      }

      #' @maestroInputs get_letters
      #' @maestroMap .input$letter
      make_message <- function(.input) {
        paste(.input$greeting, toupper(.input$letter))
      }
      
      #' @maestroInputs collect(make_message)
      capitalize <- function(.input) {
        paste(toupper(unlist(.input)))
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
  })
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$capitalize)), paste("HELLO", c("A", "B", "C")))
})

test_that("Possibly conflicting info from maestroInputs collect and maestroOutputs", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      letter_a <- function() {
        'a'
      }

      #' @maestroFrequency daily
      #' @maestroOutputs ab
      letter_b <- function() {
        'b'
      }
        
      #' @maestroInputs collect(letter_a, letter_b)
      ab <- function(.input) {
        paste0(.input$letter_a, .input$letter_b)
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
  })

  expect_equal(unlist(unname(get_artifacts(schedule)$ab)), "ab")
})

test_that("Simple pipeline, no errors", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )

  pipeline <- pipeline_list$MaestroPipelines[[1]]
  expect_s3_class(pipeline, "MaestroPipeline")

  expect_null(pipeline$get_outputs())
  expect_null(pipeline$get_inputs())

  pipeline$run(quiet = TRUE)

  expect_snapshot(pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")])
})

test_that("Pipeline with warnings", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines_run_two_warnings/pipe_warning.R")
  )
  pipeline <- pipeline_list$MaestroPipelines[[1]]

  pipeline$run(quiet = TRUE)

  expect_snapshot(pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")])
})

test_that("Pipeline with errors", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines_run_some_errors/pipe2.R")
  )
  pipeline <- pipeline_list$MaestroPipelines[[1]]

  expect_error(pipeline$run(quiet = TRUE))

  expect_snapshot(pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")])
})

test_that("Pipeline with arguments are correctly passed", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines_run_args_good/pipe1.R")
  )

  pipeline <- pipeline_list$MaestroPipelines[[1]]

  expect_snapshot(pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")])

  pipeline$run(resources = list(
    vals = 1:5
  ), quiet = TRUE)

  expect_snapshot(pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")])
})

test_that("Logging threshold", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines_run_logs_warn/pipe_warning.R")
  )

  pipeline <- pipeline_list$MaestroPipelines[[1]]

  withr::with_tempfile("log", {

    pipeline$run(
      log_file = log,
      quiet = TRUE
    )

    logs <- readLines(log)
    expect_true(!all(grepl("INFO", logs)))
    expect_true(any(grepl("WARN", logs)))
  })

  withr::with_tempfile("log", {
    pipeline$run(
      log_file = log,
      log_file_max_bytes = 1000,
      quiet = TRUE
    )

    expect_lte(file.size(log), 1000 + 100) # margin of error
  })
})

test_that("can create a schedule entry from a single well-documented fun", {
  res <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )
  expect_s3_class(res, "MaestroPipelineList")
})

test_that("can create a schedule entry from a default tagged fun", {
  res <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_default.R")
  )
  expect_s3_class(res, "MaestroPipelineList")
  expect_equal(length(res$MaestroPipelines), 1)
})

test_that("invalid tags trigger error", {
  build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_bad.R")
  ) |>
    expect_error(regexp = "Invalid maestroFrequency")
}) |>
  suppressMessages()

test_that("can create a schedule entry from a multi-function script", {
  res <- build_schedule_entry(
    test_path("test_pipelines/test_multi_fun_pipeline.R")
  )
  expect_s3_class(res, "MaestroPipelineList")
  expect_equal(length(res$MaestroPipelines), 2)
}) |>
  suppressMessages()

test_that("Errors on a pipeline with no tagged functions", {
  build_schedule_entry(
    test_path("test_pipelines_parse_all_bad/test_pipeline_no_func.R")
  ) |>
    expect_error(regexp = "tags present in")
}) |>
  suppressMessages()

test_that("Errors on pipeline that doesn't use functions and returns null", {

  expect_error({
    res <- build_schedule_entry(
      test_path("test_pipelines_parse_all_bad/tagged_but_no_func.R")
    )
  }, regexp = "has tags but no function")
}) |>
  suppressMessages()

test_that("Errors on pipeline with isolated tags", {
  expect_error({
    res <- build_schedule_entry(
      test_path("test_pipelines_parse_all_bad/tagged_but_no_func_multi.R")
    )
  }, regexp = "has tags but no function")
})

test_that("Informative error on parsing schedule entry with an error", {
  expect_error({
    res <- build_schedule_entry(
      test_path("test_pipelines_parse_all_bad/test_pipeline_script_with_error.R")
    )
  }, regexp = "Could not build")
})

test_that("Script with untagged function isn't treated as a scheduled pipeline", {
  res <- build_schedule_entry(
    test_path("test_pipelines_parse_all_good/pipe_with_custom_fun.R")
  )
  expect_equal(length(res$MaestroPipelines), 1)
})

test_that("Script with good specifiers parses well", {
  expect_no_error({
    schedule <- build_schedule_entry(
      test_path("test_pipelines_parse_all_good/specifiers.R")
    )
  })
})

test_that("Script with bad specifiers errors", {
  schedule <- build_schedule_entry(
    test_path("test_pipelines_parse_all_bad/specifiers.R")
  ) |>
    expect_error(regexp = "pipeline must have")
})

test_that("Pipeline with inputs checks for .input", {
  schedule <- build_schedule_entry(
    test_path("test_pipelines_parse_all_bad/dags.R")
  ) |>
    expect_error(regexp = "pipeline must have a parameter")

  expect_no_error({
    schedule <- build_schedule_entry(
      test_path("test_pipelines_dags_good/dags.R")
    )
  })
})

test_that("Errors if pipeline self-references as input", {
  schedule <- build_schedule_entry(
    test_path("test_pipelines_parse_all_bad/dags_self_reference_inputs.R")
  ) |>
    expect_error(regexp = "cannot contain self-references")
})

test_that("Errors if pipeline self-references as output", {
  schedule <- build_schedule_entry(
    test_path("test_pipelines_parse_all_bad/dags_self_reference_outputs.R")
  ) |>
    expect_error(regexp = "cannot contain self-references")
})

test_that("Pipelines with maestro tag get picked up", {
  res <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_maestro.R")
  )
  expect_equal(length(res$MaestroPipelines), 1)
})

test_that("Pipelines with maestroHours and maestroFrequency = 1 hour are valid", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2025-01-01 00:00:00
      #' @maestroHours 0 4 7 12 18
      hourly <- function() {

      }
      ",
      con = "pipelines/hourly.R"
    )

    expect_no_error({
      build_schedule_entry("pipelines/hourly.R")
    })
  })
})

test_that("Pipelines with maestroStartTime in HH:MM:SS and a multi-unit frequency fail", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency 2 days
      #' @maestroStartTime 00:00:00
      hourly <- function() {

      }
      ",
      con = "pipelines/hourly.R"
    )

    expect_error({
      build_schedule_entry("pipelines/hourly.R")
    }, regexp = "Cannot use a `@maestroStartTime`")
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency weekly
      #' @maestroStartTime 00:00:00
      hourly <- function() {

      }
      ",
      con = "pipelines/hourly.R"
    )

    expect_error({
      build_schedule_entry("pipelines/hourly.R")
    }, regexp = "`@maestroStartTime`")
  })
})

test_that("Minutely frequency can be combined with maestroHours and other specifiers", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 10 minutes
      #' @maestroHours 5
      minutely <- function() {

      }
      ",
      con = "pipelines/minutely.R"
    )

    expect_no_error(build_schedule_entry("pipelines/minutely.R"))
  })
})

test_that("maestroMap with multiple space-separated expressions is stored as character vector", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      c(
        "#' @maestroFrequency daily",
        "upstream <- function() {",
        "  list(ids = 1:3, labels = c('a', 'b', 'c'))",
        "}",
        "",
        "#' @maestroInputs upstream",
        "#' @maestroMap .input$ids .input$labels",
        "downstream <- function(.input) {",
        "  paste(.input$ids, .input$labels)",
        "}"
      ),
      con = "pipelines/multi_iter.R"
    )

    entry <- build_schedule_entry("pipelines/multi_iter.R")
    map <- entry$MaestroPipelines[[2]]$get_map()
    expect_equal(map, c(".input$ids", ".input$labels"))
    expect_length(map, 2L)
  })
})

test_that("maestroMap with invalid R expression errors at build_schedule_entry", {

  withr::with_tempdir({
    dir.create("pipelines")
    # Write the file directly to avoid roxygen2 interpreting escape sequences
    # in a roxygen comment string. Use a genuinely unparseable fragment.
    writeLines(
      c(
        "#' @maestroFrequency daily",
        "upstream <- function() 1:3",
        "",
        "#' @maestroInputs upstream",
        "#' @maestroMap .input$ids 1+",
        "downstream <- function(.input) .input"
      ),
      con = "pipelines/bad_iter.R"
    )

    expect_error(
      build_schedule_entry("pipelines/bad_iter.R"),
      regexp = "invalid R expression"
    )
  })
})

test_that("Simple fan out with no specified iterator", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      numbers <- function() {
        1:3
      }

      #' @maestroInputs numbers
      #' @maestroMap
      multiply <- function(.input) {
        .input * 3
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
  })

  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$multiply)), c(3, 6, 9))
  expect_equal(length(unique(status$run_id)), length(status$run_id))
})

test_that("Simple fan out into common downstream", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      numbers <- function() {
        1:3
      }

      #' @maestroInputs numbers
      #' @maestroMap
      multiply <- function(.input) {
        .input * 3
      }
        
      #' @maestroInputs multiply
      add_2 <- function(.input) {
        .input + 2
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
  })
  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$add_2)), c(5, 8, 11))
  expect_equal(length(unique(status$run_id)), length(status$run_id))
})

test_that("Simple fan out into common downstream and one error", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      numbers <- function() {
        1:3
      }

      #' @maestroInputs numbers
      #' @maestroMap
      multiply <- function(.input) {
        if (.input == 1) stop()
        .input * 3
      }
        
      #' @maestroInputs multiply
      add_2 <- function(.input) {
        .input + 2
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
    status <- get_status(schedule)
  })
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$add_2)), c(8, 11))
  expect_equal(length(unique(status$run_id)), length(status$run_id))
})

test_that("Use maestroMap to specify a particular iteration variable", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      get_letters <- function() {
        list(
          letter = letters[1:3],
          greeting = 'hello'
        )
      }

      #' @maestroInputs get_letters
      #' @maestroMap .input$letter
      make_message <- function(.input) {
        paste(.input$greeting, toupper(.input$letter))
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
    status <- get_status(schedule)
  })
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$make_message)), paste("hello", c("A", "B", "C")))
  expect_equal(length(unique(status$run_id)), length(status$run_id))
})

test_that("Fan out where upstream returns a list of S3 objects (lm)", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      fit_models <- function() {
        list(
          lm(mpg ~ wt, data = mtcars),
          lm(mpg ~ hp, data = mtcars),
          lm(mpg ~ cyl, data = mtcars)
        )
      }

      #' @maestroInputs fit_models
      #' @maestroMap
      extract_r2 <- function(.input) {
        summary(.input)$r.squared
      }",
      con = "pipelines/fanout_lm.R"
    )

    schedule <- build_schedule()
    run_schedule(schedule, orch_frequency = "1 day")
    status <- get_status(schedule)
  })

  r2_values <- unlist(unname(get_artifacts(schedule)$extract_r2))
  expect_length(r2_values, 3)
  expect_true(all(r2_values > 0 & r2_values <= 1))
  expect_snapshot(status[, c("invoked", "success")])
})

test_that("Fan where the field contains S3 objects (lm)", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      fit_models <- function() {
        list(
          model = list(
            lm(mpg ~ wt, data = mtcars),
            lm(mpg ~ hp, data = mtcars),
            lm(mpg ~ cyl, data = mtcars)
          ),
          label = c('wt', 'hp', 'cyl')
        )
      }

      #' @maestroInputs fit_models
      #' @maestroMap .input$model
      extract_r2 <- function(.input) {
        summary(.input$model)$r.squared
      }",
      con = "pipelines/fanout_lm.R"
    )

    schedule <- build_schedule()
    run_schedule(schedule, orch_frequency = "1 day")
    status <- get_status(schedule)
  })

  r2_values <- unlist(unname(get_artifacts(schedule)$extract_r2))
  expect_length(r2_values, 3)
  expect_true(all(r2_values > 0 & r2_values <= 1))
  expect_snapshot(status[, c("invoked", "success")])
})

test_that("map is misspecified name in return", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      get_letters <- function() {
        list(
          letter = letters[1:3],
          greeting = 'hello'
        )
      }

      #' @maestroInputs get_letters
      #' @maestroMap .input$asd
      make_message <- function(.input) {
        paste(.input$greeting, toupper(.input$letter))
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
    status <- get_status(schedule)
  })
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(last_run_errors()$make_message, "Error before pipeline execution: Field(s) 'asd' specified in @maestroMap not found in the output of the upstream pipeline.")
})

test_that("maestroMap with unequal length vectors errors", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      get_data <- function() {
        list(
          ids = 1:3,
          labels = c('a', 'b')
        )
      }

      #' @maestroInputs get_data
      #' @maestroMap .input$ids .input$labels
      process <- function(.input) {
        paste(.input$ids, .input$labels)
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(schedule, orch_frequency = "1 day")
    status <- get_status(schedule)
  })
  expect_match(
    last_run_errors()$process,
    regexp = "same length or length 1"
  )
})

test_that("Use maestroMap with multiple iterators", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      get_letters <- function() {
        list(
          letter = letters[1:3],
          greeting = c('hello', 'cheers', 'hi')
        )
      }

      #' @maestroInputs get_letters
      #' @maestroMap .input$letter .input$greeting
      make_message <- function(.input) {
        paste(.input$greeting, toupper(.input$letter))
      }",
      con = "pipelines/fanout.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule, orch_frequency = "1 day"
    )
    status <- get_status(schedule)
  })
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$make_message)), c("hello A", "cheers B", "hi C"))
  expect_equal(length(unique(status$run_id)), length(status$run_id))
})

#' Retrieve latest maestro build errors
#'
#' Gets the latest schedule build errors following use of `build_schedule()`. If
#' the build succeeded or `build_schedule()` has not been run it will be `NULL`.
#'
#' @return error messages
#' @export
#' @examples
#'
#' last_build_errors()
last_build_errors <- function() {
  maestro_pkgenv$last_build_errors
}

#' Retrieve latest maestro pipeline errors
#'
#' Gets the latest pipeline errors following use of `run_schedule()`. If
#' the all runs succeeded or `run_schedule()` has not been run it will be `NULL`.
#'
#' @return error messages
#' @export
#' @examples
#'
#' last_run_errors()
last_run_errors <- function() {
  maestro_pkgenv$last_run_errors
}

#' Retrieve latest maestro pipeline warnings
#'
#' Gets the latest pipeline warnings following use of `run_schedule()`. If
#' there are no warnings or `run_schedule()` has not been run it will be `NULL`.
#'
#' Note that setting `maestroLogLevel` to something greater than `WARN` will
#' ignore warnings.
#'
#' @return warning messages
#' @export
#' @examples
#'
#' last_run_warnings()
last_run_warnings <- function() {
  maestro_pkgenv$last_run_warnings
}

#' Retrieve latest maestro pipeline messages
#'
#' Gets the latest pipeline messages following use of `run_schedule()`. If
#' there are no messages or `run_schedule()` has not been run it will be `NULL`.
#'
#' Note that setting `maestroLogLevel` to something greater than `INFO` will
#' ignore messages.
#'
#' @return messages
#' @export
#' @examples
#'
#' last_run_messages()
last_run_messages <- function() {
  maestro_pkgenv$last_run_messages
}

test_that("run_schedule works on different kinds of frequencies", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good"))

  test_freqs <- c("14 days", "10 minutes", "25 mins", "1 week",
                  "1 quarter", "12 months", "1 years", "24 hours",
                  "31 days", "1 secs", "50 seconds", "daily", "hourly", "weekly", "yearly")

  purrr::walk(test_freqs, ~{
    expect_no_error({
      run_schedule(
        schedule,
        orch_frequency = .x
      ) |>
        suppressWarnings()
    })
  })
}) |>
  suppressMessages()

test_that("run_schedule errors if check_datetime is not a timestamp", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good")) |>
    suppressMessages()

  expect_error(
    run_schedule(
      schedule,
      check_datetime = "a"
    ),
    regexp = "must be a"
  )

  # But Dates work
  expect_no_error(
    run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.Date("2024-01-01")
    )
  )
}) |>
  suppressMessages()

test_that("maestro.check_datetime_override option overrides check_datetime", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good")) |>
    suppressMessages()

  override_dt <- as.POSIXct("2025-06-01 09:00:00", tz = "UTC")

  withr::with_options(
    list(maestro.check_datetime_override = override_dt),
    {
      # Argument value is ignored; option takes precedence without error
      expect_no_error(
        run_schedule(
          schedule,
          orch_frequency = "hourly",
          check_datetime = Sys.time()
        ) |>
          suppressMessages()
      )
    }
  )
})

test_that("maestro.check_datetime_override errors if not POSIXct", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good")) |>
    suppressMessages()

  withr::with_options(
    list(maestro.check_datetime_override = "not-a-datetime"),
    {
      expect_error(
        run_schedule(schedule, orch_frequency = "hourly") |>
          suppressMessages(),
        regexp = "must be a"
      )
    }
  )
})

test_that("run_schedule with quiet=TRUE prints no messages", {
  schedule <- build_schedule(test_path("test_pipelines_run_all_good")) |>
    suppressMessages()

  expect_no_message({
    run_schedule(schedule, orch_frequency = "hourly", run_all = TRUE, quiet = TRUE)
  })

  expect_type(last_run_messages(), "list")
  expect_true(rlang::is_named(last_run_messages()))
  expect_gt(length(last_run_messages()), 0)
})

test_that("run_schedule is idempotent", {
  
  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)

  run_schedule(
    schedule,
    orch_frequency = "15 minutes",
    check_datetime = as.POSIXct("2024-04-25 09:35:00", tz = "UTC"),
    quiet = TRUE
  )

  status1 <- schedule$get_status()

  run_schedule(
    schedule,
    orch_frequency = "15 minutes",
    check_datetime = as.POSIXct("2024-04-25 09:45:00", tz = "UTC"),
    quiet = TRUE
  )

  status2 <- schedule$get_status()

  expect_equal(nrow(status1), nrow(status2))
})

test_that("run_schedule timeliness checks - pipelines run when they're supposed to", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroStartTime 2024-01-01
      #' @maestroTz America/Halifax
      get_mtcars <- function() { mtcars }

      #' @maestroFrequency 3 months
      #' @maestroStartTime 2024-01-01
      wait <- function() {}

      #' @maestroFrequency 1 week
      #' @maestroStartTime 2024-01-01
      weekly <- function() {}

      #' @maestroFrequency 1 day
      #' @maestroStartTime 2024-01-01
      way_in_the_future <- function() {}

      #' @maestroFrequency weekly
      #' @maestroStartTime 2024-01-01
      weekly2 <- function() {}

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-03-01 09:35:00
      #' @maestroTz UTC
      get_mtcars2 <- function() { mtcars }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    output <- run_schedule(
      schedule,
      orch_frequency = "15 minutes",
      check_datetime = as.POSIXct("2024-04-25 09:35:00", tz = "UTC"),
      quiet = TRUE
    )
    expect_equal(output$get_status()$invoked, c(FALSE, FALSE, FALSE, FALSE, FALSE, TRUE))

    schedule <- build_schedule(quiet = TRUE)
    output <- run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-03-02 09:00:00", tz = "America/Halifax"),
      quiet = TRUE
    )
    expect_equal(output$get_status()$invoked, c(FALSE, FALSE, FALSE, FALSE, FALSE, TRUE))
  })
})

test_that("run_schedule timeliness checks - specifiers (e.g., hours, days, months)", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      #' @maestroStartTime 2024-01-01
      #' @maestroDays 1 8 24 28
      specific_days <- function() {}

      #' @maestroFrequency hourly
      #' @maestroStartTime 2024-01-01
      #' @maestroHours 0 4 8 12 16 20
      specific_hours <- function() {}

      #' @maestroFrequency biweekly
      #' @maestroStartTime 2024-01-01
      #' @maestroMonths 1 5 10
      specific_months1 <- function() {}

      #' @maestroFrequency hourly
      #' @maestroStartTime 2024-01-01
      #' @maestroDays Mon Wed Fri
      specific_days2 <- function() {}

      #' @maestroFrequency hourly
      #' @maestroStartTime 2024-01-01
      #' @maestroDays Sun Sat
      specific_days3 <- function() {}

      #' @maestroFrequency monthly
      #' @maestroStartTime 2024-01-01
      #' @maestroMonths 1 5 10
      specific_months3 <- function() {}

      #' @maestroFrequency hourly
      #' @maestroStartTime 2024-01-01
      #' @maestroHours 1 12
      #' @maestroDays 1
      #' @maestroMonths 1 5 10
      specific_multi <- function() {}
      ",
      con = "pipelines/specifiers.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    output <- run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-04-01 00:00:00", tz = "UTC"), # Monday
      quiet = TRUE
    )
    expect_equal(
      output$get_status()$invoked,
      c(TRUE, TRUE, FALSE, TRUE, FALSE, FALSE, FALSE)
    )

    schedule <- build_schedule(quiet = TRUE)
    output <- run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-05-01 00:00:00", tz = "UTC"), # Wednesday
      quiet = TRUE
    )
    expect_equal(
      output$get_status()$invoked,
      c(TRUE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE)
    )
  })
})

test_that("run_schedule timeliness checks - specifiers on a non UTC timezone", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency hourly
      #' @maestroStartTime 2025-01-01 00:00:00
      #' @maestroHours 0 4 7 12 18
      non_utc_hours <- function() {

      }
      ",
      con = "pipelines/non_utc_hours.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-01-01 07:00:00", tz = "America/Halifax"),
      quiet = TRUE
    )

    status <- get_status(schedule)
    expect_true(status$invoked[status$pipe_name == "non_utc_hours"])

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-01-01 07:00:00", tz = "UTC"),
      quiet = TRUE
    )

    status <- get_status(schedule)
    expect_false(status$invoked[status$pipe_name == "non_utc_hours"])
  })
})

test_that("run_schedule timeliness checks - non UTC daylight savings ", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency daily
      #' @maestroStartTime 2025-03-07 00:00:00
      non_utc <- function() {

      }
      ",
      con = "pipelines/non_utc.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-03-10 00:00:00", tz = "America/Halifax"),
      quiet = TRUE
    )

    status <- get_status(schedule)
    expect_true(status$invoked[status$pipe_name == "non_utc"])
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency hourly
      #' @maestroStartTime 2025-03-07 00:00:00
      #' @maestroHours 4
      non_utc <- function() {

      }
      ",
      con = "pipelines/non_utc.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-03-10 04:00:00", tz = "America/Halifax"),
      quiet = TRUE
    )

    status <- get_status(schedule)
    expect_true(status$invoked[status$pipe_name == "non_utc"])
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency hourly
      #' @maestroStartTime 2025-03-07 00:00:00
      non_utc <- function() {

      }
      ",
      con = "pipelines/non_utc.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-03-08 04:00:00", tz = "America/Halifax"),
      quiet = TRUE
    )

    status <- get_status(schedule)
    expect_true(status$invoked[status$pipe_name == "non_utc"])
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroTz America/Halifax
      #' @maestroFrequency hourly
      #' @maestroStartTime 2025-03-07 00:00:00
      non_utc <- function() {

      }
      ",
      con = "pipelines/non_utc.R"
    )

    schedule <- build_schedule(quiet = TRUE)

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-03-11 00:00:00", tz = "UTC"),
      quiet = TRUE
    )

    status <- get_status(schedule)
    expect_true(status$invoked[status$pipe_name == "non_utc"])
  })
})

test_that("run_schedule propagates warnings", {

  schedule <- build_schedule(test_path("test_pipelines_run_two_warnings"))

  expect_message({
    run_schedule(
      schedule,
      run_all = TRUE
    )
  })
  expect_gt(length(last_run_warnings()), 0)
}) |>
  suppressMessages()

test_that("run_schedule handles errors in a pipeline", {

  schedule <- build_schedule(test_path("test_pipelines_run_some_errors"))

  withr::with_tempfile("log", {
    expect_message({
      output <- run_schedule(schedule, run_all = TRUE, log_to_file = log)
    })
    status <- output$get_status()

    log_len_t1 <- length(readLines(log))
    expect_gt(log_len_t1, 0)

    errors <- last_run_errors()
    expect_type(errors, "list")
    expect_length(errors, 1)
    expect_true(all(!is.na(status$pipeline_started)))

    # Run again to ensure log file is appended to
    run_schedule(schedule, run_all = TRUE, log_to_file = log)
    log_len_t2 <- length(readLines(log))
    expect_gt(log_len_t2, log_len_t1)
  })
}) |>
  suppressMessages()

test_that("run_schedule creates maestro.log if log_to_file is TRUE", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      test <- function() {

      }
      ",
      con = "pipelines/test.R"
    )
    schedule <- build_schedule()
    run_schedule(schedule, run_all = TRUE, log_to_file = TRUE)
    expect_true(file.exists("maestro.log"))
  })
}) |>
  suppressMessages()

test_that("run_schedule doesn't create maestro.log if log_to_file is FALSE", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency hourly
      test <- function() {

      }
      ",
      con = "pipelines/test.R"
    )
    schedule <- build_schedule()
    run_schedule(schedule, run_all = TRUE, log_to_file = FALSE)
    expect_true(!file.exists("maestro.log"))
  })
}) |>
  suppressMessages()

test_that("run_schedule checks type of schedule", {
  expect_error(
    run_schedule(iris, run_all = TRUE),
    regexp = "Schedule must be an object of"
  )
}) |>
  suppressMessages()

test_that("errors on invalid orch_frequency", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)

  expect_error({
    run_schedule(schedule, orch_frequency = "1 potato")
  }, regexp = "Invalid `orch_frequency`")

  expect_error({
    run_schedule(schedule, orch_frequency = "56 weeks")
  }, regexp = "Invalid `orch_frequency`")
})

test_that("errors if resources are unnamed or non unique", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)

  expect_error({
    run_schedule(schedule, orch_frequency = "hourly", resources = list(1))
  }, regexp = "All elements")

  expect_error({
    run_schedule(schedule, orch_frequency = "hourly", resources = list(a = 1, a = 2))
  }, regexp = "All elements")
})

test_that("warns if the orch frequency is less than the highest pipe frequency", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)
  expect_warning({
    run_schedule(schedule, orch_frequency = "daily")
  })

  # Unless run_all = TRUE
  expect_no_warning({
    run_schedule(schedule, orch_frequency = "daily", run_all = TRUE)
  })
}) |>
  suppressMessages()

test_that("errors if orch_frequency is less than 1 year", {
  schedule <- build_schedule(test_path("test_pipelines_run_all_good"), quiet = TRUE)
  expect_error({
    run_schedule(
      schedule,
      orch_frequency = "2 years"
    )
  }, regexp = "Invalid `orch_frequency`")
})

test_that("maestroStartTime with HH:MM:SS runs on the expected time", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroStartTime 10:00:00
      hhmmss <- function() {

      }
      ",
      con = "pipelines/hhmmss.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct(glue::glue("{lubridate::today()} 10:00:00"), tz = "UTC"),
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$invoked)
})

test_that("maestroPriority works as expected", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroStartTime 10:00:00
      p2 <- function() {
      }

      #' @maestroFrequency 1 day
      #' @maestroStartTime 10:00:00
      #' @maestroPriority 1
      q1 <- function() {
        Sys.sleep(1)
      }
      ",
      con = "pipelines/priorities.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2025-03-17 10:00:00", tz = "UTC"),
      quiet = TRUE
    )
    status <- get_status(schedule)
  })

  expect_snapshot(status$pipe_name)
})

test_that("maestroDays works on expected days", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' test maestro pipeline
      #'
      #' @maestroFrequency daily
      #' @maestroDays Tue Thu
      #' @maestroStartTime 2026-01-19
      #' @maestroLogLevel INFO

      test <- function() {
      }
      ",
      con = "pipelines/test.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      check_datetime = as.Date("2026-01-22"), # this is Thursday - it should run
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_true(status$invoked)

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      check_datetime = as.Date("2026-01-23"), # this is Friday - it shouldn't run
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_false(status$invoked)
  })
})

test_that("Pipeline that prints curly brackets runs fine", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' test maestro pipeline
      #'
      #' @maestroFrequency daily
      test <- function() {
        message('{')
        message('\n')
      }
      ",
      con = "pipelines/test.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      quiet = TRUE,
      log_to_console = TRUE,
      log_to_file = TRUE
    )
    status <- get_status(schedule)
    expect_true(status$success)
  })
})

test_that("Old README example pipelines all run with daily orchestrator", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      #' @maestroStartTime 09:00:00
      #' @maestroTz America/Halifax
      get_mtcars <- function() {
      }

      #' @maestroFrequency daily
      #' @maestroStartTime 09:00:00
      #' @maestroTz America/Halifax
      multi_rng <- function() {
      }

      #' @maestroFrequency daily
      #' @maestroStartTime 12:30:00
      #' @maestroTz America/Halifax
      my_etl <- function() {
      }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      check_datetime = as.POSIXct("2026-04-16 11:00:00", tz = "America/Halifax"),
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_true(all(status$invoked))
  })
})

test_that("Other vignette examples", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      # ./pipelines/daily_example.R
      #' daily_example maestro pipeline
      #'
      #' @maestroFrequency 1 day
      #' @maestroStartTime 09:20:00
      daily_example <- function() {
      
        # Pipeline code
      }
      ",
      con = "pipelines/daily_example.R"
    )

    schedule <- build_schedule()

    run_schedule(
      schedule,
      orch_frequency = "1 day",
      check_datetime = as.POSIXct("2024-06-20 08:00:00", tz = "UTC"),
      quiet = TRUE
    )

    expect_true(get_status(schedule)$invoked)

    run_schedule(
      schedule,
      orch_frequency = "15 minutes",
      check_datetime = as.POSIXct("2024-06-20 08:00:00", tz = "UTC"),
      quiet = TRUE
    )

    expect_false(get_status(schedule)$invoked)
  })
})

test_that("Irregular hourly with weekly pipeline", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #'
      #' @maestroFrequency weekly
      #' @maestroStartTime Fri 04:00:00
      weekly <- function() {
      
        # Pipeline code
      }
      ",
      con = "pipelines/weekly.R"
    )

    schedule <- build_schedule()

    run_schedule(
      schedule,
      orch_frequency = "4 hours",
      check_datetime = as.POSIXct("2026-05-01 04:00:00", tz = "UTC"), # this is a Friday
      quiet = TRUE
    )

    expect_true(get_status(schedule)$invoked)
  })
})

test_that("Check validity of next_run", {

  # On exact timestamp
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroStartTime 03:00:00
      #' @maestroTz UTC
      my_pipe <- function() {
      }
      ",
      con = "pipelines/pipes.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2026-04-16 03:00:00", tz = "UTC"),
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_equal(status$next_run, as.POSIXct("2026-04-17 03:00:00", tz = "UTC"))
  })

  # After the 12:00:00 on the previous day
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 day
      #' @maestroStartTime 09:00:00
      #' @maestroTz UTC
      local <- function() {

      }
      ",
      con = "pipelines/local.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      check_datetime = as.POSIXct("2026-04-16 14:00:00", tz = "UTC"),
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_equal(status$next_run, as.POSIXct("2026-04-17 09:00:00", tz = "UTC"))
  })
})



test_that("pipeline with unmatched argument errors gracefully", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      needs_arg <- function(my_arg) {
        my_arg + 1
      }
      ",
      con = "pipelines/needs_arg.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      run_all = TRUE,
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_false(status$success)
    expect_match(last_run_errors()$needs_arg, "missing argument `my_arg`")
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      needs_arg <- function(my_arg = 2) {
        my_arg + 1
      }
      ",
      con = "pipelines/needs_arg.R"
    )

    schedule <- build_schedule(quiet = TRUE)
    run_schedule(
      schedule,
      orch_frequency = "1 day",
      run_all = TRUE,
      quiet = TRUE
    )
    status <- get_status(schedule)
    expect_true(status$success)
  })
})

#' Creates a new maestro project
#'
#' @inheritParams create_orchestrator
#' @param ... unused
#'
#' @export
#' @return invisible
#' @examples
#'
#' # Creates a new maestro project with an R orchestrator
#' if (interactive()) {
#'   new_proj_dir <- tempdir()
#'   create_maestro(new_proj_dir, type = "R", overwrite = TRUE)
#'
#'   create_maestro(new_proj_dir, type = "Quarto", overwrite = TRUE)
#' }
create_maestro <- function(path, type = "R", overwrite = FALSE, quiet = FALSE, ...) {

  type <- match.arg(type, choices = c("R", "Quarto", "RMarkdown"))

  path_to_maestro <- normalizePath(
    path,
    mustWork = FALSE
  )

  if (dir.exists(path_to_maestro)) {
    if (!overwrite) {
      cli::cli_abort(
        paste(
          "Project directory already exists. \n",
          "Set `create_maestro(overwrite = TRUE)` to overwrite anyway.\n",
          "This will remove any work in this directory. \n"
        ),
        call = NULL
      )
    } else {
      if (!quiet) cli::cli_alert_warning("Overwriting existing project.")
    }
  }

  if (!quiet) cli::cli_alert_success("Creating maestro project")

  # ensure path exists
  dir.create(path, recursive = TRUE, showWarnings = FALSE)

  create_orchestrator(
    path = file.path(path, "orchestrator"),
    type = type,
    open = FALSE,
    quiet = TRUE,
    overwrite = overwrite
  )

  create_pipeline(
    pipe_name = "my_pipe",
    pipeline_dir = file.path(path, "pipelines"),
    open = FALSE,
    quiet = TRUE,
    overwrite = overwrite
  )
}

test_that("create_maestro creates a new maestro project", {
  withr::with_tempdir({
    expect_message(create_maestro(path = ".", type = "R", overwrite = TRUE))
    expect_true(file.exists("orchestrator.R"))
    expect_true(dir.exists("pipelines"))
  })

  withr::with_tempdir({
    expect_message(create_maestro(path = ".", type = "Quarto", overwrite = TRUE))
    expect_true(file.exists("orchestrator.qmd"))
    expect_true(dir.exists("pipelines"))
  })
}) |>
  suppressMessages()

test_that("create_maestro aborts if directory already exists", {
  withr::with_tempdir({
    expect_error(
      create_maestro(path = "."),
      regexp = "Project directory already exists"
    )
  })

  # If overwrite = TRUE it will work
  withr::with_tempdir({
    expect_message(
      create_maestro(path = ".", overwrite = TRUE),
      regexp = "Overwriting existing project"
    )
  })
}) |>
  suppressMessages()

#' cli output for generate schedule table
#'
#' @param parse_succeeds list of parse results (i.e., succeeded)
#' @param parse_errors list of parse errors
#'
#' @keywords internal
#' @return cli output
maestro_parse_cli <- function(parse_succeeds, parse_errors) {

  n_fails <- length(parse_errors)
  n_succeeds <- length(parse_succeeds)

  if(n_succeeds == 0) {

    cli::cli_abort(
      c(
        "x" = "All scripts failed to parse",
        "i" = "See full error output with {.fn last_build_errors}"
      ),
      call = rlang::caller_env()
    )
  } else {

    if (n_succeeds > 0) {
      cli::cli_inform(
        c("i" = "{n_succeeds} script{?s} successfully parsed")
      )
    }

    if (n_fails > 0) {
      fail_vec <- purrr::map_chr(parse_errors, ~{
        .x$message
      }) |>
        stats::setNames("!")

      cli::cli_warn(
        c(
          "{n_fails} script{?s} failed to parse:",
          fail_vec,
          "i" = "See full error output with {.fn last_build_errors}"
        )
      )
    }
  }

  return(invisible())
}

#' @maestroFrequency daily
#' @maestroDays 1 8 24 28
specific_days <- function() {

}

#' @maestroFrequency daily
#' @maestroDays Tue Fri
specific_days2 <- function() {

}

#' @maestroFrequency hourly
#' @maestroHours 0 4 8 12 16 20
specific_hours <- function() {

}

#' @maestroFrequency biweekly
#' @maestroMonths 1 5 10
specific_months1 <- function() {

}

#' @maestroFrequency weekly
#' @maestroMonths 1 5 10
specific_months2 <- function() {

}

#' @maestroFrequency monthly
#' @maestroMonths 1 5 10
specific_months3 <- function() {

}

#' @maestroFrequency hourly
#' @maestroHours 1 12
#' @maestroDays 1
#' @maestroMonths 1 5 10
specific_multi <- function() {

}

maestro_logger <- logger::layout_glue_generator(
  format = "[{namespace}] [{level}] [{format(time, \"%Y-%m-%d %H:%M:%S\")}]: {msg}"
)

#' Convert a duration string to number of seconds
#'
#' @param time_string string like 1 day, 1 week, 12 hours, etc.
#' @keywords internal
#' @return number of seconds
convert_to_seconds <- function(time_string) {

  nunits <- parse_rounding_unit(time_string)

  # Define the conversion factors to seconds for each unit
  conversion_factors <- c(
    sec = 1, second = 1,
    min = 60, minute = 60,
    hour = 3600,
    day = 86400,
    week = 604800,
    month = 2629800,
    quarter = 7884000,
    year = 31557600
  )

  if (!nunits$unit %in% names(conversion_factors)) {
    stop("Unknown time unit")
  }

  nunits$n * conversion_factors[[nunits$unit]]
}

#' Number of days ahead to generate run sequences for a given frequency unit
#' Used by the observable get_run_sequence path.
#' @param unit character frequency unit
#' @keywords internal
#' @return integer
.run_sequence_days_out <- function(unit) {
  switch(
    unit,
    second = 3L,
    minute = 7L,
    hour = 60L,
    day = 120L,
    week = 240L,
    month = 365L * 2L,
    quarter = 365L * 4L,
    year = 365L * 10L
  )
}

#' Minimum days ahead to pre-compute run sequences on initialization
#' This small window is used at parse time; the full window is built lazily
#' when observability functions (get_run_sequence, get_slot_usage) are called.
#' @param unit character frequency unit
#' @keywords internal
#' @return integer
.run_sequence_min_days_out <- function(unit) {
  switch(
    unit,
    second = 1L,
    minute = 2L,
    7L
  )
}

valid_units <- c(
  "sec", "second", "min", "minute", "hour",
  "day", "week", "month", "quarter", "year"
)

#' Parse a time string
#'
#' @param time_string string like 1 day, daily, 1 week, 12 hours, etc.
#'
#' @keywords internal
#' @return nunit list
parse_rounding_unit <- function(time_string) {

  # Extract the number and the unit from the time string
  if (!grepl("^[0-9]", time_string)) {
    time_string <- switch (time_string,
      hourly = "1 hour",
      daily = "1 day",
      weekly = "1 week",
      biweekly = "2 weeks",
      monthly = "1 month",
      quarterly = "3 months",
      yearly = "1 year",
      time_string
    )
  }

  matches <- regmatches(time_string, regexec("([0-9]+)\\s*(\\w+)", time_string))
  number <- as.numeric(matches[[1]][2])
  unit <- matches[[1]][3]

  unit_fmt <- unit |>
    trimws()

  unit_fmt <- gsub("s$", "", x = unit_fmt)

  if (!unit_fmt %in% valid_units) {
    stop(glue::glue("Invalid rounding unit `{time_string}`."))
  }

  return(
    list(
      n = number,
      unit = unit_fmt
    )
  )
}

#' Generate a sequence of run times for a pipeline
#'
#' @param check_datetime datetime against which to check the running of pipelines (default is current system time in UTC)
#' @param pipeline_n number of units for the pipeline frequency
#' @param pipeline_unit unit for the pipeline frequency
#' @param pipeline_datetime datetime of the first time the pipeline is to run
#' @param pipeline_hours vector of integers \[0-23] corresponding to hours of day for the pipeline to run
#' @param pipeline_months vector of integers \[1-12] corresponding to months of year for the pipeline to run
#' @param pipeline_days_of_week vector of integers \[1-7] corresponding to days of week for the pipeline to run (1 = Sunday)
#' @param pipeline_days_of_month vector of integers \[1-31] corresponding to days of month for the pipeline to run
#'
#' @keywords internal
#' @return vector of timestamps or dates
get_pipeline_run_sequence <- function(
    pipeline_n,
    pipeline_unit,
    pipeline_datetime,
    check_datetime,
    pipeline_hours = 0:23,
    pipeline_days_of_week = 1:7,
    pipeline_days_of_month = 1:31,
    pipeline_months = 1:12
  ) {

  check_datetime <- lubridate::as_datetime(check_datetime)

  pipeline_unit <- switch(
    pipeline_unit,
    minutes = ,
    minute  = "min",
    seconds = ,
    second  = "sec",
    pipeline_unit
  )

  if (pipeline_datetime > check_datetime) {
    pipeline_sequence <- pipeline_datetime
  } else {
    pipeline_sequence <- seq(pipeline_datetime, check_datetime, by = paste(pipeline_n, pipeline_unit))
  }

  pipeline_sequence <- adjust_for_dst(pipeline_sequence[[1]], pipeline_sequence)

  if (!all(0:23 %in% pipeline_hours)) {
    pipeline_sequence <- pipeline_sequence[lubridate::hour(pipeline_sequence) %in% pipeline_hours]
  }

  if (!all(1:7 %in% pipeline_days_of_week)) {
    pipeline_sequence <- pipeline_sequence[lubridate::wday(pipeline_sequence, week_start = 1) %in% pipeline_days_of_week]
  }

  if (!all(1:31 %in% pipeline_days_of_month)) {
    pipeline_sequence <- pipeline_sequence[lubridate::mday(pipeline_sequence) %in% pipeline_days_of_month]
  }

  if (!all(1:12 %in% pipeline_months)) {
    pipeline_sequence <- pipeline_sequence[lubridate::month(pipeline_sequence) %in% pipeline_months]
  }

  pipeline_sequence
}

`%n%` <- function(rhs, lhs) {
  if (is.null(rhs) || length(rhs) == 0 || all(is.na(rhs))) return(lhs)
  rhs
}

#' Checks whether a DAG is valid (no loops)
#'
#' @param edges a data.frame of edges (from, to)
#'
#' @keywords internal
#' @return boolean
is_valid_dag <- function(edges) {

  # Create an adjacency list
  adj_list <- split(edges$to, edges$from)

  # Find all nodes (unique vertices)
  nodes <- unique(c(edges$from, edges$to))

  # Initialize in-degree for each node
  in_degree <- stats::setNames(rep(0, length(nodes)), nodes)

  # Calculate in-degrees
  for (to in edges$to) {
    in_degree[to] <- in_degree[to] + 1
  }

  # Topological sorting using Kahn's Algorithm
  topological_sort <- function(nodes, adj_list, in_degree) {
    sorted <- c()
    queue <- nodes[in_degree == 0]

    while (length(queue) > 0) {
      current <- queue[1]
      queue <- queue[-1]
      sorted <- c(sorted, current)

      if (!is.null(adj_list[[current]])) {
        for (neighbor in adj_list[[current]]) {
          in_degree[neighbor] <- in_degree[neighbor] - 1
          if (in_degree[neighbor] == 0) {
            queue <- c(queue, neighbor)
          }
        }
      }
    }

    length(sorted) == length(nodes)
  }

  # Check if the graph is a DAG
  is_dag <- topological_sort(nodes, adj_list, in_degree)

  is_dag
}

.unit_order <- c("second", "minute", "hour", "day", "week", "month", "quarter", "year")

.unit_aliases <- c(
  seconds = "second", sec = "second", secs = "second",
  minutes = "minute", min = "minute", mins = "minute",
  hours = "hour",
  days = "day",
  weeks = "week",
  months = "month",
  quarters = "quarter",
  years = "year"
)

standardize_units <- function(unit) {
  .unit_aliases[unit] %n% unit
}

units_lt_units <- function(u1, u2) {
  u1_ord <- factor(u1, levels = .unit_order, ordered = TRUE)
  u2_ord <- factor(u2, levels = .unit_order, ordered = TRUE)
  u1_ord < u2_ord
}

adjust_for_dst <- function(base_timestamp, adjustable_timestamps) {

  base_dst <- lubridate::dst(base_timestamp)

  adjustment <- ifelse(
    lubridate::dst(adjustable_timestamps) == base_dst,
    0,
    ifelse(base_dst, 1, -1)
  )

  adjustable_timestamps + lubridate::hours(adjustment)
}

validate_orch_frequency <- function(orch_frequency) {
  orch_nunits <- tryCatch({
    parse_rounding_unit(orch_frequency)
  }, error = \(e) {
    cli::cli_abort(
      c(
        "Invalid `orch_frequency` {orch_frequency}.",
        "i" = "Must be of the format like '1 day', '1 week', 'hourly', etc."
      ),
      call = NULL
    )
  })

  # Enforce minimum orch frequency of 1 year
  if (orch_nunits$unit == "year" && orch_nunits$n > 1) {
    cli::cli_abort(
      "Invalid `orch_frequency` {orch_frequency}. Minimum frequency is 1 year.",
      call = NULL
    )
  }

  # Additional parse using timechange to verify it isn't something like 500 days,
  # which isn't understood by timechange
  tryCatch({
    timechange::time_round(Sys.time(), paste(orch_nunits$n, orch_nunits$unit))
  }, error = \(e) {
    timechange_error_fmt <- gsub('\\..*', '', e$message)
    cli::cli_abort(
      c(
        "Invalid `orch_frequency` {orch_frequency}.
        {timechange_error_fmt}."
      ),
      call = NULL
    )
  })

  orch_nunits
}

eval_code_str <- function(code, vars = list(), inherit = rlang::caller_env()) {

  env <- rlang::env(inherit, !!!vars)

  exprs <- if (is.character(code)) rlang::parse_exprs(code) else {
    if (is.list(code)) code else list(code)
  }

  out <- NULL
  for (e in exprs) out <- rlang::eval_bare(e, env)
  invisible(out)
}

find_roots <- function(from, to, values = to) {

  parent_lookup <- from
  names(parent_lookup) <- to

  get_root <- function(node) {
    if (!node %in% names(parent_lookup)) {
      return(node)
    }
    Recall(parent_lookup[[node]])
  }

  purrr::map_chr(values, get_root)
}

make_id <- function(n = 6) {
  first <- sample(c(letters, LETTERS), 1)
  rest  <- sample(c(letters, LETTERS, 0:9), n - 1, replace = TRUE)

  paste0(c(first, rest), collapse = "")
}

# Weekday abbreviations, ordered Mon–Sun (matches lubridate week_start = 1)
.weekday_abbrs <- c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")

#' Resolve a partial \code{@maestroStartTime} string to a concrete POSIXct
#'
#' Accepts four formats:
#' \itemize{
#'   \item \code{"HH:MM:SS"} — time of day; anchored to today's date.
#'   \item \code{"Mon HH:MM:SS"} — weekday + time; anchored to that weekday of
#'     the current ISO week.
#'   \item \code{"DD HH:MM:SS"} or \code{"DD"} — month-day (+ optional time);
#'     anchored to that day of the current month.
#'   \item Full datetime string — passed through to \code{as.POSIXct()}.
#' }
#'
#' Using a recent \code{now} as the reference keeps the anchor close to the
#' present, so \code{n} in \code{.prev_on_cycle()} stays small.
#'
#' @param raw Character string from the tag value.
#' @param tz  Timezone string.
#' @param now \code{POSIXct} reference point (default \code{lubridate::now(tz)}).
#' @return A \code{POSIXct}, or \code{NA} if \code{raw} is \code{NA}/\code{""}.
#' @keywords internal
parse_maestro_start_time <- function(raw, tz, now = lubridate::now(tzone = tz)) {

  if (is.na(raw) || raw == "") return(NA)

  # HH:MM:SS
  if (grepl("^[0-9]{2}:[0-9]{2}:[0-9]{2}$", raw)) {
    now <- now - lubridate::days(1)
    date_part <- format(now, "%Y-%m-%d")
    return(as.POSIXct(paste(date_part, raw), format = "%Y-%m-%d %H:%M:%S", tz = tz))
  }

  # Mon HH:MM:SS — weekday + time
  if (grepl(sprintf("^(%s)\\s+[0-9]{2}:[0-9]{2}:[0-9]{2}$",
                    paste(.weekday_abbrs, collapse = "|")), raw)) {
    parts    <- strsplit(raw, "\\s+")[[1]]
    day_abbr <- parts[1]
    time_str <- parts[2]

    # target_dow is an offset from Monday (1 = Mon, 7 = Sun), derived purely
    # from .weekday_abbrs — not from lubridate.week.start or wday().
    # floor_date(..., week_start = 1) is also explicit, so this is
    # session-option-independent.
    target_dow  <- match(day_abbr, .weekday_abbrs)
    week_start  <- lubridate::floor_date(now, unit = "week", week_start = 1)
    anchor_date <- week_start + lubridate::days(target_dow - 1L)
    return(as.POSIXct(
      paste(format(anchor_date, "%Y-%m-%d"), time_str),
      format = "%Y-%m-%d %H:%M:%S",
      tz = tz
    ))
  }

  # DD or DD HH:MM:SS — month-day + optional time
  if (grepl("^[0-9]{1,2}(\\s+[0-9]{2}:[0-9]{2}:[0-9]{2})?$", raw)) {
    parts       <- strsplit(raw, "\\s+")[[1]]
    mday        <- as.integer(parts[1])
    time_str    <- if (length(parts) == 2) parts[2] else "00:00:00"
    month_start <- lubridate::floor_date(now, unit = "month")
    anchor_date <- month_start + lubridate::days(mday - 1L)
    return(as.POSIXct(
      paste(format(anchor_date, "%Y-%m-%d"), time_str),
      format = "%Y-%m-%d %H:%M:%S",
      tz = tz
    ))
  }

  # Full datetime — delegate to base coercion
  as.POSIXct(raw, tz = tz)
}

#' Find the most recent cycle point before a given time
#'
#' Given a repeating cycle anchored at \code{start} and stepping every
#' \code{amount} \code{unit}s, returns the latest cycle point strictly before
#' \code{current}.
#'
#' @param start \code{Date} or \code{POSIXct} defining the cycle origin.
#' @param current \code{Date} or \code{POSIXct} (coerced to match \code{start})
#'   used as the reference point. Defaults to \code{Sys.time()}.
#' @param amount Positive integer; number of \code{unit}s per cycle step.
#' @param unit Character; one of \code{"second"}, \code{"minute"},
#'   \code{"hour"}, \code{"day"}, \code{"week"}, \code{"month"},
#'   \code{"year"}. Sub-day units are not supported when \code{start} is a
#'   \code{Date}.
#'
#' @return A \code{Date} or \code{POSIXct} of the same class and timezone as
#'   \code{start}, or \code{NA} if \code{current <= start}.
#' @keywords internal
.prev_on_cycle <- function(start,
                           current = Sys.time(),
                           amount = 1L,
                           unit = c(
                             "second", "minute", "hour",
                             "day", "week",
                             "month", "year"
                           )) {
  unit <- match.arg(unit)

  amount <- as.integer(amount)
  start_is_date <- inherits(start, "Date")
  start_is_time <- inherits(start, "POSIXct")

  if (!start_is_date && !start_is_time) {
    stop("Internal error: `start` must be Date or POSIXct.")
  }

  # Coerce `current` to match `start` type
  if (start_is_date) {
    current <- as.Date(current)
    start   <- as.Date(start)
  } else {
    tz <- attr(start, "tzone")
    if (is.null(tz) || !nzchar(tz)) tz <- ""

    current <- as.POSIXct(current, tz = tz)
    start   <- as.POSIXct(start, tz = tz)
  }

  na_out <- function() {
    if (start_is_date) as.Date(NA) else as.POSIXct(NA)
  }

  if (is.na(start) || is.na(current)) return(na_out())

  if (current <= start) return(na_out())

  .days_in_month <- function(year, month) {
    y2 <- year + (month == 12L)
    m2 <- if (month == 12L) 1L else month + 1L

    first_this <- as.Date(sprintf("%04d-%02d-01", year, month))
    first_next <- as.Date(sprintf("%04d-%02d-01", y2, m2))

    as.integer(first_next - first_this)
  }

  .add_months_clamp_date <- function(date, n_months) {
    y <- as.integer(format(date, "%Y"))
    m <- as.integer(format(date, "%m"))
    d <- as.integer(format(date, "%d"))

    idx <- y * 12L + (m - 1L) + n_months
    y2 <- idx %/% 12L
    m2 <- idx %% 12L + 1L

    dim <- .days_in_month(y2, m2)
    d2 <- if (d > dim) dim else d

    as.Date(sprintf("%04d-%02d-%02d", y2, m2, d2))
  }

  .add_years_clamp_date <- function(date, n_years) {
    y <- as.integer(format(date, "%Y"))
    m <- as.integer(format(date, "%m"))
    d <- as.integer(format(date, "%d"))

    y2 <- y + n_years
    dim <- .days_in_month(y2, m)
    d2 <- if (d > dim) dim else d

    as.Date(sprintf("%04d-%02d-%02d", y2, m, d2))
  }

  .add_months_clamp_time <- function(ts, n_months) {
    tz <- attr(ts, "tzone")
    if (is.null(tz) || !nzchar(tz)) tz <- ""

    lt <- as.POSIXlt(ts, tz = tz)

    y <- lt$year + 1900L
    m <- lt$mon + 1L
    d <- lt$mday

    idx <- y * 12L + (m - 1L) + n_months
    y2 <- idx %/% 12L
    m2 <- idx %% 12L + 1L

    dim <- .days_in_month(y2, m2)
    d2 <- if (d > dim) dim else d

    ISOdatetime(y2, m2, d2, lt$hour, lt$min, lt$sec, tz = tz)
  }

  .add_years_clamp_time <- function(ts, n_years) {
    tz <- attr(ts, "tzone")
    if (is.null(tz) || !nzchar(tz)) tz <- ""

    lt <- as.POSIXlt(ts, tz = tz)

    y2 <- (lt$year + 1900L) + n_years
    m  <- lt$mon + 1L
    d  <- lt$mday

    dim <- .days_in_month(y2, m)
    d2 <- if (d > dim) dim else d

    ISOdatetime(y2, m, d2, lt$hour, lt$min, lt$sec, tz = tz)
  }

  if (start_is_date) {
    if (unit %in% c("second", "minute", "hour")) {
      stop("Internal error: sub-day units are not supported for Date inputs.")
    }

    if (unit %in% c("day", "week")) {
      step_days <- if (unit == "week") amount * 7L else amount
      delta_days <- as.integer(current - start)
      k <- (delta_days - 1L) %/% step_days
      if (k < 0L) return(na_out())
      return(start + k * step_days)
    }

    sy <- as.integer(format(start, "%Y"))
    sm <- as.integer(format(start, "%m"))
    cy <- as.integer(format(current, "%Y"))
    cm <- as.integer(format(current, "%m"))

    if (unit == "month") {
      start_idx <- sy * 12L + (sm - 1L)
      curr_idx  <- cy * 12L + (cm - 1L)

      delta_months <- curr_idx - start_idx
      n <- (delta_months %/% amount) * amount

      cand <- .add_months_clamp_date(start, n)

      if (cand >= current) {
        n2 <- n - amount
        if (n2 < 0L) return(na_out())
        cand <- .add_months_clamp_date(start, n2)
      }

      return(cand)
    }

    delta_years <- cy - sy
    n <- (delta_years %/% amount) * amount

    cand <- .add_years_clamp_date(start, n)

    if (cand >= current) {
      n2 <- n - amount
      if (n2 < 0L) return(na_out())
      cand <- .add_years_clamp_date(start, n2)
    }

    return(cand)
  }

  # POSIXct branch
  if (unit %in% c("second", "minute", "hour", "day", "week")) {
    step_secs <- switch(
      unit,
      second = amount,
      minute = amount * 60,
      hour   = amount * 3600,
      day    = amount * 86400,
      week   = amount * 7 * 86400
    )

    delta_secs <- as.numeric(difftime(current, start, units = "secs"))

    # epsilon avoids returning `current` if it lies exactly on the cycle
    k <- floor((delta_secs - 1e-9) / step_secs)
    if (is.na(k) || k < 0) return(na_out())

    return(start + k * step_secs)
  }

  sy <- as.integer(format(start, "%Y"))
  sm <- as.integer(format(start, "%m"))
  cy <- as.integer(format(current, "%Y"))
  cm <- as.integer(format(current, "%m"))

  if (unit == "month") {
    start_idx <- sy * 12L + (sm - 1L)
    curr_idx  <- cy * 12L + (cm - 1L)

    delta_months <- curr_idx - start_idx
    n <- (delta_months %/% amount) * amount

    cand <- .add_months_clamp_time(start, n)

    if (cand >= current) {
      n2 <- n - amount
      if (n2 < 0L) return(na_out())
      cand <- .add_months_clamp_time(start, n2)
    }

    return(cand)
  }

  delta_years <- cy - sy
  n <- (delta_years %/% amount) * amount

  cand <- .add_years_clamp_time(start, n)

  if (cand >= current) {
    n2 <- n - amount
    if (n2 < 0L) return(na_out())
    cand <- .add_years_clamp_time(start, n2)
  }

  cand
}

test_that("convert_to_seconds works", {

  res <- convert_to_seconds("1 week")
  expect_equal(res, unclass(lubridate::duration(1, units = "week")))

  res <- convert_to_seconds("12 hours")
  expect_equal(res, unclass(lubridate::duration(12, units = "hours")))

  res <- convert_to_seconds("25minutes")
  expect_equal(res, unclass(lubridate::duration(25, units = "minutes")))

  res <- convert_to_seconds("48    days")
  expect_equal(res, unclass(lubridate::duration(48, units = "days")))

  res <- convert_to_seconds("1 quarters")
  expect_equal(res, 7884000L)
})

test_that("parse_maestro_start_time handles HH:MM:SS", {
  now <- as.POSIXct("2026-04-14 10:00:00", tz = "UTC")
  result <- parse_maestro_start_time("04:00:00", tz = "UTC", now = now)
  expect_s3_class(result, "POSIXct")
  expect_equal(format(result, "%H:%M:%S"), "04:00:00")
  expect_equal(format(result, "%Y-%m-%d"), "2026-04-13")
})

test_that("parse_maestro_start_time handles weekday + time", {
  # 2026-04-14 is a Tuesday; Mon of that week is 2026-04-13
  now <- as.POSIXct("2026-04-14 10:00:00", tz = "UTC")

  mon <- parse_maestro_start_time("Mon 04:00:00", tz = "UTC", now = now)
  expect_s3_class(mon, "POSIXct")
  expect_equal(format(mon, "%Y-%m-%d %H:%M:%S"), "2026-04-13 04:00:00")

  wed <- parse_maestro_start_time("Wed 09:30:00", tz = "UTC", now = now)
  expect_equal(format(wed, "%Y-%m-%d %H:%M:%S"), "2026-04-15 09:30:00")

  sun <- parse_maestro_start_time("Sun 00:00:00", tz = "UTC", now = now)
  expect_equal(format(sun, "%Y-%m-%d %H:%M:%S"), "2026-04-19 00:00:00")
})

test_that("parse_maestro_start_time weekday anchor is deterministic across calls", {
  now <- as.POSIXct("2026-04-14 10:00:00", tz = "UTC")
  r1 <- parse_maestro_start_time("Mon 04:00:00", tz = "UTC", now = now)
  r2 <- parse_maestro_start_time("Mon 04:00:00", tz = "UTC", now = now)
  expect_equal(r1, r2)
})

test_that("parse_maestro_start_time handles month-day + time", {
  now <- as.POSIXct("2026-04-14 10:00:00", tz = "UTC")

  result <- parse_maestro_start_time("15 04:00:00", tz = "UTC", now = now)
  expect_s3_class(result, "POSIXct")
  expect_equal(format(result, "%Y-%m-%d %H:%M:%S"), "2026-04-15 04:00:00")

  result1 <- parse_maestro_start_time("1 00:00:00", tz = "UTC", now = now)
  expect_equal(format(result1, "%Y-%m-%d %H:%M:%S"), "2026-04-01 00:00:00")
})

test_that("parse_maestro_start_time handles bare month-day (no time)", {
  now <- as.POSIXct("2026-04-14 10:00:00", tz = "UTC")
  result <- parse_maestro_start_time("15", tz = "UTC", now = now)
  expect_s3_class(result, "POSIXct")
  expect_equal(format(result, "%Y-%m-%d %H:%M:%S"), "2026-04-15 00:00:00")
})

test_that("parse_maestro_start_time handles full datetime", {
  result <- parse_maestro_start_time("2025-06-01 08:00:00", tz = "UTC")
  expect_s3_class(result, "POSIXct")
  expect_equal(format(result, "%Y-%m-%d %H:%M:%S"), "2025-06-01 08:00:00")
})

test_that("parse_maestro_start_time returns NA for NA input", {
  result <- parse_maestro_start_time(NA, tz = "UTC")
  expect_true(is.na(result))
})

test_that("parse_rounding_unit works", {

  good_examples <- c(
    "1 days", "4 weeks", "10 months", "2 week",
    "15 minute", "30 day  ", " 20 hours", "2 years",
    "2 quarters", "19 mins"
  )

  purrr::walk(good_examples, ~{
    expect_no_error(
      parse_rounding_unit(.x)
    )
  })

  bad_examples <- c(
    "one day", "45 daysasd", "3 hrs", "4"
  )

  purrr::walk(bad_examples, ~{
    expect_error(
      parse_rounding_unit(.x),
      regexp = "Invalid"
    )
  })
})

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroOutputs high_road low_road
  start <- function() {
    c('a', 'A')
  }
  
  #' @maestroInputs start
  high_road <- function(.input) {
    toupper(.input)
  }
  
  #' @maestroInputs start
  low_road <- function(.input) {
    tolower(.input)
  }",
  con = "pipelines/dags.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' ./pipelines/dags.R
# #' @maestroOutputs high_road low_road
# start <- function() {
#   c('a', 'A')
# }
# 
# #' @maestroInputs start
# high_road <- function(.input) {
#   toupper(.input)
# }
# 
# #' @maestroInputs start
# low_road <- function(.input) {
#   tolower(.input)
# }

## ----eval=FALSE---------------------------------------------------------------
# #' Example ETL pipeline
# #' @maestroFrequency 1 day
# my_etl <- function() {
# 
#   # Pretend we're getting data from a source
#   message("Get data")
#   extracted <- mtcars
# 
#   # Transform
#   message("Transforming")
#   transformed <- extracted |>
#     dplyr::mutate(hp_deviation = hp - mean(hp))
# 
#   # Load - write to a location
#   message("Writing")
#   write.csv(transformed, file = paste0("transformed_mtcars_", Sys.Date(), ".csv"))
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(file.remove("pipelines/dags.R"))
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroFrequency 1 day
  #' @maestroOutputs transform
  extract <- function() {
    # Imagine this is something way more complicated, like a database call
    mtcars
  }
  
  #' @maestroOutputs load
  transform <- function(.input) {
    .input |> 
      dplyr::mutate(hp_deviation = hp - mean(hp))
  }
  
  #' @maestro
  load <- function(.input) {
    write.csv(.input, file = paste0('transformed_mtcars.csv'))
  }",
  con = "pipelines/etl.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency 1 day
# #' @maestroOutputs transform
# extract <- function() {
#   # Imagine this is something way more complicated, like a database call
#   mtcars
# }
# 
# #' @maestroOutputs load
# transform <- function(.input) {
#   .input |>
#     dplyr::mutate(hp_deviation = hp - mean(hp))
# }
# 
# #' @maestro
# load <- function(.input) {
#   write.csv(.input, file = paste0("transformed_mtcars.csv"))
# }

## ----cleanup, echo=FALSE, message=FALSE, warning=FALSE------------------------
invisible(unlink("pipelines", recursive = TRUE))
invisible(unlink("transformed_mtcars.csv"))

#' Get time slot usage of a schedule
#'
#' Get the number of pipelines scheduled to run for each time slot at a particular
#' slot interval. Time slots are times that the orchestrator runs and the slot interval
#' determines the level of granularity to consider.
#'
#' This function is particularly useful when you have multiple pipelines in a project
#' and you want to see what recurring time intervals may be available or underused
#' for new pipelines.
#'
#' Note that this function is intended for use in an interactive session while developing
#' a maestro project. It is not intended for use in the orchestrator.
#'
#' As an example, consider we have four pipelines running at various frequencies
#' and the orchestrator running every hour. Then let's say there's to be a new
#' pipeline that runs every day. One might ask 'what hour should I schedule this new
#' pipeline to run on?'. By using `get_slot_usage(schedule, orch_frequency = '1 hour', slot_interval = 'hour')`
#' on the existing schedule, you could identify for each hour how many pipelines
#' are already scheduled to run and choose the ones with the lowest usage.
#'
#' @inheritParams run_schedule
#' @param slot_interval a time unit indicating the interval of time to consider between slots (e.g., 'hour', 'day')
#'
#' @returns data.frame
#' @export
#'
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   get_slot_usage(
#'     schedule,
#'     orch_frequency = "1 hour",
#'     slot_interval = "hour"
#'   )
#' }
get_slot_usage <- function(schedule, orch_frequency, slot_interval = "hour") {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  if (length(schedule$PipelineList$MaestroPipelines) == 0) {
    cli::cli_inform("No pipelines in schedule.")
    return(invisible())
  }

  # Get the orchestrator nunits
  orch_nunits <- validate_orch_frequency(orch_frequency)

  window_std <- standardize_units(slot_interval)

  if (!window_std %in% valid_units) {
    cli::cli_abort(
      "`slot_interval` must be a valid time unit such as 'hour', 'day', 'week', etc."
    )
  }

  run_sequences <- schedule$PipelineList$get_run_sequences()

  run_sequences_all <- run_sequences |>
    purrr::imap(
      ~dplyr::tibble(
        pipe_name = .y,
        slot = .x
      )
    ) |>
    purrr::list_rbind()

  window_fmt <- switch (window_std,
    second = "%S",
    minute = "%M",
    hour = "%H:%M",
    day = "%d",
    week = "%a",
    month = "%b",
    quarter = "%b",
    year = "%Y"
  )

  run_sequences_all |>
    dplyr::filter(!is.na(slot)) |>
    dplyr::mutate(
      slot = format(slot, window_fmt)
    ) |>
    dplyr::distinct(pipe_name, slot) |>
    dplyr::summarise(
      n_runs = dplyr::n(),
      pipe_names = paste(pipe_name, collapse = ", "),
      .by = slot
    ) |>
    dplyr::arrange(slot)
}

#' @maestroFrequency daily
letter_a <- function() {
  'a'
}

#' @maestroFrequency daily
letter_b <- function() {
  'b'
}

#' @maestroInputs collect(letter_a, letter_b)
ab <- function(.input) {
  paste0(.input$letter_a, .input$letter_b)
}

#' @maestroInputs ab
ab_upper <- function(.input) {
  toupper(.input)
}

#' Get the run sequence of a schedule
#'
#' Retrieves the scheduled run times for a given schedule, with optional filtering
#' by number of runs and datetime range.
#'
#' @inheritParams run_schedule
#' @param n Optional positive integer. If specified, returns only the first `n` runs for each pipeline.
#' @param min_datetime Optional minimum datetime filter. Can be a `Date` or `POSIXct` object.
#'   If specified, only returns runs scheduled at or after this datetime.
#' @param max_datetime Optional maximum datetime filter. Can be a `Date` or `POSIXct` object.
#'   If specified, only returns runs scheduled at or before this datetime.
#' @param include_only_primary only primary pipelines are included 
#'   (this are pipelines that are scheduled and not downstream nodes in a DAG)
#' @param include_skipped whether to include pipelines tagged with `@maestroSkip`
#'   (default `TRUE` for backwards compatibility)
#'
#' @return A data.frame of scheduled run times with columns `pipe_name`, `scheduled_time`,
#'   and `is_primary`.
#'
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   get_run_sequence(schedule)
#'
#'   # Alternatively, use the underlying R6 method
#'   schedule$get_run_sequence()
#' }
#' @export
get_run_sequence <- function(schedule, n = NULL, min_datetime = NULL, max_datetime = NULL, include_only_primary = FALSE, include_skipped = TRUE) {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  if (!is.null(n)) {
    if (!rlang::is_scalar_integerish(n) || n < 1) {
      cli::cli_abort(
        "`n` must be a positive integer.",
        call = rlang::caller_env()
      )
    }
  }

  if (!is.null(min_datetime)) {
    if (!inherits(min_datetime, c("Date", "POSIXct", "POSIXlt"))) {
      cli::cli_abort(
        "`min_datetime` must be a Date or POSIXct object.",
        call = rlang::caller_env()
      )
    }
  }

  if (!is.null(max_datetime)) {
    if (!inherits(max_datetime, c("Date", "POSIXct", "POSIXlt"))) {
      cli::cli_abort(
        "`max_datetime` must be a Date or POSIXct object.",
        call = rlang::caller_env()
      )
    }
  }

  if (!rlang::is_scalar_logical(include_only_primary)) {
    cli::cli_abort(
      "`include_only_primary` must be a boolean.",
      call = rlang::caller_env()
    )
  }

  if (!rlang::is_scalar_logical(include_skipped)) {
    cli::cli_abort(
      "`include_skipped` must be a boolean.",
      call = rlang::caller_env()
    )
  }

  if (!is.null(min_datetime) && !is.null(max_datetime)) {
    if (min_datetime > max_datetime) {
      cli::cli_abort(
        "`min_datetime` cannot be greater than `max_datetime`.",
        call = rlang::caller_env()
      )
    }
  }

  schedule$get_run_sequence(
    n = n,
    min_datetime = min_datetime, 
    max_datetime = max_datetime,
    include_only_primary = include_only_primary,
    include_skipped = include_skipped
  )
}

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-01 09:00:00
#' @maestroTz UTC
#'
#' @export
get_mtcars <- function() {
  Sys.sleep(0.02)
  mtcars
}

#' Multiply
#'
#' @maestroFrequency 3 month
#' @maestroStartTime 1970-01-01
#' @maestroSkip
#' @maestroTz UTC
#'
#' @export
wait <- function() {
  Sys.sleep(0.01)
}

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-01 09:00:00
#' @maestroTz UTC
#'
#' @export
get_mtcars2 <- function() {
  Sys.sleep(0.02)
  mtcars
}

#' \code{maestro} package
#'
#' Lightweight pipeline orchestration in R
#'
#' Documentation: \href{https://github.com/whipson/maestro}{GitHub}
#' @keywords internal
#' @name maestro
"_PACKAGE"

## quiets concerns of R CMD check re: the .'s that appear in pipelines
if(getRversion() >= "2.15.1")  utils::globalVariables(
  c("frequency", "start_time", "tz", "skip", "log_level", "is_scheduled_now",
    "next_run", "frequency_n", "frequency_unit", "errors", "pipe_name",
    "script_path", "invoked", "success", "pipeline_started", "pipeline_ended",
    "messages", "run_time", "run_date", "run_time_15min", "n_runs",
    "hours", "days_of_week", "days_of_month", "days", "months", "slot")
)

#' Class for a schedule of pipelines
#' @import R6 tictoc utils logger lifecycle
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#' }
MaestroSchedule <- R6::R6Class(

  "MaestroSchedule",

  public = list(

    #' @field PipelineList object of type MaestroPipelineList
    PipelineList = NULL,

    #' @description
    #' Create a MaestroSchedule object
    #' @param Pipelines list of MaestroPipelines
    #' @return MaestroSchedule
    initialize = function(
      Pipelines = NULL
    ) {
      self$PipelineList <- MaestroPipelineList$new()
      pipeline_list <- purrr::map(Pipelines, ~.x$MaestroPipelines) |>
        purrr::list_flatten()
      priorities <- purrr::map(pipeline_list, ~.x$get_priority()) |>
        purrr::list_c()
      if (length(priorities) == 0) priorities <- list()
      pipeline_list <- pipeline_list[order(priorities)]
      self$PipelineList$MaestroPipelines <- pipeline_list
      self$PipelineList$n_pipelines <- length(pipeline_list)
    },

    #' @description
    #' Print the schedule object
    #' @return print
    print = function() {
      cli::cli_h3("Maestro Schedule with {length(self$PipelineList$MaestroPipelines)} pipeline{?s}: ")
      switch (private$sch_status,
        `Not Run` = cli::cli_li(cli::col_magenta(private$sch_status)),
        `Success` = cli::cli_li(cli::col_green(private$sch_status)),
        `Warning` = cli::cli_li(cli::col_yellow(private$sch_status)),
        `Error` = cli::cli_li(cli::col_red(private$sch_status)),
        cli::cli_li("Unknown")
      )
    },

    #' @description
    #' Run a MaestroSchedule
    #' @param ... arguments passed to MaestroPipelineList$run
    #' @param quiet whether or not to silence console messages
    #' @param run_all run all pipelines regardless of the schedule (default is `FALSE`) - useful for testing.
    #' @param n_show_next show the next n scheduled pipes
    #'
    #' @return invisible
    run = function(..., quiet = FALSE, run_all = FALSE, n_show_next = 5) {

      dots <- rlang::list2(..., quiet = quiet)
      check_datetime <- dots$check_datetime
      if (is.null(check_datetime)) {
        check_datetime <- lubridate::now()
      }

      is_multicore <- FALSE
      cores <- dots$cores
      if (!is.null(cores)) {
        if (cores < 1 || (cores %% 1) != 0) cli::cli_abort("`cores` must be a positive integer")
        if (cores > 1) {
          tryCatch({
            rlang::check_installed("furrr")
            is_multicore <- TRUE
          }, error = \(e) {
            cli::cli_warn("{.pkg furrr} is required for running on multiple cores.")
            cores <- 1
          })
        }
      }

      if (self$PipelineList$n_pipelines == 0) {
        cli::cli_inform("No pipelines in the schedule. Nothing to do.")
        return(invisible())
      }

      tryCatch({
        
        pipes_to_run <- self$PipelineList

        if (private$sch_status != "Not Run") {
          self$PipelineList$reset_pipelines()
        }

        if (!run_all) {
          pipes_to_run <- do.call(pipes_to_run$get_timely_pipelines, dots)
        }

        if (!quiet) {
          cli::cli_h3(
            "[{format(check_datetime, '%Y-%m-%d %H:%M:%S')}]
        Running pipelines {cli::col_green(cli::symbol$play)}")
        }

        tictoc::tic(quiet = quiet)
        pipes_that_ran <- do.call(pipes_to_run$run, dots)
        if (is_multicore) {
          self$PipelineList$update_pipelines(pipes_that_ran)
          # Collect pipelines whose inputs span independent branches were never
          # triggered inside any worker. Run them now with fully-synced state,
          # using run_pipe internally so their own downstream outputs recurse.
          do.call(self$PipelineList$run_pending_collects, dots)
        }
        elapsed <- tictoc::toc(quiet = TRUE)
        elapsed_ms <- elapsed$toc - elapsed$tic

        if (!quiet) {
          cli::cli_h3(
            "[{format(check_datetime + lubridate::milliseconds(elapsed_ms), '%Y-%m-%d %H:%M:%S')}]
            Pipeline execution completed {cli::col_silver(cli::symbol$stop)} | {elapsed$callback_msg}"
          )
        }

        run_errors <- self$PipelineList$get_errors()
        run_warnings <- self$PipelineList$get_warnings()
        run_messages <- self$PipelineList$get_messages()

        # For access via last_run_*
        maestro_pkgenv$last_run_errors <- run_errors
        maestro_pkgenv$last_run_warnings <- run_warnings
        maestro_pkgenv$last_run_messages <- run_messages

        status_table <- self$PipelineList$get_status()

        # Get the number of statuses
        invoked <- sum(status_table$invoked)
        error_count <- length(run_errors)
        success_count <- invoked - error_count
        warning_count <- length(run_warnings)

        if (!quiet) {
          cli::cli_text(
            "
        {cli::col_green(cli::symbol$tick)} {success_count} success{?es} |
        {cli::col_magenta('!')} {warning_count} warning{?s} |
        {cli::col_red(cli::symbol$cross)} {error_count} error{?s} |
        {cli::col_cyan(cli::symbol$square_small_filled)} {invoked} total
        "
          )

          if (error_count > 0) {
            cli::cli_alert_danger(
              "Use {.fn last_run_errors} to show pipeline errors."
            )
          }

          if (warning_count > 0) {
            cli::cli_alert_warning(
              "Use {.fn last_run_warnings} to show pipeline warnings."
            )
          }

          cli::cli_rule()

          # Output for showing next pipelines schedule
          if (!run_all && n_show_next > 0 && nrow(status_table) > 0) {

            next_runs_cli <- status_table |>
              dplyr::filter(!is.na(next_run)) |>
              dplyr::arrange(next_run) |>
              utils::head(n = n_show_next)

            cli::cli_h3("Next scheduled pipelines {cli::col_cyan(cli::symbol$pointer)}")
            next_run_strs <- glue::glue("{next_runs_cli$pipe_name} | {next_runs_cli$next_run}")
            cli::cli_text("Pipe name | Next scheduled run")
            cli::cli_ul(next_run_strs)
          }
        }
      }, error = function(e) {
        private$sch_status <- "Error"
        cli::cli_abort(
          "Failed to execute orchestrator with error: {e}"
        )
      }, warning = function(w) {
        private$sch_status <- "Warning"
        cli::cli_warn(
          "Orchestrator warned with {w}"
        )
      })
      private$sch_status <- "Success"

      return(invisible())
    },

    #' @description
    #' Get the schedule as a data.frame
    #' @return data.frame
    get_schedule = function() {
      self$PipelineList$get_schedule()
    },

    #' @description
    #' Get status of the pipelines as a data.frame
    #' @return data.frame
    get_status = function() {
      self$PipelineList$get_status() |> 
        dplyr::select(-internal_run_id)
    },

    #' @description
    #' Get artifacts (return values) from the pipelines
    #' @return list
    get_artifacts = function() {
      self$PipelineList$get_artifacts()
    },

    #' @description
    #' Get the network structure of the pipelines as an edge list (will be empty if there are no DAG pipelines)
    #' @return data.frame
    get_network = function() {
      self$PipelineList$get_network()
    },

    #' @description
    #' Get all pipeline flags as a long data.frame
    #' @return data.frame
    get_flags = function() {
      flag_list <- self$PipelineList$get_flags()
      purrr::imap(flag_list, ~{
        dplyr::tibble(
          pipe_name = .y,
          flag = .x
        )
      }) |>
        purrr::list_rbind()
    },

    #' @description
    #' Get all pipeline labels as a data.frame
    #' @return data.frame
    get_labels = function() {
      self$PipelineList$get_labels()
    },

    #' @description
    #' Get full sequence of scheduled executions for all pipelines
    #' @param n optional sequence limit
    #' @param min_datetime optional minimum datetime
    #' @param max_datetime optional maximum datetime
    #' @param include_only_primary only primary pipelines are included 
    #'   (this are pipelines that are scheduled and not downstream nodes in a DAG)
    #' @param include_skipped whether to include pipelines tagged with `@maestroSkip`
    #'   (default `TRUE` for backwards compatibility)
    #' @return data.frame
    get_run_sequence = function(n = NULL, min_datetime = NULL, max_datetime = NULL, include_only_primary = FALSE, include_skipped = TRUE) {

      pipeline_list <- self$PipelineList$MaestroPipelines

      if (!include_skipped) {
        pipeline_list <- purrr::discard(pipeline_list, ~isTRUE(.x$get_schedule()$skip))
      }

      run_sequence <- purrr::map(pipeline_list, ~.x$get_run_sequence(n = n, min_datetime = min_datetime, max_datetime = max_datetime)) |>
        stats::setNames(purrr::map_chr(pipeline_list, ~.x$get_schedule()$pipe_name)) |>
        purrr::discard(is.null) |>
        purrr::imap(
          ~dplyr::tibble(
            pipe_name = .y,
            scheduled_time = .x
          )
        ) |>
        purrr::list_rbind() |>
        dplyr::filter(!is.na(scheduled_time))

      network <- self$get_network()
      network$root <- find_roots(network$from, network$to)

      # Primary = not a downstream DAG node
      downstream_nodes <- network$to

      if (!include_only_primary) {
        run_sequence_dags <- NULL
        if (nrow(network) > 0) {
          run_sequence_dags <- purrr::map2(network$to, network$root, ~{
            run_sequence |>
              dplyr::filter(pipe_name == .y) |>
              dplyr::mutate(pipe_name = .x)
          }) |>
            purrr::list_rbind()
        }

        run_sequence <- run_sequence |>
          dplyr::bind_rows(run_sequence_dags) |>
          dplyr::arrange(scheduled_time)
      }

      run_sequence |>
        dplyr::mutate(is_primary = !pipe_name %in% downstream_nodes)
    }
  ),

  private = list(
    sch_status = "Not Run"
  )
)

test_that("MaestroSchedule works", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good"))

  schedule$run(
    orch_n = 1,
    orch_unit = "day"
  )

  expect_message({
    schedule$run(
      orch_n = 1,
      orch_unit = "day",
      run_all = TRUE
    )
  })

  status <- schedule$get_status()
  expect_snapshot(schedule$get_network())

  expect_s3_class(status, "data.frame")
  expect_in(
    c("pipe_name", "script_path", "invoked", "success", "pipeline_started",
      "pipeline_ended", "errors", "warnings", "messages", "next_run"
    ),
    names(status)
  )

  artifacts <- schedule$get_artifacts()

  expect_type(
    artifacts, "list"
  )
  expect_named(
    artifacts
  )
  expect_gt(length(artifacts), 0)

  expect_s3_class(status$pipeline_started, "POSIXct")
  expect_gt(nrow(status), 0)
  expect_length(last_run_errors(), 0)
  expect_length(last_run_warnings(), 0)
}) |>
  suppressMessages()

test_that("MaestroSchedule correctly returns artifacts (i.e., pipeline returns)", {

  schedule <- build_schedule(test_path("test_pipelines_run_artifacts"))

  output <- schedule$run(
    orch_n = 1,
    orch_unit = "day",
    run_all = TRUE
  )

  artifacts <- schedule$get_artifacts()

  expect_length(artifacts, 1)
  expect_equal(artifacts[[1]], "I'm an artifact")
}) |>
  suppressMessages()

test_that("MaestroSchedule works when not running all (verification of checking)", {

  schedule <- build_schedule(test_path("test_pipelines_run_all_good"))

  expect_message({
    schedule$run(
      orch_n = 1,
      orch_unit = "day"
    )}
  )
}) |>
  suppressMessages()

test_that("MaestroSchedule informs with an empty schedule", {
  schedule <- MaestroSchedule$new()
  expect_message({
    run_schedule(schedule)
  }, "No pipelines")
})

test_that("Multicore works", {

  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  future::plan(future::multisession(workers = 2))

  schedule <- build_schedule(
    test_path("test_pipelines_run_all_good"),
    cores = 2
  )

  expect_no_error({

    run_schedule(
      schedule,
      orch_frequency = "1 hour",
      cores = 2,
      log_to_console = TRUE,
      run_all = TRUE
    )
  })

  status <- get_status(schedule)

  expect_snapshot(status[, c("invoked", "success")])
}) |>
  suppressMessages()

test_that("Multicore DAGs work", {

  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  dag <- build_schedule(test_path("test_pipelines_dags_good"))
  future::plan(future::multisession(workers = 2))
  run_schedule(
    dag,
    cores = 2L
  )
  artifacts <- dag$get_artifacts()
  expect_true(all(dag$get_status()$success))
  expect_snapshot(dag$get_network())
  expect_equal(artifacts$with_inputs, "input message is: hello")
  expect_equal(artifacts$branch2, 2)
  expect_equal(artifacts$subbranch1, 2)
  expect_equal(artifacts$subbranch2, 6)
}) |>
  suppressMessages()

test_that("Simple fan out into common downstream", {

  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  future::plan(future::multisession(workers = 2))

  schedule <- build_schedule(test_path("test_pipelines_fan_out"))
  run_schedule(
    schedule,
    cores = 2L
  )
  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$add_2)), c(5, 8, 11))
  expect_equal(length(unique(status$run_id)), length(status$run_id))
})

test_that("Simple collect from distinct upstreams", {

  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  future::plan(future::multisession(workers = 2))

  schedule <- build_schedule(test_path("test_pipelines_fan_in"))
  run_schedule(
    schedule,
    cores = 2
  )

  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$ab)), "ab")
})

test_that("Collect pipeline with its own downstream output runs the downstream (multicore)", {

  # Gap #6: run_pending_collects fires the collect pipe on the main process
  # after worker sync, but doesn't recurse into the collect pipe's own outputs.
  # This test will fail until that recursion is implemented.
  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  future::plan(future::multisession(workers = 2))

  schedule <- build_schedule(test_path("test_pipelines_fan_in_chained"))
  run_schedule(schedule, cores = 2)

  status <- get_status(schedule)
  expect_true(all(status$invoked))
  expect_true(all(status$success))
  expect_equal(get_artifacts(schedule)$ab_upper, "AB")
})

test_that("Dynamic fan out followed by fan in", {

  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  future::plan(future::multisession(workers = 2))

  schedule <- build_schedule(test_path("test_pipelines_fan_out_collect"))
  run_schedule(
    schedule,
    cores = 2L
  )
  status <- get_status(schedule)
  expect_snapshot(status[, c("invoked", "success")])
  expect_equal(unlist(unname(get_artifacts(schedule)$add)), 18)
})

test_that("Collect pipeline with its own downstream output runs the downstream", {

  testthat::skip_if(Sys.getenv("MAESTRO_TEST_FUTURE") != "true")
  future::plan(future::multisession(workers = 2))

  schedule <- build_schedule(test_path("test_pipelines_fan_in_chained"))
  run_schedule(schedule, cores = 2L)

  status <- get_status(schedule)
  expect_true(all(status$invoked))
  expect_true(all(status$success))
  expect_equal(get_artifacts(schedule)$ab_upper, "AB")
})

test_that("DAGs work as expected", {
  dag <- build_schedule(test_path("test_pipelines_dags_good"))
  run_schedule(dag)
  artifacts <- dag$get_artifacts()
  expect_true(all(dag$get_status()$success))
  expect_snapshot(dag$get_network())
  expect_equal(artifacts$with_inputs, "input message is: hello")
  expect_equal(artifacts$branch2, 2)
  expect_equal(artifacts$subbranch1, 2)
  expect_equal(artifacts$subbranch2, 6)
}) |>
  suppressMessages()

test_that("get_network() standalone returns same result as R6 method", {
  dag <- build_schedule(test_path("test_pipelines_dags_good"))
  expect_equal(get_network(dag), dag$get_network())
}) |>
  suppressMessages()

test_that("get_network() errors on non-MaestroSchedule input", {
  expect_error(get_network(list()), class = "rlang_error")
})

test_that("Error messages for DAGs with nonexistent inputs/outputs", {
  expect_error({
    dag_bad <- build_schedule(test_path("test_pipelines_dags_bad"))
  }, regexp = "Pipeline")
})

test_that("Error in a DAG pipeline stops downstream computations", {
  dag_bad <- build_schedule(test_path("test_pipelines_dags_run_bad"))
  run_schedule(dag_bad)
  status <- get_status(dag_bad)
  expect_true(status$invoked[status$pipe_name == "get_num"])
  expect_true(!status$invoked[status$pipe_name == "multiply"])
}) |>
  suppressMessages()

test_that("Providing just maestroInput or just maestroOutput works fine", {

  # Full redundancy - all inputs and outputs defined
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroOutputs high_road low_road
      start <- function() {
        c('a', 'A')
      }

      #' @maestroInputs start
      high_road <- function(.input) {
        toupper(.input)
      }

      #' @maestroInputs start
      low_road <- function(.input) {
        tolower(.input)
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    network_red <- schedule$get_network()
    run_schedule(schedule)
    status_red <- get_status(schedule)

    # Just outputs
    unlink("pipelines/dags.R")
    writeLines(
      "
      #' @maestroOutputs high_road low_road
      start <- function() {
        c('a', 'A')
      }

      #' @maestroFrequency 1 day
      high_road <- function(.input) {
        toupper(.input)
      }

      #' @maestroFrequency 1 day
      low_road <- function(.input) {
        tolower(.input)
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    network_jo <- schedule$get_network()
    run_schedule(schedule)
    status_jo <- get_status(schedule)

    # Just inputs
    unlink("pipelines/dags.R")
    writeLines(
      "
      #' @maestroFrequency 1 day
      start <- function() {
        c('a', 'A')
      }

      #' @maestroInputs start
      high_road <- function(.input) {
        toupper(.input)
      }

      #' @maestroInputs start
      low_road <- function(.input) {
        tolower(.input)
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    network_ji <- schedule$get_network()
    run_schedule(schedule)
    status_ji <- get_status(schedule)

    expect_equal(network_red, network_jo)
    expect_equal(network_jo, network_ji)
    expect_equal(status_red$invoked, status_jo$invoked)
    expect_equal(status_jo$invoked, status_ji$invoked)
  })
}) |>
  suppressMessages()

test_that("Resources are passed correctly for pipes with inputs", {
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroOutputs end
      start <- function(start_val) {
        start_val
      }

      #' @maestro
      end <- function(.input, val) {
        .input * val
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule,
      resources = list(
        start_val = 2,
        val = 4
      )
    )
    out <- get_artifacts(schedule)
    expect_equal(out$end, 2 * 4)
  })
}) |>
  suppressMessages()

test_that("DAG with a loop fails validation", {

  # Loop with no primary
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "#' @maestroOutputs loopy2
      loopy1 <- function(.input = 1) {
        .input * 2
      }

      #' @maestroOutputs loopy3
      loopy2 <- function(.input = 1) {
        .input * 3
      }

      #' @maestroOutputs loopy1
      loopy3 <- function(.input = 1) {
        .input * 4
      }",
      con = "pipelines/dags.R"
    )

    expect_error({
      schedule <- build_schedule()
    }, regexp = "Invalid DAG")
  })

  # Primary followed by loop
  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "#' @maestroOutputs loopy2
      loopy1 <- function(.input = 1) {
        .input * 2
      }

      #' @maestroOutputs loopy3
      loopy2 <- function(.input = 1) {
        .input * 3
      }

      #' @maestroOutputs loopy2
      loopy3 <- function(.input = 1) {
        .input * 4
      }",
      con = "pipelines/dags.R"
    )

    expect_error({
      schedule <- build_schedule()
    }, regexp = "Invalid DAG")
  })
})

test_that("Even if a downstream pipeline is 'scheduled' it doesn't run unless the upstream component does", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 09:00:00
      #' @maestroTz UTC
      #' @maestroOutputs end
      start <- function() {
        2
      }

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 08:00:00
      #' @maestroTz UTC
      end <- function(.input) {
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-11-22 08:00:00", tz = "UTC")
    )

    expect_snapshot(schedule$get_status()$invoked)
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 09:00:00
      #' @maestroTz UTC
      #' @maestroOutputs mid
      start <- function() {
        2
      }

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 08:00:00
      #' @maestroTz UTC
      #' @maestroOutputs end
      mid <- function(.input) {
        .input / 10
      }

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 08:00:00
      #' @maestroTz UTC
      end <- function(.input) {
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-11-22 08:00:00", tz = "UTC")
    )

    expect_snapshot(schedule$get_status()$invoked)
  })
}) |>
  suppressMessages()

test_that("Even if a downstream pipeline is 'scheduled' it runs if the upstream component does", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 09:00:00
      #' @maestroTz UTC
      #' @maestroOutputs end
      start <- function() {
        2
      }

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 10:00:00
      #' @maestroTz UTC
      end <- function(.input) {
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-11-22 09:00:00", tz = "UTC")
    )

    expect_snapshot(schedule$get_status()$invoked)
  })

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 09:00:00
      #' @maestroTz UTC
      #' @maestroOutputs mid
      start <- function() {
        2
      }

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 10:00:00
      #' @maestroTz UTC
      #' @maestroOutputs end
      mid <- function(.input) {
        .input / 10
      }

      #' @maestroFrequency 1 hour
      #' @maestroStartTime 2024-11-22 10:00:00
      #' @maestroTz UTC
      end <- function(.input) {
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule,
      orch_frequency = "hourly",
      check_datetime = as.POSIXct("2024-11-22 09:00:00", tz = "UTC")
    )

    expect_snapshot(schedule$get_status()$invoked)
  })
}) |>
  suppressMessages()

test_that("Branching DAG pipelines with an error in one branch continue on second branch", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroOutputs mid
      start <- function() {
        4
      }

      #' @maestroOutputs end1 end2
      mid <- function(.input) {
        .input * 3
      }

      #' @maestro
      end1 <- function(.input) {
        stop()
        .input * 2
      }
        
      #' @maestro
      end2 <- function(.input) {
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
    expect_snapshot(status[, c("invoked", "success")])
  })
})

test_that("Branching DAG pipelines both ending in errors accurately outputs errors", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroOutputs mid
      start <- function() {
        4
      }

      #' @maestroOutputs end1 end2
      mid <- function(.input) {
        .input * 3
      }

      #' @maestro
      end1 <- function(.input) {
        stop('oops')
        .input * 2
      }
        
      #' @maestro
      end2 <- function(.input) {
        stop('oh dear')
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
    expect_snapshot(status[, c("invoked", "success")])
    expect_snapshot(last_run_errors())
  })
})

test_that("Branching and merging DAG pipelines have separate status entries for each lineage", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroOutputs mid1 mid2
      start <- function() {
        4
      }

      #' @maestroOutputs end
      mid1 <- function(.input) {
        .input * 3
      }

      #' @maestroOutputs end
      mid2 <- function(.input) {
        .input * 2
      }
        
      #' @maestro
      end <- function(.input) {
        .input * 2
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
    expect_snapshot(status[, c("invoked", "success")])
    expect_snapshot(unname(unlist(get_artifacts(schedule))))
  })
})

test_that("Branching and merging DAG pipelines use vectors for multiple errors", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroOutputs mid1 mid2
      start <- function() {
        4
      }

      #' @maestroOutputs end
      mid1 <- function(.input) {
        .input * 3
      }

      #' @maestroOutputs end
      mid2 <- function(.input) {
        .input * 2
      }
        
      #' @maestro
      end <- function(.input) {
        stop('oops')
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
    expect_snapshot(status[, c("invoked", "success")])
    expect_snapshot(unname(unlist(get_artifacts(schedule))))
    expect_snapshot(last_run_errors())
  })
})

test_that("Two separate DAGs have separate lineages", {

  withr::with_tempdir({
    dir.create("pipelines")
    writeLines(
      "
      #' @maestroFrequency daily
      #' @maestroOutputs end
      start <- function() {
        4
      }

      #' @maestro
      end <- function(.input) {
        .input * 3
      }

      #' @maestroFrequency daily
      #' @maestroPriority 1
      start2 <- function() {
        3
      }
      
      #' @maestroInputs start2
      end2 <- function(.input) {
        .input * 5
      }",
      con = "pipelines/dags.R"
    )

    schedule <- build_schedule()
    run_schedule(
      schedule
    )
    status <- get_status(schedule)
    expect_snapshot(status[, c("invoked", "success")])
  })
})

#' Visualize the schedule as a DAG
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' Create an interactive network visualization to show the dependency structure
#' of pipelines in the schedule. This is only useful if there are pipelines in
#' the schedule that take inputs/outputs from other pipelines.
#'
#' Note that running this function on a schedule with all independent pipelines
#' will produce a network visual with no connections.
#'
#' This function requires the installation of `DiagrammeR` which is not automatically
#' installed with `maestro`.
#'
#' @inheritParams run_schedule
#'
#' @return DiagrammeR visualization
#' @export
#' @keywords internal
#'
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   schedule <- run_schedule(
#'     schedule,
#'     orch_frequency = "1 day",
#'     quiet = TRUE
#'   )
#'
#'   show_network(schedule)
#' }
show_network <- function(schedule) {

  lifecycle::deprecate_warn(
    "1.1.0",
    "show_network()",
    details = "It will be removed entirely in 1.2.0."
  )

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  schedule$show_network()
}

#' Suggest orchestrator frequency based on a schedule
#'
#' Suggests a frequency to run the orchestrator based on the frequencies of the
#' pipelines in a schedule.
#'
#' This function attempts to find the smallest interval of time between all pipelines.
#' If the smallest interval is less than 15 minutes, it just uses the smallest interval.
#'
#' Note this function is intended to be used interactively when deciding how often to
#' schedule the orchestrator. Programmatic use is not recommended.
#'
#' **Deprecated.** This function is no longer actively maintained and may be removed
#' in a future release.
#' @param schedule MaestroSchedule object created by `build_schedule()`
#' @inheritParams get_pipeline_run_sequence
#'
#' @return frequency string
#' @export
#'
#' @keywords internal
#'
#' @examples
#'
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'   suggest_orch_frequency(schedule)
#' }
suggest_orch_frequency <- function(schedule, check_datetime = lubridate::now(tzone = "UTC")) {

  lifecycle::deprecate_warn("1.1.0", "suggest_orch_frequency()")

  # Check that schedule is a MaestroSchedule
  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  schedule <- schedule$get_schedule() |>
    dplyr::filter(!skip, !is.na(frequency_n))

  if (nrow(schedule) == 0) {

    cli::cli_abort(
      c(
        "No pipelines in schedule after removing skipped pipelines.",
        "i" = "Remove `maestroSkip` tags to get a suggested frequency."
      )
    )
  }

  sch_secs <- purrr::map_int(
    paste(schedule$frequency_n, schedule$frequency_unit),
    purrr::possibly(convert_to_seconds, otherwise = NA, quiet = TRUE)
  )

  if (min(sch_secs, na.rm = TRUE) <= (60 * 15)) {
    return(schedule$frequency[[which.min(sch_secs)]])
  }

  if (nrow(schedule) == 1) {
    return(schedule$frequency)
  }

  max_idx <- which.max(sch_secs)
  max_freq <- paste(schedule$frequency_n[[max_idx]], schedule$frequency_unit[[max_idx]])

  pipeline_sequences <- purrr::pmap(
    list(schedule$frequency_n, schedule$frequency_unit, schedule$start_time),
    ~{
      pipeline_sequence <- get_pipeline_run_sequence(
        ..1, ..2, ..3,
        check_datetime = check_datetime + lubridate::seconds(convert_to_seconds(max_freq))
      )

      pipeline_sequence[pipeline_sequence >= check_datetime]
    }
  ) |>
    purrr::list_c() |>
    unique() |>
    sort()

  if (length(pipeline_sequences) == 1) return(max_freq)

  pipeline_diffs <- diff(pipeline_sequences)

  min_diff_secs <- min(pipeline_diffs)

  min_diff_atts <- attributes(min_diff_secs)

  paste(round(as.numeric(min_diff_secs)), min_diff_atts$units)
}

#' @maestroFrequency daily
numbers <- function() {
  1:3
}

#' @maestroInputs numbers
#' @maestroMap
multiply <- function(.input) {
  .input * 3
}
  
#' @maestroInputs multiply
add_2 <- function(.input) {
  .input + 2
}

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-01 09:00:00
#' @maestroTz UTC
#' @maestroLogLevel INFO
#'
#' @export
get_mtcars <- function() {
  Sys.sleep(0.02)
  mtcars
}

#' Multiply
#'
#' @maestroFrequency 3 month
#' @maestroLogLevel warn
#'
#' @export
wait <- function() {
  Sys.sleep(0.01)
}

#' Add
#'
#' @maestroFrequency 3 month
#' @maestroStartTime 1970-01-01
#'
#' @export
add <- function() {
  invisible()
}

#' Something
#'
#' @maestroFrequency 3 month
#' @maestroStartTime 1970-01-01 00:00:00 ADT
#'
#' @export
something <- function() {
  invisible()
}

#' Daily
#'
#' @maestroFrequency daily
#' @maestroStartTime 1970-01-01 00:00:00 ADT
#'
#' @export
something2 <- function() {
  invisible()
}

#' Get the flags of pipelines in a MaestroSchedule object
#'
#' Creates a long data.frame where each row is a flag for each pipeline.
#'
#' @inheritParams run_schedule
#'
#' @return data.frame
#' @export
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   get_flags(schedule)
#'
#'   # Alternatively, use the underlying R6 method
#'   schedule$get_flags()
#' }
get_flags <- function(schedule) {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  schedule$get_flags()
}

#' @maestroFrequency
data <- iris |>
  dplyr::filter() |>
  dplyr::summarise()

#' @maestroFrequency
data |>
  head(n = 6)

#' @maestroFrequency
this_one_is_good <- function() {
  # But the fact that there are bad ones above makes it bad
  iris
}

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroTz UTC
#'
#' @export
get_mtcars <- function() {
  mtcars
}

#' Multiply
#'
#' @maestroFrequency 3 months
#'
#' @export
multiply <- function(val, by) {
  val * by
}

#' Bad
#'
#' @maestroTz BLA
something_else <- function() {

}

# Simple 1 -> 1 -----------------------------------------------------------
#' @maestroInputs primary
with_inputs <- function(.input) {
  paste("input message is:", .input)
}

#' @maestroFrequency daily
#' @maestroOutputs with_inputs
primary <- function() {
  "hello"
}


# Tree with Two Branches x 2 ----------------------------------------------
#' @maestroFrequency daily
#' @maestroOutputs branch1 branch2
trunk <- function() {
  1
}

#' @maestroInputs trunk
#' @maestroOutputs subbranch1
branch1 <- function(.input) {
  .input * 1
}

#' @maestroInputs trunk
#' @maestroOutputs subbranch2
branch2 <- function(.input) {
  .input * 2
}

#' @maestroInputs branch1
subbranch1 <- function(.input) {
  .input * 2
}

#' @maestroInputs branch2
subbranch2 <- function(.input) {
  .input * 3
}

# Split and Unite ---------------------------------------------------------

#' @maestroOutputs high_road low_road
start <- function() {
  c("a", "A")
}

#' @maestroInputs start
#' @maestroOutputs end
high_road <- function(.input) {
  toupper(.input)
}

#' @maestroInputs start
#' @maestroOutputs end
low_road <- function(.input) {
  tolower(.input)
}

#' @maestroInputs high_road low_road
end <- function(.input) {
  c(.input, "b")
}

# Simple 1 -> 1 -----------------------------------------------------------
#' @maestroInputs get_num
multiply <- function(.input) {
  .input * 4
}

#' @maestroFrequency daily
#' @maestroOutputs multiply
get_num <- function() {
  stop("oops")
}

#' Get the network structure of pipelines in a MaestroSchedule object
#'
#' Returns the pipeline dependency structure as an edge list data.frame.
#' Each row represents a directed dependency between two pipelines. The
#' result will be empty if there are no DAG pipelines in the schedule.
#'
#' @inheritParams run_schedule
#'
#' @return data.frame with columns `from` and `to`
#' @export
#' @examples
#' if (interactive()) {
#'   pipeline_dir <- tempdir()
#'   create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
#'   schedule <- build_schedule(pipeline_dir = pipeline_dir)
#'
#'   get_network(schedule)
#'
#'   # Alternatively, use the underlying R6 method
#'   schedule$get_network()
#' }
get_network <- function(schedule) {

  if (!"MaestroSchedule" %in% class(schedule)) {
    cli::cli_abort(
      c("Schedule must be an object of {.cls MaestroSchedule} and not an object of class {.cls {class(schedule)}}.",
        "i" = "Use {.fn build_schedule} to create a valid schedule."),
      call = rlang::caller_env()
    )
  }

  schedule$get_network()
}

#' Parse and validate tags then create and populate MaestroPipelineList
#'
#' @param script_path path to script
#'
#' @keywords internal
#'
#' @return MaestroPipelineList R6 class
build_schedule_entry <- function(script_path) {

  # Current list of maestro tags and their equivalent table names
  maestro_tag_names <- list(
    frequency = "maestroFrequency",
    start_time = "maestroStartTime",
    tz = "maestroTz",
    skip = "maestroSkip",
    log_level = "maestroLogLevel",
    hours = "maestroHours",
    days = "maestroDays",
    months = "maestroMonths",
    inputs = "maestroInputs",
    outputs = "maestroOutputs",
    maestro = "maestro",
    priority = "maestroPriority",
    flags = "maestroFlags",
    run_if = "maestroRunIf",
    map = "maestroMap"
  )

  # List of maestro tags that can be used more than once
  maestro_tag_names_mult <- list(
    labels = "maestroLabel"
  )

  # Initial Validation ------------------------------------------------------

  # Get all the roxygen tags - this actually executes the script, which is fine
  # if it's functions but can be problematic if it includes other code that
  # evaluates immediately
  tag_list <- tryCatch({
    parse_env <- new.env()
    parse <- function() roxygen2::parse_file(script_path)
    environment(parse) <- parse_env
    parse()
  }, error = \(e) {
    cli::cli_abort(
      c("Could not build schedule entry with error: {e$message}",
        "i" = "Pipeline scripts should not typically include code that's run
        outside of a function. Be sure to wrap code in a function block."),
      call = NULL
    )
  }, warning = \(w) {
    cli::cli_abort(w$message, call = NULL)
  }) |>
    suppressMessages()

  # Get specifically the tags used by maestro
  withCallingHandlers({
    maestro_tag_vals <- purrr::map(tag_list, ~{
      tag <- .x
      single_vals <- purrr::map(
        maestro_tag_names,
        ~{
          roxygen2::block_get_tag_value(tag, .x)
        }
      )
      multi_vals <- purrr::map(
        maestro_tag_names_mult,
        ~{
          roxygen2::block_get_tags(tag, .x) |> 
            purrr::map("val")
        }
      )
      append(single_vals, multi_vals)
    })
  },
  purrr_error_indexed = function(err) {
    rlang::cnd_signal(err$parent)
  })

  if (length(maestro_tag_vals) == 0) {
    cli::cli_abort(
      c("No {.pkg maestro} tags present in {basename(script_path)}.",
        "i" = "A valid pipeline must have at least one function with one or
        more {.pkg maestro} tags: e.g., `#' @maestroFrequency 1 day`."),
      call = NULL
    )
  }

  # Get pipe names from the function name and check
  withCallingHandlers({
    pipe_names <- purrr::imap(tag_list, ~{

      obj_class <- class(.x$object)

      # Check that it is a function
      if (!"function" %in% obj_class) {
        cli::cli_abort(
          c("{basename(script_path)} line {(.x$line)} has tags but no function. Be sure place
          tags above the function you want to schedule."),
          call = NULL
        )
      }

      # Return the name
      .x$object$topic
    })
  }, purrr_error_indexed = function(err) {
    rlang::cnd_signal(err$parent)
  })

  # Create MaestroPipelineList ----------------------------------------------

  maestro_pipeline_list <- MaestroPipelineList$new()

  withCallingHandlers({
    pipelines <- purrr::pmap(list(tag_list, pipe_names, maestro_tag_vals), ~{
      block  <- ..1
      .x <- ..2
      .y <- ..3

      # Validate inputs
      if (!all(is.null(.y$inputs))) {
        if (.x %in% .y$inputs) {
          cli::cli_abort(
            c("`@maestroInput` cannot contain self-references. Pipeline {.pkg .x} in {basename(script_path)}
              contains an input with the same name."),
            call = NULL
          )
        }
        params <- roxygen2::block_get_tag_value(block, ".formals")
        if (!".input" %in% params) {
          cli::cli_abort(
            c("If specifying `@maestroInputs` the pipeline must have a parameter named `.input`",
              "i" = "Example: {.code pipeline <- function(.input) ...}"),
            call = NULL
          )
        }
      }

      # Validate outputs
      if (!all(is.null(.y$outputs))) {
        if (.x %in% .y$outputs) {
          cli::cli_abort(
            c("`@maestroOutput` cannot contain self-references. Pipeline {.pkg .x} in {basename(script_path)}
              contains an output with the same name."),
            call = NULL
          )
        }
      }

      freq_nunits <- if (!is.null(.y$frequency)) parse_rounding_unit(.y$frequency) else NULL

      if (!is.null(freq_nunits)) {

        freq_unit <- freq_nunits$unit

        # Validate hours
        if (
          !all(is.null(.y$hours)) && !freq_unit %in% c("second", "minute", "hour")
        ) {
          cli::cli_abort(
            c("If specifying `@maestroHours` the pipeline must have a `@maestroFrequency` of at least 'hourly'.",
              "i" = "Issue is with pipeline named {.x}."),
            call = NULL
          )
        }

        # Validate days
        if (!all(is.null(.y$days)) &&
            !freq_unit %in% c("second", "minute", "hour", "day")) {
          cli::cli_abort(
            c("If specifying `@maestroDays` the pipeline must have a `@maestroFrequency` of at least 'daily'.",
              "i" = "Issue is with pipeline named {.x}."),
            call = NULL
          )
        }

        # Validate months
        if (!all(is.null(.y$months)) && !freq_unit %in%
            c("second", "minute", "hour", "day", "week", "month")) {
          cli::cli_abort(
            c("If specifying `@maestroMonths` the pipeline must have a `@maestroFrequency` of
            at least 'monthly'.",
              "i" = "Issue is with pipeline named {.x}."),
            call = NULL
          )
        }
      }

      tz <- .y$tz %n% "UTC"

      .weekday_abbrs <- c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
      is_time_only   <- !is.null(.y$start_time) &&
        grepl("^[0-9]{2}:[0-9]{2}:[0-9]{2}$", .y$start_time %n% "")
      is_weekday_fmt <- !is.null(.y$start_time) &&
        grepl(
          sprintf("^(%s)\\s+[0-9]{2}:[0-9]{2}:[0-9]{2}$",
                  paste(.weekday_abbrs, collapse = "|")),
          .y$start_time %n% ""
        )
      is_monthday_fmt <- !is.null(.y$start_time) &&
        grepl("^[0-9]{1,2}(\\s+[0-9]{2}:[0-9]{2}:[0-9]{2})?$",
              .y$start_time %n% "")

      if (!is.null(freq_nunits)) {

        if (is_time_only) {
          if (freq_nunits$n > 1 && freq_nunits$unit == "day") {
            cli::cli_abort(
              c("Cannot use a `@maestroStartTime` with format %H:%M:%S in combination with
                a multi-unit daily `@maestroFrequency`.",
                "i" = "Instead use something like `daily` or `1 day`."),
              call = NULL
            )
          }
          if (freq_nunits$unit %in% c("week", "month", "year")) {
            cli::cli_abort(
              c("`@maestroStartTime` with format %H:%M:%S is only valid for `@maestroFrequency` of minute, hour, or day."),
              call = NULL
            )
          }
        }

        if (is_weekday_fmt && !freq_nunits$unit %in% c("week")) {
          cli::cli_abort(
            c("`@maestroStartTime` with weekday format (e.g., `Mon 04:00:00`) is only valid for weekly `@maestroFrequency`.",
              "i" = "Issue is with pipeline named {.x}."),
            call = NULL
          )
        }

        if (is_monthday_fmt && !freq_nunits$unit %in% c("month")) {
          cli::cli_abort(
            c("`@maestroStartTime` with month-day format (e.g., `15 04:00:00`) is only valid for monthly `@maestroFrequency`.",
              "i" = "Issue is with pipeline named {.x}."),
            call = NULL
          )
        }
      }

      # Pass the raw start_time string through; MaestroPipeline resolves it
      start_time_raw <- .y$start_time %n% NA_character_

      if (!is.null(.y$inputs)) {
        inputs <- .y$inputs$inputs
        is_collect <- .y$inputs$is_collect
      } else {
        inputs <- NULL
        is_collect <- FALSE
      }

      # Parse and validate @maestroMap
      map <- if (!is.null(.y$map)) {
        exprs <- .y$map  # character vector, one element per iterator
        # Validate that every element is parseable R
        bad <- purrr::keep(exprs, \(e) {
          inherits(tryCatch(str2lang(e), error = \(err) err), "error")
        })
        if (length(bad) > 0) {
          cli::cli_abort(
            c(
              "`@maestroMap` contains invalid R expression{?s}: {.code {bad}}",
              "i" = "Issue is with pipeline named {.x}."
            ),
            call = NULL
          )
        }
        exprs
      } else {
        NULL
      }

      # Create the new pipeline
      MaestroPipeline$new(
        script_path = script_path,
        pipe_name = .x,
        frequency = .y$frequency %n% "daily",
        start_time_raw = start_time_raw,
        tz = tz,
        skip = .y$skip %n% FALSE,
        log_level = .y$log_level %n% "INFO",
        hours = .y$hours %n% 0:23,
        days = .y$days %n% NULL,
        months = .y$months %n% 1:12,
        inputs = inputs,
        outputs = .y$outputs %n% NULL,
        priority = as.numeric(.y$priority %n% Inf),
        flags = .y$flags %n% character(),
        run_if = .y$run_if %n% NULL,
        is_collect = is_collect,
        map = map,
        labels = .y$labels
      )
    })
  }, purrr_error_indexed = function(err) {
    rlang::cnd_signal(err$parent)
  })

  maestro_pipeline_list$MaestroPipelines <- pipelines
  maestro_pipeline_list$n_pipelines <- length(pipelines)

  maestro_pipeline_list
}

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroTz America/Halifax
get_mtcars <- function() {
  Sys.sleep(0.02)
  mtcars
}

#' Multiply
#'
#' @maestroFrequency 3 month
wait <- function() {
  Sys.sleep(0.01)
}

#' @maestroFrequency 1 week
weekly <- function() {
  1 + 1
}

#' @maestroFrequency 1 day
way_in_the_future <- function() {
  invisible()
}

#' @maestroFrequency weekly
weekly2 <- function() {
  1 + 1
}

#' Class for a list of MaestroPipelines
#' A MaestroPipelineList is created when there are multiple maestro pipelines in a
#' single script
#' @keywords internal
MaestroPipelineList <- R6::R6Class(

  "MaestroPipelineList",

  public = list(

    #' @field MaestroPipelines list of pipelines
    #' @field n_pipelines number of pipelines in the list
    MaestroPipelines = list(),
    n_pipelines = 0L,

    #' @description
    #' Create a MaestroPipelineList object
    #' @param MaestroPipelines list of MaestroPipelines
    #' @param network initialize a network
    #' @return MaestroPipelineList
    initialize = function(MaestroPipelines = list(), network = NULL) {
      self$n_pipelines <- length(MaestroPipelines)
      self$MaestroPipelines <- MaestroPipelines
      private$network <- network
    },

    #' @description
    #' Print the MaestroPipelineList
    #' @return print
    print = function() {
      cli::cli_h3("Maestro Pipelines List with {length(self$MaestroPipelines)} pipeline{?s}")
    },

    #' @description
    #' Add pipelines to the list
    #' @param MaestroPipelines list of MaestroPipelines
    #' @return invisible
    add_pipelines = function(MaestroPipelines = NULL) {
      if (inherits(MaestroPipelines, "MaestroPipeline")) {
        self$n_pipelines <- self$n_pipelines + 1
        self$MaestroPipelines <- append(self$MaestroPipelines, MaestroPipelines)
      } else {
        purrr::walk(MaestroPipelines$MaestroPipelines, ~{
          self$n_pipelines <- self$n_pipelines + 1
          self$MaestroPipelines <- append(self$MaestroPipelines, .x)
        })
      }
      invisible()
    },

    #' @description
    #' Update pipelines in a list
    #' @param MaestroPipelines list of MaestroPipelines
    #' @return invisible
    update_pipelines = function(MaestroPipelines = NULL) {
      pipe_names <- self$get_pipe_names()
      if (inherits(MaestroPipelines, "MaestroPipeline")) {
        pipe_to_update_name <- MaestroPipelines$get_pipe_name()
        idx_to_update <- which(pipe_names == pipe_to_update_name)
        self$MaestroPipelines[[idx_to_update]] <- MaestroPipelines
      } else {
        purrr::walk(MaestroPipelines, ~{
          pipe_to_update_name <- .x$get_pipe_name()
          idx_to_update <- which(pipe_names == pipe_to_update_name)
          self$MaestroPipelines[[idx_to_update]] <- .x
        })
      }
      invisible()
    },
    
    #' @description
    #' Get names of the pipelines in the list arranged by priority
    #' @return character
    get_pipe_names = function() {
      purrr::map_chr(self$MaestroPipelines, ~.x$get_pipe_name())
    },

    #' @description
    #' Get a MaestroPipeline by its name
    #' @param pipe_name name of the pipeline
    #' @return MaestroPipeline
    get_pipe_by_name = function(pipe_name) {
      names <- self$get_pipe_names()
      name_idx <- which(names %in% pipe_name)
      if (length(name_idx) == 0) {
        cli::cli_abort("No pipeline named {pipe_name} in {.cls MaestroPipelineList}")
      }
      self$MaestroPipelines[[name_idx]]
    },

    #' @description
    #' Get a MaestroPipelineList with selected pipelines
    #' @param pipe_names names of the pipelines
    #' @return MaestroPipelineList
    get_pipes_by_name = function(pipe_names) {
      names <- self$get_pipe_names()
      names_idx <- which(names %in% pipe_names)
      if (length(names_idx) == 0) {
        cli::cli_abort("No pipelines named {pipe_names} in {.cls MaestroPipelineList}")
      }
      self$MaestroPipelines[names_idx]
    },

    #' @description
    #' Get priorities
    #' @return numeric
    get_priorities = function() {
      purrr::map_dbl(self$MaestroPipelines, ~.x$get_priority())
    },

    #' @description
    #' Get the schedule as a data.frame
    #' @return data.frame
    get_schedule = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_schedule()) |>
        purrr::list_rbind()
    },

    #' @description
    #' Get a new MaestroPipelineList containing only those pipelines scheduled to run
    #' @param ... arguments passed to self$check_timeliness
    #' @return MaestroPipelineList
    get_timely_pipelines = function(...) {
      dots <- rlang::list2(...)
      timely_pipelines_idx <- do.call(self$check_timeliness, dots)
      MaestroPipelineList$new(self$MaestroPipelines[timely_pipelines_idx], network = private$network)
    },

    #' @description
    #' Get pipelines that are primary (i.e., don't have an inputting pipeline)
    #' @return list of MaestroPipelines
    get_primary_pipes = function() {
      network <- self$get_network()
      names <- self$get_pipe_names()
      primary_pipelines_idx <- which(!names %in% network$to)
      self$MaestroPipelines[primary_pipelines_idx]
    },

    #' @description
    #' Check whether pipelines in the list are scheduled to run based on orchestrator frequency and current time
    #' @param ... arguments passed to self$check_timeliness
    #' @return logical
    check_timeliness = function(...) {
      dots <- rlang::list2(...)
      purrr::map_lgl(self$MaestroPipelines, ~do.call(.x$check_timeliness, dots))
    },

    #' @description
    #' Get status of the pipelines as a data.frame
    #' @return data.frame
    get_status = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_status()) |>
        purrr::list_rbind()
    },

    #' @description
    #' Get list of errors from the pipelines
    #' @return list
    get_errors = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_errors()) |>
        purrr::set_names(self$get_pipe_names()) |> 
        purrr::discard(is.null)
    },

    #' @description
    #' Get list of warnings from the pipelines
    #' @return list
    get_warnings = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_warnings()) |>
        purrr::set_names(self$get_pipe_names()) |> 
        purrr::discard(is.null)
    },

    #' @description
    #' Get list of messages from the pipelines
    #' @return list
    get_messages = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_messages()) |>
        purrr::set_names(self$get_pipe_names()) |> 
        purrr::discard(is.null)
    },

    #' @description
    #' Get artifacts (return values) from the pipelines
    #' @return list
    get_artifacts = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_artifacts()) |>
        stats::setNames(self$get_pipe_names()) |>
        purrr::discard(is.null) |> 
        purrr::discard(~length(.x) == 0)
    },

    #' @description
    #' Get run sequences from the pipelines
    #' @param n optional sequence limit
    #' @param min_datetime optional minimum datetime
    #' @param max_datetime optional maximum datetime
    #' @return list
    get_run_sequences = function(n = NULL, min_datetime = NULL, max_datetime = NULL) {
      purrr::map(self$MaestroPipelines, ~.x$get_run_sequence(n = n, min_datetime = min_datetime, max_datetime = max_datetime)) |>
        stats::setNames(self$get_pipe_names()) |>
        purrr::discard(is.null)
    },

    #' @description
    #' Get the flags of the pipelines as a named list
    #' @return list
    get_flags = function() {
      purrr::map(self$MaestroPipelines, ~.x$get_flags()) |>
        stats::setNames(self$get_pipe_names())
    },

    #' @description
    #' Get the labels of the pipelines as a data.frame
    #' @return data.frame
    get_labels = function() {
      purrr::map(self$MaestroPipelines, ~{
        .x$get_labels() |> 
          dplyr::mutate(pipe_name = .x$get_pipe_name(), .before = 0)
      }) |>
        purrr::list_rbind()
    },

    #' @description
    #' Get the network structure as a edge list
    #' @return data.frame
    get_network = function() {
      if (!is.null(private$network)) return(private$network)

      network <- dplyr::tibble(
        from = character(),
        to = character()
      )

      if (length(self$MaestroPipelines) > 0) {
        network <- purrr::map(self$MaestroPipelines, ~{
          network_dat <- dplyr::tibble(
            from = character(),
            to = character()
          )
          to <- .x$get_outputs()
          from <- .x$get_inputs()
          this <- .x$get_pipe_name()
          if (!is.null(to)) {
            network_dat <- network_dat |>
              dplyr::bind_rows(
                dplyr::tibble(
                  from = this,
                  to = to
                )
              )
          }

          if (!is.null(from)) {
            network_dat <- network_dat |>
              dplyr::bind_rows(
                dplyr::tibble(
                  from = from,
                  to = this
                )
              )
          }
          network_dat
        }) |>
          purrr::list_rbind() |>
          dplyr::distinct(from, to)
      }

      private$network <- network
      network
    },

    #' @description
    #' Validates whether all inputs and outputs exist and that the network is a valid DAG
    #' @return warning or invisible
    validate_network = function() {

      pipe_names <- self$get_pipe_names()

      inputs <- purrr::map(self$MaestroPipelines, ~.x$get_inputs()) |>
        purrr::set_names(pipe_names) |>
        purrr::keep(~!is.null(.x))

      outputs <- purrr::map(self$MaestroPipelines, ~.x$get_outputs()) |>
        purrr::set_names(pipe_names) |>
        purrr::keep(~!is.null(.x))

      if (length(inputs) > 0) {
        withCallingHandlers({
          purrr::iwalk(inputs, ~{
            if (!all(.x %in% pipe_names)) {
              invalid <- .x[which(!.x %in% pipe_names)]
              cli::cli_abort(
                "Pipeline {.pkg {.y}} references non-existent input pipeline{?s} {.pkg {invalid}}.",
                call = NULL
              )
            }
          })
        }, purrr_error_indexed = function(err) {
          rlang::cnd_signal(err$parent)
        })
      }

      if (length(outputs) > 0) {
        withCallingHandlers({
          purrr::iwalk(outputs, ~{
            if (!all(.x %in% pipe_names)) {
              invalid <- .x[which(!.x %in% pipe_names)]
              cli::cli_abort(
                "Pipeline {.pkg {.y}} references non-existent output pipeline{?s} {.pkg {invalid}}.",
                call = NULL
              )
            }
          })
        }, purrr_error_indexed = function(err) {
          rlang::cnd_signal(err$parent)
        })
      }

      withCallingHandlers({
        purrr::walk(self$MaestroPipelines, ~{
          pipe <- .x
          if (pipe$get_is_collect()) {
            n <- length(pipe$get_inputs())
            if (n < 2L) {
              single_is_iter <- n == 1L &&
                pipe$get_inputs()[[1]] %in% pipe_names &&
                self$get_pipe_by_name(pipe$get_inputs()[[1]])$get_is_map()
              if (!single_is_iter) {
                cli::cli_abort(
                  "Pipeline {.pkg {pipe$get_pipe_name()}} uses {.code collect()} but references {n} input{?s}. {.code collect()} requires at least two input pipelines or a single {.code maestroMap} pipeline.",
                  call = NULL
                )
              }
            }
          }
        })
      }, purrr_error_indexed = function(err) {
        rlang::cnd_signal(err$parent)
      })

      network <- self$get_network()

      if (nrow(network) > 0) {

        # Ensure that the outputs and inputs are reflected in the pipelines
        purrr::walk2(network$from, network$to, ~{
          from_pipe <- self$get_pipe_by_name(.x)
          to_pipe <- self$get_pipe_by_name(.y)

          cur_inputs <- to_pipe$get_inputs()
          cur_outputs <- from_pipe$get_outputs()

          from_pipe$update_outputs(unique(c(cur_outputs, .y)))
          to_pipe$update_inputs(unique(c(cur_inputs, .x)))
        })

        if (!is_valid_dag(network)) {
          cli::cli_abort(
            "Invalid DAG detected. Ensure there are no cycles in the DAG.",
            call = NULL
          )
        }
      }
    },

    #' @description
    #' Runs all the pipelines in the list
    #' @param ... arguments passed to MaestroPipeline$run
    #' @param cores if using multicore number of cores to run in (uses `furrr`)
    #' @param pipes_to_run an optional vector of pipe names to run. If `NULL` defaults to all primary pipelines
    #'
    #' @return invisible
    run = function(..., cores = 1L, pipes_to_run = NULL) {
      dots <- rlang::list2(...)

      # Parallelization
      mapper_fun <- function(...) {
        purrr::map(...)
      }

      if (cores > 1) {
        mapper_fun <- function(...) {
          furrr::future_map(
            ..., 
            .options = furrr::furrr_options(
              packages = c("maestro", "logger"),
              stdout = FALSE, 
              seed = NULL
            )
          )
        }
      }

      if (is.null(pipes_to_run)) {
        pipes_to_run <- self$get_primary_pipes()
      } else {
        pipes_to_run <- self$get_pipes_by_name(pipes_to_run)
      }
      network <- self$get_network()

      run_pipe <- function(
        pipe, 
        .input = NULL, 
        depth = -1, 
        input_run_id = NA_character_, 
        run_results = list(),
        lineage = character(),
        iter = NULL,
        pre_error = NULL,
        ...
      ) {

        run_id <- make_id()
        depth <- min(depth + 1, 6)
        
        tryCatch({
          do.call(
            pipe$run, 
            append(
              dots, 
              list(
                .input = .input, 
                run_id = run_id, 
                input_run_id = input_run_id,
                depth = depth,
                lineage = lineage,
                iter = iter,
                pre_error = pre_error
              )
            )
          )
        }, error = \(e) {
          return(pipe)
        })
        
        lineage <- append(lineage, pipe$get_pipe_name())
        run_results <- append(run_results, pipe)
        
        .input <- pipe$get_returns()
        out_names <- network$to[network$from == pipe$get_pipe_name()]
        
        if (pipe$get_status_chr() %in% c("Error", "Not Run")) {
          return(run_results)
        }
        
        if (length(out_names) == 0) return(run_results)

        for (i in out_names) {
          pipe <- self$get_pipe_by_name(i)

          # For collect() fan-in, resolve_collect_input checks all readiness
          collect_input <- NULL
          if (pipe$get_is_collect()) {
            if (pipe$get_status_chr() != "Not Run") next
            collect_input <- private$resolve_collect_input(pipe, network)
            if (is.null(collect_input)) next
          }

          # Use the collected named list as .input when this is a collect pipe,
          # otherwise fall through to the normal .input from the current pipe.
          effective_input <- if (!is.null(collect_input)) collect_input else .input

          if (pipe$get_is_map()) {
            pre_error <- NULL
            map_over <- pipe$get_map()
            scatter_vec <- purrr::map(
              stats::setNames(map_over, map_over),
              \(expr) eval(str2lang(expr), envir = list(.input = effective_input))
            )
            fields <- sub("^\\.input\\$", "", names(scatter_vec))

            if (length(map_over) == 1 && trimws(map_over) == ".input") {
              scatter_vec <- scatter_vec[[1]]
              iter_names <- as.character(seq_len(length(scatter_vec)))
              scatter_input <- purrr::map(scatter_vec, ~.x) |> 
                stats::setNames(iter_names)

            } else {
              vec_lengths <- lengths(scatter_vec)
              non_scalar_lengths <- unique(vec_lengths[vec_lengths > 1])
              if (length(non_scalar_lengths) > 1) {
                pre_error <- simpleError(
                  paste0(
                    "@maestroMap vectors must all be the same length or length 1. ",
                    "Got lengths: ",
                    paste(paste0(fields, "=", vec_lengths), collapse = ", ")
                  )
                )
              }

              # Check for any NULL results (field not found)
              null_fields <- names(scatter_vec)[purrr::map_lgl(scatter_vec, is.null)]
              if (length(null_fields) > 0) {
                bad_fields <- sub("^\\.input\\$", "", null_fields)
                pre_error <- simpleError(
                  paste0(
                    "Field(s) '", paste(bad_fields, collapse = "', '"),
                    "' specified in @maestroMap not found in ",
                    "the output of the upstream pipeline."
                  )
                )
              }

              if (is.null(pre_error)) {
                n_iter <- max(vec_lengths)
                # Recycle length-1 vectors
                scatter_vec <- purrr::map(scatter_vec, \(v) if (length(v) == 1) rep(v, n_iter) else v)

                iter_names <- as.character(seq_len(n_iter))
                scatter_input <- purrr::map(seq_len(n_iter), \(i) {
                  inp <- effective_input
                  for (f in seq_along(fields)) {
                    inp[[fields[[f]]]] <- scatter_vec[[f]][[i]]
                  }
                  inp
                }) |>
                  stats::setNames(iter_names)
              } else {
                scatter_input <- list(structure(list("0"), names = "N"))
              }              
            }

            pipe$set_n_expected_iterations(length(scatter_input))
            purrr::iwalk(scatter_input, ~{
              res <- run_pipe(
                pipe,
                .input = .x,
                depth = depth,
                run_id = run_id,
                input_run_id = run_id,
                run_results = run_results,
                lineage = lineage,
                iter = .y,
                pre_error = pre_error
              )

              run_results <<- append(run_results, res)
            })
          } else if (pipe$get_is_collect()) {
            in_names <- network$from[network$to == pipe$get_pipe_name()]
            in_pipes  <- purrr::map(in_names, self$get_pipe_by_name)
            collect_input_run_id <- purrr::map(in_pipes, ~{
              st <- .x$get_status()
              st$run_id[!is.na(st$run_id) & st$success]
            }) |>
              purrr::list_c() |>
              paste(collapse = ", ")

            collect_lineage_prefix <- purrr::map(in_pipes, ~{
              st <- .x$get_status()
              st$lineage[!is.na(st$lineage) & st$success]
            }) |>
              purrr::list_c() |>
              unique() |>
              paste(collapse = "&")

            run_results <- run_pipe(
              pipe,
              .input = effective_input,
              depth = depth,
              input_run_id = collect_input_run_id,
              run_results = run_results,
              lineage = collect_lineage_prefix,
              pre_error = pre_error
            )
          } else {
            run_results <- run_pipe(
              pipe,
              .input = effective_input,
              depth = depth,
              run_id = run_id,
              input_run_id = run_id,
              run_results = run_results,
              lineage = lineage,
              pre_error = pre_error
            )
          }
        }

        run_results
      }
      
      # Run the pipelines
      run_res <- mapper_fun(
        pipes_to_run,
        purrr::safely(~{
          run_pipe(
            .x, 
            run_results = list(),
            lineage = character()
          )
        }, quiet = TRUE)
      )

      purrr::map(run_res, "result") |>
        purrr::list_flatten()
    },

    #' @description
    #' Run any collect pipelines that are ready but were skipped during a
    #' parallel run. Called on the main process after update_pipelines() has
    #' synced worker state back. Uses run_pipe internally so that the collect
    #' pipe's own downstream outputs are recursed into normally.
    #' @param ... arguments forwarded to MaestroPipeline$run (same dots as run())
    #' @return list of MaestroPipeline objects that were run
    run_pending_collects = function(...) {
      dots <- rlang::list2(...)
      network <- self$get_network()

      collect_pipes <- purrr::keep(self$MaestroPipelines, ~.x$get_is_collect())
      pending <- purrr::keep(collect_pipes, ~.x$get_status_chr() == "Not Run")
      if (length(pending) == 0) return(invisible(list()))

      run_pipe <- function(
        pipe,
        .input = NULL,
        depth = -1,
        input_run_id = NA_character_,
        run_results = list(),
        lineage = character(),
        iter = NULL,
        pre_error = NULL,
        ...
      ) {
        run_id <- make_id()
        depth <- min(depth + 1, 6)

        tryCatch({
          do.call(
            pipe$run,
            append(
              dots,
              list(
                .input = .input,
                run_id = run_id,
                input_run_id = input_run_id,
                depth = depth,
                lineage = lineage,
                iter = iter,
                pre_error = pre_error
              )
            )
          )
        }, error = \(e) return(pipe))

        lineage <- append(lineage, pipe$get_pipe_name())
        run_results <- append(run_results, pipe)

        .input <- pipe$get_returns()
        out_names <- network$to[network$from == pipe$get_pipe_name()]

        if (pipe$get_status_chr() %in% c("Error", "Not Run")) return(run_results)
        if (length(out_names) == 0) return(run_results)

        for (i in out_names) {
          pipe <- self$get_pipe_by_name(i)

          collect_input <- NULL
          if (pipe$get_is_collect()) {
            if (pipe$get_status_chr() != "Not Run") next
            collect_input <- private$resolve_collect_input(pipe, network)
            if (is.null(collect_input)) next
          }

          effective_input <- if (!is.null(collect_input)) collect_input else .input

          run_results <- run_pipe(
            pipe,
            .input = effective_input,
            depth = depth,
            input_run_id = run_id,
            run_results = run_results,
            lineage = lineage
          )
        }

        run_results
      }

      purrr::map(pending, ~{
        collect_input <- private$resolve_collect_input(.x, network)
        if (is.null(collect_input)) return(list())
        in_names <- network$from[network$to == .x$get_pipe_name()]
        in_pipes  <- purrr::map(in_names, self$get_pipe_by_name)
        collect_input_run_id <- purrr::map(in_pipes, ~{
          st <- .x$get_status()
          st$run_id[!is.na(st$run_id) & st$success]
        }) |>
          purrr::list_c() |>
          paste(collapse = ", ")
        collect_lineage_prefix <- purrr::map(in_pipes, ~{
          st <- .x$get_status()
          st$lineage[!is.na(st$lineage) & st$success]
        }) |>
          purrr::list_c() |>
          unique() |>
          paste(collapse = "&")
        run_pipe(
          .x,
          .input = collect_input,
          input_run_id = collect_input_run_id,
          run_results = list(),
          lineage = collect_lineage_prefix
        )
      }) |>
        purrr::list_flatten()
    },
    
    #' @description
    #' Resets the run time attributes
    reset_pipelines = function() {
      purrr::walk(self$MaestroPipelines, ~.x$reset_run_time_attributes())
    }
  ),

  private = list(
    network = NULL,

    resolve_collect_input = function(pipe, network) {
      i <- pipe$get_pipe_name()
      in_names <- network$from[network$to == i]
      in_pipes <- purrr::map(in_names, self$get_pipe_by_name) |>
        stats::setNames(in_names)

      # All non-iter inputs must have succeeded
      non_iter_not_ready <- purrr::map_lgl(in_pipes, ~{
        !.x$get_is_map() && .x$get_status_chr() != "Success"
      })
      if (any(non_iter_not_ready)) return(NULL)

      # iter inputs must have all iterations finished (succeeded or errored)
      iter_not_ready <- purrr::map_lgl(in_pipes, ~{
        if (!.x$get_is_map()) return(FALSE)

        expected_n <- .x$get_n_expected_iterations()
        if (is.null(expected_n)) {

          scatter_source_name <- network$from[network$to == .x$get_pipe_name()]
          if (length(scatter_source_name) == 0) return(FALSE)
          scatter_source <- self$get_pipe_by_name(scatter_source_name[[1]])
          expected_n <- length(scatter_source$get_returns())
        }
        finished_n <- .x$get_n_invocations()
        finished_n < expected_n
      })
      if (any(iter_not_ready)) return(NULL)

      # Skip if every iteration of every iter input errored
      iter_all_failed <- purrr::map_lgl(in_pipes, ~{
        .x$get_is_map() && .x$get_n_artifacts() == 0L
      })
      if (any(iter_all_failed)) return(NULL)

      purrr::map(in_pipes, ~{
        if (.x$get_is_map()) .x$get_all_returns() else .x$get_returns()
      }) |>
        stats::setNames(in_names)
    }
  )
)

test_that("Simple pipeline list, no errors", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )

  expect_s3_class(pipeline_list, "MaestroPipelineList")
  expect_snapshot(pipeline_list)
  expect_type(pipeline_list$get_pipe_names(), "character")
  expect_s3_class(pipeline_list$get_pipe_by_name("get_mtcars"), "MaestroPipeline")
  expect_error(
    pipeline_list$get_pipe_by_name("asdasd"),
    regexp = "No pipeline"
  )

  pipeline_list$run(quiet = TRUE)
  pipeline <- pipeline_list$MaestroPipelines[[1]]
  expect_snapshot(pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")])

  expect_type(
    pipeline_list$check_timeliness(orch_n = 1, orch_unit = "day"),
    "logical"
  )

  expect_s3_class(
    pipeline_list$get_timely_pipelines(orch_n = 1, orch_unit = "day"),
    "MaestroPipelineList"
  )
})

test_that("Errors are handled", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines_run_some_errors/pipe2.R")
  )

  expect_no_error(pipeline_list$run(quiet = TRUE))
})

test_that("Populate on instantiation", {

  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )

  pipelines <- pipeline_list$MaestroPipelines

  new_pipeline_list <- MaestroPipelineList$new(pipelines)
  expect_s3_class(new_pipeline_list, "MaestroPipelineList")
  expect_s3_class(new_pipeline_list$MaestroPipelines[[1]], "MaestroPipeline")
})

test_that("Populate after instantiation", {
  pipeline_list <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )

  pipelines <- pipeline_list$MaestroPipelines

  new_pipeline_list <- MaestroPipelineList$new()
  new_pipeline_list$add_pipelines(pipelines[[1]])
  expect_s3_class(new_pipeline_list, "MaestroPipelineList")
  expect_s3_class(new_pipeline_list$MaestroPipelines[[1]], "MaestroPipeline")
  expect_equal(new_pipeline_list$n_pipelines, 1)
})

test_that("Attribute n_pipelines is valid", {
  pipeline_list1 <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )

  expect_equal(pipeline_list1$n_pipelines, 1)

  pipeline_list2 <- build_schedule_entry(
    test_path("test_pipelines/test_multi_fun_pipeline.R")
  )

  expect_equal(pipeline_list2$n_pipelines, 2)

  new_pipe <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_hours_good.R")
  )

  pipeline_list2$add_pipelines(new_pipe)

  expect_equal(pipeline_list2$n_pipelines, 3)
})

test_that("Can add two MaestroPipelineLists", {
  pipeline_list1 <- build_schedule_entry(
    test_path("test_pipelines/test_pipeline_daily_good.R")
  )

  pipeline_list2 <- build_schedule_entry(
    test_path("test_pipelines/test_multi_fun_pipeline.R")
  )

  pipeline_list1$add_pipelines(pipeline_list2)
  expect_equal(pipeline_list1$n_pipelines, 3)
})

## ----echo=FALSE---------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroFrequency daily
  get_letters <- function() {
    c('a', 'b', 'c')
  }

  #' @maestroInputs get_letters
  #' @maestroMap
  shout <- function(.input) {
    toupper(.input)
  }",
  con = "pipelines/fanout.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency daily
# get_letters <- function() {
#   c("a", "b", "c")
# }
# 
# #' @maestroInputs get_letters
# #' @maestroMap
# shout <- function(.input) {
#   toupper(.input)
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(file.remove("pipelines/fanout.R"))
writeLines(
  "
  #' @maestroFrequency daily
  get_letters <- function() {
    list(
      letter = letters[1:3],
      greeting = 'hello'
    )
  }

  #' @maestroInputs get_letters
  #' @maestroMap .input$letter
  make_message <- function(.input) {
    paste(.input$greeting, toupper(.input$letter))
  }",
  con = "pipelines/fanout.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency daily
# get_letters <- function() {
#   list(
#     letter = letters[1:3],
#     greeting = "hello"
#   )
# }
# 
# #' @maestroInputs get_letters
# #' @maestroMap .input$letter
# make_message <- function(.input) {
#   paste(.input$greeting, toupper(.input$letter))
# }

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency daily
# get_data <- function() {
#   list(
#     letter = letters[1:3],
#     greeting = c("hello", "cheers", "hi")
#   )
# }
# 
# #' @maestroInputs get_data
# #' @maestroMap .input$letter .input$greeting
# make_message <- function(.input) {
#   paste(.input$greeting, toupper(.input$letter))
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(file.remove("pipelines/fanout.R"))
writeLines(
  "
  #' @maestroFrequency daily
  letter_a <- function() 'a'

  #' @maestroFrequency daily
  letter_b <- function() 'b'

  #' @maestroInputs collect(letter_a, letter_b)
  combine <- function(.input) {
    paste0(.input$letter_a, .input$letter_b)
  }",
  con = "pipelines/fanin.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency daily
# letter_a <- function() "a"
# 
# #' @maestroFrequency daily
# letter_b <- function() "b"
# 
# #' @maestroInputs collect(letter_a, letter_b)
# combine <- function(.input) {
#   paste0(.input$letter_a, .input$letter_b)
# }

## ----echo=FALSE, warning=FALSE, message=FALSE---------------------------------
invisible(file.remove("pipelines/fanin.R"))
writeLines(
  "
  #' @maestroFrequency daily
  numbers <- function() 1:3

  #' @maestroInputs numbers
  #' @maestroMap
  multiply <- function(.input) .input * 3

  #' @maestroInputs collect(multiply)
  add <- function(.input) {
    sum(unlist(.input))
  }",
  con = "pipelines/fanin.R"
)

## ----eval=FALSE---------------------------------------------------------------
# #' @maestroFrequency daily
# numbers <- function() 1:3
# 
# #' @maestroInputs numbers
# #' @maestroMap
# multiply <- function(.input) .input * 3
# 
# #' @maestroInputs collect(multiply)
# add <- function(.input) {
#   sum(unlist(.input))
# }

## ----cleanup2, echo=FALSE, message=FALSE, warning=FALSE-----------------------
invisible(unlink("pipelines", recursive = TRUE))

#' Simple mtcars print function
#'
#' This is a function that runs every hour starting at
#' 2024-03-01 09:00:00
#'
#' @maestroFrequency 1 day
#' @maestroLogLevel Ghost
#'
#' @export
get_mtcars <- function() {
  Sys.sleep(0.02)
  mtcars
}

#' @maestroFrequency 1 day
#' @maestroLogLevel WARN
pipe3 <- function() {
  message("Hide me")

  warning("Oops")
  iris
}

#' @maestroLogLevel WARN
pipe4 <- function() {
  message("Hide me too")

  warning("Another oops")
  rnorm(10)
}

---
title: "Logging"
vignette: >
  %\VignetteIndexEntry{Logging}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| echo: false
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")
```

Logging is critical for monitoring pipelines in a deployed environment. Logging is a way of tracking error messages, warnings, and other user generated informational messages. Maestro provides several options for logging these conditions from pipelines - logging to a file and/or logging to the console. Maestro leverages [logger](https://daroczig.github.io/logger/) for managing and formatting logs.

## Conditions in R: Message, Warning, Error

R has three types of conditions (increasing in severity): message, warning, and error. Maestro makes use of this system for reporting the statuses of pipelines and for managing logs. For this reason, we recommend using `message()` for adding informational logs to pipelines in contrast to `print()` and `cat()`.

There are also several logging packages for R including `logger` and `logr`. As we'll see later, these can be integrated with maestro as well with some special modifications.

## Log to File

Maestro allows for a log file to be continuously appended to with pipeline logs. Conventionally, we give it a name of `maestro.log`, but this can be any text file. Here we create a sample set of 3 pipelines, one with a message, one with a warning, and another with an error.

```{r}
#| echo: false
#| output: false
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency hourly
  hello <- function() {
    message('hello')
  }
  
  #' @maestroFrequency hourly
  uhh <- function() {
    warning('this could be a problem')
  }
  
  #' @maestroFrequency hourly
  oh_no <- function() {
    stop('this is bad')
  }",
  con = "pipelines/logs.R"
)
```

```{r}
#| eval: false
#' pipelines/logs.R
#' @maestroFrequency hourly
hello <- function() {
  message('hello')
}

#' @maestroFrequency hourly
uhh <- function() {
  warning('this could be a problem')
}

#' @maestroFrequency hourly
oh_no <- function() {
  stop('this is bad')
}
```

Now we run the orchestrator. For demo purposes, we'll set `run_all = TRUE` to allow all pipelines to run regardless of scheduling. We set `log_to_file = TRUE` argument to tell it to log to a file. This creates (and later appends to) a generic 'maestro.log' file in the project directory. We can see the typical output of a maestro schedule as the pipeline runs.

```{asciicast}
# orchestrator.R
library(maestro)

schedule <- build_schedule(quiet = TRUE)

status <- run_schedule(
  schedule, 
  run_all = TRUE,
  log_to_file = TRUE
)
```

Now let's take a look at the log file. We can see the logs formatted with the name of the pipeline from where the message came, the type of log (INFO, WARN, or ERROR), the timestamp, and the message itself.

```{asciicast}
readLines("maestro.log")
```

## Log to Console

We can also have the logs printed directly to the console using the `log_to_console` argument.

```{asciicast}
# orchestrator.R
schedule <- build_schedule(quiet = TRUE)

status <- run_schedule(
  schedule, 
  run_all = TRUE,
  log_to_console = TRUE
)
```

Now the logs have been interwoven with the output from maestro. Both logging options operate independently, so it is possible to log to a file and to the console.

```{r}
#| echo: false
#| output: false
file.remove("maestro.log")
```

## Log Levels

Maestro uses the concept of log levels (also known as thresholds) to allow users to suppress logs that do not meet a severity threshold. If, for example, you were only concerned with error messages and wanted to ignore warnings and info messages, you would use the `maestroLogLevel` tag for the relevant pipelines.

```{r}
#| echo: false
#| output: false
unlink("pipelines", recursive = TRUE)
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency hourly
  #' @maestroLogLevel ERROR
  hello <- function() {
    message('hello')
  }
  
  #' @maestroFrequency hourly
  #' @maestroLogLevel ERROR
  uhh <- function() {
    warning('this could be a problem')
  }
  
  #' @maestroFrequency hourly
  #' @maestroLogLevel ERROR
  oh_no <- function() {
    stop('this is bad')
  }",
  con = "pipelines/logs.R"
)
```

```{r}
#| eval: false
#' pipelines/logs.R
#' @maestroFrequency hourly
#' @maestroLogLevel ERROR
hello <- function() {
  message('hello')
}

#' @maestroFrequency hourly
#' @maestroLogLevel ERROR
uhh <- function() {
  warning('this could be a problem')
}

#' @maestroFrequency hourly
#' @maestroLogLevel ERROR
oh_no <- function() {
  stop('this is bad')
}
```

Now, only the error messages are displayed and logged.

```{asciicast}
# orchestrator.R
schedule <- build_schedule(quiet = TRUE)

status <- run_schedule(
  schedule, 
  run_all = TRUE,
  log_to_console = TRUE
)
```

By default, pipelines use a log level of INFO, which means that all messages, warnings, and errors are logged.

## Other Loggers

If you wish to use other logging libraries, we recommend using [logger](https://daroczig.github.io/logger/). You can put these statements anywhere in your pipelines and they'll propagate into the logs that maestro uses. Best approach is to use the `namespace` argument and reference the name of the pipeline.

```{r}
#| echo: false
#| output: false
unlink("pipelines", recursive = TRUE)
dir.create("pipelines")
writeLines(
  "
  #' @maestroFrequency hourly
  hello <- function() {
    logger::log_info('hi', namespace = 'hello')
  }
  
  #' @maestroFrequency hourly
  uhh <- function() {
    logger::log_warn('this could be a problem', namespace = 'uhh')
  }
  
  #' @maestroFrequency hourly
  oh_no <- function() {
    logger::log_error('this is bad', namespace = 'oh_no')
  }",
  con = "pipelines/logs.R"
)
```

```{r}
#| eval: false
#' pipelines/logs.R
#' @maestroFrequency hourly
hello <- function() {
  logger::log_info("hi", namespace = "hello")
}

#' @maestroFrequency hourly
uhh <- function() {
  logger::log_warn("this could be a problem", namespace = "uhh")
}

#' @maestroFrequency hourly
oh_no <- function() {
  logger::log_error("this is bad", namespace = "oh_no")
}
```

```{asciicast}
# orchestrator.R
schedule <- build_schedule(quiet = TRUE)

status <- run_schedule(
  schedule, 
  run_all = TRUE,
  log_to_console = TRUE
)
```

Note that logger just creates messages and does not actually trigger conditions. This impacts how the statuses of pipelines appear. It is important to use match namespace with the name of your pipeline (i.e., the function name) for the logs to appear in a log file.

```{r cleanup}
#| echo: false
#| output: false
unlink("pipelines", recursive = TRUE)
file.remove("maestro.log")
```

<div align="center">

  <h1>maestro</h1>

  <p>coming: when it's ready...</p>

</div>

**maestro** is a tool designed to streamline and accelerate the fine-tuning process for
multimodal models. It provides ready-to-use recipes for fine-tuning popular
vision-language models (VLMs) such as **Florence-2**, **PaliGemma**, and
**Qwen2-VL** on downstream vision-language tasks.

## install

Pip install the supervision package in a
[**Python>=3.8**](https://www.python.org/) environment.

```bash
pip install maestro
```

## quickstart

### CLI

VLMs can be fine-tuned on downstream tasks directly from the command line with
`maestro` command:

```bash
maestro florence2 train --dataset='<DATASET_PATH>' --epochs=10 --batch-size=8
```

### SDK

Alternatively, you can fine-tune VLMs using the Python SDK, which accepts the same
arguments as the CLI example above:

```python
from maestro.trainer.common import MeanAveragePrecisionMetric
from maestro.trainer.models.florence_2 import train, Configuration

config = Configuration(
    dataset='<DATASET_PATH>',
    epochs=10,
    batch_size=8,
    metrics=[MeanAveragePrecisionMetric()]
)

train(config)
```

---
title: "Advanced Scheduling"
vignette: >
  %\VignetteIndexEntry{Advanced Scheduling}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| echo: false
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")
```

This vignette covers more advanced concepts and examples related to scheduling. At it's core, maestro adheres to two keys principles:

1.  Stateless: It does not need to be continuously running - it can be run in a serverless architecture
2.  Use of *rounded scheduling*: The timeliness of pipeline executions depends on how often you run your orchestrator

Here, we'll explain these concepts more deeply in the context of maestro and give some concrete examples.

## Stateless Execution

Maestro takes a unique approach to scheduling compared to other orchestration tools. Whereas most schedulers involve a continuously running program to monitor the time and execute jobs when the current time is right, maestro is designed to run intermittently. It also doesn't need to save or cache data between executions - in other words, it's *stateless*.

This design has several benefits; namely, you can run it in a serverless way which saves on compute resources. However, to achieve this it takes some shortcuts which may mean that precise timeliness is lost. This will become clearer in our examples.

## Rounded Scheduling

The timeliness of a pipeline is measured in how close the scheduled execution time is to the actual execution time. **Maestro is only as timely as it needs to be relative to the unit of time you are interested in**. This is the concept of *rounded scheduling*. If you run your orchestrator once daily, then the timeliness of the pipelines will be within the nearest day - it doesn't care that you specified a pipeline to run at exactly 09:21:20 each day. If you run it every 10 minutes, then the timeliness of the pipelines will be within the nearest 10 minute interval.

Let's look at some examples:

First, we'll consider one pipeline that is scheduled to run daily at 09:20:00 and we'll configure the orchestrator to run daily.

```{r, echo=FALSE, warning=FALSE, message=FALSE}
dir.create("pipelines")
writeLines(
  "
  # ./pipelines/daily_example.R
  #' daily_example maestro pipeline
  #'
  #' @maestroFrequency 1 day
  #' @maestroStartTime 09:20:00
  daily_example <- function() {
  
    # Pipeline code
  }
  ",
  con = "pipelines/daily_example.R"
)
```

```{r, eval=FALSE}
# ./pipelines/daily_example.R
#' daily_example maestro pipeline
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 09:20:00
daily_example <- function() {

  # Pipeline code
}
```

For demonstration purposes, I'll manually set the check time to be 08:00:00 UTC (this is the time maestro will use to compare against the scheduled time). In practice, you almost always want this to be the system time using either `Sys.time()` or `lubridate::now()`.

```{asciicast}
# ./orchestrator.R

library(maestro)

schedule <- build_schedule()

status <- run_schedule(
  schedule,
  orch_frequency = "1 day",
  check_datetime = as.POSIXct("2024-06-20 08:00:00", tz = "UTC")
)
```

We can see that the pipeline executed even though the current time was not 09:20:00. This is because we set the orchestrator to run daily and so it considers it close enough to within a day.

Let's see what happens if we up the frequency of the orchestrator:

```{asciicast}
# ./orchestrator.R
status <- run_schedule(
  schedule,
  orch_frequency = "15 minutes",
  check_datetime = as.POSIXct("2024-06-20 08:00:00", tz = "UTC")
)
```

It was skipped because it wasn't within a 15 minute degree of difference but the output tells us that will next run at `2024-06-20 09:15:00`.\
\
The takeaway message is that the timeliness of the pipeline depends on how frequently the orchestrator runs. Remember that when you declare the `orch_frequency = "15 minutes"` that is essentially a contract stating that you *will* run it every 15 minutes - maestro does not do this for you. If you run the orchestrator more or less frequently than you said you would unexpected things will happen. Specifically, if you run it more frequently than stated, your pipelines will run more often than expected, likewise less frequently than stated means that pipelines won't run as often.

## *How often should I schedule my orchestrator?*

If you have a single pipeline or even multiple pipelines that all run at the same time this is an easy question to answer. In practice (and in our experience using maestro in production) you have multiple pipelines that run at different intervals and different times. Maybe some run hourly, some run daily, and others run monthly.

### Example 1

Let's say we have three pipelines with the following frequencies and start times:

```{r, echo=FALSE}
data.frame(
  name = paste0("pipe", 1:3),
  frequency = c("1 hour", "2 days", "4 months"),
  start_time = as.POSIXct(c("2024-06-18 12:30:00", "2024-06-18 06:00:00", "2024-06-20 00:00:00"))
)
```

A good starting point is to schedule it as often as the highest frequency pipeline in the project - so `1 hour` in the above example. If you run it on \*:30:00 minute each day, pipe1 will execute nearly exactly at the scheduled time and the other pipelines will be executed 30 minutes early. If you're comfortable with this margin of error then it's no big deal, but if not then an orchestrator frequency of `30 minutes` will ensure all pipelines run as scheduled exactly.

Let's see another example:

### Example 2

```{r, echo=FALSE}
data.frame(
  name = paste0("pipe", 4:6),
  frequency = c("1 hour", "1 hour", "1 hour"),
  start_time = as.POSIXct(c("2024-06-18 00:00:00", "2024-06-18 00:10:00", "2024-06-18 00:20:00"))
)
```

All three pipelines are hourly but they start on different 10-minute intervals. If we run the orchestrator at `1 hour` they'll all execute at the same time. If it is important that they execute at different times, we should set it to `10 minutes`.

So a good heuristic is to run it as often as the smallest interval of time difference between any pipeline. This is pretty good so long as we don't run it so often that the pipelines can't complete before the next execution time. We don't recommend running the orchestrator more frequently than every 5 minutes unless you're confident that your pipelines are fast to execute.[^1]

[^1]: Make sure you consider the amount of time it takes to execute `build_schedule()` and that you account for the additional work done in `run_schedule()` not related to your actual pipeline logic.

## Irregular Schedules

It is also possible to run pipelines on an irregular schedule by specifying hours, days of week, days of month, or months of the year:

### Hours of the day

If your pipeline runs hourly at a minimum, you can run it on specific hours. This could be useful, for example, if you have pipelines you only want to run during business hours. Use the `maestroHours` tag with integers corresponding to hours [0-23] separated by white space. Be sure to use the corresponding `maestroFrequency`.

```{r, eval=FALSE}
#' specific_hours maestro pipeline
#'
#' @maestroFrequency hourly
#' @maestroHours 6 9 12 15 18
some_hours_example <- function() {

  # Pipeline code
}
```

### Days of week/days of month

If your pipeline runs at least daily, you can run it on specific days of week or days of month. If doing days of week use strings like Mon, Tue, Wed, etc. If using days of month use integers [1-31]. You cannot specify both days of week and days of month.

```{r, eval=FALSE}
#' specific_days_of_week maestro pipeline
#'
#' @maestroFrequency daily
#' @maestroDays Mon Tue Wed Thu Fri
some_dow_example <- function() {

  # Pipeline code
}

#' specific_days_of_week maestro pipeline
#'
#' @maestroFrequency daily
#' @maestroDays 1 14 28
some_dom_example <- function() {

  # Pipeline code
}
```

### Months of year

If your pipeline runs at least monthly, you can run it on specific months. Use integers [1-12] to specify the months.

```{r, eval=FALSE}
#' specific_months maestro pipeline
#'
#' @maestroFrequency months
#' @maestroMonths 1 7 10
some_months <- function() {

  # Pipeline code
}
```

Note that you can use combinations of these specifiers depending on the minimum frequency of the pipeline. For example, an hourly pipeline can run on specific hours and days. If you wanted a pipeline to only run on business hours including not on weekends. It could look like this:

```{r, eval=FALSE}
#' business_hours maestro pipeline
#'
#' @maestroFrequency hourly
#' @maestroHours 9 10 11 12 13 14 15 16 17
#' @maestroDays Mon Tue Wed Thu Fri
during_work_hours_example <- function() {

  # Pipeline code
}
```

## Final Remarks

When it comes time to deploy your project make sure that whatever you use to actually run your project (e.g., cron, TaskScheduler, Google Cloud Scheduler, etc.) is indeed running it at the same frequency as you have in the orchestrator. It's best to stick to using whole units of time rather than fractional units - if my orchestrator runs every 15 minutes I like to have it run on the 00:00, 15:00, 30:00, and 45:00 minutes. It makes reasoning about the scheduling simpler.

```{r cleanup, echo=FALSE, message=FALSE, warning=FALSE}
unlink("pipelines", recursive = TRUE)
```

---
title: "Deployment"
vignette: >
  %\VignetteIndexEntry{Deployment}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
eval: false
---

Maestro is only of practical use if it's deployed. It doesn't matter if this deployment is cloud or on a local server, but the goal is to have it run automatically on a regular cadence. In this way, maestro can be the backbone of a small-moderate enterprise's data orchestration.

Maestro deployment is not all that different from deploying any R script. The main difference is that you need to deploy the entire project (pipelines + orchestrator) and then target specifically the orchestrator script to execute repeatedly.

There are numerous avenues for deploying a maestro project, so here we'll just look a handful of options. Here we will only focus on the execution of maestro and not on side effects like storage and networking (in practice, you'd probably have your pipelines do something like read from an external source and load to a data warehouse).

Our example to deploy will be as simple as it gets - just the bare bones skeleton generated from `maestro::create_maestro()`. This builds an empty maestro project with a single pipeline and an orchestrator script. It is sufficient for testing but does nothing of interest.

## Local Deployment using Cron/Task Manager

The easiest way to deploy maestro is to have it run on your own local machine and triggered on a schedule using Cron (Mac/Linux)/Task Scheduler (Windows).

### Cron

If your OS is Linux or Mac you can use Cron scheduling. The [cronR](https://github.com/bnosac/cronR) package allows you to create and manage cron scheduled tasks from within R. Create a new script at the root level of the project:

```{r}
#| eval: false
library(cronR)
library(here) # makes it easier to deal with path issues

cmd <- cron_rscript(
  here("orchestrator.R"), 
  workdir = here()
)

cron_add(
  cmd, 
  frequency = "daily",
  id = "maestro",
  ask = FALSE
)
```

### Task Scheduler

If your OS is Windows you can leverage Task Scheduler. The [taskscheduleR](https://github.com/bnosac/taskscheduleR) package allows you to create and manage scheduled tasks from within R. Create a new R script at the root level of the project:

```{r}
#| eval: false
library(taskscheduleR)
library(here) # makes it easier to deal with path issues

taskscheduler_create(
  taskname = "maestro", 
  rscript = here("orchestrator.R"), 
  schedule = "DAILY",
  exec_path = here(),
  startdate = format(Sys.Date() - 1, "%m/%d/%Y")
)
```

### Limitations

While this is the easiest approach it has a number of limitations. First, these will only run if the machine is constantly running (i.e., not in hibernation or sleep mode); so if you're doing this on a desktop or laptop you'll probably need to go through some extra steps to allow it to run even when closed.

Second, even if you have continuous availability, it is hard to monitor the success of the deployment. There are GUI-based managers for Cron and Task Scheduler, but they're a far cry from true observability.

Third, it is difficult to version this kind of deployment. If you need to make changes or reversions you can quickly get lost in a swamp of stale and aborted deployments.

All said, this kind of deployment is probably only advisable for testing and proof-of-concept. There are probably more sophisticated Cron wrappers or substitutes that make local deployment more feasible (i.e., include monitoring, CI/CD, etc.), but we won't focus on those here.

## Posit Connect

Seasoned R users will likely be familiar with [Posit](https://posit.co/) and its hosting platform [Posit Connect](https://posit.co/products/enterprise/connect). While traditionally seen as a hosting environment for dashboards, APIs, and reports, it is actually perfectly capable of data orchestration with maestro.

::: callout-note
A Posit Connect license and instance is required for this type of deployment. User must also have Publisher privileges to deploy to Connect.
:::

The key difference for a Posit Connect deployment is the use of [Quarto](https://quarto.org/) for the orchestrator.[^1] There's no need for any visual embellishments in this Quarto doc (but, hey, here's a pretty good place for some observability metrics and graphics!), all you need is at least one R chunk containing the orchestrator code.

[^1]: This feels a bit odd given that Quarto is conventionally used for publishing reports, presentations, and dashboards. That said, there's nothing stopping you from executing whatever R code you want in a Quarto document, and Posit Connect won't let you "deploy" a stand-alone R script anyway. RMarkdown is also a valid option although we'll stick with Quarto here.

Here's a basic .qmd script that could be deployed on Posit Connect.

```` markdown
---
title: "Maestro"
format: html
resources:
  - pipelines/*
  - renv.lock
---

```{{r}}
library(maestro)

schedule <- build_schedule()

run_schedule(
  schedule,
  orch_frequency = "1 hour"
)
```
````

You can then use the push deployment via RStudio or Positron. In RStudio it looks like this:

![](images/maestro-publish.png)

Make sure you select **Publish with source code** so that you can trigger the orchestrator to run on a schedule in Connect.

Once it has been published you can set the schedule using the Schedule tab. Be sure to use the same schedule frequency as your orchestration frequency:

![](images/maestro-posit-scheduled.png)

This is a relatively easy way to deploy a maestro project assuming you have a Posit Connect license.

::: callout-note
[Posit Connect Cloud](https://connect.posit.cloud/), a managed instance of Posit Connect, may be another option for deploying maestro projects, although currently this has not be tested.
:::

### Limitations

While a more robust deployment compared to local Cron/TaskManager, observability is still *initially* limited to manually viewing the rendered Quarto doc. With a bit of extra tinkering, however, you can trigger some side effects in the orchestration script such as sending the [pipeline statuses to a database for display in a dashboard](https://whipson.github.io/data-in-flight/posts/you-dont-need-airflow/main.html#monitoring), or [triggering an email](https://docs.posit.co/connect/user/quarto/#email-customization) to be sent if any pipelines fail.

On the other hand, you may not appreciate the additional overhead that comes with needing to execute the orchestrator as a Quarto doc. Personally, I don't think this creates a meaningful cost in terms of run time or compute (assuming you're not triggering too many side effects after the orchestration), but it does require a Quarto installation on the Posit Connect instance, which could be a non-starter in some cases.

Finally, if your Posit Connect instance is deployed on a server managed by someone other than yourself you may run into issues if the server becomes unavailable or you cannot debug problems at the server level. In particular, viewing pipeline logs can be difficult since the logs get stored at the project level and are thus not accessible in Connect. These instead will need to be sent somewhere they can be inspected.

## Cloud

There are virtually limitless options for deploying in the cloud. In fact, the options described above can also be applied to a cloud-based server (e.g., spin up a Linux server and use Cron to run the schedule and just keep the server running indefinitely).

Here we'll focus on a specific pattern of maestro deployment that works well in the cloud both in terms of cost and ease of deployment. We'll avoid tutorials with specific cloud providers and instead generalize the types of services needed while providing examples of popular specific services. If you're looking for a walk through for a cloud-based deployment, check out this [blog post](https://whipson.github.io/data-in-flight/posts/maestro-gcp-deployment/maestro-gcp-deployment.html) using Google Cloud Platform (GCP).

### Serverless Cloud

The advantage of maestro is its serverless design. This means that for virtually all use cases you don't need to have a server running continuously. Instead, you can only run the server *when the orchestrator needs to run*. This usually results in cost savings.

For a basic maestro cloud deployment, you need two services:

1.  Serverless execution environment (i.e., something to run the code)
2.  Scheduler (i.e., something to tell the code when to run)

Some cloud providers may bundle these services together, others may offer them as separate services.

For example, in Amazon Web Services (AWS), the serverless execution environment would be a Lambda function and the scheduler could be EventBridge. In GCP it would be Cloud Run scheduled via Cloud Scheduler. In Azure it would be Azure Functions configured with a Timer Trigger, and so on.

As usual, be sure to use the same scheduling frequency as the orchestrator frequency.

### Limitations

The only limitation in the cloud is your skill/time and your budget. Often these two limitations are inversely correlated: managed cloud services like Digital Ocean ease the pain of deployment at the cost of taking more of your money. Availability is not likely to be an issue, but custom tinkering will be needed to set up observability. Finally, it is worth considering best practices such as CI/CD to make it easier to update the maestro project. This will also require additional cloud services and therefore cost.

## Final Considerations

It's important to remember that deploying a maestro project is not all that different from deploying a single R script. It just has a few extra files that need to go with it. One recommendation for deployment is to take advantage of maestro's serverless nature. If you only need to run your orchestrator daily, don't deploy it where you are charged for something that runs continuously (that's 23 hours a day you're paying for nothing).

It's also important to consider that deployment *for you* likely means more than just maestro running in isolation. Maestro really only makes sense if it's doing something like extracting data from a source and inserting it into a database, or generating and storing predictions from a model. All of this lies outside maestro's functionality but is nevertheless part of the entire deployment stack.

## Overview

Florence-2 is a lightweight vision-language model open-sourced by Microsoft under the
MIT license. The model demonstrates strong zero-shot and fine-tuning capabilities
across tasks such as captioning, object detection, grounding, and segmentation.

<iframe loading="lazy" width="720" height="405" src="https://www.youtube.com/embed/i3KjYgxNH6w" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share" allowfullscreen> </iframe>
*Florence-2: Fine-tune Microsoft’s Multimodal Model.*

## Architecture

The model takes images and task prompts as input, generating the desired results in
text format. It uses a DaViT vision encoder to convert images into visual token
embeddings. These are then concatenated with BERT-generated text embeddings and
processed by a transformer-based multi-modal encoder-decoder to generate the response.

![florence-2-architecture](https://storage.googleapis.com/com-roboflow-marketing/maestro/florence-2-architecture.webp)
*Overview of Florence-2 architecture. Source: Florence-2: Advancing a Unified Representation for a Variety of Vision Tasks.*


## Fine-tuning Examples

### Dataset Format

The Florence-2 model expects a specific dataset structure for training and evaluation.
The dataset should be organized into train, test, and validation splits, with each
split containing image files and an `annotations.jsonl` file.

```
dataset/
├── train/
│   ├── 123e4567-e89b-12d3-a456-426614174000.png
│   ├── 987f6543-a21c-43c3-a562-926514273001.png
│   ├── ...
│   ├── annotations.jsonl
├── test/
│   ├── 456b7890-e32d-44f5-b678-724564172002.png
│   ├── 678c1234-e45b-67f6-c789-813264172003.png
│   ├── ...
│   ├── annotations.jsonl
└── valid/
    ├── 789d2345-f67c-89d7-e891-912354172004.png
    ├── 135e6789-d89f-12e3-f012-456464172005.png
    ├── ...
    └── annotations.jsonl
```

Depending on the vision task being performed, the structure of the `annotations.jsonl`
file will vary slightly.

!!! warning
    The dataset samples shown below are formatted for improved readability, with each
    JSON structure spread across multiple lines. In practice, the `annotations.jsonl`
    file must contain each JSON object on a single line, without any line breaks
    between the key-value pairs. Make sure to adhere to this structure to avoid parsing
    errors during model training.

=== "Object Detection"

    ```txt
    {
        "image":"123e4567-e89b-12d3-a456-426614174000.png",
        "prefix":"<OD>",
        "suffix":"9 of clubs<loc_138><loc_100><loc_470><loc_448>10 of clubs<loc_388><loc_145><loc_670><loc_453>"
    }
    {
        "image":"987f6543-a21c-43c3-a562-926514273001.png",
        "prefix":"<OD>",
        "suffix":"5 of clubs<loc_554><loc_2><loc_763><loc_467>6 of clubs<loc_399><loc_79><loc_555><loc_466>"
    }
    ...
    ```

=== "Visual Question Answering (VQA)"

    ```txt
    {
        "image":"123e4567-e89b-12d3-a456-426614174000.png",
        "prefix":"<VQA> Is the value of Favorable 38 in 2015?",
        "suffix":"Yes"
    }
    {
        "image":"987f6543-a21c-43c3-a562-926514273001.png",
        "prefix":"<VQA> How many values are below 40 in Unfavorable graph?",
        "suffix":"6"
    }
    ...
    ```

=== "Object Character Recognition (OCR)"

    ```txt
    {
        "image":"123e4567-e89b-12d3-a456-426614174000.png",
        "prefix":"<OCR>",
        "suffix":"ke begherte Die mi"
    }
    {
        "image":"987f6543-a21c-43c3-a562-926514273001.png",
        "prefix":"<OCR>",
        "suffix":"mi uort in de middelt"
    }
    ...
    ```

### CLI

!!! tip
    Depending on the GPU you are using, you may need to adjust the `batch-size` to
    ensure that your model trains within memory limits. For larger GPUs with more
    memory, you can increase the batch size for better performance.

!!! tip
    Depending on the vision task you are executing, you may need to select different
    vision metrics. For example, tasks like object detection typically use
    `mean_average_precision`, while VQA and OCR tasks use metrics like
    `word_error_rate` and `character_error_rate`.

!!! tip
    You may need to use different learning rates depending on the task. We have found
    that lower learning rates work better for tasks like OCR or VQA, as these tasks
    require more precision.


=== "Object Detection"

    ```bash
    maestro florence2 train --dataset='<DATASET_PATH>' \
    --epochs=10 --batch-size=8 --lr=5e-6 --metrics=mean_average_precision
    ```

=== "Visual Question Answering (VQA)"

    ```bash
    maestro florence2 train --dataset='<DATASET_PATH>' \
    --epochs=10 --batch-size=8 --lr=1e-6 \
    --metrics=word_error_rate, character_error_rate
    ```

=== "Object Character Recognition (OCR)"

    ```bash
    maestro florence2 train --dataset='<DATASET_PATH>' \
    --epochs=10 --batch-size=8 --lr=1e-6 \
    --metrics=word_error_rate, character_error_rate
    ```

### SDK

=== "Object Detection"

    ```python
    from maestro.trainer.common import MeanAveragePrecisionMetric
    from maestro.trainer.models.florence_2 import train, Configuration

    config = Configuration(
        dataset='<DATASET_PATH>',
        epochs=10,
        batch_size=8,
        lr=5e-6,
        metrics=[MeanAveragePrecisionMetric()]
    )

    train(config)
    ```

=== "Visual Question Answering (VQA)"

    ```python
    from maestro.trainer.common import WordErrorRateMetric, CharacterErrorRateMetric
    from maestro.trainer.models.florence_2 import train, Configuration

    config = Configuration(
        dataset='<DATASET_PATH>',
        epochs=10,
        batch_size=8,
        lr=1e-6,
        metrics=[WordErrorRateMetric(), CharacterErrorRateMetric()]
    )

    train(config)
    ```

=== "Object Character Recognition (OCR)"

    ```python
    from maestro.trainer.common import WordErrorRateMetric, CharacterErrorRateMetric
    from maestro.trainer.models.florence_2 import train, Configuration

    config = Configuration(
        dataset='<DATASET_PATH>',
        epochs=10,
        batch_size=8,
        lr=1e-6,
        metrics=[WordErrorRateMetric(), CharacterErrorRateMetric()]
    )

    train(config)
    ```

## API

<div class="md-typeset">
    <h2><a href="#maestro.trainer.models.florence_2.core.Configuration">Configuration</a></h2>
</div>

:::maestro.trainer.models.florence_2.core.Configuration

<div class="md-typeset">
    <h2><a href="#maestro.trainer.models.florence_2.core.train">train</a></h2>
</div>

:::maestro.trainer.models.florence_2.core.train

<div class="md-typeset">
    <h2><a href="#maestro.trainer.models.florence_2.core.evaluate">evaluate</a></h2>
</div>

:::maestro.trainer.models.florence_2.core.evaluate

---
title: "Maestro Tag Reference"
vignette: >
  %\VignetteIndexEntry{Maestro Tag Reference}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
eval: false
toc: true
---

Maestro tags are roxygen2 comments for configuring the scheduling and execution of pipelines.

Maestro tags follow the format: `#' @maestroTagName value`

Some tags may not take a value.

Maestro tags must be written above the function that is to be included as a pipeline. A typical pipeline with tags could look like this:

``` r
#' @maestroFrequency 1 hour
#' @maestroStartTime 12:30:00
#' @maestroLogLevel WARN
my_pipeline <- function() {

  # Pipeline code
}
```

Below are descriptions of all the tags currently available in maestro along with examples.

# Basic Tags

## maestroFrequency

How often to run the pipeline. This tag takes a time unit indicating how long to wait between subsequent runs of the pipeline. Acceptable values include an integer value followed by one of minute(s), hour(s), day(s), week(s), month(s), and year(s). Note that some combinations of integer + unit may be invalid. Adverbs like 'hourly', 'daily', 'weekly', etc. are also valid.

Default: `1 day`

Examples:

-   `#' @maestroFrequency 1 hour`

-   `#' @maestroFrequency daily`

-   `#' @maestroFrequency 2 weeks`

-   `#' @maestroFrequency 3 months`

## maestroStartTime

Timestamp, date, or time corresponding to the start time of the pipeline. This also sets the cadence of the pipeline in some cases. For instance, if the start time is `2025-03-18 03:00:00` and the frequency is daily, the pipeline will run at 03:00 every day. A value in the future prevents the pipeline from running until that time has been reached.

For pipelines running less frequently than daily, partial anchor formats are supported to make it easier to express natural cycle points:

- **`HH:MM:SS`** — time of day, for `minute`, `hour`, or `day` frequencies.
- **`Mon HH:MM:SS`** — weekday abbreviation + time, for `week` frequencies. Valid abbreviations: `Mon`, `Tue`, `Wed`, `Thu`, `Fri`, `Sat`, `Sun`.
- **`DD HH:MM:SS`** or **`DD`** — month-day (+ optional time), for `month` frequencies.

Default: Current date-time floored to the pipeline's frequency unit.

Examples:

-   `#' @maestroStartTime 2025-02-05 12:00:00`

-   `#' @maestroStartTime 2025-01-01`

-   `#' @maestroStartTime 08:00:00`

-   `#' @maestroStartTime Mon 04:00:00`

-   `#' @maestroStartTime Wed 09:30:00`

-   `#' @maestroStartTime 15 04:00:00`

-   `#' @maestroStartTime 1`

## maestroTz

Timezone in which the maestroStartTime is to be considered. Takes any valid timezone string that can be found in `OlsonNames()`.

Default: `UTC`

Examples:

-   `#' @maestroTz Europe/Paris`

-   `#' @maestroTz America/Halifax`

-   `#' @maestroTz US/Pacific`

## maestroLogLevel

Minimum logging threshold for messages, warnings, and errors that come from the pipeline. These levels correspond to those in `logger:::log_levels_supported` but most commonly used are ERROR, WARN, INFO. For example, if you use `WARN` then any messages of lower urgency (i.e., INFO) will be suppressed, but errors will be logged.

Default: `INFO`

Examples:

-   `#' @maestroLogLevel ERROR`

-   `#' @maestroLogLevel WARN`

## maestroFlags

Arbitrary labeling tags which are then made accessible via `get_flags()`. A pipeline can have multiple tags separated by spaces.

Default: 

Examples

-   `#' @maestroFlags critical aviation`

-   `#' @maestroFlags time-series`

------------------------------------------------------------------------

# Specifier Tags

## maestroHours

Specific hours of the day in which to run the pipeline. This only applies for pipelines that run at an hourly or minutely frequency. Acceptable values are integers from 0-23 separated by spaces. If empty, pipeline runs at all hours. This tag uses the timezone specified by `maestroTz` (will be UTC if empty).

Default:

Examples:

-   `#' @maestroHours 1 4 7 10`

-   `#' @maestroHours 0 5 20`

## maestroDays

Specific days of week or days of month on which to run the pipeline. This only applies for pipelines that run at a minutely, hourly, or daily frequency. Acceptable values are either integers from 1-31 or day of week strings like Mon, Tue, Wed, etc. These two options cannot be used in combination.

Default:

Examples:

-   `#' @maestroDays 1 11 21 31`

-   `#' @maestroDays Mon Tue Wed Thu Fri`

## maestroMonths

Specific months of the year on which to run the pipeline. This only applies for pipelines that do run at least monthly. Acceptable values are integers (1-12) corresponding to the month of the year (e.g., 1 = January, 2 = February, etc.).

Default:

Examples:

-   `#' @maestroMonths 3 8 12`

-   `#' @maestroMonths 10`

------------------------------------------------------------------------

# DAG Tags

## maestroInputs

For a DAG style pipeline, the names of pipelines that input into the pipeline. These names must match the function name defining the inputting pipeline. Multiple pipelines can be used as inputs and the input value is used in the target pipeline via the required `.input` parameter. Note that this tag could be redundant if the inputting pipeline uses `maestroOutputs`.

Default:

Examples:

-   `#' @maestroInputs extract verify`

## maestroOutputs

For a DAG style pipeline, the names of pipelines that receive the return value of this pipeline as input. These names must match the function name defining the outputting pipeline. Multiple pipelines can be outputted into. The return value of the pipeline will be passed into the receiving pipeline. Note that this tag could be redundant if pipeline to be inputted into uses `maestroInputs`.

Default:

Examples:

-   `#' @maestroOutputs transform`

------------------------------------------------------------------------

# Advanced Tags

## maestroSkip

Flags a pipeline to never be executed even if it is scheduled to run. This can be useful when developing or testing a pipeline. This tag takes no value, instead the presence of the tag indicates whether it is skipped. This tag is ignored when `run_schedule(..., run_all = TRUE)` or when using `invoke()`.

Default:

Examples:

-   `#' @maestroSkip`

## maestroPriority

Determines the order in which pipelines that run at the same scheduled instance are executed. Values are positive integers from 1 to N. Order is determined in descending order such that 1 indicates the highest priority. Pipelines with the same priority run in the order in which `build_schedule()` parses the pipeline (usually alphabetical according to file name and then line number within file). By default, all pipelines are given equal priority.

Default:

Examples:

-   `#' @maestroPriority 1`
-   `#' @maestroPriority 3`

## maestroRunIf

Sets the pipeline to be conditional based on a TRUE/FALSE value returned from an R expression. R code following the tag
must evaluate to a single TRUE or FALSE, where TRUE causes the pipeline to run (assuming it meets scheduling criteria). If 
the pipeline is part of a DAG it can use the upstream return value as .input. Variables passed through the `resources` argument of `run_schedule()` are also accessible in expression following the tag.

Default:

Examples

-   `#' @maestroRunIf sample(c(TRUE, FALSE), 1)`
-   ```
    #' @maestroRunIf
    #' rand_bool <- sample(c(TRUE, FALSE), 1)
    #' !rand_bool
    ```
-   ```
    #' @maestroInputs init_pipeline
    #' @maestroRunIf .input
    ```

## maestro

Generic tag for identifying a maestro pipeline with all the defaults. Useful when you want a pipeline to be scheduled via maestro that accepts all default tag values. Only use this tag if you have no other maestro tags. The tag takes no value.

Default:

Examples:

-   `#' @maestro`

# Contributing to Multimodal Maestro 🛠️

Thank you for your interest in contributing to Multimodal Maestro (`MM` for short)!

Multimodal Maestro gives you more control over Large Multimodal Models (LMMs) by providing the utilities you need to effectively prompt the models.

We are actively improving this library to add more prompting techniques.

## Contribution Guidelines

We welcome contributions to:

1. Add a new feature to the library (guidance below).
2. Improve our documentation and add examples to make it clear how to leverage the Multimodal Maestro library.
3. Report bugs and issues in the project.
4. Submit a request for a new feature.
5. Improve our test coverage.

### Contributing Features ✨

Before you contribute a new feature, consider submitting an Issue to discuss the feature so the community can weigh in and assist.

## How to Contribute Changes

First, fork this repository to your own GitHub account. Click "fork" in the top corner of the `multimodal-maestro` repository to get started:

![Forking the repository](https://media.roboflow.com/fork.png)

![Creating a repository fork](https://media.roboflow.com/create_fork.png)

Then, run `git clone` to download the project code to your computer.

Move to a new branch using the `git checkout` command:

```bash
git checkout -b <your_branch_name>
```

The name you choose for your branch should describe the change you want to make (i.e. `add-few-shot-prompting`).

Make any changes you want to the project code, then run the following commands to commit your changes:

```bash
git add .
git commit -m "Your commit message"
git push -u origin main
```

## 🎨 Code quality

### Pre-commit tool

This project utilizes the [pre-commit](https://pre-commit.com/) tool to maintain code quality and consistency. Before submitting a pull request or making any commits, it is important to run the pre-commit tool to ensure that your changes meet the project's guidelines.

Furthermore, we have integrated a pre-commit GitHub Action into our workflow. This means that with every pull request opened, the pre-commit checks will be automatically enforced, streamlining the code review process and ensuring that all contributions adhere to our quality standards.

To run the pre-commit tool, follow these steps:

1. Install pre-commit by running the following command: `poetry install`. It will not only install pre-commit but also install all the deps and dev-deps of project

2. Once pre-commit is installed, navigate to the project's root directory.

3. Run the command `pre-commit run --all-files`. This will execute the pre-commit hooks configured for this project against the modified files. If any issues are found, the pre-commit tool will provide feedback on how to resolve them. Make the necessary changes and re-run the pre-commit command until all issues are resolved.

4. You can also install pre-commit as a git hook by execute `pre-commit install`. Every time you made `git commit` pre-commit run automatically for you.

### Docstrings

All new functions and classes in `multimodal-maestro` should include docstrings. This is a prerequisite for any new functions and classes to be added to the library.

`multimodal-maestro` adheres to the [Google Python docstring style](https://google.github.io/styleguide/pyguide.html#383-functions-and-methods). Please refer to the style guide while writing docstrings for your contribution.

Then, go back to your fork of the `multimodal-maestro` repository, click "Pull Requests", and click "New Pull Request".

![Opening a pull request](https://media.roboflow.com/open_pr.png)

Make sure the `base` branch is `develop` before submitting your PR.

On the next page, review your changes then click "Create pull request":

![Configuring a pull request](https://media.roboflow.com/create_pr_submit.png)

Next, write a description for your pull request, and click "Create pull request" again to submit it for review:

![Submitting a pull request](https://media.roboflow.com/write_pr.png)

When creating new functions, please ensure you have the following:

1. Docstrings for the function and all parameters.
2. Unit tests for the function.
3. Examples in the documentation for the function.
4. Created an entry in our docs to autogenerate the documentation for the function.
5. Please share a Google Colab with minimal code to test new feature or reproduce PR whenever it is possible. Please ensure that Google Colab can be accessed without any issue.

When you submit your Pull Request, you will be asked to sign a Contributor License Agreement (CLA) by the `cla-assistant` GitHub bot. We can only respond to PRs from contributors who have signed the project CLA.

All pull requests will be reviewed by the maintainers of the project. We will provide feedback and ask for changes if necessary.

PRs must pass all tests and linting requirements before they can be merged.

## 📝 documentation

The `multimodal-maestro` documentation is stored in a folder called `docs`. The project documentation is built using `mkdocs`.

To run the documentation, install the project requirements with `poetry install dev`. Then, run `mkdocs serve` to start the documentation server.

You can learn more about mkdocs on the [mkdocs website](https://www.mkdocs.org/).

## 🧪 tests

[`pytests`](https://docs.pytest.org/en/7.1.x/) is used to run our tests.

## 📄 license

By contributing, you agree that your contributions will be licensed under an [MIT license](https://github.com/roboflow/multimodal-maestro/blob/develop/LICENSE.md).

## Object Detection

Object Detection is a core computer vision task where a model is trained to identify and locate multiple objects within an image by drawing bounding boxes around them. In the context of Vision-Language Models (VLMs), object detection is enhanced by the model's ability to not only recognize objects but also describe them in natural language. VLMs can provide additional context by naming objects, detailing attributes (such as color, size, or type), and offering richer descriptions of the scene. This fusion of vision and language supports more detailed and semantically aware detection, where object recognition can be linked to more complex visual understanding tasks.

## Visual Question Answering (VQA)

Visual Question Answering (VQA) merges vision and language by requiring a model to analyze an image and answer questions about its content. VLMs excel in VQA because they jointly understand both the visual components of an image and the linguistic details of the question. This allows the model to perform tasks like answering "How many dogs are there?" or "Is the person in the image wearing glasses?" with high accuracy. VQA is a key task for VLMs, demonstrating their ability to reason about complex visual scenes while considering natural language prompts.

## Object Character Recognition (OCR)

Object Character Recognition (OCR) involves detecting and recognizing text within an image, often from signs, documents, or other real-world scenes. With VLMs, OCR capabilities go beyond simple text extraction. These models understand the context in which the text appears, enabling them to answer questions, perform translations, or incorporate textual information into broader visual tasks. This contextual awareness makes VLMs particularly adept at handling tasks like reading and interpreting text embedded in images, and answering questions like "What does the sign say?" or "Translate the text in the image."

<div class="md-typeset">
    <h2><a href="#maestro.trainer.common.MeanAveragePrecisionMetric">MeanAveragePrecisionMetric</a></h2>
</div>

:::maestro.trainer.common.MeanAveragePrecisionMetric

<div class="md-typeset">
    <h2><a href="#maestro.trainer.common.WordErrorRateMetric">WordErrorRateMetric</a></h2>
</div>

:::maestro.trainer.common.WordErrorRateMetric

<div class="md-typeset">
    <h2><a href="#maestro.trainer.common.CharacterErrorRateMetric">CharacterErrorRateMetric</a></h2>
</div>

:::maestro.trainer.common.CharacterErrorRateMetric

---
title: "Quick Start"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Quick Start}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| echo: false

knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    options(maestro.check_datetime_override = as.POSIXct("2026-04-21 03:00:00", tz = "UTC"))
    set.seed(1) 
  })
)

options(asciicast_theme = "pkgdown")
```


A common task in data engineering is to automate, schedule, and monitor multiple data processing pipelines. This is called **orchestration**. `maestro` is an R package that helps orchestrate data pipelines.

A fully realized `maestro` project involves the following components and actions:

1.  Collection of **pipelines** (R functions to be orchestrated, such as batch ETL jobs)

2.  **Orchestrator** - an R script or Quarto doc that orchestrates the pipelines and monitors them

3.  A process external to R to schedule the orchestrator (e.g., cron, Posit Connect).

### Project Setup

```{r}
library(maestro)
```

Create a `maestro` project in an existing project or a new project using `create_maestro()` or the New Project wizard in RStudio. This creates the orchestrator script and the folder of pipelines with one sample pipeline. Your project should look something like this:

```{r echo=FALSE, error=FALSE, warning=FALSE, message=FALSE}
dir.create("pipelines")
writeLines(
  "
  #' my_pipe maestro pipeline
  #'
  #' @maestroFrequency 1 day
  #' @maestroStartTime 03:00:00
  #' @maestroTz UTC
  #' @maestroLogLevel INFO
  
  my_pipe <- function() {
  
    # Pipeline code
  }",
  con = "pipelines/my_pipe.R"
)
```

```         
maestro_project
├── maestro_project.Rproj
├── orchestrator.R
└── pipelines
    ├── my_pipe.R
    └── another_pipe.R
```

## Pipelines

Pipelines are the jobs you want to automate, schedule, and monitor. For the most part, they're regular R functions with a special sprinkling of comments.

### Anatomy of a Pipeline

A pipeline is simply an R function with decorators called maestro tags. Maestro tags are special code comments used for communicating the scheduling and configuration of a pipeline to the orchestrator. Let's take a quick look at the sample `my_pipe.R`:

```{r eval=FALSE}
#' my_pipe maestro pipeline
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 03:00:00
#' @maestroTz UTC
#' @maestroLogLevel INFO

my_pipe <- function() {

  # Pipeline code
}
```

`my_pipe` is a function with an empty body - so right now it won't do anything. The comments above are interpreted by `maestro` as "this function is scheduled to run every day at 03:00:00 UTC time".

maestroFrequency and maestroStartTime are the most important tags for scheduling. Frequency is how often you want the pipeline to run and can be formatted as a single string like hourly, daily, weekly, biweekly, etc. or with a number and a unit (e.g., 1 day, 3 hours, etc.).

Note that you don't need to provide all these tags. A single maestro tag is enough to distinguish it as a pipeline. Pipelines missing tags will use consistent defaults (e.g., if `maestroFrequency` is missing the default is 1 day/daily).

In most use cases, the actual code inside of `my_pipe` would be to run an ETL job (extract data from a source, transform it, and load it into a file system or database). In technical terms, it's the *side effect* of the code and not its return value that is important.

Here's a more realistic, albeit impractical, example:

```{r}
#' my_pipe maestro pipeline
#'
#' @maestroFrequency 1 day
#' @maestroStartTime 03:00:00
#' @maestroTz UTC
#' @maestroLogLevel INFO

my_pipe <- function() {

  random_data <- data.frame(
    letters = sample(letters, 10),
    numbers = sample.int(10)
  )
  
  write.csv(random_data, file = tempfile())
}
```

### Adding New Pipelines

A project with a single pipeline is ok, but in `maestro` is more useful when you have multiple jobs to run. You can add more pipelines to your pipelines directory manually or use `create_pipeline()`:

```{r example-pipe}
#| eval: false
create_pipeline(
  pipe_name = "another_pipeline",
  pipeline_dir = "pipelines",
  frequency = "1 hour",
  tz = "America/Halifax",
  log_level = "ERROR"
)
```

## Orchestrator

The orchestrator is the process that schedules and monitors the pipelines.

### Anatomy of the Orchestrator

The orchestrator can be an R script, Quarto/RMarkdown doc, but here we'll use a regular R script. Here is where you'll run `maestro` functions. The two main functions are `build_schedule()` and `run_schedule()`.

```{asciicast}
library(maestro)

schedule <- build_schedule()

output <- run_schedule(
  schedule,
  orch_frequency = "1 hour"
)
```

Building the schedule gets `maestro` to look through the pipelines in the pipelines folder and creates a schedule object. Then, you pass that to `run_schedule()` along with how often the orchestrator is supposed to run. It is important to tell `maestro` how often it'll be checking the pipelines using the `orch_frequency` parameter. Here, we're informing it that the orchestrator is running every 1 hour.

Importantly, it isn't `maestro`'s job to actually run it this often - it's *your* job to make sure it runs at that frequency (e.g., deploying it via cron or some cloud environment where code can be scheduled).[^1]

[^1]: The decision around how often to run the orchestrator depends on the frequencies of the pipelines in the project. The simplest guideline is to take the frequency of your most often recurring pipeline and split that in half. So for example, if my most often running pipeline runs every 1 day then my orchestrator should run every 12 hours.

```{r cleanup, echo=FALSE, message=FALSE, warning=FALSE}
unlink("pipelines", recursive = TRUE)
```

---
title: "Directed Acyclic Graphs (DAGs)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Directed Acyclic Graphs (DAGs)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r echo=FALSE}
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")
```


A [directed acyclic graph](https://en.wikipedia.org/wiki/Directed_acyclic_graph) or DAG is a kind of network graph where nodes are connected by edges, and these connections cannot loop back or cycle. Most data orchestration tools lay out a data pipeline as a DAG where data is passed from one function to the next until it reaches the end. This allows for more module, single-purpose functions and can make it easier to identify where errors are occurring.

You can create DAG pipelines in maestro using the `maestroInputs` and/or `maestroOutputs` tags. Let's see this in action.

```{r echo=FALSE, warning=FALSE, message=FALSE}
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroOutputs high_road low_road
  start <- function() {
    c('a', 'A')
  }
  
  #' @maestroInputs start
  high_road <- function(.input) {
    toupper(.input)
  }
  
  #' @maestroInputs start
  low_road <- function(.input) {
    tolower(.input)
  }",
  con = "pipelines/dags.R"
)
```

We'll create three simple pipelines. `start` outputs a vector, `high_road` takes an input and makes it all uppercase, `low_road` makes the input all lowercase. We use the `maestroOutputs` tag to indicate the names of the downstream pipelines (i.e., these pipelines use the output of the target pipeline as input) and we use the `maestroInputs` tag to indicate the names of pipelines that are used as input.[^1]

[^1]: Specifying the outputs and inputs is redundant. You can specify just the outputs or just the inputs if you like, but make sure all pipelines are identified as maestro pipelines by including at least one maestro tag (you could make use of the catch-all `@maestro` tag for this.

Note the use of `.input` as a parameter for all pipelines that receive an input. It is important to have this here to enable the passing of data from inputs to outputs. It must be named `.input`.

```{r eval=FALSE}
#' ./pipelines/dags.R
#' @maestroOutputs high_road low_road
start <- function() {
  c('a', 'A')
}

#' @maestroInputs start
high_road <- function(.input) {
  toupper(.input)
}

#' @maestroInputs start
low_road <- function(.input) {
  tolower(.input)
}
```

Now we'll create and run the schedule. Notice that the output in the console will reflect the network structure of the DAG.

```{asciicast}
# ./orchestrator.R
library(maestro)

schedule <- build_schedule(quiet = TRUE)

status <- run_schedule(
  schedule,
  run_all = TRUE
)

get_artifacts(schedule)
```

## ETL Example

A great case for using DAGs is with ETL/ELT pipelines. Each component of extract, transform, and load could be a single element in the DAG. Consider the example on the home page:

```{r eval=FALSE}
#' Example ETL pipeline
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-25 12:30:00
my_etl <- function() {
  
  # Pretend we're getting data from a source
  message("Get data")
  extracted <- mtcars
  
  # Transform
  message("Transforming")
  transformed <- extracted |> 
    dplyr::mutate(hp_deviation = hp - mean(hp))
  
  # Load - write to a location
  message("Writing")
  write.csv(transformed, file = paste0("transformed_mtcars_", Sys.Date(), ".csv"))
}
```

It's pretty concise, so we probably wouldn't bother breaking it apart in practice, but let's do it for illustrative purposes (and also get rid of the messaging).

```{r echo=FALSE, warning=FALSE, message=FALSE}
invisible(file.remove("pipelines/dags.R"))
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroFrequency 1 day
  #' @maestroStartTime 2024-03-25 12:30:00
  #' @maestroOutputs transform
  extract <- function() {
    # Imagine this is something way more complicated, like a database call
    mtcars
  }
  
  #' @maestroOutputs load
  transform <- function(.input) {
    .input |> 
      dplyr::mutate(hp_deviation = hp - mean(hp))
  }
  
  #' @maestro
  load <- function(.input) {
    write.csv(.input, file = paste0('transformed_mtcars.csv'))
  }",
  con = "pipelines/etl.R"
)
```

```{r eval=FALSE}
#' @maestroFrequency 1 day
#' @maestroStartTime 2024-03-25 12:30:00
#' @maestroOutputs transform
extract <- function() {
  # Imagine this is something way more complicated, like a database call
  mtcars
}

#' @maestroOutputs load
transform <- function(.input) {
  .input |> 
    dplyr::mutate(hp_deviation = hp - mean(hp))
}

#' @maestro
load <- function(.input) {
  write.csv(.input, file = paste0("transformed_mtcars.csv"))
}
```

```{asciicast}
library(maestro)

schedule <- build_schedule(quiet = TRUE)

status <- run_schedule(
  schedule,
  run_all = TRUE
)
```

```{r cleanup, echo=FALSE, message=FALSE, warning=FALSE}
invisible(unlink("pipelines", recursive = TRUE))
invisible(unlink("transformed_mtcars.csv"))
```

---
title: "Motivation and Concepts"
vignette: >
  %\VignetteIndexEntry{Motivation and Concepts}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## What is maestro?

`maestro` is an R package for creating and orchestrating many data pipelines in R. If you have several batch jobs/pipelines that you want to schedule and monitor from within a single R project, then `maestro` is for you. All you do is *decorate* R functions with special `roxygen2` tags and then execute an orchestrator script.

The diagram below imagines a weather-based maestro project with four distinct pipeline scripts (one of which has a dependency on an upstream pipeline). In this scenario, every 15 minutes the orchestrator runs and evaluates which pipelines are scheduled to run; then it kicks off those pipelines.

```{=html}
<div id="maestro-orch">
<style>
  #maestro-orch { margin: 1.5rem 0; }
  #maestro-orch svg { display: block; width: 100%; max-width: 700px; margin: 0 auto; }
  #maestro-orch .orch-ring {
    fill: none; stroke: #9012C7; stroke-width: 2; opacity: 0;
    transform-box: fill-box; transform-origin: center;
  }
  #maestro-orch .orch-ring.fire { animation: moRingOut 0.72s ease-out forwards; }
  @keyframes moRingOut {
    0%   { opacity: 0.5; transform: scale(1.0); }
    100% { opacity: 0.0; transform: scale(2.1); }
  }
  #maestro-orch .node-box {
    fill: #fff; stroke: #ddd0ed; stroke-width: 1.5;
    transition: fill 0.32s, stroke 0.32s, filter 0.32s;
  }
  #maestro-orch .node-box.active {
    fill: #9012C7; stroke: #9012C7;
    filter: drop-shadow(0 3px 14px rgba(144,18,199,0.32));
  }
  #maestro-orch .down-box {
    fill: #fff; stroke: #ddd0ed; stroke-width: 1.5;
    transition: fill 0.32s, stroke 0.32s, filter 0.32s;
  }
  #maestro-orch .down-box.active {
    fill: #c060e8; stroke: #9012C7;
    filter: drop-shadow(0 3px 10px rgba(144,18,199,0.25));
  }
  #maestro-orch .branch {
    stroke: #e8ddf5; stroke-width: 2; fill: none;
    marker-end: url(#mo-arr-off); transition: stroke 0.32s, filter 0.32s;
  }
  #maestro-orch .branch.active {
    stroke: #9012C7; marker-end: url(#mo-arr-on);
    filter: drop-shadow(0 0 5px rgba(144,18,199,0.38));
  }
  #maestro-orch .tl-segment { stroke: #e8ddf5; stroke-width: 2; transition: stroke 0.3s; }
  #maestro-orch .tl-segment.active { stroke: #9012C7; }
  #maestro-orch .tl-node {
    fill: #e4d6f5; stroke: #d4c4ea; stroke-width: 1.5;
    transition: fill 0.25s, stroke 0.25s;
  }
  #maestro-orch .tl-node.active { fill: #9012C7; stroke: #9012C7; }
  #maestro-orch .tl-label { fill: #c0a8d8; font-size: 9.5px; letter-spacing: 0.05em; transition: fill 0.25s; }
  #maestro-orch .tl-label.active { fill: #9012C7; font-weight: 700; }
  #maestro-orch .clock-face { fill: none; stroke: rgba(255,255,255,0.48); stroke-width: 1.5; }
  #maestro-orch .clock-tick { stroke: rgba(255,255,255,0.55); stroke-width: 1.5; stroke-linecap: round; }
  #maestro-orch .clock-hand { stroke: #fff; stroke-width: 2.2; stroke-linecap: round; }
  #maestro-orch .clock-dot  { fill: rgba(255,255,255,0.9); }
  #maestro-orch .badge-bg { fill: #ede5f8; transition: fill 0.32s; }
  #maestro-orch .badge-bg.active { fill: rgba(255,255,255,0.16); }
  #maestro-orch .badge-text {
    font-family: 'Courier New', Courier, monospace; font-size: 7.8px;
    fill: #8858aa; transition: fill 0.32s;
  }
  #maestro-orch .badge-text.active { fill: rgba(255,255,255,0.85); }
  #maestro-orch .node-name {
    font-size: 11px; font-weight: 500; fill: #7a5c96; transition: fill 0.32s;
  }
  #maestro-orch .node-name.active { fill: #fff; }
  #maestro-orch .wordmark { fill: #c8b2de; font-size: 11px; letter-spacing: 0.22em; font-weight: 500; }
  #maestro-orch .orch-label { fill: #6b24a0; font-size: 9.5px; font-weight: 600; letter-spacing: 0.1em; }
</style>

<svg width="700" height="410" viewBox="0 0 700 410">
<defs>
  <marker id="mo-arr-off" markerWidth="7" markerHeight="6" refX="6.5" refY="3" orient="auto" markerUnits="strokeWidth">
    <path d="M0,0.5 L7,3 L0,5.5 Z" fill="#e8ddf5"/>
  </marker>
  <marker id="mo-arr-on" markerWidth="7" markerHeight="6" refX="6.5" refY="3" orient="auto" markerUnits="strokeWidth">
    <path d="M0,0.5 L7,3 L0,5.5 Z" fill="#9012C7"/>
  </marker>
  <filter id="mo-card-shadow" x="-10%" y="-15%" width="120%" height="140%">
    <feDropShadow dx="0" dy="2" stdDeviation="4" flood-color="rgba(100,50,140,0.10)"/>
  </filter>
  <pattern id="mo-pts" x="0" y="0" width="26" height="26" patternUnits="userSpaceOnUse">
    <circle cx="13" cy="13" r="1.1" fill="#f0e8f8"/>
  </pattern>
</defs>
  <rect width="700" height="410" rx="16" fill="#faf7fd"/>
  <rect width="700" height="410" rx="16" fill="url(#mo-pts)"/>
  <text class="wordmark" x="350" y="26" text-anchor="middle">MAESTRO</text>
  <line id="mo-tls0" class="tl-segment" x1="158" y1="56" x2="298" y2="56"/>
  <line id="mo-tls1" class="tl-segment" x1="318" y1="56" x2="452" y2="56"/>
  <line id="mo-tls2" class="tl-segment" x1="472" y1="56" x2="592" y2="56"/>
  <circle id="mo-tn0" class="tl-node" cx="147" cy="56" r="11"/>
  <circle id="mo-tn1" class="tl-node" cx="308" cy="56" r="11"/>
  <circle id="mo-tn2" class="tl-node" cx="462" cy="56" r="11"/>
  <circle id="mo-tn3" class="tl-node" cx="603" cy="56" r="11"/>
  <text id="mo-tl0" class="tl-label" x="147" y="78" text-anchor="middle">:00</text>
  <text id="mo-tl1" class="tl-label" x="308" y="78" text-anchor="middle">:15</text>
  <text id="mo-tl2" class="tl-label" x="462" y="78" text-anchor="middle">:30</text>
  <text id="mo-tl3" class="tl-label" x="603" y="78" text-anchor="middle">:45</text>
  <text class="tl-label" x="222" y="78" text-anchor="middle" opacity="0.7">15m</text>
  <text class="tl-label" x="385" y="78" text-anchor="middle" opacity="0.7">15m</text>
  <text class="tl-label" x="527" y="78" text-anchor="middle" opacity="0.7">15m</text>
  <circle id="mo-or1" class="orch-ring" cx="112" cy="246" r="52"/>
  <circle id="mo-or2" class="orch-ring" cx="112" cy="246" r="60"/>
  <circle id="mo-orch-body" cx="112" cy="246" r="44" fill="#9012C7"
          style="filter: drop-shadow(0 3px 12px rgba(144,18,199,0.28)); transition: filter 0.2s;"/>
  <circle class="clock-face" cx="112" cy="246" r="20"/>
  <line class="clock-tick" x1="112" y1="228" x2="112" y2="231"/>
  <line class="clock-tick" x1="130" y1="246" x2="127" y2="246"/>
  <line class="clock-tick" x1="112" y1="264" x2="112" y2="261"/>
  <line class="clock-tick" x1="94"  y1="246" x2="97"  y2="246"/>
  <line id="mo-orch-min" class="clock-hand" x1="112" y1="246" x2="112" y2="228"/>
  <circle class="clock-dot" cx="112" cy="246" r="2.5"/>
  <text class="orch-label" x="112" y="304" text-anchor="middle">ORCHESTRATOR</text>
  <line id="mo-bra" class="branch" x1="151" y1="226" x2="291" y2="158"/>
  <line id="mo-brb" class="branch" x1="156" y1="246" x2="291" y2="246"/>
  <line id="mo-brc" class="branch" x1="151" y1="267" x2="291" y2="334"/>
  <line id="mo-brd" class="branch" x1="414" y1="246" x2="471" y2="246"/>
  <rect id="mo-ta" class="node-box" x="294" y="118" width="120" height="72" rx="12" filter="url(#mo-card-shadow)"/>
  <rect id="mo-ta-badge" class="badge-bg" x="300" y="124" width="108" height="28" rx="7"/>
  <text id="mo-ta-badge-t" class="badge-text" x="354" y="135" text-anchor="middle">
    <tspan x="354" dy="0">@maestroFrequency</tspan><tspan x="354" dy="12">15 minutes</tspan>
  </text>
  <text id="mo-ta-name" class="node-name" x="354" y="174" text-anchor="middle">temp_readings</text>
  <rect id="mo-tb" class="node-box" x="294" y="210" width="120" height="72" rx="12" filter="url(#mo-card-shadow)"/>
  <rect id="mo-tb-badge" class="badge-bg" x="300" y="216" width="108" height="28" rx="7"/>
  <text id="mo-tb-badge-t" class="badge-text" x="354" y="227" text-anchor="middle">
    <tspan x="354" dy="0">@maestroFrequency</tspan><tspan x="354" dy="12">30 minutes</tspan>
  </text>
  <text id="mo-tb-name" class="node-name" x="354" y="267" text-anchor="middle">weather_bulletins</text>
  <rect id="mo-tc" class="node-box" x="294" y="304" width="120" height="72" rx="12" filter="url(#mo-card-shadow)"/>
  <rect id="mo-tc-badge" class="badge-bg" x="300" y="310" width="108" height="28" rx="7"/>
  <text id="mo-tc-badge-t" class="badge-text" x="354" y="321" text-anchor="middle">
    <tspan x="354" dy="0">@maestroFrequency</tspan><tspan x="354" dy="12">1 hour</tspan>
  </text>
  <text id="mo-tc-name" class="node-name" x="354" y="361" text-anchor="middle">satellite_imagery</text>
  <rect id="mo-td" class="down-box" x="472" y="218" width="104" height="56" rx="12" filter="url(#mo-card-shadow)"/>
  <rect id="mo-td-badge" class="badge-bg" x="478" y="224" width="92" height="26" rx="7"/>
  <text id="mo-td-badge-t" class="badge-text" x="524" y="234" text-anchor="middle">
    <tspan x="524" dy="0">@maestroInputs</tspan><tspan x="524" dy="12">weather_bulletins</tspan>
  </text>
  <text id="mo-td-name" class="node-name" x="524" y="262" text-anchor="middle">send_alert</text>
</svg>

<script>
(function() {
  const sleep = ms => new Promise(r => setTimeout(r, ms));
  const $  = id => document.getElementById(id);
  function cls(state, name, ...ids) {
    ids.forEach(id => { const el = $(id); if (el) el.classList.toggle(name, state); });
  }
  function setFill(fill, ...ids) {
    ids.forEach(id => { const el = $(id); if (el) el.setAttribute('fill', fill); });
  }
  const TASK_DEF = { name: '#7a5c96', badge: '#ede5f8', badgeT: '#8858aa' };
  const TASK_ACT = { name: '#fff',    badge: 'rgba(255,255,255,0.16)', badgeT: 'rgba(255,255,255,0.85)' };
  const DOWN_DEF = { name: '#7a5c96', badge: '#ede5f8', badgeT: '#8858aa' };
  const DOWN_ACT = { name: '#fff',    badge: 'rgba(255,255,255,0.16)', badgeT: 'rgba(255,255,255,0.85)' };
  function setNode(id, colors) {
    const active = colors === TASK_ACT || colors === DOWN_ACT;
    cls(active, 'active', id, `${id}-badge`, `${id}-badge-t`);
    setFill(colors.name,   `${id}-name`);
    setFill(colors.badge,  `${id}-badge`);
    setFill(colors.badgeT, `${id}-badge-t`);
  }
  function setTimeline(tick) {
    for (let i = 0; i < 4; i++) {
      cls(i === tick, 'active', `mo-tn${i}`, `mo-tl${i}`);
      if (i < 3) cls(i < tick, 'active', `mo-tls${i}`);
    }
  }
  function pulseOrch() {
    const body = $('mo-orch-body');
    body.style.filter = 'drop-shadow(0 0 22px rgba(144,18,199,0.62))';
    setTimeout(() => { body.style.filter = 'drop-shadow(0 3px 12px rgba(144,18,199,0.28))'; }, 480);
    ['mo-or1','mo-or2'].forEach((id, i) => {
      const el = $(id); el.classList.remove('fire'); void el.offsetWidth;
      setTimeout(() => el.classList.add('fire'), i * 110);
    });
  }
  const MIN_POS = [
    { x2: 112, y2: 228 },
    { x2: 130, y2: 246 },
    { x2: 112, y2: 264 },
    { x2: 94,  y2: 246 },
  ];
  function setMinHand(tick) {
    const p = MIN_POS[tick];
    $('mo-orch-min').setAttribute('x2', p.x2);
    $('mo-orch-min').setAttribute('y2', p.y2);
  }
  const FIRES = [
    { a: true,  b: true,  c: true  },
    { a: true,  b: false, c: false },
    { a: true,  b: true,  c: false },
    { a: true,  b: false, c: false },
  ];
  const STEP = 310, HOLD = 820, GAP = 280, SETTLE = 150;
  async function runTick(i) {
    const f = FIRES[i];
    setTimeline(i); setMinHand(i); pulseOrch();
    await sleep(300);
    if (f.a) cls(true, 'active', 'mo-bra');
    if (f.b) cls(true, 'active', 'mo-brb');
    if (f.c) cls(true, 'active', 'mo-brc');
    await sleep(STEP);
    if (f.a) setNode('mo-ta', TASK_ACT);
    await sleep(80);
    if (f.b) setNode('mo-tb', TASK_ACT);
    await sleep(80);
    if (f.c) setNode('mo-tc', TASK_ACT);
    if (f.b) {
      await sleep(STEP);
      cls(true, 'active', 'mo-brd');
      await sleep(STEP * 0.5);
      setNode('mo-td', DOWN_ACT);
    }
    await sleep(HOLD);
    cls(false, 'active', 'mo-bra', 'mo-brb', 'mo-brc', 'mo-brd');
    setNode('mo-ta', TASK_DEF); setNode('mo-tb', TASK_DEF);
    setNode('mo-tc', TASK_DEF); setNode('mo-td', DOWN_DEF);
    await sleep(SETTLE);
  }
  async function loop() {
    for (let i = 0; i < 4; i++) { await runTick(i); await sleep(GAP); }
    await sleep(500); loop();
  }
  setTimeout(loop, 600);
})();
</script>
</div>
```

## Why do I need maestro?

Running data pipelines is an essential component of data engineering. It is not unusual to have dozens of pipelines that need to run at different frequencies, and when you go to deploy these pipelines scheduling and monitoring them quickly becomes unwieldy. Perhaps you've considered moving to heftier orchestration suites such as Airflow, Dagster, and others which require learning entirely new skills and pose their own challenges with deployment. `maestro` allows you to orchestrate your pipelines entirely in R. All you then need is an environment to deploy your `maestro` project.

## Pipelines

A pipeline is some process that takes raw data (often from an external source) and moves it somewhere else often transforming it along the way. Think of a pipeline as a factory assembly line where data is the raw material. As this data travels along the pipeline, it undergoes various transformations—such as cleaning, aggregation, and analysis—making it increasingly refined and valuable. The refined product is then stored in a new location where it can be used either by an end consumer or another pipeline. The prototypical type of pipeline in data engineering is ETL (**E**xtract, **T**ransform, **L**oad), where data is extracted from a source, transformed, then loaded into storage.

### Scheduled Batch Processing

The pipeline needs to run regularly and automatically to process new data. Most analytic workloads undergo batch processing - the processing of data in discrete timed batches. In scheduled batch processing, you as the engineer decide how often you want your pipeline to run (every day at 12:00?, every hour on the 15th minute?).

In `maestro` a pipeline is an R function with `roxygen2` comments for scheduling and configuration:

```{r, eval=FALSE}
#' my_pipe maestro pipeline
#'
#' @maestroFrequency daily
#' @maestroStartTime 02:00:00

my_pipe <- function() {

  random_data <- data.frame(
    letters = sample(letters, 10),
    numbers = sample.int(10)
  )
  
  write.csv(random_data, file = tempfile())
}
```

## Orchestrator

An orchestrator is a process that triggers pipelines to run. Think of it as the factory manager who turns on various assembly lines as needed. It also monitors all the pipelines to ensure smooth operation. Just like the factory manager, the orchestrator operates in "shifts" and so needs to be scheduled to perform it's job too.

### Rounded Scheduling

Importantly, `maestro` needs to know how often you're going to run the orchestrator. Unlike most orchestration tools out there, `maestro` isn't intended to be continuously running, which saves you on compute resources. But this means that pipelines won't necessarily run *exactly* when they're scheduled to. This is a concept we call *rounded scheduling.*

Let’s say we have a pipeline scheduled to run hourly on the 02 minute mark (e.g., 01:02, 02:02, etc.), and our orchestrator runs every hour on the 00 minute. When the orchestrator runs, it’ll be slightly before the pipeline scheduled time, but it’ll trigger the pipeline anyway because it’s close enough within the frequency of the orchestrator. If instead our orchestrator ran every 15 minutes, it’d still only execute the pipeline once in the hour. But if we underprovisioned the orchestrator and ran it only every day, then the pipeline would only execute once a day. So an important guideline is that the orchestrator needs to run at least as frequency as your highest frequency pipeline.

In `maestro` an orchestrator is an R script or Quarto like this:

```{r, eval=FALSE}
library(maestro)

schedule <- build_schedule()

run_schedule(
  schedule,
  orch_frequency = "1 hour"
)
```

By passing the `orch_frequency = "1 hour"` to `run_schedule()`, we're saying that we intend to run the orchestrator every 1 hour.

## Comparison with other packages

### {R} targets

[targets](https://docs.ropensci.org/targets/index.html) is a "pipeline tool for statistics and data science in R". If you have multiple connected components of a pipeline, `targets` skips computation of tasks that are up-to-date. `targets` seems to be primarily used for projects with a single output (e.g., model, document) where there are multiple steps that cumulatively take a long time to complete. In contrast, `maestro` is focused on projects with multiple *independent* pipelines. Moreover, `maestro` pipelines are primarily used when the *up-to-dateness* of the source data is unknown (e.g., coming from an API or database), unlike in `targets` where it determines the *up-to-dateness* based on the contents of a file.

### {Python} dagster

[Dagster](https://dagster.io/) is an "open source orchestration platform for the development, production, and observation of data assets". Like `maestro`, dagster uses decorators (special comments) to configure *data assets* (functions). Unlike `maestro`, dagster is primarily for chaining together dependent components of a multi-step pipeline - a DAG. It also supports a developer UI and is more fully developed than `maestro` at the current time.

DAGs are supported in `maestro` by defining `maestroInputs` and `maestroOutputs` tags, but `maestro` is still predominately geared toward projects with multiple pipelines running simultaneously or on different schedules.

## When to not use maestro?

While `maestro` can be used for almost any data engineering task that can be performed in R, there are cases where it is less appropriate to use it.

### Streaming and Event-driven

`maestro` does not support streaming (i.e., continuous) or event-driven pipelines. Only batch processes can be run in `maestro`.

### Hundreds of pipelines

Although there is no hard limit to the number of pipelines you can run in `maestro` (and there are ways of maximizing its efficiency as the number of pipelines increases, such as using multiple cores), we advise against using `maestro` to run *this* many pipelines - at least not in a single project. There are several reasons for this: (1) the orchestrator execution time will be become a problem even with multiple cores; (2) organizing and keeping track of this many pipelines in a single R project becomes difficult; (3) the number of dependencies to manage in the project will likely balloon.

If you wish to continue using `maestro` in this scenario, then our recommendation is to split the jobs into multiple projects all running on `maestro`.

Nevertheless, if you have hundreds of jobs to run it's likely an indicator that your enterprise has matured out of `maestro` into something a bit more sophisticated.

### High frequency jobs

If you have pipelines that need to run every minute or less you may want to look for a solution that supports near real time or real time data processing. The orchestrator may have trouble keeping up if it's scheduled to run this often.

### Multiple languages (R + Python)

`maestro` is for R pipelines only. Using `reticulate` may help with Python in a pinch though.

---
title: "Advanced DAGs - Dynamic Fan-out and Collect"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Advanced DAGs - Dynamic Fan-out and Collect}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r echo=FALSE}
knitr::opts_chunk$set(
  collapse = FALSE,
  comment = "",
  out.width = "100%",
  cache = FALSE,
  asciicast_knitr_output = "html"
)

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  same_process = TRUE,
  startup = quote({
    library(maestro)
    set.seed(1) 
  })
)
options(asciicast_theme = "pkgdown")
```

More complex types of DAGs involve dynamically spawning new pipelines (dynamic fan-out) based on a list or vector and/or collecting inputs from multiple pipelines into a single pipeline.

## Dynamic Fan-out

Sometimes an upstream pipeline returns a collection of values and you want to run a downstream pipeline **once per element** — a pattern called fan-out or scatter. Add `@maestroMap` to the downstream pipeline to enable this. An empty `@maestroMap` tag iterates over each element of the upstream return value directly.

```
get_letters
  |-shout[1]
  |-shout[2]
  |-shout[3]
```

```{r echo=FALSE, warning=FALSE, message=FALSE}
invisible(dir.create("pipelines"))
writeLines(
  "
  #' @maestroFrequency daily
  get_letters <- function() {
    c('a', 'b', 'c')
  }

  #' @maestroInputs get_letters
  #' @maestroMap
  shout <- function(.input) {
    toupper(.input)
  }",
  con = "pipelines/fanout.R"
)
```

```{r eval=FALSE}
#' @maestroFrequency daily
get_letters <- function() {
  c("a", "b", "c")
}

#' @maestroInputs get_letters
#' @maestroMap
shout <- function(.input) {
  toupper(.input)
}
```

`shout` will execute three times — once for `"a"`, once for `"b"`, once for `"c"` — and the CLI output labels each branch with its iteration index in square brackets.

```{asciicast}
library(maestro)

schedule <- build_schedule(quiet = TRUE)

run_schedule(schedule, run_all = TRUE)

get_artifacts(schedule)
```

Note there is no `@maestroOutputs` equivalent for defining dynamic fan-out. Here, you must use `@maestroInputs` combined with `@maestroMap`.

### Iterating over a field of a list

When the upstream pipeline returns a **named list**, use `@maestroMap` to select which field to scatter over. The full list remains available as `.input` inside each branch, so other fields are still accessible.

```{r echo=FALSE, warning=FALSE, message=FALSE}
invisible(file.remove("pipelines/fanout.R"))
writeLines(
  "
  #' @maestroFrequency daily
  get_letters <- function() {
    list(
      letter = letters[1:3],
      greeting = 'hello'
    )
  }

  #' @maestroInputs get_letters
  #' @maestroMap .input$letter
  make_message <- function(.input) {
    paste(.input$greeting, toupper(.input$letter))
  }",
  con = "pipelines/fanout.R"
)
```

```{r eval=FALSE}
#' @maestroFrequency daily
get_letters <- function() {
  list(
    letter = letters[1:3],
    greeting = "hello"
  )
}

#' @maestroInputs get_letters
#' @maestroMap .input$letter
make_message <- function(.input) {
  paste(.input$greeting, toupper(.input$letter))
}
```

`make_message` runs once per element of `letter`, producing `"hello A"`, `"hello B"`, `"hello C"`. The `greeting` field is available in every branch because the full list is passed as `.input` each time.

```{asciicast}
library(maestro)

schedule <- build_schedule(quiet = TRUE)

run_schedule(schedule, run_all = TRUE)

get_artifacts(schedule)
```

If the field name in `@maestroMap` does not exist in the upstream return value, maestro records an informative error on the downstream pipeline rather than silently producing zero branches.

### Iterating over multiple fields simultaneously

You can supply multiple space-separated expressions to `@maestroMap` to zip across several fields at once — similar to `purrr::pmap()`. Each iteration receives `.input` with all specified fields replaced by their i-th element.

```{r eval=FALSE}
#' @maestroFrequency daily
get_data <- function() {
  list(
    letter = letters[1:3],
    greeting = c("hello", "cheers", "hi")
  )
}

#' @maestroInputs get_data
#' @maestroMap .input$letter .input$greeting
make_message <- function(.input) {
  paste(.input$greeting, toupper(.input$letter))
}
```

This produces `"hello A"`, `"cheers B"`, `"hi C"` — each branch receives a distinct `(letter, greeting)` pair.

All vectors must be the same length, or length 1 (in which case the scalar is recycled across all iterations). Mismatched lengths produce a pipeline error.

## Fan-in (Collect)

Fan-in is the complement of fan-out: multiple upstream pipelines are gathered into a single downstream pipeline. Wrap one or more upstream names with `collect()` in `@maestroInputs` to enable this.

```
letter_a ─┐
           |-+combine
letter_b ─┘
```

The downstream pipeline receives a named list as `.input`, where each name corresponds to an upstream pipeline and each value is that pipeline's return value.

```{r echo=FALSE, warning=FALSE, message=FALSE}
invisible(file.remove("pipelines/fanout.R"))
writeLines(
  "
  #' @maestroFrequency daily
  letter_a <- function() 'a'

  #' @maestroFrequency daily
  letter_b <- function() 'b'

  #' @maestroInputs collect(letter_a, letter_b)
  combine <- function(.input) {
    paste0(.input$letter_a, .input$letter_b)
  }",
  con = "pipelines/fanin.R"
)
```

```{r eval=FALSE}
#' @maestroFrequency daily
letter_a <- function() "a"

#' @maestroFrequency daily
letter_b <- function() "b"

#' @maestroInputs collect(letter_a, letter_b)
combine <- function(.input) {
  paste0(.input$letter_a, .input$letter_b)
}
```

`combine` fires only after both `letter_a` and `letter_b` have succeeded. Inside `combine`, `.input$letter_a` is `"a"` and `.input$letter_b` is `"b"`. Collect pipelines are shown with a `|-+` prefix in the CLI to distinguish them from regular downstream pipelines.

```{asciicast}
library(maestro)

schedule <- build_schedule(quiet = TRUE)

run_schedule(schedule, run_all = TRUE)

get_status(schedule)[, c("pipe_name", "invoked", "success", "input_run_id", "lineage")]
```

If any upstream pipeline fails, the collect pipeline will not fire. The failed pipeline's run ID is also excluded from `input_run_id` in `get_status()`.

### Fan-out into Fan-in

`@maestroMap` and `collect()` compose naturally. An upstream pipeline can fan out with `@maestroMap`, and a downstream pipeline can gather all successful iterations back together with `collect()`. Note that in the dynamic fan-out to fan-in case, the downstream pipeline will run if at least one upstream iteration has succeeded.

```
numbers
  |-multiply[1] ─┐
  |-multiply[2] ──|-+add
  |-multiply[3] ─┘
```

```{r echo=FALSE, warning=FALSE, message=FALSE}
invisible(file.remove("pipelines/fanin.R"))
writeLines(
  "
  #' @maestroFrequency daily
  numbers <- function() 1:3

  #' @maestroInputs numbers
  #' @maestroMap
  multiply <- function(.input) .input * 3

  #' @maestroInputs collect(multiply)
  add <- function(.input) {
    sum(unlist(.input))
  }",
  con = "pipelines/fanin.R"
)
```

```{r eval=FALSE}
#' @maestroFrequency daily
numbers <- function() 1:3

#' @maestroInputs numbers
#' @maestroMap
multiply <- function(.input) .input * 3

#' @maestroInputs collect(multiply)
add <- function(.input) {
  sum(unlist(.input))
}
```

Here `multiply` executes three times (once per element of `1:3`), then `add` collects all three results and sums them. The `.input` received by `add` is a list of the successful iteration return values.

```{asciicast}
library(maestro)

schedule <- build_schedule(quiet = TRUE)

run_schedule(schedule, run_all = TRUE)

get_artifacts(schedule)$add
```

```{r cleanup2, echo=FALSE, message=FALSE, warning=FALSE}
invisible(unlink("pipelines", recursive = TRUE))
```

# get_schedule returns a data.frame

    Code
      schedule
    Message
      
      -- Maestro Schedule with 7 pipelines:  
      * Not Run

# get_schedule works with DAG schedules

    Code
      get_schedule(schedule)[, c("script_path", "pipe_name", "frequency", "tz",
        "skip", "log_level")]
    Output
      # A tibble: 11 x 6
         script_path                     pipe_name   frequency tz    skip  log_level
         <chr>                           <chr>       <chr>     <chr> <lgl> <chr>    
       1 test_pipelines_dags_good/dags.R with_inputs <NA>      <NA>  FALSE INFO     
       2 test_pipelines_dags_good/dags.R primary     daily     UTC   FALSE INFO     
       3 test_pipelines_dags_good/dags.R trunk       daily     UTC   FALSE INFO     
       4 test_pipelines_dags_good/dags.R branch1     <NA>      <NA>  FALSE INFO     
       5 test_pipelines_dags_good/dags.R branch2     <NA>      <NA>  FALSE INFO     
       6 test_pipelines_dags_good/dags.R subbranch1  <NA>      <NA>  FALSE INFO     
       7 test_pipelines_dags_good/dags.R subbranch2  <NA>      <NA>  FALSE INFO     
       8 test_pipelines_dags_good/dags.R start       daily     UTC   FALSE INFO     
       9 test_pipelines_dags_good/dags.R high_road   <NA>      <NA>  FALSE INFO     
      10 test_pipelines_dags_good/dags.R low_road    <NA>      <NA>  FALSE INFO     
      11 test_pipelines_dags_good/dags.R end         <NA>      <NA>  FALSE INFO     

# maestroStartTime with HH:MM:SS runs on the expected time

    Code
      status$invoked
    Output
      [1] TRUE

# maestroPriority works as expected

    Code
      status$pipe_name
    Output
      [1] "q1" "p2"

# Simple pipeline list, no errors

    Code
      pipeline_list
    Message
      
      -- Maestro Pipelines List with 1 pipeline 

---

    Code
      pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")]
    Output
      # A tibble: 1 x 5
        invoked success errors warnings messages
        <lgl>   <lgl>    <int>    <int>    <int>
      1 TRUE    TRUE         0        0        0

# Simple collect from distinct upstreams

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 3 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   

# Dynamic fan out followed by fan in

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 5 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   

# Fan-out collect with iterateOver

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 5 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   

# get_slot_usage works as expected

    Code
      avail_hour
    Output
      # A tibble: 24 x 3
         slot  n_runs pipe_names               
         <chr>  <int> <chr>                    
       1 00:00      2 hourly1, hourly2         
       2 01:00      2 hourly1, hourly3         
       3 02:00      2 hourly1, hourly2         
       4 03:00      1 hourly1                  
       5 04:00      3 hourly1, hourly2, hourly3
       6 05:00      1 hourly1                  
       7 06:00      2 hourly1, hourly2         
       8 07:00      2 hourly1, hourly3         
       9 08:00      2 hourly1, hourly2         
      10 09:00      1 hourly1                  
      # i 14 more rows

---

    Code
      avail_day
    Output
      # A tibble: 31 x 3
         slot  n_runs pipe_names               
         <chr>  <int> <chr>                    
       1 01         3 hourly1, hourly2, hourly3
       2 02         3 hourly1, hourly2, hourly3
       3 03         3 hourly1, hourly2, hourly3
       4 04         3 hourly1, hourly2, hourly3
       5 05         3 hourly1, hourly2, hourly3
       6 06         3 hourly1, hourly2, hourly3
       7 07         3 hourly1, hourly2, hourly3
       8 08         3 hourly1, hourly2, hourly3
       9 09         3 hourly1, hourly2, hourly3
      10 10         3 hourly1, hourly2, hourly3
      # i 21 more rows

# get_slot_usage works with variety of frequencies

    Code
      avail_hour
    Output
      # A tibble: 24 x 3
         slot  n_runs pipe_names
         <chr>  <int> <chr>     
       1 00:00      1 hourly1   
       2 01:00      1 hourly1   
       3 02:00      1 hourly1   
       4 03:00      1 hourly1   
       5 04:00      1 hourly1   
       6 05:00      1 hourly1   
       7 06:00      1 hourly1   
       8 07:00      1 hourly1   
       9 08:00      1 hourly1   
      10 09:00      1 hourly1   
      # i 14 more rows

---

    Code
      avail_day
    Output
      # A tibble: 31 x 3
         slot  n_runs pipe_names              
         <chr>  <int> <chr>                   
       1 01         3 hourly1, daily1, weekly1
       2 02         3 hourly1, daily1, weekly1
       3 03         3 hourly1, daily1, weekly1
       4 04         3 hourly1, daily1, weekly1
       5 05         3 hourly1, daily1, weekly1
       6 06         3 hourly1, daily1, weekly1
       7 07         3 hourly1, daily1, weekly1
       8 08         3 hourly1, daily1, weekly1
       9 09         3 hourly1, daily1, weekly1
      10 10         3 hourly1, daily1, weekly1
      # i 21 more rows

---

    Code
      avail_week
    Output
      # A tibble: 7 x 3
        slot  n_runs pipe_names              
        <chr>  <int> <chr>                   
      1 Fri        2 hourly1, daily1         
      2 Mon        2 hourly1, daily1         
      3 Sat        2 hourly1, daily1         
      4 Sun        2 hourly1, daily1         
      5 Thu        3 hourly1, daily1, weekly1
      6 Tue        2 hourly1, daily1         
      7 Wed        2 hourly1, daily1         

# DAGs work as expected

    Code
      dag$get_network()
    Output
      # A tibble: 9 x 2
        from      to         
        <chr>     <chr>      
      1 primary   with_inputs
      2 trunk     branch1    
      3 trunk     branch2    
      4 branch1   subbranch1 
      5 branch2   subbranch2 
      6 start     high_road  
      7 start     low_road   
      8 high_road end        
      9 low_road  end        

# Even if a downstream pipeline is 'scheduled' it doesn't run unless the upstream component does

    Code
      schedule$get_status()$invoked
    Output
      [1] FALSE FALSE

---

    Code
      schedule$get_status()$invoked
    Output
      [1] FALSE FALSE FALSE

# Even if a downstream pipeline is 'scheduled' it runs if the upstream component does

    Code
      schedule$get_status()$invoked
    Output
      [1] TRUE TRUE

---

    Code
      schedule$get_status()$invoked
    Output
      [1] TRUE TRUE TRUE

# Branching DAG pipelines with an error in one branch continue on second branch

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    FALSE  
      4 TRUE    TRUE   

# Branching DAG pipelines both ending in errors accurately outputs errors

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    FALSE  
      4 TRUE    FALSE  

---

    Code
      last_run_errors()
    Output
      $end1
      [1] "oops"
      
      $end2
      [1] "oh dear"
      

# Branching and merging DAG pipelines have separate status entries for each lineage

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 5 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   

---

    Code
      unname(unlist(get_artifacts(schedule)))
    Output
      [1]  4 12  8 24 16

# Branching and merging DAG pipelines use vectors for multiple errors

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 5 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    FALSE  
      5 TRUE    FALSE  

---

    Code
      unname(unlist(get_artifacts(schedule)))
    Output
      [1]  4 12  8

---

    Code
      last_run_errors()
    Output
      $end
      [1] "oops" "oops"
      

# Two separate DAGs have separate lineages

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   

# Multicore works

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 7 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   
      6 TRUE    TRUE   
      7 TRUE    TRUE   

# Multicore DAGs work

    Code
      dag$get_network()
    Output
      # A tibble: 9 x 2
        from      to         
        <chr>     <chr>      
      1 primary   with_inputs
      2 trunk     branch1    
      3 trunk     branch2    
      4 branch1   subbranch1 
      5 branch2   subbranch2 
      6 start     high_road  
      7 start     low_road   
      8 high_road end        
      9 low_road  end        

# Simple fan out into common downstream

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 7 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   
      6 TRUE    TRUE   
      7 TRUE    TRUE   

# Simple collect from distinct upstreams

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 3 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   

# Dynamic fan out followed by fan in

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 5 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   

# Simple pipeline, no errors

    Code
      pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")]
    Output
      # A tibble: 1 x 5
        invoked success errors warnings messages
        <lgl>   <lgl>    <int>    <int>    <int>
      1 TRUE    TRUE         0        0        0

# Pipeline with warnings

    Code
      pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")]
    Output
      # A tibble: 1 x 5
        invoked success errors warnings messages
        <lgl>   <lgl>    <int>    <int>    <int>
      1 TRUE    TRUE         0        1        0

# Pipeline with errors

    Code
      pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")]
    Output
      # A tibble: 1 x 5
        invoked success errors warnings messages
        <lgl>   <lgl>    <int>    <int>    <int>
      1 TRUE    FALSE        1        0        0

# Pipeline with arguments are correctly passed

    Code
      pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")]
    Output
      # A tibble: 1 x 5
        invoked success errors warnings messages
        <lgl>   <lgl>    <int>    <int>    <int>
      1 FALSE   NA           0        0        0

---

    Code
      pipeline$get_status()[c("invoked", "success", "errors", "warnings", "messages")]
    Output
      # A tibble: 1 x 5
        invoked success errors warnings messages
        <lgl>   <lgl>    <int>    <int>    <int>
      1 TRUE    TRUE         0        0        0

# get_flags works as expected

    Code
      flags_df
    Output
      # A tibble: 5 x 2
        pipe_name flag        
        <chr>     <chr>       
      1 p1        experimental
      2 p1        dev         
      3 p2        critical    
      4 p2        aviation    
      5 p3        critical    

# returns empty data.frame if there are no flags

    Code
      flags_df
    Output
      # A tibble: 0 x 2
      # i 2 variables: pipe_name <chr>, flag <chr>

# Conditional pipes work with DAG pipelines via .input

    Code
      status$invoked
    Output
      [1] TRUE TRUE

---

    Code
      status$invoked
    Output
      [1]  TRUE FALSE

---

    Code
      status$invoked
    Output
      [1]  TRUE FALSE

# DAGs with a conditional pipe in the middle by default halt further execution of the DAG

    Code
      status$invoked
    Output
      [1]  TRUE FALSE FALSE

# Conditional pipes work using resources

    Code
      status$invoked
    Output
      [1]  TRUE FALSE

# Conditions with errors are handled

    Code
      status$success
    Output
      [1] NA

---

    Code
      status$invoked
    Output
      [1] TRUE

---

    Code
      last_run_errors()
    Output
      $p1
      <simpleError in rlang::eval_bare(e, env): Error evaluating condition: oh no>
      

# Conditions that don't return a single boolean are handled

    Code
      status$success
    Output
      [1] NA

---

    Code
      status$invoked
    Output
      [1] TRUE

---

    Code
      last_run_errors()
    Output
      $p1
      <simpleError: Error evaluating condition: `c(1, 2)` did not return a single boolean.>
      

# Empty maestroRunIf is ignored

    Code
      status$success
    Output
      [1] TRUE

---

    Code
      status$invoked
    Output
      [1] TRUE

# Branching pipelines execute with conditionals

    Code
      status$invoked
    Output
      [1]  TRUE  TRUE FALSE

# invoke gives informative error message on failure

    Code
      schedule$get_status()[, c("invoked", "success")]
    Output
      # A tibble: 1 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    FALSE  

# invoke triggers DAG pipelines

    Code
      schedule$get_status()[, c("invoked", "success")]
    Output
      # A tibble: 2 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   

# Simple fan out with no specified iterator

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   

# Simple fan out into common downstream

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 7 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   
      6 TRUE    TRUE   
      7 TRUE    TRUE   

# Simple fan out into common downstream and one error

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 6 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    FALSE  
      3 TRUE    TRUE   
      4 TRUE    TRUE   
      5 TRUE    TRUE   
      6 TRUE    TRUE   

# Use maestroMap to specify a particular iteration variable

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   

# Fan out where upstream returns a list of S3 objects (lm)

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   

# Fan where the field contains S3 objects (lm)

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   

# map is misspecified name in return

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 2 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    FALSE  

# Use maestroMap with multiple iterators

    Code
      status[, c("invoked", "success")]
    Output
      # A tibble: 4 x 2
        invoked success
        <lgl>   <lgl>  
      1 TRUE    TRUE   
      2 TRUE    TRUE   
      3 TRUE    TRUE   
      4 TRUE    TRUE   

# can get a single labeled pipeline

    Code
      schedule$get_labels()
    Output
      # A tibble: 1 x 3
        pipe_name label  value         
        <chr>     <chr>  <chr>         
      1 labelled  domain transportation

# can get a double labeled pipeline

    Code
      schedule$get_labels()
    Output
      # A tibble: 2 x 3
        pipe_name label  value         
        <chr>     <chr>  <chr>         
      1 labelled  domain transportation
      2 labelled  author will          

# can get a double labeled pipeline with the same key

    Code
      schedule$get_labels()
    Output
      # A tibble: 2 x 3
        pipe_name label  value
        <chr>     <chr>  <chr>
      1 labelled  author will 
      2 labelled  author kaz  

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_network.R
\name{get_network}
\alias{get_network}
\title{Get the network structure of pipelines in a MaestroSchedule object}
\usage{
get_network(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
data.frame with columns \code{from} and \code{to}
}
\description{
Returns the pipeline dependency structure as an edge list data.frame.
Each row represents a directed dependency between two pipelines. The
result will be empty if there are no DAG pipelines in the schedule.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  get_network(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_network()
}
}
\seealso{
\code{\link[=show_network]{show_network()}} which is deprecated in favour of this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_orchestrator.R
\name{create_orchestrator}
\alias{create_orchestrator}
\title{Create a new orchestrator}
\usage{
create_orchestrator(
  path,
  type = c("R", "Quarto", "RMarkdown"),
  open = interactive(),
  quiet = FALSE,
  overwrite = FALSE
)
}
\arguments{
\item{path}{file path for the orchestrator script}

\item{type}{file type for the orchestrator (supports R, Quarto, and RMarkdown)}

\item{open}{whether or not to open the script upon creation}

\item{quiet}{whether to silence messages in the console (default = \code{FALSE})}

\item{overwrite}{whether to overwrite an existing orchestrator or maestro project}
}
\value{
invisible
}
\description{
Create a new orchestrator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_schedule.R
\name{get_schedule}
\alias{get_schedule}
\title{Get the schedule from a MaestroSchedule object}
\usage{
get_schedule(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
data.frame
}
\description{
A schedule is represented as a table where each row is a pipeline and
the columns contain scheduling parameters such as the frequency and start time.
}
\details{
The schedule table is used internally in a MaestroSchedule object but can be
accessed using this function or accessing the R6 method of the MaestroSchedule
object.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  get_schedule(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_schedule()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MaestroPipelineList.R
\name{MaestroPipelineList}
\alias{MaestroPipelineList}
\title{Class for a list of MaestroPipelines
A MaestroPipelineList is created when there are multiple maestro pipelines in a
single script}
\description{
Class for a list of MaestroPipelines
A MaestroPipelineList is created when there are multiple maestro pipelines in a
single script
}
\keyword{internal}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{MaestroPipelines}}{list of pipelines}

    \item{\code{n_pipelines}}{number of pipelines in the list}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MaestroPipelineList-initialize}{\code{MaestroPipelineList$new()}}
    \item \href{#method-MaestroPipelineList-print}{\code{MaestroPipelineList$print()}}
    \item \href{#method-MaestroPipelineList-add_pipelines}{\code{MaestroPipelineList$add_pipelines()}}
    \item \href{#method-MaestroPipelineList-update_pipelines}{\code{MaestroPipelineList$update_pipelines()}}
    \item \href{#method-MaestroPipelineList-get_pipe_names}{\code{MaestroPipelineList$get_pipe_names()}}
    \item \href{#method-MaestroPipelineList-get_pipe_by_name}{\code{MaestroPipelineList$get_pipe_by_name()}}
    \item \href{#method-MaestroPipelineList-get_pipes_by_name}{\code{MaestroPipelineList$get_pipes_by_name()}}
    \item \href{#method-MaestroPipelineList-get_priorities}{\code{MaestroPipelineList$get_priorities()}}
    \item \href{#method-MaestroPipelineList-get_schedule}{\code{MaestroPipelineList$get_schedule()}}
    \item \href{#method-MaestroPipelineList-get_timely_pipelines}{\code{MaestroPipelineList$get_timely_pipelines()}}
    \item \href{#method-MaestroPipelineList-get_primary_pipes}{\code{MaestroPipelineList$get_primary_pipes()}}
    \item \href{#method-MaestroPipelineList-check_timeliness}{\code{MaestroPipelineList$check_timeliness()}}
    \item \href{#method-MaestroPipelineList-get_status}{\code{MaestroPipelineList$get_status()}}
    \item \href{#method-MaestroPipelineList-get_errors}{\code{MaestroPipelineList$get_errors()}}
    \item \href{#method-MaestroPipelineList-get_warnings}{\code{MaestroPipelineList$get_warnings()}}
    \item \href{#method-MaestroPipelineList-get_messages}{\code{MaestroPipelineList$get_messages()}}
    \item \href{#method-MaestroPipelineList-get_artifacts}{\code{MaestroPipelineList$get_artifacts()}}
    \item \href{#method-MaestroPipelineList-get_run_sequences}{\code{MaestroPipelineList$get_run_sequences()}}
    \item \href{#method-MaestroPipelineList-get_flags}{\code{MaestroPipelineList$get_flags()}}
    \item \href{#method-MaestroPipelineList-get_labels}{\code{MaestroPipelineList$get_labels()}}
    \item \href{#method-MaestroPipelineList-get_network}{\code{MaestroPipelineList$get_network()}}
    \item \href{#method-MaestroPipelineList-validate_network}{\code{MaestroPipelineList$validate_network()}}
    \item \href{#method-MaestroPipelineList-run}{\code{MaestroPipelineList$run()}}
    \item \href{#method-MaestroPipelineList-run_pending_collects}{\code{MaestroPipelineList$run_pending_collects()}}
    \item \href{#method-MaestroPipelineList-reset_pipelines}{\code{MaestroPipelineList$reset_pipelines()}}
    \item \href{#method-MaestroPipelineList-clone}{\code{MaestroPipelineList$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-initialize}{}}}
\subsection{\code{MaestroPipelineList$new()}}{
  Create a MaestroPipelineList object
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$new(MaestroPipelines = list(), network = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{MaestroPipelines}}{list of MaestroPipelines}
      \item{\code{network}}{initialize a network}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroPipelineList
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-print"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-print}{}}}
\subsection{\code{MaestroPipelineList$print()}}{
  Print the MaestroPipelineList
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$print()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    print
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-add_pipelines"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-add_pipelines}{}}}
\subsection{\code{MaestroPipelineList$add_pipelines()}}{
  Add pipelines to the list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$add_pipelines(MaestroPipelines = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{MaestroPipelines}}{list of MaestroPipelines}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-update_pipelines"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-update_pipelines}{}}}
\subsection{\code{MaestroPipelineList$update_pipelines()}}{
  Update pipelines in a list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$update_pipelines(MaestroPipelines = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{MaestroPipelines}}{list of MaestroPipelines}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_pipe_names"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_pipe_names}{}}}
\subsection{\code{MaestroPipelineList$get_pipe_names()}}{
  Get names of the pipelines in the list arranged by priority
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_pipe_names()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    character
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_pipe_by_name"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_pipe_by_name}{}}}
\subsection{\code{MaestroPipelineList$get_pipe_by_name()}}{
  Get a MaestroPipeline by its name
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_pipe_by_name(pipe_name)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{pipe_name}}{name of the pipeline}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroPipeline
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_pipes_by_name"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_pipes_by_name}{}}}
\subsection{\code{MaestroPipelineList$get_pipes_by_name()}}{
  Get a MaestroPipelineList with selected pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_pipes_by_name(pipe_names)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{pipe_names}}{names of the pipelines}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroPipelineList
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_priorities"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_priorities}{}}}
\subsection{\code{MaestroPipelineList$get_priorities()}}{
  Get priorities
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_priorities()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    numeric
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_schedule"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_schedule}{}}}
\subsection{\code{MaestroPipelineList$get_schedule()}}{
  Get the schedule as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_schedule()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_timely_pipelines"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_timely_pipelines}{}}}
\subsection{\code{MaestroPipelineList$get_timely_pipelines()}}{
  Get a new MaestroPipelineList containing only those pipelines scheduled to run
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_timely_pipelines(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{arguments passed to self$check_timeliness}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroPipelineList
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_primary_pipes"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_primary_pipes}{}}}
\subsection{\code{MaestroPipelineList$get_primary_pipes()}}{
  Get pipelines that are primary (i.e., don't have an inputting pipeline)
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_primary_pipes()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list of MaestroPipelines
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-check_timeliness"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-check_timeliness}{}}}
\subsection{\code{MaestroPipelineList$check_timeliness()}}{
  Check whether pipelines in the list are scheduled to run based on orchestrator frequency and current time
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$check_timeliness(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{arguments passed to self$check_timeliness}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    logical
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_status"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_status}{}}}
\subsection{\code{MaestroPipelineList$get_status()}}{
  Get status of the pipelines as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_status()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_errors"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_errors}{}}}
\subsection{\code{MaestroPipelineList$get_errors()}}{
  Get list of errors from the pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_errors()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_warnings"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_warnings}{}}}
\subsection{\code{MaestroPipelineList$get_warnings()}}{
  Get list of warnings from the pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_warnings()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_messages"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_messages}{}}}
\subsection{\code{MaestroPipelineList$get_messages()}}{
  Get list of messages from the pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_messages()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_artifacts"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_artifacts}{}}}
\subsection{\code{MaestroPipelineList$get_artifacts()}}{
  Get artifacts (return values) from the pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_artifacts()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_run_sequences"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_run_sequences}{}}}
\subsection{\code{MaestroPipelineList$get_run_sequences()}}{
  Get run sequences from the pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_run_sequences(
  n = NULL,
  min_datetime = NULL,
  max_datetime = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{optional sequence limit}
      \item{\code{min_datetime}}{optional minimum datetime}
      \item{\code{max_datetime}}{optional maximum datetime}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_flags"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_flags}{}}}
\subsection{\code{MaestroPipelineList$get_flags()}}{
  Get the flags of the pipelines as a named list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_flags()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_labels"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_labels}{}}}
\subsection{\code{MaestroPipelineList$get_labels()}}{
  Get the labels of the pipelines as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_labels()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-get_network"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-get_network}{}}}
\subsection{\code{MaestroPipelineList$get_network()}}{
  Get the network structure as a edge list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$get_network()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-validate_network"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-validate_network}{}}}
\subsection{\code{MaestroPipelineList$validate_network()}}{
  Validates whether all inputs and outputs exist and that the network is a valid DAG
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$validate_network()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    warning or invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-run"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-run}{}}}
\subsection{\code{MaestroPipelineList$run()}}{
  Runs all the pipelines in the list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$run(..., cores = 1L, pipes_to_run = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{arguments passed to MaestroPipeline$run}
      \item{\code{cores}}{if using multicore number of cores to run in (uses \code{furrr})}
      \item{\code{pipes_to_run}}{an optional vector of pipe names to run. If \code{NULL} defaults to all primary pipelines}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-run_pending_collects"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-run_pending_collects}{}}}
\subsection{\code{MaestroPipelineList$run_pending_collects()}}{
  Run any collect pipelines that are ready but were skipped during a
parallel run. Called on the main process after update_pipelines() has
synced worker state back. Uses run_pipe internally so that the collect
pipe's own downstream outputs are recursed into normally.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$run_pending_collects(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{arguments forwarded to MaestroPipeline$run (same dots as run())}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list of MaestroPipeline objects that were run
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-reset_pipelines"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-reset_pipelines}{}}}
\subsection{\code{MaestroPipelineList$reset_pipelines()}}{
  Resets the run time attributes
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$reset_pipelines()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipelineList-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipelineList-clone}{}}}
\subsection{\code{MaestroPipelineList$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipelineList$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{convert_to_seconds}
\alias{convert_to_seconds}
\title{Convert a duration string to number of seconds}
\usage{
convert_to_seconds(time_string)
}
\arguments{
\item{time_string}{string like 1 day, 1 week, 12 hours, etc.}
}
\value{
number of seconds
}
\description{
Convert a duration string to number of seconds
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maestro_tags.R
\name{maestro_tags}
\alias{maestro_tags}
\title{Maestro Tags}
\description{
maestro tags are roxygen2 comments for configuring the scheduling and execution
of pipelines.
}
\details{
maestro tags follow the format: \verb{#' @maestroTagName value}

Some tags may not take a value.

maestro tags must be written above the function that is to be included as a pipeline.
A typical pipeline with tags could look like this:

\if{html}{\out{<div class="sourceCode">}}\preformatted{#' @maestroFrequency 1 hour
#' @maestroStartTime 12:30:00
#' @maestroLogLevel WARN
my_pipeline <- function() \{

   # Pipeline code
\}
}\if{html}{\out{</div>}}

Below are descriptions of all the tags currently available in maestro along
with examples.
}
\section{maestroFrequency}{
How often to run the pipeline. This tag takes a time unit indicating how long
to wait between subsequent runs of the pipeline. Acceptable values include
an integer value followed by one of minute(s), hour(s), day(s), week(s), month(s),
and year(s). Note that some combinations of integer + unit may be invalid.
Adverbs like 'hourly', 'daily', 'weekly', etc. are also valid.

Default: \verb{1 day}

Examples:
\itemize{
\item \verb{#' @maestroFrequency 1 hour}
\item \verb{#' @maestroFrequency daily}
\item \verb{#' @maestroFrequency 2 weeks}
\item \verb{#' @maestroFrequency 3 months}
}
}

\section{maestroStartTime}{
Timestamp, date, or time corresponding to the start time of the pipeline. This
also sets the cadence of the pipeline in some cases. For instance, if the start time
is \verb{2025-03-18 03:00:00} and the frequency is daily, the pipeline will run at 03:00
every day. A value in the future prevents the pipeline from running until that time
has been reached.

For pipelines with a frequency lower than daily, partial anchor formats are supported
to make it easier to express natural cycle points without choosing a specific date:
\itemize{
\item \strong{\code{HH:MM:SS}} — time-of-day anchor for \code{minute}, \code{hour}, or \code{day} frequencies.
e.g., \code{08:00:00} runs every day at 8am.
\item \strong{\verb{Mon HH:MM:SS}} (weekday abbreviation + time) — anchor for \code{week} or multi-week
frequencies. Resolved to that weekday of the current ISO week.
e.g., \verb{Mon 04:00:00} with \verb{@maestroFrequency 1 week} runs every Monday at 4am.
Valid abbreviations: \code{Mon}, \code{Tue}, \code{Wed}, \code{Thu}, \code{Fri}, \code{Sat}, \code{Sun}.
\item \strong{\verb{DD HH:MM:SS}} or \strong{\code{DD}} (month-day + optional time) — anchor for \code{month}
frequencies. Resolved to that day of the current month.
e.g., \verb{15 04:00:00} with \verb{@maestroFrequency 1 month} runs on the 15th of every month at 4am.
}

Default: \verb{2024-01-01 00:00:00}

Examples:
\itemize{
\item \verb{#' @maestroStartTime 2025-02-05 12:00:00}
\item \verb{#' @maestroStartTime 2025-01-01}
\item \verb{#' @maestroStartTime 08:00:00}
\item \verb{#' @maestroStartTime Mon 04:00:00}
\item \verb{#' @maestroStartTime Wed 09:30:00}
\item \verb{#' @maestroStartTime 15 04:00:00}
\item \verb{#' @maestroStartTime 1}
}
}

\section{maestroTz}{
Timezone in which the maestroStartTime is to be considered. Takes any valid timezone
string that can be found in \code{OlsonNames()}.

Default: \code{UTC}

Examples:
\itemize{
\item \verb{#' @maestroTz Europe/Paris}
\item \verb{#' @maestroTz America/Halifax}
\item \verb{#' @maestroTz US/Pacific}
}
}

\section{maestroLogLevel}{
Minimum logging threshold for messages, warnings, and errors that come from the pipeline.
These levels correspond to those in \code{logger:::log_levels_supported} but most commonly
used are ERROR, WARN, INFO. For example, if you use \code{WARN} then any messages of lower
urgency (i.e., INFO) will be suppressed, but errors will be logged.

Default: \code{INFO}

Examples:
\itemize{
\item \verb{#' @maestroLogLevel ERROR}
\item \verb{#' @maestroLogLevel WARN}
}
}

\section{maestroHours}{
Specific hours of the day in which to run the pipeline. This only applies for pipelines that run
at an hourly or minutely frequency. Acceptable values are integers from 0-23 separated
by spaces. If empty, pipeline runs at all hours. This tag uses the timezone specified by
\code{maestroTz} (will be UTC if empty).

Default:

Examples:
\itemize{
\item \verb{#' @maestroHours 1 4 7 10}
\item \verb{#' @maestroHours 0 5 20}
}
}

\section{maestroDays}{
Specific days of week or days of month on which to run the pipeline. This only applies
for pipelines that run at a minutely, hourly, or daily frequency. Acceptable values are
either integers from 1-31 or day of week strings like Mon, Tue, Wed, etc. These two
options cannot be used in combination.

Default:

Examples
\itemize{
\item \verb{#' @maestroDays 1 11 21 31}
\item \verb{#' @maestroDays Mon Tue Wed Thu Fri}
}
}

\section{maestroMonths}{
Specific months of the year on which to run the pipeline. This only applies for
pipelines that do run at least monthly. Acceptable values are integers (1-12) corresponding
to the month of the year (e.g., 1 = January, 2 = February, etc.).

Default:

Examples:
\itemize{
\item \verb{#' @maestroMonths 3 8 12}
\item \verb{#' @maestroMonths 10}
}
}

\section{maestroInputs}{
For a DAG style pipeline, the names of pipelines that input into the pipeline.
These names must match the function name defining the inputting pipeline. Multiple
pipelines can be used as inputs and the input value is used in the target pipeline
via the required \code{.input} parameter. Note that this tag could be redundant if the
inputting pipeline uses \code{maestroOutputs}.

To enable \strong{fan-in (collect)}, wrap one or more input names with \code{collect()}.
The downstream pipeline receives a named list as \code{.input}, where each name
corresponds to an upstream pipeline and each value is that pipeline's return
value. All listed upstream pipelines must have succeeded before the collect
pipeline fires. Works with both static multi-source inputs and dynamic fan-out
followed by fan-in (\verb{@maestroMap} → \code{collect()}): in the latter case \code{.input}
is a list of all successful iteration results.

Default:

Examples:
\itemize{
\item \verb{#' @maestroInputs extract verify}
\item \verb{#' @maestroInputs collect(letter_a, letter_b)}
\item \verb{#' @maestroInputs collect(multiply)} \emph{(collect all iterations of a \verb{@maestroMap} upstream)}
}
}

\section{maestroOutputs}{
For a DAG style pipeline, the names of pipelines that receive the return value of
this pipeline as input. These names must match the function name defining the
outputting pipeline. Multiple pipelines can be outputted into. The return value
of the pipeline will be passed into the receiving pipeline. Note that this tag
could be redundant if pipeline to be inputted into uses \code{maestroInputs}.

Default:

Examples:
\itemize{
\item \verb{#' @maestroOutputs transform}
}
}

\section{maestroSkip}{
Flags a pipeline to never be executed even if it is scheduled to run. This can
be useful when developing or testing a pipeline. This tag takes no value, instead
the presence of the tag indicates whether it is skipped. This tag is ignored when
\code{run_schedule(..., run_all = TRUE)} or when using \code{invoke()}.

Default:

Examples:
\itemize{
\item \verb{#' @maestroSkip}
}
}

\section{maestroRunIf}{
An R expression that is evaluated at run time to determine whether the pipeline
should execute. The expression must return a single \code{TRUE} or \code{FALSE}. If the
expression returns \code{FALSE}, the pipeline is silently skipped for that run. Resources
passed via \code{run_schedule(..., resources = list(...))} are available inside the
expression, as is \code{.input} for DAG pipelines.

Default:

Examples:
\itemize{
\item \verb{#' @maestroRunIf Sys.getenv("RUN_PIPELINE") == "true"}
\item \verb{#' @maestroRunIf lubridate::wday(lubridate::now()) == 2}
\item \verb{#' @maestroRunIf !is.null(.input)}
}
}

\section{maestroMap}{
Enables \strong{dynamic fan-out (scatter)}: the downstream pipeline executes once
per element of the upstream return value. When the tag value is empty, each
element of the upstream return value is passed directly as \code{.input}. When one
or more R expressions referencing \code{.input} fields are provided (space-separated),
those fields are scattered over and zipped together \code{pmap}-style: each iteration
receives \code{.input} with all specified fields replaced by their i-th element.
The full list remains accessible as \code{.input} in every branch.

Vectors must all have the same length, unless a vector has length 1, in which
case it is recycled across all iterations. Mismatched lengths produce a
pipeline error.

Requires \verb{@maestroInputs} to reference an upstream pipeline.

Default:

Examples:
\itemize{
\item \verb{#' @maestroMap} \emph{(iterate over each element of the upstream return value)}
\item \verb{#' @maestroMap .input$items}
\item \verb{#' @maestroMap .input$ids .input$labels}
\item \verb{#' @maestroMap .input$model .input$label .input$threshold}
}
}

\section{maestroPriority}{
Determines the order in which pipelines that run at the same scheduled instance
are executed. Values are positive integers from 1 to N. Order is determined in
descending order such that 1 indicates the highest priority. Pipelines with the
same priority run in the order in which \code{build_schedule()} parses the pipeline
(usually alphabetical according to file name and then line number within file).
By default, all pipelines are given equal priority.

Default:

Examples:
\itemize{
\item \verb{#' @maestroPriority 1}
\item \verb{#' @maestroPriority 3}
}
}

\section{maestroFlags}{
Arbitrary strings which are then made accessible via \code{get_flags()}. A pipeline
can have multiple tags separated by spaces.

Default:

Examples:
\itemize{
\item \verb{#' @maestroFlags critical etl cloud}
\item \verb{#' @maestroFlags aviation}
}
}

\section{maestroLabel}{
Key-value pairs used for attributing metadata for a pipeline. Multiple labels
can be assigned to a pipeline using multiple instances of \verb{@maestroLabel}. These
labels can be extracted as a data.frame. Each label must follow the pattern of
\code{key} \code{value}.

Default:

Examples:
\itemize{
\item \verb{#' @maestroLabel domain transportation}
\item \verb{#' @maestroLabel author will.hipson}
}
}

\section{maestro}{
Generic tag for identifying a maestro pipeline with all the defaults. Useful
when you want a pipeline to be scheduled via maestro that accepts all default
tag values. Only use this tag if you have no other maestro tags. The tag takes
no value.

Default:

Examples:
\itemize{
\item \verb{#' @maestro}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maestro.R
\docType{package}
\name{maestro}
\alias{maestro-package}
\alias{maestro}
\title{\code{maestro} package}
\description{
Lightweight pipeline orchestration in R
}
\details{
Documentation: \href{https://github.com/whipson/maestro}{GitHub}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/whipson/maestro}
  \item \url{https://whipson.github.io/maestro/}
  \item Report bugs at \url{https://github.com/whipson/maestro/issues}
}

}
\author{
\strong{Maintainer}: Will Hipson \email{will.e.hipson@gmail.com} (\href{https://orcid.org/0000-0002-3931-2189}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Will Hipson \email{will.e.hipson@gmail.com} (\href{https://orcid.org/0000-0002-3931-2189}{ORCID}) [copyright holder]
  \item Ryan Garnett \email{ryangarnett78@gmail.com} [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_status.R
\name{get_status}
\alias{get_status}
\title{Get the statuses of the pipelines in a MaestroSchedule object}
\usage{
get_status(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
data.frame
}
\description{
A status data.frame contains the names and locations of the pipelines as
well as information around whether they were invoked, the status (error, warning, etc.),
and the run time.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  schedule <- run_schedule(
    schedule,
    orch_frequency = "1 day",
    quiet = TRUE
  )

  get_status(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_status()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_maestro.R
\name{create_maestro}
\alias{create_maestro}
\title{Creates a new maestro project}
\usage{
create_maestro(path, type = "R", overwrite = FALSE, quiet = FALSE, ...)
}
\arguments{
\item{path}{file path for the orchestrator script}

\item{type}{file type for the orchestrator (supports R, Quarto, and RMarkdown)}

\item{overwrite}{whether to overwrite an existing orchestrator or maestro project}

\item{quiet}{whether to silence messages in the console (default = \code{FALSE})}

\item{...}{unused}
}
\value{
invisible
}
\description{
Creates a new maestro project
}
\examples{

# Creates a new maestro project with an R orchestrator
if (interactive()) {
  new_proj_dir <- tempdir()
  create_maestro(new_proj_dir, type = "R", overwrite = TRUE)

  create_maestro(new_proj_dir, type = "Quarto", overwrite = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/latest_getters.R
\name{last_run_warnings}
\alias{last_run_warnings}
\title{Retrieve latest maestro pipeline warnings}
\usage{
last_run_warnings()
}
\value{
warning messages
}
\description{
Gets the latest pipeline warnings following use of \code{run_schedule()}. If
there are no warnings or \code{run_schedule()} has not been run it will be \code{NULL}.
}
\details{
Note that setting \code{maestroLogLevel} to something greater than \code{WARN} will
ignore warnings.
}
\examples{

last_run_warnings()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.prev_on_cycle}
\alias{.prev_on_cycle}
\title{Find the most recent cycle point before a given time}
\usage{
.prev_on_cycle(
  start,
  current = Sys.time(),
  amount = 1L,
  unit = c("second", "minute", "hour", "day", "week", "month", "year")
)
}
\arguments{
\item{start}{\code{Date} or \code{POSIXct} defining the cycle origin.}

\item{current}{\code{Date} or \code{POSIXct} (coerced to match \code{start})
used as the reference point. Defaults to \code{Sys.time()}.}

\item{amount}{Positive integer; number of \code{unit}s per cycle step.}

\item{unit}{Character; one of \code{"second"}, \code{"minute"},
\code{"hour"}, \code{"day"}, \code{"week"}, \code{"month"},
\code{"year"}. Sub-day units are not supported when \code{start} is a
\code{Date}.}
}
\value{
A \code{Date} or \code{POSIXct} of the same class and timezone as
\code{start}, or \code{NA} if \code{current <= start}.
}
\description{
Given a repeating cycle anchored at \code{start} and stepping every
\code{amount} \code{unit}s, returns the latest cycle point strictly before
\code{current}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_schedule.R
\name{run_schedule}
\alias{run_schedule}
\title{Run a schedule}
\usage{
run_schedule(
  schedule,
  orch_frequency = "1 day",
  check_datetime = lubridate::now(tzone = "UTC"),
  resources = list(),
  run_all = FALSE,
  n_show_next = 5,
  cores = 1,
  log_file_max_bytes = 1e+06,
  quiet = FALSE,
  log_to_console = FALSE,
  log_to_file = FALSE
)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}

\item{orch_frequency}{of the orchestrator, a single string formatted like "1 day", "2 weeks", "hourly", etc.}

\item{check_datetime}{datetime against which to check the running of pipelines (default is current system time in UTC)}

\item{resources}{named list of shared resources made available to pipelines as needed}

\item{run_all}{run all pipelines regardless of the schedule (default is \code{FALSE}) - useful for testing.
Does not apply to pipes with a \code{maestroSkip} tag. Conditional pipelines using \code{maestroRunIf} still behave according to their condition.}

\item{n_show_next}{show the next n scheduled pipes}

\item{cores}{number of cpu cores to run if running in parallel. If > 1, \code{furrr} is used and a multisession plan must be executed in the orchestrator (see details)}

\item{log_file_max_bytes}{numeric specifying the maximum number of bytes allowed in the log file before purging the log (within a margin of error)}

\item{quiet}{silence metrics to the console (default = \code{FALSE}). Note this does not affect messages generated from pipelines when \code{log_to_console = TRUE}.}

\item{log_to_console}{whether or not to include pipeline messages, warnings, errors to the console (default = \code{FALSE}) (see Logging & Console Output section)}

\item{log_to_file}{either a boolean to indicate whether to create and append to a \code{maestro.log} or a character path to a specific log file. If \code{FALSE} or \code{NULL} it will not log to a file.}
}
\value{
MaestroSchedule object
}
\description{
Given a schedule in a \code{maestro} project, runs the pipelines that are scheduled to execute
based on the current time.
}
\details{
\subsection{Pipeline schedule logic}{

The function \code{run_schedule()} examines each pipeline in the schedule table and determines
whether it is scheduled to run at the current time using some simple time arithmetic. We assume
\code{run_schedule(schedule, check_datetime = Sys.time())}, but this need not be the case.
}

\subsection{Output}{

\code{run_schedule()} returns the same MaestroSchedule object with modified attributes. Use \code{get_status()}
to examine the status of each pipeline and use \code{get_artifacts()} to get any return values from the
pipelines as a list.
}

\subsection{Pipelines with arguments (resources)}{

If a pipeline takes an argument that doesn't include a default value, these can be supplied
in the orchestrator via \code{run_schedule(resources = list(arg1 = val))}. The name of the argument
used by the pipeline must match the name of the argument in the list. Currently, each named
resource must refer to a single object. In other words, you can't have two pipes using
the same argument but requiring different values.
}

\subsection{Running in parallel}{

Pipelines can be run in parallel using the \code{cores} argument. First, you must run \code{future::plan(future::multisession)}
in the orchestrator. Then, supply the desired number of cores to the \code{cores} argument. Note that
console output appears different in multicore mode.
}

\subsection{Logging & Console Output}{

By default, \code{maestro} suppresses pipeline messages, warnings, and errors from appearing in the console, but
messages coming from \code{print()} and other console logging packages like \code{cli} and \code{logger} are not suppressed
and will be interwoven into the output generated from \code{run_schedule()}. Messages from \code{cat()} and related functions are always suppressed
due to the nature of how those functions operate with standard output.

Users are advised to make use of R's \code{message()}, \code{warning()}, and \code{stop()} functions in their pipelines
for managing conditions. Use \code{log_to_console = TRUE} to print these to the console.

Note that \code{log_to_console} does not work when running multicore. In the case of multicore, it is recommended to use
\code{log_to_file}.

Maestro can generate a log file that is appended to each time the orchestrator is run. Use \code{log_to_file = TRUE} or \code{log_to_file = '[path-to-file]'} and
maestro will create/append to a file in the project directory.
This log file will be appended to until it exceeds the byte size defined in \code{log_file_max_bytes} argument after which
the log file is deleted.
}
}
\examples{

if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  # Runs the schedule every 1 day
  run_schedule(
    schedule,
    orch_frequency = "1 day",
    quiet = TRUE
  )

  # Runs the schedule every 15 minutes
  run_schedule(
    schedule,
    orch_frequency = "15 minutes",
    quiet = TRUE
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_labels.R
\name{get_labels}
\alias{get_labels}
\title{Get the labels of pipelines in a MaestroSchedule object}
\usage{
get_labels(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
data.frame
}
\description{
Creates a data.frame of labels for all labelled pipelines in the schedule.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  get_labels(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_labels()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_pipeline_run_sequence}
\alias{get_pipeline_run_sequence}
\title{Generate a sequence of run times for a pipeline}
\usage{
get_pipeline_run_sequence(
  pipeline_n,
  pipeline_unit,
  pipeline_datetime,
  check_datetime,
  pipeline_hours = 0:23,
  pipeline_days_of_week = 1:7,
  pipeline_days_of_month = 1:31,
  pipeline_months = 1:12
)
}
\arguments{
\item{pipeline_n}{number of units for the pipeline frequency}

\item{pipeline_unit}{unit for the pipeline frequency}

\item{pipeline_datetime}{datetime of the first time the pipeline is to run}

\item{check_datetime}{datetime against which to check the running of pipelines (default is current system time in UTC)}

\item{pipeline_hours}{vector of integers [0-23] corresponding to hours of day for the pipeline to run}

\item{pipeline_days_of_week}{vector of integers [1-7] corresponding to days of week for the pipeline to run (1 = Sunday)}

\item{pipeline_days_of_month}{vector of integers [1-31] corresponding to days of month for the pipeline to run}

\item{pipeline_months}{vector of integers [1-12] corresponding to months of year for the pipeline to run}
}
\value{
vector of timestamps or dates
}
\description{
Generate a sequence of run times for a pipeline
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MaestroPipeline.R
\name{MaestroPipeline}
\alias{MaestroPipeline}
\title{Class for an individual maestro pipeline
A pipeline is defined as a single R script with a schedule or input}
\description{
Class for an individual maestro pipeline
A pipeline is defined as a single R script with a schedule or input
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MaestroPipeline-initialize}{\code{MaestroPipeline$new()}}
    \item \href{#method-MaestroPipeline-print}{\code{MaestroPipeline$print()}}
    \item \href{#method-MaestroPipeline-run}{\code{MaestroPipeline$run()}}
    \item \href{#method-MaestroPipeline-get_pipe_name}{\code{MaestroPipeline$get_pipe_name()}}
    \item \href{#method-MaestroPipeline-get_frequency_nunits}{\code{MaestroPipeline$get_frequency_nunits()}}
    \item \href{#method-MaestroPipeline-get_schedule}{\code{MaestroPipeline$get_schedule()}}
    \item \href{#method-MaestroPipeline-check_timeliness}{\code{MaestroPipeline$check_timeliness()}}
    \item \href{#method-MaestroPipeline-get_status}{\code{MaestroPipeline$get_status()}}
    \item \href{#method-MaestroPipeline-get_status_chr}{\code{MaestroPipeline$get_status_chr()}}
    \item \href{#method-MaestroPipeline-get_outputs}{\code{MaestroPipeline$get_outputs()}}
    \item \href{#method-MaestroPipeline-get_inputs}{\code{MaestroPipeline$get_inputs()}}
    \item \href{#method-MaestroPipeline-get_priority}{\code{MaestroPipeline$get_priority()}}
    \item \href{#method-MaestroPipeline-get_returns}{\code{MaestroPipeline$get_returns()}}
    \item \href{#method-MaestroPipeline-get_artifacts}{\code{MaestroPipeline$get_artifacts()}}
    \item \href{#method-MaestroPipeline-get_n_artifacts}{\code{MaestroPipeline$get_n_artifacts()}}
    \item \href{#method-MaestroPipeline-get_n_invocations}{\code{MaestroPipeline$get_n_invocations()}}
    \item \href{#method-MaestroPipeline-get_all_returns}{\code{MaestroPipeline$get_all_returns()}}
    \item \href{#method-MaestroPipeline-get_errors}{\code{MaestroPipeline$get_errors()}}
    \item \href{#method-MaestroPipeline-get_warnings}{\code{MaestroPipeline$get_warnings()}}
    \item \href{#method-MaestroPipeline-get_messages}{\code{MaestroPipeline$get_messages()}}
    \item \href{#method-MaestroPipeline-get_flags}{\code{MaestroPipeline$get_flags()}}
    \item \href{#method-MaestroPipeline-get_labels}{\code{MaestroPipeline$get_labels()}}
    \item \href{#method-MaestroPipeline-get_is_collect}{\code{MaestroPipeline$get_is_collect()}}
    \item \href{#method-MaestroPipeline-get_map}{\code{MaestroPipeline$get_map()}}
    \item \href{#method-MaestroPipeline-get_is_map}{\code{MaestroPipeline$get_is_map()}}
    \item \href{#method-MaestroPipeline-get_n_expected_iterations}{\code{MaestroPipeline$get_n_expected_iterations()}}
    \item \href{#method-MaestroPipeline-set_n_expected_iterations}{\code{MaestroPipeline$set_n_expected_iterations()}}
    \item \href{#method-MaestroPipeline-update_inputs}{\code{MaestroPipeline$update_inputs()}}
    \item \href{#method-MaestroPipeline-update_outputs}{\code{MaestroPipeline$update_outputs()}}
    \item \href{#method-MaestroPipeline-reset_run_time_attributes}{\code{MaestroPipeline$reset_run_time_attributes()}}
    \item \href{#method-MaestroPipeline-get_run_sequence}{\code{MaestroPipeline$get_run_sequence()}}
    \item \href{#method-MaestroPipeline-clone}{\code{MaestroPipeline$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-initialize}{}}}
\subsection{\code{MaestroPipeline$new()}}{
  Create a new Pipeline object
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$new(
  script_path,
  pipe_name,
  frequency = NA_character_,
  start_time_raw = NA_character_,
  tz = NA_character_,
  hours = NULL,
  days = NULL,
  months = NULL,
  skip = FALSE,
  log_level = "INFO",
  inputs = NULL,
  outputs = NULL,
  priority = Inf,
  flags = c(),
  run_if = NULL,
  is_collect = FALSE,
  map = NULL,
  labels = list()
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{script_path}}{path to the script}
      \item{\code{pipe_name}}{name of the pipeline}
      \item{\code{frequency}}{frequency of the pipeline (e.g., 1 day)}
      \item{\code{start_time_raw}}{start time as a raw string from the @maestroStartTime tag}
      \item{\code{tz}}{time zone of the pipeline}
      \item{\code{hours}}{specific hours of the day}
      \item{\code{days}}{specific days of week or month}
      \item{\code{months}}{specific months of year}
      \item{\code{skip}}{whether to skip the pipeline regardless of scheduling}
      \item{\code{log_level}}{log level of the pipeline}
      \item{\code{inputs}}{names of pipelines that this pipeline is dependent on for input}
      \item{\code{outputs}}{names of pipelines for which this pipeline is a dependency}
      \item{\code{priority}}{priority of the pipeline}
      \item{\code{flags}}{arbitrary pipelines flags}
      \item{\code{run_if}}{string representing an R expression that can be evaluated and returns TRUE or FALSE; or NULL}
      \item{\code{is_collect}}{logical; TRUE when @maestroInputs uses collect() fan-in marker}
      \item{\code{map}}{named list of key=expr_string pairs from @maestroMap, or NULL}
      \item{\code{labels}}{list of key-value pairs for pipeline labeling}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroPipeline object
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-print"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-print}{}}}
\subsection{\code{MaestroPipeline$print()}}{
  Prints the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$print()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-run"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-run}{}}}
\subsection{\code{MaestroPipeline$run()}}{
  Runs the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$run(
  resources = list(),
  log_file = tempfile(),
  quiet = FALSE,
  log_file_max_bytes = 1e+06,
  .input = NULL,
  depth = 0,
  log_to_console = FALSE,
  run_id = NA_character_,
  input_run_id = NA_character_,
  lineage = c(),
  iter = NULL,
  pre_error = NULL,
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{resources}}{named list of arguments and values to pass to the pipeline}
      \item{\code{log_file}}{path to the log file for logging}
      \item{\code{quiet}}{whether to silence console output}
      \item{\code{log_file_max_bytes}}{maximum bytes of the log file before trimming}
      \item{\code{.input}}{input values from upstream pipelines}
      \item{\code{depth}}{number of inputting pipelines above the current}
      \item{\code{log_to_console}}{whether or not to output statements in the console (FALSE is to suppress and append to log)}
      \item{\code{run_id}}{unique id for the run}
      \item{\code{input_run_id}}{unique id of the run that inputted into the current run (NA if there is no input)}
      \item{\code{lineage}}{character vector of upstream pipeline names ordered from first to latest (or empty if no upstream pipes)}
      \item{\code{iter}}{iteration number for dynamic fanout}
      \item{\code{pre_error}}{trigger error from outside execution}
      \item{\code{...}}{additional arguments (unused)}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_pipe_name"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_pipe_name}{}}}
\subsection{\code{MaestroPipeline$get_pipe_name()}}{
  Get the pipeline name
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_pipe_name()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    pipeline_name
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_frequency_nunits"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_frequency_nunits}{}}}
\subsection{\code{MaestroPipeline$get_frequency_nunits()}}{
  Get the frequency n and unit as a list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_frequency_nunits()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list with n and unit
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_schedule"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_schedule}{}}}
\subsection{\code{MaestroPipeline$get_schedule()}}{
  Get the schedule as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_schedule()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-check_timeliness"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-check_timeliness}{}}}
\subsection{\code{MaestroPipeline$check_timeliness()}}{
  Check whether a pipeline is scheduled to run based on orchestrator frequency and current time
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$check_timeliness(
  orch_unit,
  orch_n,
  check_datetime = lubridate::now(),
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{orch_unit}}{unit of the orchestrator (e.g., day)}
      \item{\code{orch_n}}{number of units of the frequency}
      \item{\code{check_datetime}}{datetime against which to check the timeliness of the pipeline (should almost always be now)}
      \item{\code{...}}{unused}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroPipeline
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_status"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_status}{}}}
\subsection{\code{MaestroPipeline$get_status()}}{
  Get status of the pipeline as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_status()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_status_chr"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_status_chr}{}}}
\subsection{\code{MaestroPipeline$get_status_chr()}}{
  Get status of the pipeline as a string
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_status_chr()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    character
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_outputs"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_outputs}{}}}
\subsection{\code{MaestroPipeline$get_outputs()}}{
  Names of pipelines that receive input from this pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_outputs()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    character
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_inputs"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_inputs}{}}}
\subsection{\code{MaestroPipeline$get_inputs()}}{
  Names of pipelines that input into this pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_inputs()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    character
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_priority"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_priority}{}}}
\subsection{\code{MaestroPipeline$get_priority()}}{
  Get priority of the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_priority()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    numeric
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_returns"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_returns}{}}}
\subsection{\code{MaestroPipeline$get_returns()}}{
  Get immediate return values from the pipeline for downstream pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_returns()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_artifacts"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_artifacts}{}}}
\subsection{\code{MaestroPipeline$get_artifacts()}}{
  Get artifacts (return values) from the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_artifacts()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_n_artifacts"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_n_artifacts}{}}}
\subsection{\code{MaestroPipeline$get_n_artifacts()}}{
  Get the number of completed artifact runs (iterations) for this pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_n_artifacts()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    integer
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_n_invocations"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_n_invocations}{}}}
\subsection{\code{MaestroPipeline$get_n_invocations()}}{
  Get the number of times this pipeline was invoked (successes + errors).
For @maestroMap pipelines this equals the number of iterations that have
finished, regardless of outcome.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_n_invocations()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    integer
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_all_returns"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_all_returns}{}}}
\subsection{\code{MaestroPipeline$get_all_returns()}}{
  Get all artifact values as a plain unnamed list, regardless of how many
iterations have run. Used by collect() to gather each-pipe outputs.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_all_returns()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_errors"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_errors}{}}}
\subsection{\code{MaestroPipeline$get_errors()}}{
  Get list of errors from the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_errors()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_warnings"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_warnings}{}}}
\subsection{\code{MaestroPipeline$get_warnings()}}{
  Get list of warnings from the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_warnings()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_messages"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_messages}{}}}
\subsection{\code{MaestroPipeline$get_messages()}}{
  Get list of messages from the pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_messages()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_flags"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_flags}{}}}
\subsection{\code{MaestroPipeline$get_flags()}}{
  Get the flags of a pipeline as a vector
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_flags()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    character
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_labels"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_labels}{}}}
\subsection{\code{MaestroPipeline$get_labels()}}{
  Get the labels of a pipeline as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_labels()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_is_collect"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_is_collect}{}}}
\subsection{\code{MaestroPipeline$get_is_collect()}}{
  Get whether the pipeline uses \code{collect} for fan in
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_is_collect()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    logical
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_map"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_map}{}}}
\subsection{\code{MaestroPipeline$get_map()}}{
  Get the .input value to iterate over
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_map()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    character
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_is_map"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_is_map}{}}}
\subsection{\code{MaestroPipeline$get_is_map()}}{
  Get whether the pipeline uses an iterator for dynamic fan out
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_is_map()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    logical
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_n_expected_iterations"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_n_expected_iterations}{}}}
\subsection{\code{MaestroPipeline$get_n_expected_iterations()}}{
  Get the number of iterations expected for this @maestroMap pipeline.
Set by run_pipe() just before the scatter loop so that resolve_collect_input()
can compare against get_n_invocations() without depending on the upstream
return value's length (which is wrong when a field selector is used).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_n_expected_iterations()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    integer or NULL
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-set_n_expected_iterations"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-set_n_expected_iterations}{}}}
\subsection{\code{MaestroPipeline$set_n_expected_iterations()}}{
  Record the number of iterations that will be dispatched for this @maestroMap pipeline.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$set_n_expected_iterations(n)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{integer}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-update_inputs"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-update_inputs}{}}}
\subsection{\code{MaestroPipeline$update_inputs()}}{
  Update the inputs of a pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$update_inputs(inputs)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{inputs}}{character vector of inputting pipeline names}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    vector
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-update_outputs"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-update_outputs}{}}}
\subsection{\code{MaestroPipeline$update_outputs()}}{
  Update the outputs of a pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$update_outputs(outputs)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{outputs}}{character vector of outputting pipeline names}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    vector
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-reset_run_time_attributes"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-reset_run_time_attributes}{}}}
\subsection{\code{MaestroPipeline$reset_run_time_attributes()}}{
  Resets run time attributes
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$reset_run_time_attributes()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-get_run_sequence"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-get_run_sequence}{}}}
\subsection{\code{MaestroPipeline$get_run_sequence()}}{
  Get the run sequence of a pipeline
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$get_run_sequence(
  n = NULL,
  min_datetime = NULL,
  max_datetime = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{optional sequence limit}
      \item{\code{min_datetime}}{optional minimum datetime}
      \item{\code{max_datetime}}{optional maximum datetime}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    vector
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroPipeline-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroPipeline-clone}{}}}
\subsection{\code{MaestroPipeline$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroPipeline$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_valid_dag}
\alias{is_valid_dag}
\title{Checks whether a DAG is valid (no loops)}
\usage{
is_valid_dag(edges)
}
\arguments{
\item{edges}{a data.frame of edges (from, to)}
}
\value{
boolean
}
\description{
Checks whether a DAG is valid (no loops)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_run_sequence.R
\name{get_run_sequence}
\alias{get_run_sequence}
\title{Get the run sequence of a schedule}
\usage{
get_run_sequence(
  schedule,
  n = NULL,
  min_datetime = NULL,
  max_datetime = NULL,
  include_only_primary = FALSE,
  include_skipped = TRUE
)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}

\item{n}{Optional positive integer. If specified, returns only the first \code{n} runs for each pipeline.}

\item{min_datetime}{Optional minimum datetime filter. Can be a \code{Date} or \code{POSIXct} object.
If specified, only returns runs scheduled at or after this datetime.}

\item{max_datetime}{Optional maximum datetime filter. Can be a \code{Date} or \code{POSIXct} object.
If specified, only returns runs scheduled at or before this datetime.}

\item{include_only_primary}{only primary pipelines are included
(this are pipelines that are scheduled and not downstream nodes in a DAG)}

\item{include_skipped}{whether to include pipelines tagged with \verb{@maestroSkip}
(default \code{TRUE} for backwards compatibility)}
}
\value{
A data.frame of scheduled run times with columns \code{pipe_name}, \code{scheduled_time},
and \code{is_primary}.
}
\description{
Retrieves the scheduled run times for a given schedule, with optional filtering
by number of runs and datetime range.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  get_run_sequence(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_run_sequence()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.run_sequence_min_days_out}
\alias{.run_sequence_min_days_out}
\title{Minimum days ahead to pre-compute run sequences on initialization
This small window is used at parse time; the full window is built lazily
when observability functions (get_run_sequence, get_slot_usage) are called.}
\usage{
.run_sequence_min_days_out(unit)
}
\arguments{
\item{unit}{character frequency unit}
}
\value{
integer
}
\description{
Minimum days ahead to pre-compute run sequences on initialization
This small window is used at parse time; the full window is built lazily
when observability functions (get_run_sequence, get_slot_usage) are called.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/latest_getters.R
\name{last_run_messages}
\alias{last_run_messages}
\title{Retrieve latest maestro pipeline messages}
\usage{
last_run_messages()
}
\value{
messages
}
\description{
Gets the latest pipeline messages following use of \code{run_schedule()}. If
there are no messages or \code{run_schedule()} has not been run it will be \code{NULL}.
}
\details{
Note that setting \code{maestroLogLevel} to something greater than \code{INFO} will
ignore messages.
}
\examples{

last_run_messages()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/suggest_orch_frequency.R
\name{suggest_orch_frequency}
\alias{suggest_orch_frequency}
\title{Suggest orchestrator frequency based on a schedule}
\usage{
suggest_orch_frequency(
  schedule,
  check_datetime = lubridate::now(tzone = "UTC")
)
}
\arguments{
\item{schedule}{MaestroSchedule object created by \code{build_schedule()}}

\item{check_datetime}{datetime against which to check the running of pipelines (default is current system time in UTC)}
}
\value{
frequency string
}
\description{
Suggests a frequency to run the orchestrator based on the frequencies of the
pipelines in a schedule.
}
\details{
This function attempts to find the smallest interval of time between all pipelines.
If the smallest interval is less than 15 minutes, it just uses the smallest interval.

Note this function is intended to be used interactively when deciding how often to
schedule the orchestrator. Programmatic use is not recommended.

\strong{Deprecated.} This function is no longer actively maintained and may be removed
in a future release.
}
\examples{

if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)
  suggest_orch_frequency(schedule)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_pipeline.R
\name{create_pipeline}
\alias{create_pipeline}
\title{Create a new pipeline in a pipelines directory}
\usage{
create_pipeline(
  pipe_name,
  pipeline_dir = "pipelines",
  frequency = "1 day",
  start_time = Sys.Date(),
  tz = "UTC",
  log_level = "INFO",
  quiet = FALSE,
  open = interactive(),
  overwrite = FALSE,
  skip = FALSE,
  inputs = NULL,
  outputs = NULL,
  priority = NULL
)
}
\arguments{
\item{pipe_name}{name of the pipeline and function}

\item{pipeline_dir}{directory containing the pipeline scripts}

\item{frequency}{how often the pipeline should run (e.g., 1 day, daily, 3 hours, 4 months). Fills in maestroFrequency tag}

\item{start_time}{start time of the pipeline schedule. Fills in maestroStartTime tag}

\item{tz}{timezone that pipeline will be scheduled in. Fills in maestroTz tag}

\item{log_level}{log level for the pipeline (e.g., INFO, WARN, ERROR). Fills in maestroLogLevel tag}

\item{quiet}{whether to silence messages in the console (default = \code{FALSE})}

\item{open}{whether or not to open the script upon creation}

\item{overwrite}{whether or not to overwrite an existing pipeline of the same name and location.}

\item{skip}{whether to skip the pipeline when running in the orchestrator (default = \code{FALSE})}

\item{inputs}{vector of names of pipelines that input into this pipeline (default = \code{NULL} for no inputs)}

\item{outputs}{vector of names of pipelines that receive output from this pipeline (default = \code{NULL} for no outputs)}

\item{priority}{a single positive integer corresponding to the order in which this pipeline will be invoked in the presence of other simultaneously invoked pipelines.}
}
\value{
invisible
}
\description{
Allows the creation of new pipelines (R scripts) and fills in the maestro tags as specified.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline(
    "extract_data",
    pipeline_dir = pipeline_dir,
    frequency = "1 hour",
    open = FALSE,
    quiet = TRUE,
    overwrite = TRUE
  )

  create_pipeline(
    "new_job",
    pipeline_dir = pipeline_dir,
    frequency = "20 minutes",
    start_time = as.POSIXct("2024-06-21 12:20:00"),
    log_level = "ERROR",
    open = FALSE,
    quiet = TRUE,
    overwrite = TRUE
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invoke.R
\name{invoke}
\alias{invoke}
\title{Manually run a pipeline regardless of schedule}
\usage{
invoke(
  schedule,
  pipe_name,
  resources = list(),
  ...,
  quiet = TRUE,
  log_to_console = FALSE
)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}

\item{pipe_name}{name of a single pipe name from the schedule}

\item{resources}{named list of shared resources made available to pipelines as needed}

\item{...}{other arguments passed to \code{run_schedule()}}

\item{quiet}{silence metrics to the console (default = \code{FALSE}). Note this does not affect messages generated from pipelines when \code{log_to_console = TRUE}.}

\item{log_to_console}{whether or not to include pipeline messages, warnings, errors to the console (default = \code{FALSE}) (see Logging & Console Output section)}
}
\value{
invisible
}
\description{
Instantly run a single pipeline from the schedule. This is useful for testing
purposes or if you want to just run something one-off.
}
\details{
Scheduling parameters such as the frequency, start time, and specifiers are ignored.
The pipeline will be run even if \code{maestroSkip} is present. If the pipeline is a DAG
pipeline, \code{invoke} will attempt to execute the full DAG.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  invoke(schedule, "my_new_pipeline")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{parse_rounding_unit}
\alias{parse_rounding_unit}
\title{Parse a time string}
\usage{
parse_rounding_unit(time_string)
}
\arguments{
\item{time_string}{string like 1 day, daily, 1 week, 12 hours, etc.}
}
\value{
nunit list
}
\description{
Parse a time string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_slot_usage.R
\name{get_slot_usage}
\alias{get_slot_usage}
\title{Get time slot usage of a schedule}
\usage{
get_slot_usage(schedule, orch_frequency, slot_interval = "hour")
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}

\item{orch_frequency}{of the orchestrator, a single string formatted like "1 day", "2 weeks", "hourly", etc.}

\item{slot_interval}{a time unit indicating the interval of time to consider between slots (e.g., 'hour', 'day')}
}
\value{
data.frame
}
\description{
Get the number of pipelines scheduled to run for each time slot at a particular
slot interval. Time slots are times that the orchestrator runs and the slot interval
determines the level of granularity to consider.
}
\details{
This function is particularly useful when you have multiple pipelines in a project
and you want to see what recurring time intervals may be available or underused
for new pipelines.

Note that this function is intended for use in an interactive session while developing
a maestro project. It is not intended for use in the orchestrator.

As an example, consider we have four pipelines running at various frequencies
and the orchestrator running every hour. Then let's say there's to be a new
pipeline that runs every day. One might ask 'what hour should I schedule this new
pipeline to run on?'. By using \code{get_slot_usage(schedule, orch_frequency = '1 hour', slot_interval = 'hour')}
on the existing schedule, you could identify for each hour how many pipelines
are already scheduled to run and choose the ones with the lowest usage.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  get_slot_usage(
    schedule,
    orch_frequency = "1 hour",
    slot_interval = "hour"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{parse_maestro_start_time}
\alias{parse_maestro_start_time}
\title{Resolve a partial \code{@maestroStartTime} string to a concrete POSIXct}
\usage{
parse_maestro_start_time(raw, tz, now = lubridate::now(tzone = tz))
}
\arguments{
\item{raw}{Character string from the tag value.}

\item{tz}{Timezone string.}

\item{now}{\code{POSIXct} reference point (default \code{lubridate::now(tz)}).}
}
\value{
A \code{POSIXct}, or \code{NA} if \code{raw} is \code{NA}/\code{""}.
}
\description{
Accepts four formats:
\itemize{
\item \code{"HH:MM:SS"} — time of day; anchored to today's date.
\item \code{"Mon HH:MM:SS"} — weekday + time; anchored to that weekday of
the current ISO week.
\item \code{"DD HH:MM:SS"} or \code{"DD"} — month-day (+ optional time);
anchored to that day of the current month.
\item Full datetime string — passed through to \code{as.POSIXct()}.
}
}
\details{
Using a recent \code{now} as the reference keeps the anchor close to the
present, so \code{n} in \code{.prev_on_cycle()} stays small.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/latest_getters.R
\name{last_run_errors}
\alias{last_run_errors}
\title{Retrieve latest maestro pipeline errors}
\usage{
last_run_errors()
}
\value{
error messages
}
\description{
Gets the latest pipeline errors following use of \code{run_schedule()}. If
the all runs succeeded or \code{run_schedule()} has not been run it will be \code{NULL}.
}
\examples{

last_run_errors()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MaestroSchedule.R
\name{MaestroSchedule}
\alias{MaestroSchedule}
\title{Class for a schedule of pipelines}
\description{
Class for a schedule of pipelines
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)
}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{PipelineList}}{object of type MaestroPipelineList}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MaestroSchedule-initialize}{\code{MaestroSchedule$new()}}
    \item \href{#method-MaestroSchedule-print}{\code{MaestroSchedule$print()}}
    \item \href{#method-MaestroSchedule-run}{\code{MaestroSchedule$run()}}
    \item \href{#method-MaestroSchedule-get_schedule}{\code{MaestroSchedule$get_schedule()}}
    \item \href{#method-MaestroSchedule-get_status}{\code{MaestroSchedule$get_status()}}
    \item \href{#method-MaestroSchedule-get_artifacts}{\code{MaestroSchedule$get_artifacts()}}
    \item \href{#method-MaestroSchedule-get_network}{\code{MaestroSchedule$get_network()}}
    \item \href{#method-MaestroSchedule-get_flags}{\code{MaestroSchedule$get_flags()}}
    \item \href{#method-MaestroSchedule-get_labels}{\code{MaestroSchedule$get_labels()}}
    \item \href{#method-MaestroSchedule-get_run_sequence}{\code{MaestroSchedule$get_run_sequence()}}
    \item \href{#method-MaestroSchedule-clone}{\code{MaestroSchedule$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-initialize}{}}}
\subsection{\code{MaestroSchedule$new()}}{
  Create a MaestroSchedule object
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$new(Pipelines = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{Pipelines}}{list of MaestroPipelines}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    MaestroSchedule
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-print"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-print}{}}}
\subsection{\code{MaestroSchedule$print()}}{
  Print the schedule object
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$print()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    print
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-run"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-run}{}}}
\subsection{\code{MaestroSchedule$run()}}{
  Run a MaestroSchedule
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$run(..., quiet = FALSE, run_all = FALSE, n_show_next = 5)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{arguments passed to MaestroPipelineList$run}
      \item{\code{quiet}}{whether or not to silence console messages}
      \item{\code{run_all}}{run all pipelines regardless of the schedule (default is \code{FALSE}) - useful for testing.}
      \item{\code{n_show_next}}{show the next n scheduled pipes}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    invisible
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_schedule"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_schedule}{}}}
\subsection{\code{MaestroSchedule$get_schedule()}}{
  Get the schedule as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_schedule()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_status"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_status}{}}}
\subsection{\code{MaestroSchedule$get_status()}}{
  Get status of the pipelines as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_status()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_artifacts"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_artifacts}{}}}
\subsection{\code{MaestroSchedule$get_artifacts()}}{
  Get artifacts (return values) from the pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_artifacts()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    list
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_network"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_network}{}}}
\subsection{\code{MaestroSchedule$get_network()}}{
  Get the network structure of the pipelines as an edge list (will be empty if there are no DAG pipelines)
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_network()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_flags"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_flags}{}}}
\subsection{\code{MaestroSchedule$get_flags()}}{
  Get all pipeline flags as a long data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_flags()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_labels"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_labels}{}}}
\subsection{\code{MaestroSchedule$get_labels()}}{
  Get all pipeline labels as a data.frame
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_labels()}
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-get_run_sequence"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-get_run_sequence}{}}}
\subsection{\code{MaestroSchedule$get_run_sequence()}}{
  Get full sequence of scheduled executions for all pipelines
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$get_run_sequence(
  n = NULL,
  min_datetime = NULL,
  max_datetime = NULL,
  include_only_primary = FALSE,
  include_skipped = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{optional sequence limit}
      \item{\code{min_datetime}}{optional minimum datetime}
      \item{\code{max_datetime}}{optional maximum datetime}
      \item{\code{include_only_primary}}{only primary pipelines are included
(this are pipelines that are scheduled and not downstream nodes in a DAG)}
      \item{\code{include_skipped}}{whether to include pipelines tagged with \verb{@maestroSkip}
(default \code{TRUE} for backwards compatibility)}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    data.frame
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MaestroSchedule-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MaestroSchedule-clone}{}}}
\subsection{\code{MaestroSchedule$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MaestroSchedule$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show_network.R
\name{show_network}
\alias{show_network}
\title{Visualize the schedule as a DAG}
\usage{
show_network(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
DiagrammeR visualization
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

Create an interactive network visualization to show the dependency structure
of pipelines in the schedule. This is only useful if there are pipelines in
the schedule that take inputs/outputs from other pipelines.

Note that running this function on a schedule with all independent pipelines
will produce a network visual with no connections.

This function requires the installation of \code{DiagrammeR} which is not automatically
installed with \code{maestro}.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  schedule <- run_schedule(
    schedule,
    orch_frequency = "1 day",
    quiet = TRUE
  )

  show_network(schedule)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.run_sequence_days_out}
\alias{.run_sequence_days_out}
\title{Number of days ahead to generate run sequences for a given frequency unit
Used by the observable get_run_sequence path.}
\usage{
.run_sequence_days_out(unit)
}
\arguments{
\item{unit}{character frequency unit}
}
\value{
integer
}
\description{
Number of days ahead to generate run sequences for a given frequency unit
Used by the observable get_run_sequence path.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_schedule.R
\name{build_schedule}
\alias{build_schedule}
\title{Build a schedule}
\usage{
build_schedule(pipeline_dir = "./pipelines", cores = 1L, quiet = FALSE)
}
\arguments{
\item{pipeline_dir}{path to directory containing the pipeline scripts}

\item{cores}{number of cpu cores to use when parsing pipeline scripts. If > 1,
\code{furrr} is used and a multisession plan must be set in the orchestrator
(e.g. \code{future::plan(future::multisession)}). See \code{run_schedule()} for details.}

\item{quiet}{silence metrics to the console (default = \code{FALSE})}
}
\value{
MaestroSchedule
}
\description{
Builds a \code{MaestroSchedule} object for use in \code{run_schedule()}.
}
\details{
This function parses the maestro tags of functions located in \code{pipeline_dir} which is
conventionally called 'pipelines'. An orchestrator requires a \code{MaestroSchedule}
to determine which pipelines are to run and when. Each pipeline in the schedule
is a parsed function and its scheduling parameters such as its frequency.

The \code{MaestroSchedule} is mostly intended to be passed directly to \code{run_schedule()}.
In other words, it is not recommended to make changes to it.
}
\examples{

# Creating a temporary directory for demo purposes! In practice, just
# create a 'pipelines' directory at the project level.

if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  build_schedule(pipeline_dir = pipeline_dir)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_schedule_entry.R
\name{build_schedule_entry}
\alias{build_schedule_entry}
\title{Parse and validate tags then create and populate MaestroPipelineList}
\usage{
build_schedule_entry(script_path)
}
\arguments{
\item{script_path}{path to script}
}
\value{
MaestroPipelineList R6 class
}
\description{
Parse and validate tags then create and populate MaestroPipelineList
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_cli.R
\name{maestro_parse_cli}
\alias{maestro_parse_cli}
\title{cli output for generate schedule table}
\usage{
maestro_parse_cli(parse_succeeds, parse_errors)
}
\arguments{
\item{parse_succeeds}{list of parse results (i.e., succeeded)}

\item{parse_errors}{list of parse errors}
}
\value{
cli output
}
\description{
cli output for generate schedule table
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_artifacts.R
\name{get_artifacts}
\alias{get_artifacts}
\title{Get the artifacts (return values) of the pipelines in a MaestroSchedule object.}
\usage{
get_artifacts(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
named list
}
\description{
Artifacts are return values from pipelines. They are accessible as a named list
where the names correspond to the names of the pipeline.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  schedule <- run_schedule(
    schedule,
    orch_frequency = "1 day",
    quiet = TRUE
  )

  get_artifacts(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_artifacts()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_flags.R
\name{get_flags}
\alias{get_flags}
\title{Get the flags of pipelines in a MaestroSchedule object}
\usage{
get_flags(schedule)
}
\arguments{
\item{schedule}{object of type MaestroSchedule created using \code{build_schedule()}}
}
\value{
data.frame
}
\description{
Creates a long data.frame where each row is a flag for each pipeline.
}
\examples{
if (interactive()) {
  pipeline_dir <- tempdir()
  create_pipeline("my_new_pipeline", pipeline_dir, open = FALSE)
  schedule <- build_schedule(pipeline_dir = pipeline_dir)

  get_flags(schedule)

  # Alternatively, use the underlying R6 method
  schedule$get_flags()
}
}
